(* *********************************************************************)
(*                                                                     *)
(*              The Compcert verified compiler                         *)
(*                                                                     *)
(*          Xavier Leroy, INRIA Paris-Rocquencourt                     *)
(*          Jacques-Henri Jourdan, INRIA Paris-Rocquencourt            *)
(*          Vincent Laporte, Université de Rennes 1                    *)
(*                                                                     *)
(*  Copyright Institut National de Recherche en Informatique et en     *)
(*  Automatique and Université de Rennes 1.  All rights reserved.      *)
(*  This file is distributed under the terms of the INRIA              *)
(*  Non-Commercial License Agreement.                                  *)
(*                                                                     *)
(* *********************************************************************)

Require Coqlib.
Require Wfsimpl.
Require AST.
Require Iteration.
Require Floats.
Require SelectLong.
Require Selection.
Require RTLgen.
Require Inlining.
Require ValueDomain.
Require Tailcall.
Require Allocation.
Require Bounds.
Require Ctypes.
Require Csyntax.
Require Ctyping.
Require Clight.
Require Compiler.
Require Parser.
Require Initializers.
Require Int31.
Require PrintPos.
Require CsharpminorAnalysis.
Require PedraExtract.
Require UnLoad.

(* Standard lib *)
Require Import ExtrOcamlBasic.
Require Import ExtrOcamlString.

(* Arithmetics *)
Extract Inductive BinPos.positive => "PNZimpl.P.t"
  ["PNZimpl.P.xI" "PNZimpl.P.xO" "PNZimpl.P.xH"] "PNZimpl.P.dispatch".
Extract Inductive BinNums.N => "PNZimpl.N.t"
  ["PNZimpl.N.n0" "PNZimpl.N.npos"] "PNZimpl.N.dispatch".
Extract Inductive BinInt.Z => "PNZimpl.Z.t"
  ["PNZimpl.Z.z0" "PNZimpl.Z.zpos" "PNZimpl.Z.zneg"] "PNZimpl.Z.dispatch".

Extract Constant BinPos.Pos.compare_cont => "PNZimpl.P.compare_cont".
Extract Constant BinPos.Pos.eq_dec =>       "PNZimpl.P.eq".
Extract Constant BinPos.Pos.eqb =>          "PNZimpl.P.eq".
Extract Constant BinPos.Pos.add =>          "PNZimpl.P.add".
Extract Constant BinPos.Pos.mul =>          "PNZimpl.P.mul".
Extract Constant BinPos.Pos.succ =>         "PNZimpl.P.succ".
Extract Constant BinPos.Pos.pow =>          "PNZimpl.P.pow".

Extract Constant BinNat.N.land =>     "PNZimpl.N.logand".
Extract Constant BinNat.N.lxor =>     "PNZimpl.N.logxor".
Extract Constant BinNat.N.lor =>      "PNZimpl.N.logor".
Extract Constant BinNat.N.shiftr =>   "PNZimpl.N.shiftr".
Extract Constant BinNat.N.shiftl =>   "PNZimpl.N.shiftl".
Extract Constant BinNat.N.add =>      "PNZimpl.N.add".
Extract Constant BinNat.N.mul =>      "PNZimpl.N.mul".
Extract Constant BinNat.N.succ =>     "PNZimpl.N.succ".
Extract Constant BinNat.N.div_eucl => "PNZimpl.N.div_eucl".
Extract Constant BinNat.N.pow =>      "PNZimpl.N.pow".

Extract Constant BinInt.Z.add =>      "PNZimpl.Z.add".
Extract Constant BinInt.Z.sub =>      "PNZimpl.Z.sub".
Extract Constant BinInt.Z.opp =>      "PNZimpl.Z.neg".
Extract Constant BinInt.Z.mul =>      "PNZimpl.Z.mul".
Extract Constant BinInt.Z.land =>     "PNZimpl.Z.logand".
Extract Constant BinInt.Z.lxor =>     "PNZimpl.Z.logxor".
Extract Constant BinInt.Z.lor =>      "PNZimpl.Z.logor".
Extract Constant BinInt.Z.shiftr =>   "PNZimpl.Z.shiftr".
Extract Constant BinInt.Z.shiftl =>   "PNZimpl.Z.shiftl".
Extract Constant BinInt.Z.succ =>     "PNZimpl.Z.succ".
Extract Constant BinInt.Z.pred =>     "PNZimpl.Z.pred".
Extract Constant BinInt.Z.div_eucl => "PNZimpl.Z.div_eucl".
Extract Constant BinInt.Z.quotrem =>  "PNZimpl.Z.quotrem".
Extract Constant BinInt.Z.pow =>      "PNZimpl.Z.pow".

(* Coqlib *)
Extract Inlined Constant Coqlib.proj_sumbool => "(fun x -> x)".

(* Wfsimpl *)
Extraction Inline Wfsimpl.Fix Wfsimpl.Fixm.

(* AST *)
Extract Constant AST.ident_of_string =>
  "fun s -> Camlcoq.intern_string (Camlcoq.camlstring_of_coqstring s)".

(* Memory - work around an extraction bug. *)
Extraction NoInline Memory.Mem.valid_pointer.

(* Errors *)
Extraction Inline Errors.bind Errors.bind2.

(* Iteration *)

Extract Constant Iteration.GenIter.iterate =>
  "let rec iter f a =
     match f a with Coq_inl b -> Some b | Coq_inr a' -> iter f a'
   in iter".

(* Selection *)

Extract Constant Selection.compile_switch => "Switchaux.compile_switch".

(* RTLgen *)
Extract Constant RTLgen.more_likely => "RTLgenaux.more_likely".
Extraction Inline RTLgen.ret RTLgen.error RTLgen.bind RTLgen.bind2.

(* Inlining *)
Extract Inlined Constant Inlining.should_inline => "Inliningaux.should_inline".
Extraction Inline Inlining.ret Inlining.bind.

(* Value Analysis *)
Require Debug.
Extract Constant Debug.print_string => "fun x -> print_endline (Camlcoq.camlstring_of_coqstring x)".
Extract Constant Debug.string_of_ident => "fun x -> Camlcoq.(coqstring_of_camlstring (extern_atom x))".

(* Csharpminor Iterator *)
Extract Constant CsharpminorIter.fuel => "let rec omega = S omega in omega".

(* Allocation *)
Extract Constant Allocation.regalloc => "Regalloc.regalloc".

(* Linearize *)
Extract Constant Linearize.enumerate_aux => "Linearizeaux.enumerate_aux".

(* SimplExpr *)
Extract Constant SimplExpr.first_unused_ident => "Camlcoq.first_unused_ident".
Extraction Inline SimplExpr.ret SimplExpr.error SimplExpr.bind SimplExpr.bind2.

(* Compopts *)
Extract Constant Compopts.optim_for_size =>
  "fun _ -> !Clflags.option_Osize".
Extract Constant Compopts.va_strict =>
  "fun _ -> false".
Extract Constant Compopts.propagate_float_constants =>
  "fun _ -> !Clflags.option_ffloatconstprop >= 1".
Extract Constant Compopts.generate_float_constants =>
  "fun _ -> !Clflags.option_ffloatconstprop >= 2".
Extract Constant Compopts.optim_tailcalls =>
  "fun _ -> !Clflags.option_ftailcalls".
Extract Constant Compopts.optim_constprop =>
  "fun _ -> !Clflags.option_fconstprop".
Extract Constant Compopts.optim_CSE =>
  "fun _ -> !Clflags.option_fcse".
Extract Constant Compopts.optim_redundancy =>
  "fun _ -> !Clflags.option_fredundancy".
Extract Constant Compopts.thumb =>
  "fun _ -> !Clflags.option_mthumb".

(* Compiler *)
Extract Constant Compiler.print_Clight => "PrintClight.print_if".
Extract Constant Compiler.print_Cminor => "PrintCminor.print_if".
Extract Constant Compiler.print_RTL => "PrintRTL.print_if".
Extract Constant Compiler.print_LTL => "PrintLTL.print_if".
Extract Constant Compiler.print_Mach => "PrintMach.print_if".
Extract Constant Compiler.print => "fun (f: 'a -> unit) (x: 'a) -> f x; x".
Extract Constant Compiler.time  => "Timing.time_coq".

(*Extraction Inline Compiler.apply_total Compiler.apply_partial.*)

(* Cabs *)
Extract Constant Cabs.cabsloc => 
"{ lineno : int;
   filename: string;
   byteno: int;
   ident : int;
 }".
Extract Constant Cabs.string => "String.t".
Extract Constant Cabs.char_code => "int64".

(* Int31 *)
Extract Inductive Int31.digits => "bool" [ "false" "true" ].
Extract Inductive Int31.int31 => "int" [ "Camlcoq.Int31.constr" ] "Camlcoq.Int31.destr".
Extract Constant Int31.twice => "Camlcoq.Int31.twice".
Extract Constant Int31.twice_plus_one => "Camlcoq.Int31.twice_plus_one".
Extract Constant Int31.compare31 => "Camlcoq.Int31.compare".
Extract Constant Int31.On => "0".
Extract Constant Int31.In => "1".

(* Processor-specific extraction directives *)

Load extractionMachdep.

(* Avoid name clashes *)
Extraction Blacklist List String Int ToString.

(* Cutting the dependency to R. *)
Extract Inlined Constant Fcore_defs.F2R => "fun _ -> assert false".
Extract Inlined Constant Fappli_IEEE.FF2R => "fun _ -> assert false".
Extract Inlined Constant Fappli_IEEE.B2R => "fun _ -> assert false".
Extract Inlined Constant Fappli_IEEE.round_mode => "fun _ -> assert false".
Extract Inlined Constant Fcalc_bracket.inbetween_loc => "fun _ -> assert false".

(* Needed in Coq 8.4 to avoid problems with Function definitions. *)
Set Extraction AccessOpaque.

(* Go! *)

Cd "extraction".

Separate Extraction
   Compiler.transf_c_program Compiler.transf_cminor_program
   Cexec.do_initial_state Cexec.do_step Cexec.at_final_state
   Ctypes.merge_attributes Ctypes.remove_attributes Ctypes.build_composite_env
   Initializers.transl_init Initializers.constval
   Csyntax.Eindex Csyntax.Epreincr
   Ctyping.typecheck_program
   Ctyping.epostincr Ctyping.epostdecr Ctyping.epreincr Ctyping.epredecr
   Clight.make_program
   Conventions1.dummy_int_reg Conventions1.dummy_float_reg
   RTL.instr_defs RTL.instr_uses
   Machregs.mregs_for_operation Machregs.mregs_for_builtin
   Machregs.two_address_op Machregs.is_stack_reg
   AST.signature_main
   Parser.translation_unit_file
   CsharpminorAnalysis.vanalysis
   Debugging.traceCmp
   UnLoad.doit.
