(* *********************************************************************)
(*                                                                     *)
(*              The Verasco static analyzer                            *)
(*                                                                     *)
(*          Jacques-Henri Jourdan, INRIA Paris-Rocquencourt            *)
(*                                                                     *)
(*  Copyright Institut National de Recherche en Informatique et en     *)
(*  Automatique and Université de Rennes 1.  All rights reserved.      *)
(*  This file is distributed under the terms of the GNU General        *)
(*  Public License as published by the Free Software Foundatio,        *)
(*  either version 2 of the License, or (at your option) any later     *)
(*  version.  This file is also distributed under the terms of the     *)
(*  INRIA Non-Commercial License Agreement.                            *)
(*                                                                     *)
(* *********************************************************************)

Require Import Utf8.
Require Import IdealExpr.
Require Import ZArith.
Require Import Coqlib.
Require Import ZIntervals FloatIntervals.
Require Import Integers.
Require Import AdomLib.
Require Import AbIdealEnv.
Require Import Maps ShareTree Util TreeAl.
Require Import Reals.
Require Import Fcore_Raux.
Require Import Fcore_defs.
Require Import Fcore_generic_fmt.
Require Import Fcore_float_prop.
Require Import Fappli_IEEE.
Require Import Fappli_IEEE_bits.
Require Import Fappli_IEEE_extra.
Require Import Fourier.
Require Import Floats.
Require Import FloatLib.
Require Import Psatz.
Require Import LinearQuery.
Require Import FactMsgHelpers.

Transparent Float.add Float.sub Float.mul Float.div Float.neg Float.zero.

Local Existing Instance itv_gamma_R.
Local Existing Instance itv_gamma_oR.

Module Tp := Tree_Properties(T).
Module Tsp := ShareTree_Properties(T).

Definition t := unit.
Instance gamma_t : gamma_op t (var -> ideal_num) := fun _ _ => True.

Lemma list_norepet_rev:
  ∀ l, list_norepet (A:=var) l -> list_norepet (rev l).
Proof.
  induction l. auto. intros. inv H.
  simpl. apply list_norepet_append_commut. constructor. 2:auto.
  rewrite <- in_rev. auto.
Qed.

Definition deletek {A} k (l:list (var * A)) :=
  List.filter (fun kv => negb (eq_dec (fst kv) k)) l.

Lemma deletek_eq:
  ∀ A k l, ~List.In k (List.map fst l) ->
         deletek (A:=A) k l = l.
Proof.
  induction l as [|[]]. auto.
  simpl. destruct T.elt_eq. now intuition.
  intros. simpl. rewrite IHl. now auto. now intuition.
Qed.

Lemma deletek_in:
  ∀ A k l k'v',
    List.In k'v' (deletek (A:=A) k l) <->
    List.In k'v' l /\ (fst k'v') <> k.
Proof.
  induction l as [|[]]. simpl. now intuition. destruct k'v'.
  simpl. destruct T.elt_eq; simpl; setoid_rewrite IHl; intuition congruence.
Qed.

Lemma deletek_norepet:
  ∀ A k l,
    list_norepet (List.map fst l) ->
    list_norepet (List.map fst (deletek (A:=A) k l)).
Proof.
  induction l as [|[]]. now constructor.
  intros. inv H. simpl. destruct T.elt_eq. now auto.
  simpl. constructor. 2:now auto.
  contradict H2. rewrite in_map_iff in *. setoid_rewrite deletek_in in H2.
  firstorder.
Qed.

Lemma eval_linear_expr_exten:
  ∀ (P1 P2:R->Prop) ρ l1 l2,
    list_norepet (List.map fst l1) ->
    list_norepet (List.map fst l2) ->
    (∀ (k:var) v, List.In (k, v) l1 <-> List.In (k, v) l2) ->
  (P1 ⊆ P2) ->
  ∀ x,
    fold_right
      (fun (kitvs:var * (itv*itv)) X =>
         let '(k, (itvα, itvk)) := kitvs in
         fun y =>
           match rρ ρ k with
           | None => False
           | Some kx => (∃ α, α ∈ γ itvα /\ X (y - α * kx)) /\
                        rρ ρ k ∈ γ itvk
           end)%R
      P1 l1 x ->
    fold_right
      (fun (kitvs:var * (itv*itv)) X =>
         let '(k, (itvα, itvk)) := kitvs in
         fun y =>
           match rρ ρ k with
           | None => False
           | Some kx => (∃ α, α ∈ γ itvα /\ X (y - α * kx)) /\
                        rρ ρ k ∈ γ itvk
           end)%R
      P2 l2 x.
Proof.
  induction l1 as [|[k [itvα itvk]]].
  - intros. destruct l2 as [|[k v]]. apply H2, H3.
    destruct (H1 k v) as [_ []]. left. auto.
  - intros. inv H. simpl in H3. destruct (rρ ρ k) as [kx|] eqn:EQkx. 2:contradiction.
    destruct H3 as [(αk & Hα & HαHx) Hty].
    apply IHl1 with (l2:=deletek k l2) in HαHx.
    + assert (List.In (k, (itvα, itvk)) l2). { apply H1. left. auto. } clear l1 IHl1 H1 H6 H7.
      revert x HαHx. induction l2 as [|[k' [itvα' itvk']] ?]; inv H0; intros. destruct H.
      simpl in HαHx. destruct @eq_dec. subst.
      * assert ((itvα, itvk)  = (itvα', itvk')).
        { destruct H. congruence. clear - H H4. destruct H4.
          apply in_map with (f:=fst) in H. auto. }
        inv H0. clear H. simpl. rewrite EQkx. split. eexists. split. eauto.
        rewrite deletek_eq in HαHx by auto. auto. auto.
      * simpl in *. destruct (rρ ρ k'). 2:now auto.
        destruct HαHx as [(α' & Hα' & Hα'kx) Hty']. split. eexists. split. eauto.
        apply IHl2. auto. destruct H. congruence. auto.
        replace (x - α' * r - αk * kx)%R with (x - αk * kx - α' * r)%R by ring. auto.
        auto.
    + now auto.
    + apply deletek_norepet. auto.
    + intros. rewrite deletek_in, <- H1. simpl.
      rewrite in_map_iff in H6. clear -H6. firstorder. congruence.
    + now auto.
Qed.

Lemma eval_T_linear_expr_exten:
  ∀ ρ i g1 g2,
    (∀ k, T.get k g1 = T.get k g2) ->
    ∀ x, eval_T_linear_expr ρ (i, g1) x <-> eval_T_linear_expr ρ (i, g2) x.
Proof.
  intros ρ i g1 g2 Hg1g2. unfold eval_T_linear_expr.
  split; apply eval_linear_expr_exten;
  try (intros; now auto);
  try (apply T.elements_keys_norepet); intros;
  split; intro;
  apply T.elements_correct; apply T.elements_complete in H; rewrite <- H; auto.
Qed.

Lemma eval_T_linear_expr_set_None:
  ∀ ρ i g (k:var) kx,
  ∀ itvs α, α ∈ γ (fst itvs) ->
    rρ ρ k = Some kx ->
    rρ ρ k ∈ γ (snd itvs) ->
    T.get k g = None ->
  ∀ x, eval_T_linear_expr ρ (i, g) x ->
       eval_T_linear_expr ρ (i, T.set k itvs g) (x+α*kx).
Proof.
  intros ρ i g k kx itv α Hα Hkx Hty Hkg x.
  unfold eval_T_linear_expr. intros Hx.
  eapply eval_linear_expr_exten with (l1:=(k, itv)::T.elements g).
  - simpl. constructor. rewrite in_map_iff. intro.
    destruct H as ([] & ? & ?). simpl in H. subst.
    apply T.elements_complete in H0. congruence. apply T.elements_keys_norepet.
  - apply T.elements_keys_norepet.
  - intros. simpl. split; intro. destruct H.
    + inv H. apply T.elements_correct. apply T.gss.
    + apply T.elements_correct. rewrite T.gso.
      apply T.elements_complete. auto.
      intro. subst. apply T.elements_complete in H. congruence.
    + apply T.elements_complete in H. rewrite T.gsspec in H.
      destruct T.elt_eq. inv H. now auto.
      right. apply T.elements_correct. auto.
  - eauto.
  - destruct itv. simpl. rewrite Hkx. split. eexists. split. eauto.
    replace (x + α * kx - α * kx)%R with x by ring. auto. rewrite <- Hkx. auto.
Qed.

Lemma eval_T_linear_expr_rem:
  ∀ ρ i g (k:var) kx, rρ ρ k = Some kx ->
  ∀ itvs, T.get k g = Some itvs ->
  ∀ x, eval_T_linear_expr ρ (i, g) x ->
  (∃ α, α ∈ γ (fst itvs) /\
        eval_T_linear_expr ρ (i, T.remove k g) (x-α*kx)) /\
  rρ ρ k ∈ γ (snd itvs).
Proof.
  intros ρ i g k kx Hkx itv Hitv x.
  unfold eval_T_linear_expr. intros Hx. unfold eval_linear_expr in Hx.
  eapply eval_linear_expr_exten with (l2:=(k, itv)::deletek k (T.elements g)) in Hx.
  - simpl in Hx. rewrite Hkx in Hx. destruct itv.
    destruct Hx as [(α & Hα & Hxα) Hty]. split. exists α. split. now eauto.
    eapply eval_linear_expr_exten, Hxα.
    apply deletek_norepet. apply T.elements_keys_norepet. apply T.elements_keys_norepet.
    simpl. intros. rewrite deletek_in. simpl.
    split; intros.
    apply T.elements_correct. rewrite T.gro by intuition. apply T.elements_complete. apply H.
    apply T.elements_complete in H. rewrite T.grspec in H.
    destruct T.elt_eq. discriminate. split; [|auto]. apply T.elements_correct, H. eauto.
    rewrite Hkx. auto.
  - apply T.elements_keys_norepet.
  - simpl. constructor.
    rewrite in_map_iff. setoid_rewrite deletek_in. clear. now firstorder.
    apply deletek_norepet. apply T.elements_keys_norepet.
  - intros. simpl. rewrite deletek_in. simpl. split; intro.
    destruct (T.elt_eq k k0). left. apply T.elements_complete in H.
    subst k. rewrite H in Hitv. inv Hitv. auto. auto.
    destruct H as [|[]]. inv H. apply T.elements_correct. auto. auto.
  - auto.
Qed.

Lemma eval_T_linear_expr_set_Some:
  ∀ ρ i g (k:var) kx, rρ ρ k = Some kx ->
  ∀ itvs, T.get k g = Some itvs ->
  ∀ x, eval_T_linear_expr ρ (i, g) x ->
  ∃ α0, α0 ∈ γ (fst itvs) /\
        ∀ itv1 itv1' α1, α1 ∈ γ itv1 -> rρ ρ k ∈ γ itv1' ->
        eval_T_linear_expr ρ (i, T.set k (itv1, itv1') g) (x+(α1-α0)*kx).
Proof.
  intros ρ i g k kx Hkx itv0 Hitv0 x Hx.
  eapply eval_T_linear_expr_rem in Hx; eauto.
  destruct Hx as [(α & Hα & Hxα) Hty]. exists α. split. auto. intros itv1 ty α1 Hty' Hα1.
  apply eval_T_linear_expr_exten with (g1 := T.set k (itv1, ty) (T.remove k g)).
  intros. rewrite !T.gsspec, T.grspec. destruct T.elt_eq; auto.
  replace (x + (α1 - α) * kx)%R with (x - α * kx + α1 * kx)%R by ring.
  apply eval_T_linear_expr_set_None; auto using T.grs.
Qed.

Lemma eval_linear_expr_change_cst:
  ∀ ρ i0 g,
  ∀ x, eval_linear_expr ρ (i0, g) x ->
  ∃ α0, α0 ∈ γ i0 /\
        ∀ i1 α1, α1 ∈ γ i1 ->
        eval_linear_expr ρ (i1, g) (x+α1-α0).
Proof.
  intros ρ i0 g. induction g as [|(? & ? & ?)]; simpl.
  intros. eexists. split. apply H. intros. replace (x + α1 - x)%R with α1 by ring. auto.
  unfold eval_linear_expr. simpl. destruct (rρ ρ v). 2:contradiction.
  intros x [(α & Hα & Hxα) Hty]. apply IHg in Hxα. clear IHg.
  destruct Hxα as (α' & Hα' & Hxα').
  eexists. split. eauto. split. eexists. split. eauto.
  replace (x + α1 - α' - α * r)%R with (x - α * r + α1 - α')%R by ring.
  apply Hxα'. auto. auto.
Qed.

Definition add_T_linear_expr (e1 e2:T_linear_expr) : T_linear_expr :=
  (add_itv (fst e1) (fst e2),
   T.combine
     (fun ity1 ity2 =>
        match ity1, ity2 with
        | None, None => None
        | Some i, None => Some i
        | None, Some i => Some i
        | Some (i1, ty), Some (i2, _) => Some (add_itv i1 i2, ty)
        end)
     (snd e1) (snd e2)).

Definition add_T_linear_expr_alt (e1 e2:T_linear_expr) : T_linear_expr :=
  (add_itv (fst e1) (fst e2),
   T.fold
     (fun (g:T.t _) (k:var) ity1 =>
        match T.get k g with
        | None => T.set k ity1 g
        | Some (i2, _) => T.set k (add_itv (fst ity1) i2, snd ity1) g
        end) (snd e1) (snd e2)).

Lemma add_T_linear_expr_alt_eq:
  ∀ e1 e2 k,
    T.get k (snd (add_T_linear_expr_alt e1 e2)) =
    T.get k (snd (add_T_linear_expr e1 e2)).
Proof.
  intros. unfold add_T_linear_expr, add_T_linear_expr_alt. simpl.
  rewrite T.gcombine, T.fold_spec, <- fold_left_rev_right by reflexivity.
  pose proof @T.elements_correct _ (snd e1) k.
  setoid_rewrite in_rev in H.
  pose proof @T.elements_keys_norepet _ (snd e1). apply list_norepet_rev in H0.
  rewrite <- map_rev in H0.
  pose proof @T.elements_complete _ (snd e1) k.
  setoid_rewrite in_rev in H1.
  revert H H1. generalize (T.get k (snd e1)).
  induction (rev (T.elements (snd e1))).
  simpl; intros. destruct o. edestruct H. auto. destruct T.get; now auto.
  inv H0. specialize (IHl H3). simpl. intros.
  destruct a as [k' [i ty]]. simpl. change T.elt with var in *.
  destruct (T.get k'
             (fold_right
                (λ (y:var * (itv*itv)) (x:T.t _), match T.get (fst y) x with
                        | Some (i2, _) =>
                          T.set (fst y) (add_itv (fst (snd y)) i2, snd (snd y)) x
                        | None => T.set (fst y) (snd y) x
                        end) (snd e2) l))
  as [[]|] eqn:EQ; rewrite T.gsspec; destruct (T.elt_eq k k').
  - subst. rewrite (H1 (i, ty)) by auto.
    rewrite (IHl None) in EQ. destruct T.get as [[]|]; congruence. discriminate.
    intros. destruct H2. apply in_map with (f:=fst) in H0. now auto.
  - apply IHl.
    intros. edestruct H. eauto. congruence. auto. auto.
  - subst. rewrite (H1 (i, ty)) by auto.
    rewrite (IHl None) in EQ. destruct T.get; congruence. discriminate.
    intros. destruct H2. apply in_map with (f:=fst) in H0. now auto.
  - apply IHl.
    intros. edestruct H. eauto. congruence. auto. auto.
Qed.

Lemma add_T_linear_expr_alt_correct:
  ∀ ρ e1 e2 x1 x2,
    eval_T_linear_expr ρ e1 x1 ->
    eval_T_linear_expr ρ e2 x2 ->
    eval_T_linear_expr ρ (add_T_linear_expr_alt e1 e2) (x1+x2).
Proof.
  unfold eval_T_linear_expr at 1, add_T_linear_expr_alt. intros ρ e1 e2 x1 x2.
  rewrite !T.fold_spec. revert e2 x2 x1.
  induction (T.elements (snd e1)) as [|(k & itvα & itvk)]; simpl; intros.
  - destruct e2. unfold eval_T_linear_expr in H0. apply eval_linear_expr_change_cst in H0.
    destruct H0 as (α0 & Hα0 & Hα0x2).
    replace (x1+x2)%R with (x2 + (x1+α0) - α0)%R by ring.
    apply Hα0x2. apply add_itv_correct; auto.
  - unfold eval_linear_expr in H. simpl in H.
    destruct (rρ ρ k) eqn:EQrρ. 2:contradiction.
    destruct H as [(α & Hα & Hαx2) Hty].
    replace (x1+x2)%R with (x1-α*r+(x2+α*r))%R by ring.
    eapply (IHl (_, _)). eauto. destruct e2. simpl.
    destruct (T.get k t0) as [[]|] eqn:EQ.
    + eapply eval_T_linear_expr_set_Some in EQ. 2:eauto. 2:eauto.
      destruct EQ as (α0 & Hα0 & Hα0x2).
      replace (x2 + α*r)%R with (x2 + (α + α0 - α0) * r)%R by ring.
      apply Hα0x2. apply add_itv_correct; auto.
      rewrite EQrρ. auto.
    + apply eval_T_linear_expr_set_None; auto. rewrite EQrρ. auto.
Qed.

Lemma add_T_linear_expr_correct:
  ∀ ρ e1 e2 x1 x2,
    eval_T_linear_expr ρ e1 x1 ->
    eval_T_linear_expr ρ e2 x2 ->
    eval_T_linear_expr ρ (add_T_linear_expr e1 e2) (x1+x2).
Proof.
  intros. eapply eval_T_linear_expr_exten.
  intro. symmetry. apply add_T_linear_expr_alt_eq.
  apply add_T_linear_expr_alt_correct; auto.
Qed.

Definition opp_T_linear_expr (e:T_linear_expr) : T_linear_expr :=
  (opp_itv (fst e), T.map1 (fun ity => (opp_itv (fst ity), snd ity)) (snd e)).

Lemma opp_T_linear_expr_correct:
  ∀ ρ e x,
    eval_T_linear_expr ρ e x ->
    eval_T_linear_expr ρ (opp_T_linear_expr e) (-x).
Proof.
  unfold eval_T_linear_expr. intros ρ e x Hx.
  eapply eval_linear_expr_exten
  with (l1 := List.map (fun ki => (fst ki, (opp_itv (fst (snd ki)), snd (snd ki)))) (T.elements (snd e))).
  - rewrite map_map. apply T.elements_keys_norepet.
  - apply T.elements_keys_norepet.
  - intros. unfold opp_T_linear_expr. simpl.
    pose proof (fun A g k v => conj (@T.elements_complete A g k v)
                                    (@T.elements_correct _ g k v):(_ <-> _)).
    rewrite in_map_iff. split; intros.
    + apply T.elements_correct. rewrite T.gmap1.
      destruct H0 as ([] & ? & ?). inv H0.
      apply T.elements_complete in H1. rewrite H1. auto.
    + apply T.elements_complete in H0. rewrite T.gmap1 in H0.
      destruct (T.get k (snd e)) as [[]|] eqn:EQ. 2:discriminate. inv H0.
      eexists (_, (_, _)). split. reflexivity. apply T.elements_correct. auto.
  - eauto.
  - revert x Hx. induction (T.elements (snd e)) as [|[k [itvα itvk]]]; simpl; intros.
    apply opp_itv_correct, Hx. unfold eval_linear_expr in *. simpl in Hx.
    destruct (rρ ρ k). 2:now auto.
    destruct Hx as [(α & Hα & Hxα) Hty].
    split. eexists. split. apply opp_itv_correct. apply Hα.
    replace (-x - - α*r)%R with (-(x-α*r))%R by ring. now auto. auto.
Qed.

Definition mult_T_linear_expr (i:itv) (e:T_linear_expr) : T_linear_expr :=
  (mult_itv i (fst e), T.map1 (fun itvs => (mult_itv i (fst itvs), snd itvs)) (snd e)).

Lemma mult_T_linear_expr_correct:
  ∀ ρ e x i y,
    eval_T_linear_expr ρ e x ->
    y ∈ γ i ->
    eval_T_linear_expr ρ (mult_T_linear_expr i e) (y * x).
Proof.
  unfold eval_T_linear_expr. intros ρ e x i y Hx Hy.
  eapply eval_linear_expr_exten
  with (l1 := List.map (fun ki => (fst ki, (mult_itv i (fst (snd ki)), snd (snd ki))))
                       (T.elements (snd e))).
  - rewrite map_map. apply T.elements_keys_norepet.
  - apply T.elements_keys_norepet.
  - intros. unfold mult_T_linear_expr. simpl.
    pose proof (fun A g k v => conj (@T.elements_complete A g k v)
                                    (@T.elements_correct _ g k v):(_ <-> _)).
    rewrite in_map_iff. setoid_rewrite H. rewrite T.gmap1. split.
    intros ([] & ? & ?). inv H0. erewrite T.elements_complete by apply H1. eauto.
    intros. destruct (T.get k (snd e)) as [[]|] eqn:EQ; inv H0.
    eexists (_, (_, _)). split. 2:apply H, EQ. auto.
  - eauto.
  - revert x Hx. induction (T.elements (snd e)) as [|[k [itvα itvk]]]; simpl; intros.
    apply mult_itv_correct, Hx. auto.
    unfold eval_linear_expr in Hx. simpl in Hx. destruct (rρ ρ k). 2:now auto.
    destruct Hx as [(α & Hα & Hxα) HH].
    split. eexists. split. apply mult_itv_correct. eauto. apply Hα.
    replace (y * x - y*α*r)%R with (y*(x-α*r))%R by ring. now auto. auto.
Qed.

Definition inv_itv (i:itv) : itv+⊤ :=
  let 'ITV a b := i in
  if Fleb (B754_finite 53 1024 false 1 (-1074) eq_refl) a ||
     Fleb b (B754_finite 53 1024 true 1 (-1074) eq_refl) then
    Just (ITV (Bdiv 53 1024 eq_refl eq_refl Float.binop_pl mode_DN one b)
              (Bdiv 53 1024 eq_refl eq_refl Float.binop_pl mode_UP one a))
  else All.

Lemma inv_itv_correct:
  ∀ i x, x ∈ γ i -> (/x)%R ∈ γ (inv_itv i).
Proof.
  intros. destruct i. unfold inv_itv.
  destruct H.
  destruct (Fleb (B754_finite 53 1024 false 1 (-1074) eq_refl) f) eqn:EQ1.
  2:destruct (Fleb f0 (B754_finite 53 1024 true 1 (-1074) eq_refl)) eqn:EQ2.
  - pose proof Bdiv_correct 53 1024 eq_refl eq_refl Float.binop_pl mode_UP one f.
    pose proof Bdiv_correct 53 1024 eq_refl eq_refl Float.binop_pl mode_DN one f0.
    destruct f as [|[]| |[]]; try discriminate.
    assert (B2R 53 1024 (B754_finite 53 1024 false m e e0) ≠ 0%R).
    { intro. apply F2R_eq_0_reg in H3. discriminate. }
    specialize (H1 H3). clear H3.
    unfold RF_le, Float.neg in H0. rewrite B2R_Bopp, is_finite_Bopp in H0.
    unfold is_finite in H0. apply Ropp_le_cancel in H0.
    assert (0 < B2R 53 1024 (B754_finite 53 1024 false m e e0))%R.
    { apply F2R_gt_0_compat, eq_refl. }
    split; unfold RF_le, Float.neg.
    + clear H2. Rlt_bool_case H1.
      * destruct H1 as (-> & -> & _).
        eapply Rle_trans. apply Rinv_le, H0. now auto.
        eapply Rle_trans. apply round_UP_pt. apply (fexp_correct 53 1024), eq_refl.
        right. f_equal. compute. Psatz.lra.
      * destruct (Bdiv 53 1024 eq_refl eq_refl Float.binop_pl mode_UP one) as [| |? []|];
        inv H1. reflexivity.
    + unfold RF_le in H.
      destruct (is_finite 53 1024 f0) eqn:FINf0.
      2:subst f0; simpl; rewrite <- Ropp_0;
      eapply Ropp_le_contravar, Rlt_le, Rinv_0_lt_compat, Rlt_le_trans, H0; now auto.
      assert (B2R _ _ f0 <> 0)%R by Psatz.lra.
      rewrite is_finite_Bopp, B2R_Bopp.
      specialize (H2 H4). Rlt_bool_case H2.
      * destruct H2 as (-> & -> & _).
        eapply Ropp_le_contravar, Rle_trans, Rinv_le, H. 2:Psatz.lra.
        eapply Rle_trans. 2:apply round_DN_pt. 2:apply (fexp_correct 53 1024), eq_refl.
        right. f_equal. compute. Psatz.lra.
      * destruct f0; try (exfalso; apply H4, eq_refl).
        assert (b = false).
        { clear - H H0 H3. destruct b. 2:reflexivity.
          assert (0 < B2R 53 1024 (B754_finite 53 1024 true m0 e1 e2))%R by Psatz.lra.
          apply F2R_gt_0_reg in H1. discriminate. }
        subst b.
        destruct (Bdiv 53 1024 eq_refl eq_refl Float.binop_pl mode_DN one) as [| |? []|]; inv H2.
        apply Ropp_le_contravar. rewrite Rabs_right in H5.
        eapply Rle_trans. eapply Rle_trans, H5. compute; Psatz.lra.
        eapply Rle_trans. 2:apply Rinv_le, H. 2:Psatz.lra.
        simpl. unfold F2R at 1, Rdiv. simpl.
        rewrite Rinv_r, Rmult_1_l by Psatz.lra.
        apply round_DN_pt, (fexp_correct 53 1024), eq_refl.
        apply Rle_ge, round_ge_generic.
        apply fexp_correct, eq_refl. apply valid_rnd_round_mode. apply generic_format_0.
        apply Rle_mult_inv_pos. apply F2R_ge_0_compat. discriminate.
        apply F2R_gt_0_compat. reflexivity.
  - clear EQ1.
    pose proof Bdiv_correct 53 1024 eq_refl eq_refl Float.binop_pl mode_UP one f.
    pose proof Bdiv_correct 53 1024 eq_refl eq_refl Float.binop_pl mode_DN one f0.
    destruct f0 as [|[]| |[]]; try discriminate.
    assert (B2R 53 1024 (B754_finite 53 1024 true m e e0) ≠ 0%R).
    { intro. apply F2R_eq_0_reg in H3. discriminate. }
    specialize (H2 H3). clear H3.
    unfold RF_le, Float.neg in H, H0. rewrite B2R_Bopp, is_finite_Bopp in H0.
    unfold is_finite in H.
    assert (B2R 53 1024 (B754_finite 53 1024 true m e e0) < 0)%R.
    { apply F2R_lt_0_compat, eq_refl. }
    split; unfold RF_le, Float.neg.
    + destruct (is_finite 53 1024 f) eqn:FINf.
      2:replace f with (B754_infinity 53 1024 true)
        by (destruct f as [|[]| |]; inv H0; reflexivity);
        simpl; eapply Ropp_le_cancel; rewrite Ropp_inv_permute, Ropp_0 by Psatz.lra;
        apply Rlt_le, Rinv_0_lt_compat; Psatz.lra.
      assert (B2R _ _ f <> 0)%R by Psatz.lra.
      specialize (H1 H4). Rlt_bool_case H1.
      * destruct H1 as (-> & -> & _).
        eapply Rle_trans.
        apply Ropp_le_cancel. rewrite (Ropp_inv_permute x), Ropp_inv_permute.
        apply Rinv_le, H0. Psatz.lra. Psatz.lra. Psatz.lra.
        simpl. unfold Rdiv, F2R. rewrite Rinv_r, Rmult_1_l.
        apply round_UP_pt. apply (fexp_correct 53 1024), eq_refl. simpl. Psatz.lra.
      * destruct f; try (exfalso; apply H4, eq_refl).
        assert (b = true).
        { clear - H H0 H3. destruct b. reflexivity.
          assert (B2R 53 1024 (B754_finite 53 1024 false m0 e1 e2) < 0)%R by Psatz.lra.
          apply F2R_lt_0_reg in H1. discriminate. }
        subst b.
        destruct (Bdiv 53 1024 eq_refl eq_refl Float.binop_pl mode_UP one) as [| |? []|]; inv H1.
        rewrite Rabs_left1 in H5.
        eapply Rle_trans. 2:eapply Rle_trans. 2:apply Ropp_le_contravar, H5.
        2:compute; Psatz.lra.
        eapply Rle_trans. apply Ropp_le_cancel.
        rewrite Ropp_inv_permute, (Ropp_inv_permute x). apply Rinv_le, H0.
        Psatz.lra. Psatz.lra. Psatz.lra.
        rewrite Ropp_involutive.
        simpl. unfold F2R at 2, Rdiv. simpl.
        rewrite Rinv_r, Rmult_1_l by Psatz.lra.
        apply round_UP_pt, (fexp_correct 53 1024), eq_refl.
        apply round_le_generic.
        apply fexp_correct, eq_refl. apply valid_rnd_round_mode. apply generic_format_0.
        simpl. unfold Rdiv, F2R at 1. rewrite Rinv_r, Rmult_1_l.
        apply Ropp_le_cancel. rewrite Ropp_0, Ropp_inv_permute.
        apply Rlt_le, Rinv_0_lt_compat, Ropp_0_gt_lt_contravar, F2R_lt_0_compat, eq_refl.
        intro. apply F2R_eq_0_reg in H1. discriminate. simpl. Psatz.lra.
    + clear H1. Rlt_bool_case H2.
      * rewrite B2R_Bopp, is_finite_Bopp. destruct H2 as (-> & -> & _).
        rewrite Ropp_inv_permute, <- round_UP_opp by Psatz.lra.
        eapply Rle_trans. apply Rinv_le, Ropp_le_contravar, H.
        rewrite <- Ropp_0. apply Ropp_lt_contravar, F2R_lt_0_compat. reflexivity.
        eapply Rle_trans. apply round_UP_pt. apply (fexp_correct 53 1024), eq_refl.
        right. f_equal. rewrite <- Ropp_inv_permute. compute. Psatz.lra.
        intro. apply F2R_eq_0_reg in H2. discriminate.
      * destruct (Bdiv 53 1024 eq_refl eq_refl Float.binop_pl mode_DN one) as [| |? []|];
        inv H2. reflexivity.
  - exact I.
Qed.

Definition join_itv (i1 i2:itv) : itv :=
  match i1, i2 with
  | ITV m1 M1, ITV m2 M2 =>
    ITV (if Fleb m1 m2 || is_nan _ _ m2 then m1 else m2)
        (if Fleb M1 M2 || is_nan _ _ M1 then M2 else M1)
  end.

Lemma join_itv_correct:
  ∀ i1 i2 (x:R), x ∈ (γ i1 ∪ γ i2) -> x ∈ γ (join_itv i1 i2).
Proof.
  intros. destruct i1, i2, H as [[]|[]]; split.
  - destruct (Fleb f0 f2) eqn:FLEB.
    2:destruct f0; try discriminate; now auto.
    simpl. unfold RF_le in H |- *.
    destruct (is_finite 53 1024 f0) eqn:FINf0, (is_finite 53 1024 f2) eqn:FINf2.
    + rewrite Fleb_Rle in FLEB by auto. Rle_bool_case FLEB. Psatz.lra. discriminate.
    + destruct f2 as [|[]| |], f0 as [| | |[]]; try discriminate; auto.
    + subst f0. destruct f2; discriminate.
    + subst f0. destruct f2 as [|[]| |]; try discriminate; auto.
  - unfold RF_le, Float.neg in *. rewrite is_finite_Bopp, B2R_Bopp in *.
    destruct (Fleb f f1) eqn:FLEB.
    destruct f; try discriminate; now auto.
    destruct (is_finite 53 1024 f) eqn:FINf, (is_finite 53 1024 f1) eqn:FINf1; simpl.
    + rewrite is_finite_not_is_nan, FINf1 by auto.
      rewrite Fleb_Rle in FLEB by auto. Rle_bool_case FLEB. discriminate. Psatz.lra.
    + destruct f1 as [|[]| |], f as [| | |[]]; try discriminate; simpl; auto.
    + destruct f as [|[]| |]; inv H0. rewrite is_finite_not_is_nan, FINf1 by auto.
      destruct f1; discriminate.
    + destruct f as [|[]| |], f1 as [|[]| |]; try discriminate. reflexivity.
  - destruct (Fleb f0 f2) eqn:FLEB. now auto.
    simpl. unfold RF_le in H |- *.
    destruct (is_finite 53 1024 f0) eqn:FINf0, (is_finite 53 1024 f2) eqn:FINf2.
    + rewrite is_finite_not_is_nan, FINf0 by auto.
      rewrite Fleb_Rle in FLEB by auto. Rle_bool_case FLEB. discriminate. Psatz.lra. 
    + subst f2. destruct f0 as [| | |[]]; discriminate.
    + destruct f0 as [|[]| |], f2; try discriminate; simpl; auto.
    + subst f2. destruct f0 as [|[]| |]; try discriminate; simpl; auto.
  - unfold RF_le, Float.neg in H0 |- *. rewrite is_finite_Bopp, B2R_Bopp in H0 |- *.
    destruct (Fleb f f1) eqn:FLEB. simpl. 2:destruct f1; try discriminate; now auto.
    destruct (is_finite 53 1024 f) eqn:FINf, (is_finite 53 1024 f1) eqn:FINf1; simpl.
    + rewrite Fleb_Rle in FLEB by auto. Rle_bool_case FLEB. Psatz.lra. discriminate.
    + destruct f1 as [|[]| |]; try discriminate. destruct f as [| | |[]]; try discriminate.
    + destruct f1, f as [|[]| |]; try discriminate; simpl; auto.
    + destruct f1 as [|[]| |]; try discriminate.
      destruct f as [|[]| |]; try discriminate; now auto.
Qed.

Definition double_mult_eps_itv : itv :=
  ITV (B754_finite 53 1024 false 9007199254740991 (-53) eq_refl)
      (B754_finite 53 1024 false 4503599627370497 (-52) eq_refl).

Definition double_add_eps : float :=
  B754_finite 53 1024 false 1 (-1074) eq_refl.

Definition double_add_eps_itv : itv :=
  ITV (Bopp _ _ Float.neg_pl double_add_eps) double_add_eps.

Lemma double_round_NE_bounds:
  ∀ x,
    let r := round radix2 (Fcore_FLT.FLT_exp (3 - 1024 - 53) 53) (round_mode mode_NE) x in
    (∃ α, α ∈ γ double_mult_eps_itv /\ r = x*α)%R \/
    (r-x)%R ∈ γ double_add_eps_itv.
Proof.
  intros.
  destruct (Req_dec x 0).
  - subst r x. rewrite round_0 by apply valid_rnd_round_mode.
    left. exists 1%R. compute; Psatz.lra.
  - pose proof @Fcore_ulp.ulp_half_error radix2 (Fcore_FLT.FLT_exp (3 - 1024 - 53) 53)
         (fexp_correct 53 1024 eq_refl) (negb ∘ Zeven) x.
    change (Znearest (negb ∘ Zeven)) with (round_mode mode_NE) in H0. fold r in H0.
    revert H0. generalize r. clear r. intro r.
    unfold Fcore_ulp.ulp, canonic_exp, Fcore_FLT.FLT_exp.
    apply Z.max_case_strong.
    + left. exists (r/x)%R. split. 2:field; auto.
      replace (ln_beta radix2 x - 53) with (ln_beta radix2 x - 1 + - 52) in H1 by ring.
      rewrite bpow_plus in H1.
      assert (bpow radix2 (ln_beta radix2 x-1) <= Rabs x)%R.
      { apply Rnot_lt_le. intro. apply ln_beta_le_bpow in H2. 2:now auto. Psatz.lia. }
      change (bpow radix2 (-52)) with (/4503599627370496)%R in H1.
      assert (Rabs (r-x) <= /9007199254740992 * Rabs x)%R by Psatz.lra.
      clear -H3 H.
      assert (Rabs ((r-x)/x) <= / 9007199254740992)%R.
      { unfold Rdiv. rewrite Rabs_mult, Rabs_Rinv by auto.
        eapply Rmult_le_reg_r. apply (Rabs_pos_lt _ H).
        rewrite Rmult_assoc, Rinv_l. Psatz.lra. apply Rabs_no_R0, H. }
      apply Rabs_le_inv in H0. replace ((r-x)/x)%R with (r/x-1)%R  in H0 by (field; auto).
      split; unfold RF_le; simpl; unfold F2R, Z2R, bpow; simpl; Psatz.lra.
    + right. apply Rabs_le_inv in H1. destruct H1.
      split; eapply Rle_trans. apply H2. 2:apply Ropp_le_contravar, H1.
      compute. left. Psatz.lra. compute. left. Psatz.lra.
Qed.

Definition double_round_T_linear_expr (e:T_linear_expr) : T_linear_expr :=
  (join_itv (add_itv double_add_eps_itv (fst e))
            (mult_itv double_mult_eps_itv (fst e)),
   T.map1 (fun itvs => (mult_itv double_mult_eps_itv (fst itvs), snd itvs)) (snd e)).

Lemma double_round_T_linear_expr_correct:
  ∀ e ρ x, eval_T_linear_expr ρ e x ->
   let r := round radix2 (Fcore_FLT.FLT_exp (3 - 1024 - 53) 53) (round_mode mode_NE) x in
   eval_T_linear_expr ρ (double_round_T_linear_expr e) r.
Proof.
  intros.
  pose proof (double_round_NE_bounds x). fold r in H0. destruct H0 as [(α & Hα & Hxα)|].
  - pose proof (mult_T_linear_expr_correct ρ e x double_mult_eps_itv α H Hα).
    replace (α * x)%R with r in H0 by (rewrite Hxα; ring).
    unfold eval_T_linear_expr in *.
    eapply eval_linear_expr_exten, H0.
    apply T.elements_keys_norepet. apply T.elements_keys_norepet.
    split; auto. intros; apply join_itv_correct; auto.
  - revert H. destruct e.
    unfold double_round_T_linear_expr, eval_T_linear_expr, mult_T_linear_expr, fst, snd.
    intros.
    eapply eval_linear_expr_exten
    with (l1:=List.map (fun ki => (fst ki, (mult_itv double_mult_eps_itv (fst (snd ki)), snd (snd ki))))
                       (T.elements t0)).
    + rewrite map_map. apply T.elements_keys_norepet.
    + apply T.elements_keys_norepet.
    + intros. rewrite in_map_iff. split; intros.
      * apply T.elements_correct. rewrite T.gmap1.
        destruct H1 as ([] & A & B). apply pair_eq_inv in A. destruct A.
        simpl in H1. subst. unfold snd. apply T.elements_complete in B. rewrite B.
        auto.
      * apply T.elements_complete in H1. rewrite T.gmap1 in H1.
        destruct (T.get k t0) as [[]|] eqn:EQ. 2:discriminate.
        exists (k, (i0, i1)). split. f_equal. unfold option_map in H1.
        unfold snd, fst. congruence.
        apply T.elements_correct. auto.
    + eauto.
    + unfold eval_linear_expr in H. unfold fst, snd in *.
      revert H0 H. generalize r. clear r. revert x.
      induction (T.elements t0) as [|(? & ? & ?)].
      * unfold map, fold_right.
        intros. apply join_itv_correct. left.
        eapply add_itv_correct in H. 2:apply H0. replace (r-x+x)%R with r in H by ring. auto.
      * unfold map, fold_right. intros. destruct (rρ ρ e) eqn:EQ. 2:now auto.
        destruct H as [(? & ? & ?) ?]. split. 2:now auto.
        eexists. split. apply mult_itv_correct with (x:=1%R), H.
        compute; Psatz.lra. eapply IHl, H1.
        replace (r - 1 * x0 * r0 - (x - x0 * r0))%R with (r-x)%R by ring. auto.
Qed.

Definition intervalize_linear_expr (e:linear_expr) : itv :=
  List.fold_left
    (fun itv itvs =>
       let '(_, (itvα, itvk)) := itvs in
       add_itv (mult_itv itvα itvk) itv) (snd e) (fst e).

Lemma intervalize_linear_expr_correct :
  ∀ e ρ x,
    eval_linear_expr ρ e x ->
    x ∈ γ (intervalize_linear_expr e).
Proof.
  intros.
  unfold eval_linear_expr in H. unfold intervalize_linear_expr.
  revert x H. generalize (fst e). induction (snd e) as [|(k & itvα & itvk)]. now auto.
  intros. apply IHl.
  simpl in H. destruct (rρ ρ k) eqn:EQ. 2:contradiction.
  destruct H as [(α & Hα & Hαx) HH]. clear -Hα Hαx HH.
  revert x Hαx. induction l as [|(k' & itvα' & itvk')].
  - simpl. intros. replace x with (α*r+(x-α*r))%R by ring.
    apply add_itv_correct. apply mult_itv_correct; auto. auto.
  - simpl. destruct (rρ ρ k'). 2:now auto.
    intros x [(α' & Hα' & Hαx') HH']. split. 2:auto.
    eexists. split. eauto. apply IHl.
    replace (x - α' * r0 - α * r)%R with (x - α * r - α' * r0)%R by ring.
    auto.
Qed.

Definition intervalize_T_linear_expr (e:T_linear_expr) : itv :=
  T.fold1 (fun itv itvs => add_itv (mult_itv (fst itvs) (snd itvs)) itv) (snd e) (fst e).

Lemma intervalize_T_linear_expr_correct :
  ∀ e ρ x,
    eval_T_linear_expr ρ e x ->
    x ∈ γ (intervalize_T_linear_expr e).
Proof.
  intros.
  unfold eval_T_linear_expr in H. unfold intervalize_T_linear_expr.
  rewrite T.fold1_spec. change T.elt with var.
  apply intervalize_linear_expr_correct in H.
  unfold intervalize_linear_expr in H.
  match goal with
  | H : γ ?A x |- γ ?B x => assert (A = B)
  end. 2:congruence.
  simpl. clear H. rewrite <- !fold_left_rev_right.
  induction (rev (A:=(var*_)) (T.elements (snd e))) as [|(? & ? & ?)]. reflexivity.
  simpl. congruence.
Qed.

Definition double_round_over_T_linear_expr (e:T_linear_expr) : option T_linear_expr :=
  let le := double_round_T_linear_expr e in
  let 'ITV a b := intervalize_T_linear_expr le in
  if is_finite _ _ a && is_finite _ _ b then Some le else None.

Lemma double_round_over_T_linear_expr_correct:
  ∀ e ρ x, eval_T_linear_expr ρ e x ->
  let r := round radix2 (Fcore_FLT.FLT_exp (3 - 1024 - 53) 53) (round_mode mode_NE) x in
  ∀ e', double_round_over_T_linear_expr e = Some e' ->
  ∀ f sgn, (if Rlt_bool (Rabs r) (bpow radix2 1024) then
             B2R _ _ f = r /\ is_finite _ _ f = true
            else B2FF _ _ f = binary_overflow 53 1024 mode_NE sgn) ->
  ∃ y, r_of_inum (INf f) = Some y /\ eval_T_linear_expr ρ e' y.
Proof.
  unfold double_round_over_T_linear_expr. intros.
  apply double_round_T_linear_expr_correct in H.
  rewrite Rlt_bool_true in H1.
  - unfold r_of_inum. destruct H1 as (? & ->). eexists. split. now eauto.
    destruct intervalize_T_linear_expr, is_finite, is_finite; inv H0. congruence.
  - apply intervalize_T_linear_expr_correct in H.
    destruct intervalize_T_linear_expr, H. unfold RF_le in H, H2.
    unfold Float.neg in H2. rewrite is_finite_Bopp, B2R_Bopp in H2.
    destruct (is_finite _ _ f0), (is_finite _ _ f1); inv H0.
    apply Rabs_lt. split.
    + eapply Rlt_le_trans, Ropp_le_cancel, H2.
      apply (fun H => proj1 (Rabs_lt_inv _ _ H)), abs_B2R_lt_emax.
    + eapply Rle_lt_trans, Rabs_lt_inv, abs_B2R_lt_emax. apply H.
Qed.

Definition single_mult_eps_itv : itv :=
  ITV (B754_finite 53 1024 false 9007198717870087 (-53) eq_refl)
      (B754_finite 53 1024 false 4503599895805949 (-52) eq_refl).

Definition single_add_eps : float :=
  B754_finite 53 1024 false 4503599627370496 (-202) eq_refl.

Definition single_add_eps_itv : itv :=
  ITV (Bopp _ _ Float.neg_pl single_add_eps) single_add_eps.

Lemma single_round_NE_bounds:
  ∀ x,
    let r := round radix2 (Fcore_FLT.FLT_exp (3 - 128 - 24) 24) (round_mode mode_NE) x in
    (∃ α, α ∈ γ single_mult_eps_itv /\ r = x*α)%R \/
    (r-x)%R ∈ γ single_add_eps_itv.
Proof.
  intros.
  destruct (Req_dec x 0).
  - subst r x. rewrite round_0 by apply valid_rnd_round_mode.
    left. exists 1%R. compute; Psatz.lra.
  - pose proof @Fcore_ulp.ulp_half_error radix2 (Fcore_FLT.FLT_exp (3 - 128 - 24) 24)
         (fexp_correct 24 128 eq_refl) (negb ∘ Zeven) x.
    change (Znearest (negb ∘ Zeven)) with (round_mode mode_NE) in H0. fold r in H0.
    revert H0. unfold Fcore_ulp.ulp, canonic_exp, Fcore_FLT.FLT_exp.
    apply Z.max_case_strong.
    + intro. assert (r <> 0)%R.
      { intro. destruct (ln_beta radix2 x).
        eapply exp_small_round_0 in H1.
        2:apply fexp_correct, eq_refl. 2:apply valid_rnd_round_mode. 2:eauto.
        simpl in H0. unfold Fcore_FLT.FLT_exp in H1. Psatz.lia. }
      left. exists (r/x)%R. split. 2:field; auto.
      replace (ln_beta radix2 x - 24) with (ln_beta radix2 x - 1 + - 23) in H2 by ring.
      rewrite bpow_plus in H2.
      split.
      * unfold RF_le. simpl. unfold F2R. simpl.
        destruct (Rle_lt_dec (r/x)%R 1). Psatz.lra.
        rewrite <- Rabs_pos_eq in r0. 2:Psatz.lra.
        unfold Rdiv in r0. rewrite Rabs_mult, Rabs_Rinv in r0 by auto.
        eapply Rle_trans in H2. 2:apply Rabs_triang_inv2.
        apply Rmult_lt_compat_r with (r:=Rabs x) in r0.
        2:apply Rabs_pos_lt; now auto.
        rewrite Rmult_assoc, Rinv_l in r0 by (apply Rabs_no_R0; auto).
        rewrite Rabs_pos_eq in H2 by Psatz.lra.
        assert (bpow radix2 (ln_beta radix2 x - 1) <= Rabs x * / (1+/16777216))%R.
        { apply Rnot_lt_le. intro. clear H1. subst r.
          rewrite <- Fcore_rnd_ne.round_NE_abs in * by apply fexp_correct, eq_refl.
          rewrite <- ln_beta_abs in H0.
          remember (round radix2 (Fcore_FLT.FLT_exp (3 - 128 - 24) 24)
                          (Znearest (negb ∘ Zeven)) (Rabs x)).
          unfold round, Fcore_FLT.FLT_exp, scaled_mantissa, canonic_exp, F2R in Heqr.
          rewrite Z.max_l, ln_beta_abs in Heqr by auto. clear H0 H2. simpl in Heqr.
          unfold ln_beta_val in *. destruct (ln_beta radix2 x) as [log [Hlog _]]. auto.
          assert (-(/2) < Rabs x * bpow radix2 (- (log - 24)) - bpow radix2 23 < /2)%R.
          { split.
            - replace 23 with ((log-1)+(-(log-24))) by ring.
              rewrite bpow_plus, <- Rmult_minus_distr_r.
              eapply Rlt_le_trans with (r2:=0%R). Psatz.lra.
              apply Rmult_le_pos. Psatz.lra. apply bpow_ge_0.
            - apply Rmult_lt_reg_r with (r:=bpow radix2 (log-1)). apply bpow_gt_0.
              rewrite Rmult_minus_distr_r, Rmult_assoc, <- bpow_plus.
              replace (- (log - 24) + (log-1)) with 23 by ring.
              revert H3. generalize (bpow radix2 (log - 1)).
              compute. intros. Psatz.lra. }
          apply Rabs_lt in H0. rewrite <- Z2R_Zpower with (e:=23) in H0 by omega.
          erewrite Znearest_imp in Heqr by eauto.
          rewrite Z2R_Zpower, <- bpow_plus in Heqr by omega.
          replace (23+(log-24)) with (log-1) in Heqr by ring.
          Psatz.lra. }
        change (bpow radix2 (-23)) with (/8388608)%R in H2.
        assert (Rabs r <= Rabs x * (1+/16777216 * / (1+/16777216)))%R by Psatz.lra.
        apply Rmult_le_compat_r with (r:=(/Rabs x)%R) in H4.
        eapply Rle_trans. 2:eapply Rle_trans. 2:apply H4.
        rewrite <- Rabs_Rinv, <- Rabs_mult by auto. now apply Rle_abs.
        rewrite Rmult_comm, <- Rmult_assoc, Rinv_l by (apply Rabs_no_R0; auto).
        Psatz.lra.
        apply Rlt_le. apply Rinv_0_lt_compat, Rabs_pos_lt, H.
      * assert (bpow radix2 (ln_beta radix2 x-1) <= Rabs r)%R.
        { apply Rnot_lt_le. intro. apply ln_beta_le_bpow in H3. 2:auto.
          apply ln_beta_round_ge in H1. fold r in H1. Psatz.lia.
          apply fexp_correct, eq_refl. apply valid_rnd_round_mode. }
        change (bpow radix2 (-23)) with (/8388608)%R in H2.
        assert (Rabs (r-x) <= /16777216 * Rabs r)%R by Psatz.lra.
        clear - H H4 H1.
        assert (Rabs ((r-x)/r) <= / 16777216)%R.
        { unfold Rdiv. rewrite Rabs_mult, Rabs_Rinv by auto.
          apply Rmult_le_reg_r with (r:=Rabs r). apply (Rabs_pos_lt _ H1).
          rewrite Rmult_assoc, Rinv_l. Psatz.lra. apply Rabs_no_R0, H1. }
        replace ((r-x)/r)%R with (1-x/r)%R in H0 by (field; auto).
        apply Rabs_le_inv in H0.
        assert (1 - / 16777216 <= x / r <= 1 + / 16777216)%R by Psatz.lra. clear H0 H4.
        destruct H2. apply Rinv_le in H2. 2:Psatz.lra. clear H0.
        rewrite Rinv_Rdiv in H2 by auto. unfold RF_le. simpl. unfold F2R.
        simpl. Psatz.lra.
    + right. apply Rabs_le_inv in H1. destruct H1.
      split; eapply Rle_trans. apply H2. 2:apply Ropp_le_contravar, H1.
      compute. Psatz.lra. compute. Psatz.lra.
Qed.

Definition single_round_T_linear_expr (e:T_linear_expr) : T_linear_expr :=
  (join_itv (add_itv single_add_eps_itv (fst e))
            (mult_itv single_mult_eps_itv (fst e)),
   T.map1 (fun itvs => (mult_itv single_mult_eps_itv (fst itvs), snd itvs)) (snd e)).

Lemma single_round_T_linear_expr_correct:
  ∀ e ρ x, eval_T_linear_expr ρ e x ->
   let r := round radix2 (Fcore_FLT.FLT_exp (3 - 128 - 24) 24) (round_mode mode_NE) x in
   eval_T_linear_expr ρ (single_round_T_linear_expr e) r.
Proof.
  intros.
  pose proof (single_round_NE_bounds x). fold r in H0. destruct H0 as [(α & Hα & Hxα)|].
  - pose proof (mult_T_linear_expr_correct ρ e x single_mult_eps_itv α H Hα).
    replace (α * x)%R with r in H0 by (rewrite Hxα; ring).
    unfold eval_T_linear_expr in *.
    eapply eval_linear_expr_exten, H0.
    apply T.elements_keys_norepet.
    apply T.elements_keys_norepet.
    split; auto. intros; apply join_itv_correct; auto.
  - revert H. destruct e.
    unfold single_round_T_linear_expr, eval_T_linear_expr, mult_T_linear_expr, fst, snd.
    intros.
    eapply eval_linear_expr_exten
    with (l1:=List.map (fun ki => (fst ki, (mult_itv single_mult_eps_itv (fst (snd ki)), snd (snd ki))))
                       (T.elements t0)).
    + rewrite map_map. apply T.elements_keys_norepet.
    + apply T.elements_keys_norepet.
    + intros. rewrite in_map_iff. split; intros.
      * apply T.elements_correct. rewrite T.gmap1.
        destruct H1 as ([] & A & B). apply pair_eq_inv in A. destruct A.
        simpl in H1. subst. unfold snd. apply T.elements_complete in B. rewrite B.
        auto.
      * apply T.elements_complete in H1. rewrite T.gmap1 in H1.
        destruct (T.get k t0) as [[]|] eqn:EQ. 2:discriminate.
        exists (k, (i0, i1)). split. f_equal. unfold option_map in H1.
        unfold snd, fst. congruence.
        apply T.elements_correct. auto.
    + eauto.
    + unfold eval_linear_expr in H. unfold fst, snd in *.
      revert H0 H. generalize r. clear r. revert x.
      induction (T.elements t0) as [|(? & ? & ?)].
      * unfold map, fold_right.
        intros. apply join_itv_correct. left.
        eapply add_itv_correct in H. 2:apply H0. replace (r-x+x)%R with r in H by ring. auto.
      * unfold map, fold_right. intros. destruct (rρ ρ e) eqn:EQ. 2:now auto.
        destruct H as [(? & ? & ?) ?]. split. 2:now auto.
        eexists. split. apply mult_itv_correct with (x:=1%R), H.
        compute; Psatz.lra. eapply IHl, H1.
        replace (r - 1 * x0 * r0 - (x - x0 * r0))%R with (r-x)%R by ring. auto.
Qed.

Definition single_ub : float :=
  B754_finite 53 1024 false 9007199254740991 75 eq_refl.

Definition single_lb : float :=
  B754_finite 53 1024 true 9007199254740991 75 eq_refl.

Definition single_round_over_T_linear_expr (e:T_linear_expr) : option T_linear_expr :=
  let le := single_round_T_linear_expr e in
  let 'ITV a b := intervalize_T_linear_expr le in
  if Fleb single_lb a && Fleb b single_ub then Some le else None.

Lemma single_round_over_T_linear_expr_correct:
  ∀ e ρ x, eval_T_linear_expr ρ e x ->
  let r := round radix2 (Fcore_FLT.FLT_exp (3 - 128 - 24) 24) (round_mode mode_NE) x in
  ∀ e', single_round_over_T_linear_expr e = Some e' ->
  ∀ f sgn, (if Rlt_bool (Rabs r) (bpow radix2 128) then
             B2R _ _ f = r /\ is_finite _ _ f = true
            else B2FF _ _ f = binary_overflow 24 128 mode_NE sgn) ->
  ∃ y, r_of_inum (INf (Float.of_single f)) = Some y /\ eval_T_linear_expr ρ e' y.
Proof.
  unfold single_round_over_T_linear_expr. intros.
  apply single_round_T_linear_expr_correct in H.
  rewrite Rlt_bool_true in H1.
  - unfold r_of_inum. edestruct floatofsingle_exact.
    destruct H1. rewrite H2, H3, H1 by auto. eexists. split. now eauto.
    destruct intervalize_T_linear_expr, @Fleb, @Fleb; inv H0. congruence.
  - apply intervalize_T_linear_expr_correct in H.
    destruct intervalize_T_linear_expr, H. unfold RF_le in H, H2.
    unfold Float.neg in H2. rewrite is_finite_Bopp, B2R_Bopp in H2.
    destruct (Fleb single_lb f0) eqn:Ff0, (Fleb f1 single_ub) eqn:Ff1; inv H0.
    assert (is_finite _ _ f0 = true).
    { destruct f0 as [|[]| |]; try reflexivity; discriminate. }
    assert (is_finite _ _ f1 = true).
    { destruct f1 as [|[]| |]; try reflexivity; discriminate. }
    rewrite Fleb_Rle in Ff0 by auto. rewrite Fleb_Rle in Ff1 by auto.
    Rle_bool_case Ff0; Rle_bool_case Ff1; try discriminate.
    rewrite H0 in H2. rewrite H3 in H.
    apply Rabs_lt. split.
    + eapply Rlt_le_trans, Ropp_le_cancel, H2. eapply Rlt_le_trans, H4. compute. Psatz.lra.
    + eapply Rle_lt_trans. apply H. eapply Rle_lt_trans. apply H5. compute. Psatz.lra.
Qed.

Definition ZofB_T_linear_expr (e:T_linear_expr) : T_linear_expr :=
  (add_itv (fst e) (ITV (Float.neg one) one), snd e).

Lemma Telements_empty:
  T.elements (T.empty (itv*itv)) = List.nil.
Proof.
  pose proof @T.elements_complete _ (T.empty (itv * itv)).
  destruct T.elements as [|[]]. now auto.
  specialize (H _ _ (or_introl eq_refl)). rewrite T.gempty in H. discriminate.
Qed.

Lemma Telements_singleton:
  ∀ x y,
    T.elements (T.set x y (T.empty (itv*itv))) = (x,y)::List.nil.
Proof.
  intros.
  pose proof T.elements_complete (T.set x y (T.empty (itv*itv))).
  pose proof @T.elements_correct _ (T.set x y (T.empty (itv*itv))).
  pose proof T.elements_keys_norepet (T.set x y (T.empty (itv*itv))).
  destruct T.elements as [|[x1 y1][|[x2 y2]]].
  - destruct (H0 _ _ (T.gss _ _ _)).
  - destruct (H0 _ _ (T.gss _ _ _)) as [|[]]. f_equal. auto.
  - pose proof H _ _ (or_introl eq_refl).
    pose proof H _ _ (or_intror (or_introl eq_refl)).
    rewrite T.gsspec, T.gempty in H2, H3.
    destruct T.elt_eq, T.elt_eq; inv H2; inv H3.
    inv H1. destruct H4. left. auto.
Qed.

Lemma ZofB_T_linear_expr_correct:
  ∀ e ρ x, eval_T_linear_expr ρ e x ->
   eval_T_linear_expr ρ (ZofB_T_linear_expr e) (Z2R (Ztrunc x)).
Proof.
  intros. destruct e.
  replace (Z2R (Ztrunc x)) with (x+(Z2R (Ztrunc x)-x))%R by ring.
  eapply eval_T_linear_expr_exten, add_T_linear_expr_correct with (e1:=(_, _)) (e2:=(_, _)).
  - intros. rewrite T.gcombine by auto. simpl. rewrite T.gempty.
    destruct T.get as [[]|]; auto.
  - auto.
  - unfold eval_T_linear_expr. simpl. rewrite Telements_empty.
    split; unfold Ztrunc, RF_le; destruct (Rlt_bool_spec x 0); simpl;
    unfold F2R, Zceil, Zfloor; simpl;
    rewrite ?Z2R_opp, Z2R_minus, Rinv_r, Z2R_IZR  by Psatz.lra; simpl;
    destruct (archimed (-x)), (archimed x); Psatz.lra.
Qed.

Fixpoint linearize_expr (c:query_chan) (e:iexpr) : option T_linear_expr :=
  match e with
  | IEvar _ x =>
    match itv_of_itvabs (c.(get_itv) e) with
    | All => None
    | Just (ITV a b as i) =>
      if is_finite _ _ a && is_finite _ _ b then
        Some (ITV (B754_zero _ _ false) (B754_zero _ _ false),
              T.set x (ITV one one, i) (T.empty _))
      else None
    end
  | IEconst (INz z) => Some (itv_of_z z, T.empty _)
  | IEconst (INf f) =>
    if is_finite _ _ f then Some (ITV f f, T.empty _) else None
  | IEZitv itv => Some (itv_of_zitv itv, T.empty _)
  | IEunknown _ => None
  | IEunop (IOneg | IOnegf) e' =>
    do_opt le' <- linearize_expr c e';
    Some (opp_T_linear_expr le')
  (* | IEunop IOnot e' => TODO *)
  | IEunop IOsingleoffloat e' =>
    do_opt le' <- linearize_expr c e';
    single_round_over_T_linear_expr le'
  | IEunop IOfloatofz e' =>
    do_opt le' <- linearize_expr c e';
    double_round_over_T_linear_expr le'
  | IEunop IOsingleofz e' =>
    do_opt le' <- linearize_expr c e';
    single_round_over_T_linear_expr le'
  | IEunop IOzoffloat e' =>
    do_opt le' <- linearize_expr c e';
    Some (ZofB_T_linear_expr le')
  | IEbinop IOadd e1 e2 =>
    do_opt le1 <- linearize_expr c e1;
    do_opt le2 <- linearize_expr c e2;
    Some (add_T_linear_expr le1 le2)
  | IEbinop IOsub e1 e2 =>
    do_opt le1 <- linearize_expr c e1;
    do_opt le2 <- linearize_expr c e2;
    Some (add_T_linear_expr le1 (opp_T_linear_expr le2))
  | IEbinop IOmul e1 e2 =>
    (* TODO : better heuristic for choosing wich factor to intervalize. *)
    match e2 with
    | IEconst (INz z) =>
      do_opt le' <- linearize_expr c e1;
        Some (mult_T_linear_expr (itv_of_z z) le')
    | IEZitv i =>
      do_opt le' <- linearize_expr c e1;
      Some (mult_T_linear_expr (itv_of_zitv i) le')
    | _ =>
      match c.(get_itv) e1 with
      | NotBot (Just (IdealIntervals.Az i)) =>
        do_opt le' <- linearize_expr c e2;
        Some (mult_T_linear_expr (itv_of_zitv i) le')
      | _ => None
      end
    end
  | IEbinop IOdiv e1 e2 =>
    do_opt le' <- linearize_expr c e1;
    match c.(get_itv) e2 with
    | NotBot (Just (IdealIntervals.Az i)) =>
      match inv_itv (itv_of_zitv i) with
      | Just i' => Some (ZofB_T_linear_expr (mult_T_linear_expr i' le'))
      | All => None
      end
    | _ => None
    end
  (* | IEbinop (IOshl|IOshr) e1 e2 => TODO *)
  | IEbinop IOaddf e1 e2 =>
    do_opt le1 <- linearize_expr c e1;
    do_opt le2 <- linearize_expr c e2;
    double_round_over_T_linear_expr (add_T_linear_expr le1 le2)
  | IEbinop IOsubf e1 e2 =>
    do_opt le1 <- linearize_expr c e1;
    do_opt le2 <- linearize_expr c e2;
    double_round_over_T_linear_expr (add_T_linear_expr le1 (opp_T_linear_expr le2))
  | IEbinop IOmulf e1 e2 =>
    (* TODO : better heuristic for choosing wich factor to intervalize. *)
    match e2 with
    | IEconst (INf f) =>
      if is_finite _ _ f then
        do_opt le' <- linearize_expr c e1;
        double_round_over_T_linear_expr (mult_T_linear_expr (ITV f f) le')
      else
        None
    | _ =>
      match itv_of_itvabs (c.(get_itv) e1) with
      | Just (ITV a b as i) =>
        do_opt le' <- linearize_expr c e2;
        double_round_over_T_linear_expr (mult_T_linear_expr i le')
      | _ => None
      end
    end
  | IEbinop IOdivf e1 e2 =>
    do_opt le' <- linearize_expr c e1;
    match itv_of_itvabs (c.(get_itv) e2) with
    | Just (ITV a b as i) =>
      match inv_itv i with
      | Just i' => double_round_over_T_linear_expr (mult_T_linear_expr i' le')
      | All => None
      end
    | _ => None
    end
  | _ =>
    match itv_of_itvabs (c.(get_itv) e) with
    | Just i => Some (i, T.empty _)
    | _ => None
    end
  end.

Lemma linearize_expr_correct:
  ∀ c e le, linearize_expr c e = Some le ->
  ∀ ρ x, ρ ∈ γ c -> eval_iexpr ρ e x ->
  ∃ y, r_of_inum x = Some y /\ eval_T_linear_expr ρ le y.
Proof.
  induction e as [| | | |[]|[]]; simpl; intros le Hle ρ x Hρ Hx;
  assert (Hx':=Hx); inv Hx';
  (eapply get_itv_correct, itv_of_itvabs_correct in Hx; [|apply Hρ]).
  - (* IEvar *)
    apply get_itv_correct with (e:=IEvar i v) (a:=ρ v) in Hρ. 2:constructor; now auto.
    apply itv_of_itvabs_correct in Hρ. destruct itv_of_itvabs; inv Hle.
    unfold eval_T_linear_expr. destruct i0.
    destruct (is_finite _ _ f) eqn:FINf, (is_finite _ _ f0) eqn:FINf0; inv H0. simpl.
    rewrite Telements_singleton.
    unfold eval_linear_expr. simpl. unfold rρ. destruct r_of_inum. 2:now destruct Hρ.
    eexists. split. now eauto. split. 2:now auto.
    exists 1%R. unfold RF_le. compute; Psatz.lra.
  - (* IEconst *)
    unfold r_of_inum. destruct x; [destruct (is_finite 53 1024 f) eqn:FINf|]; inv Hle; eexists;
    (split; [now eauto|]); unfold eval_T_linear_expr, eval_linear_expr;
    simpl; rewrite Telements_empty; simpl.
    unfold RF_le, Float.neg. rewrite is_finite_Bopp, B2R_Bopp, FINf. Psatz.lra.
    apply itv_of_z_correct.
  - (* IEZitv *)
    inv Hle. eexists. split. now eauto.
    unfold eval_T_linear_expr. simpl. rewrite Telements_empty.
    apply itv_of_zitv_correct, H0.
  - (* IEunknown *)
    discriminate.
  - (* IOneg *)
    inv H3. destruct (linearize_expr c e); inv Hle.
    edestruct IHe as (y & Hxy & Hley); eauto. inv Hxy. clear IHe.
    eexists. split. now eauto. rewrite Z2R_opp. apply opp_T_linear_expr_correct, Hley.
  - (* IOnot *)
    destruct itv_of_itvabs; inv Hle. destruct r_of_inum. 2:contradiction.
    unfold eval_T_linear_expr. simpl. rewrite Telements_empty. eauto.
  - (* IOnegf *)
    inv H3. destruct (linearize_expr c e); inv Hle.
    edestruct IHe as (y & Hxy & Hley); eauto. clear IHe.
    unfold r_of_inum, Float.neg in *. rewrite is_finite_Bopp, B2R_Bopp in *.
    destruct (is_finite _ _ f). 2:discriminate.
    eexists. split. eauto. apply opp_T_linear_expr_correct. congruence.
  - (* IOabsf *)
    destruct itv_of_itvabs; inv Hle. destruct r_of_inum. 2:contradiction.
    unfold eval_T_linear_expr. simpl. rewrite Telements_empty. eauto.
  - (* IOsingleoffloat *)
    inv H3. destruct (linearize_expr c e); inv Hle.
    edestruct IHe as (y & Hxy & Hley); eauto. clear IHe.
    eapply single_round_over_T_linear_expr_correct in H0. now eauto. now eauto.
    simpl in Hxy. destruct (is_finite _ _ f) eqn:FINf; inv Hxy.
    pose proof (Bconv_correct 53 1024 24 128 eq_refl eq_refl Float.to_single_pl mode_NE f FINf).
    Rlt_bool_case H.
    destruct H as (? & ? & _); eauto. eauto.
  - (* IOfloatofz *)
    inv H3. destruct (linearize_expr c e); inv Hle.
    edestruct IHe as (y & Hxy & Hley); eauto. clear IHe.
    eapply double_round_over_T_linear_expr_correct in H0. now eauto. now eauto.
    inv Hxy. pose proof BofZ_correct 53 1024 eq_refl eq_refl i.
    Rlt_bool_case H. destruct H as (? & ? & _); eauto. eauto.
  - (* IOsingleofz *)
    inv H3. destruct (linearize_expr c e); inv Hle.
    edestruct IHe as (y & Hxy & Hley); eauto. clear IHe.
    eapply single_round_over_T_linear_expr_correct in H0. now eauto. now eauto. inv Hxy.
    pose proof (BofZ_correct 24 128 eq_refl eq_refl i).
    Rlt_bool_case H.
    destruct H as (? & ? & _); eauto. eauto.
  - (* IOzoffloat *)
    inv H3. destruct (linearize_expr c e); inv Hle.
    edestruct IHe as (y & Hxy & Hley); eauto. clear IHe.
    rewrite ZofB_correct in H. simpl in Hxy. destruct is_finite; inv Hxy. inv H.
    eapply ZofB_T_linear_expr_correct in Hley. simpl. eauto.
  - (* IOadd *)
    inv H5. destruct (linearize_expr c e1), (linearize_expr c e2); inv Hle.
    edestruct IHe1 as (y1 & Hxy1 & Hley1); eauto. clear IHe1.
    edestruct IHe2 as (y2 & Hxy2 & Hley2); eauto. clear IHe2.
    inv Hxy1. inv Hxy2. simpl. eexists. split. eauto.
    rewrite Z2R_plus. eapply add_T_linear_expr_correct; auto.
  - (* IOsub *)
    inv H5. destruct (linearize_expr c e1), (linearize_expr c e2); inv Hle.
    edestruct IHe1 as (y1 & Hxy1 & Hley1); eauto. clear IHe1.
    edestruct IHe2 as (y2 & Hxy2 & Hley2); eauto. clear IHe2.
    inv Hxy1. inv Hxy2. simpl. eexists. split. eauto.
    unfold Zminus. rewrite Z2R_plus, Z2R_opp. eapply add_T_linear_expr_correct. auto.
    apply opp_T_linear_expr_correct, Hley2.
  - (* IOmul *)
    inv H5. simpl. eexists. split. eauto. rewrite Z2R_mult.
    assert (match get_itv c e1 with
            | Bot => None
            | NotBot All => None
            | NotBot (Just (IdealIntervals.Az i0)) =>
                match linearize_expr c e2 with
                | Some le' => Some (mult_T_linear_expr (itv_of_zitv i0) le')
                | None => None
                end
            | NotBot (Just (IdealIntervals.Af _)) => None
            end = Some le ->
            eval_T_linear_expr ρ le (Z2R i1 * Z2R i2)).
    { clear Hle. intro. pose proof (get_itv_correct Hρ H2).
      destruct (get_itv c e1) as [|[|[]]]; try discriminate.
      destruct (linearize_expr c e2); inv H.
      edestruct IHe2 as (y2 & Hxy2 & Hley2); eauto. clear IHe2. inv Hxy2.
      apply mult_T_linear_expr_correct. auto.
      apply itv_of_zitv_correct. auto. }
    destruct e2; auto. destruct i; auto.
    rewrite Rmult_comm. destruct (linearize_expr c e1); inv Hle.
    edestruct IHe1 as (y1 & Hxy1 & Hley1); eauto. clear IHe1. inv Hxy1.
    apply mult_T_linear_expr_correct. auto. inv H4. apply itv_of_z_correct.
    rewrite Rmult_comm. destruct (linearize_expr c e1); inv Hle.
    edestruct IHe1 as (y1 & Hxy1 & Hley1); eauto. clear IHe1. inv Hxy1.
    apply mult_T_linear_expr_correct. auto. inv H4. apply itv_of_zitv_correct, H3.
  - (* IOdiv *)
    inv H5. destruct (linearize_expr c e1). 2:discriminate.
    pose proof (get_itv_correct Hρ H4).
    destruct (get_itv c e2) as [|[|[]]]; try discriminate.
    edestruct IHe1 as (y & Hxy & Hley); eauto. clear IHe2 IHe1.
    inv Hxy. simpl. eexists. split. eauto.
    apply itv_of_zitv_correct, inv_itv_correct in H0. destruct inv_itv; inv Hle.
    eapply mult_T_linear_expr_correct in H0. 2:apply Hley.
    apply ZofB_T_linear_expr_correct in H0.
    replace (i1 ÷ i2) with (Ztrunc (/ Z2R i2 * Z2R i1)). auto.
    rewrite Rmult_comm, Z.quot_div by auto. rewrite <- Zfloor_div by Psatz.lia.
    unfold Rdiv, Ztrunc, Zceil. destruct i1, i2; try congruence; simpl.
    + rewrite Rmult_0_l. apply (Ztrunc_Z2R 0).
    + rewrite Rmult_0_l. apply (Ztrunc_Z2R 0).
    + rewrite Rlt_bool_false. destruct Zfloor; auto.
      apply Rle_mult_inv_pos.
      apply (Z2R_le 0 (Z.pos _)). discriminate.
      apply (Z2R_lt 0 (Z.pos _)). reflexivity.
    + rewrite <- Ropp_inv_permute, Ropp_mult_distr_r_reverse, Rlt_bool_true,
                 Ropp_involutive.
      destruct Zfloor; auto.
      apply Ropp_lt_gt_0_contravar, Rlt_mult_inv_pos.
      apply (Z2R_lt 0 (Z.pos _)). reflexivity.
      apply (Z2R_lt 0 (Z.pos _)). reflexivity.
      intro. apply (eq_Z2R (Z.pos _) 0) in H1. discriminate.
    + rewrite Ropp_mult_distr_l_reverse, Rlt_bool_true, Ropp_involutive.
      destruct Zfloor; auto.
      apply Ropp_lt_gt_0_contravar, Rlt_mult_inv_pos.
      apply (Z2R_lt 0 (Z.pos _)). reflexivity.
      apply (Z2R_lt 0 (Z.pos _)). reflexivity.
    + rewrite Ropp_mult_distr_l_reverse, <- Ropp_inv_permute, Ropp_mult_distr_r_reverse,
              Ropp_involutive, Rlt_bool_false.
      destruct Zfloor; auto.
      apply Rle_mult_inv_pos.
      apply (Z2R_le 0 (Z.pos _)). discriminate.
      apply (Z2R_lt 0 (Z.pos _)). reflexivity.
      intro. apply (eq_Z2R (Z.pos _) 0) in H1. discriminate.
  - (* IOmod *)
    destruct itv_of_itvabs; inv Hle. destruct r_of_inum. 2:contradiction.
    unfold eval_T_linear_expr. simpl. rewrite Telements_empty. eauto.
  - (* IOand *)
    destruct itv_of_itvabs; inv Hle. destruct r_of_inum. 2:contradiction.
    unfold eval_T_linear_expr. simpl. rewrite Telements_empty. eauto.
  - (* IOor *)
    destruct itv_of_itvabs; inv Hle. destruct r_of_inum. 2:contradiction.
    unfold eval_T_linear_expr. simpl. rewrite Telements_empty. eauto.
  - (* IOxor *)
    destruct itv_of_itvabs; inv Hle. destruct r_of_inum. 2:contradiction.
    unfold eval_T_linear_expr. simpl. rewrite Telements_empty. eauto.
  - (* IOshl *)
    destruct itv_of_itvabs; inv Hle. destruct r_of_inum. 2:contradiction.
    unfold eval_T_linear_expr. simpl. rewrite Telements_empty. eauto.
  - (* IOshr *)
    destruct itv_of_itvabs; inv Hle. destruct r_of_inum. 2:contradiction.
    unfold eval_T_linear_expr. simpl. rewrite Telements_empty. eauto.
  - (* IOcmp *)
    destruct itv_of_itvabs; inv Hle. destruct r_of_inum. 2:contradiction.
    unfold eval_T_linear_expr. simpl. rewrite Telements_empty. eauto.
  - (* IOaddf *)
    inv H5. destruct (linearize_expr c e1), (linearize_expr c e2); inv Hle.
    edestruct IHe1 as (y1 & Hxy1 & Hley1); eauto. clear IHe1.
    edestruct IHe2 as (y2 & Hxy2 & Hley2); eauto. clear IHe2.
    eapply double_round_over_T_linear_expr_correct in H0.
    now eauto. apply add_T_linear_expr_correct; now eauto.
    unfold r_of_inum in Hxy1, Hxy2. unfold Float.add.
    destruct (is_finite _ _ f1) eqn:FINf1, (is_finite _ _ f2) eqn:FINf2;
    inv Hxy1; inv Hxy2.
    pose proof (Bplus_correct 53 1024 eq_refl eq_refl Float.binop_pl mode_NE
                              _ _ FINf1 FINf2).
    Rlt_bool_case H. destruct H as (? & ? & _). auto.
    destruct H. eauto.
  - (* IOsubf *)
    inv H5. destruct (linearize_expr c e1), (linearize_expr c e2); inv Hle.
    edestruct IHe1 as (y1 & Hxy1 & Hley1); eauto. clear IHe1.
    edestruct IHe2 as (y2 & Hxy2 & Hley2); eauto. clear IHe2.
    eapply double_round_over_T_linear_expr_correct in H0.
    now eauto. apply add_T_linear_expr_correct, opp_T_linear_expr_correct; now eauto.
    unfold r_of_inum in Hxy1, Hxy2. unfold Float.sub.
    destruct (is_finite _ _ f1) eqn:FINf1, (is_finite _ _ f2) eqn:FINf2;
    inv Hxy1; inv Hxy2. fold (B2R 53 1024 f1 - B2R 53 1024 f2)%R.
    pose proof (Bminus_correct 53 1024 eq_refl eq_refl Float.binop_pl mode_NE
                              _ _ FINf1 FINf2).
    Rlt_bool_case H. destruct H as (? & ? & _). auto.
    destruct H. eauto.
  - (* IOmulf *)
    inv H5. unfold Float.mul.
    assert (match itv_of_itvabs (get_itv c e1) with
            | All => None
            | Just (ITV a b as i1) =>
              match linearize_expr c e2 with
              | Some le' =>
                double_round_over_T_linear_expr (mult_T_linear_expr i1 le')
              | None => None
              end
            end = Some le ->
              ∃ y : R,
                r_of_inum
                  (INf (Bmult 53 1024 eq_refl eq_refl Float.binop_pl mode_NE f1 f2)) =
                Some y ∧ eval_T_linear_expr ρ le y).
    { intro. clear Hle IHe1. pose proof Hρ. eapply get_itv_correct in Hρ. 2:now eauto.
      apply itv_of_itvabs_correct in Hρ.
      destruct (itv_of_itvabs (get_itv c e1)) as [|[]]. discriminate.
      destruct (linearize_expr c e2). 2:discriminate.
      edestruct IHe2 as (y & Hxy & Hley); eauto. clear IHe2.
      simpl in Hρ. destruct (is_finite _ _ f1) eqn:FINf1. 2:contradiction.
      eapply double_round_over_T_linear_expr_correct in H. eauto.
      apply mult_T_linear_expr_correct; eauto.
      simpl in Hxy. destruct (is_finite _ _ f2) eqn:FINf2; inv Hxy.
      pose proof Bmult_correct 53 1024 eq_refl eq_refl Float.binop_pl mode_NE f1 f2.
      rewrite FINf2, FINf1 in H1.
      Rlt_bool_case H1. destruct H1 as (? & ? & _); now eauto. eauto. }
    destruct e2; eauto. clear H IHe2. inv H4.
    destruct (is_finite _ _ f2) eqn:FINf2. 2:discriminate.
    destruct (linearize_expr c e1). 2:discriminate.
    edestruct IHe1 as (y & Hxy & Hley); eauto. clear IHe1.
    eapply double_round_over_T_linear_expr_correct in Hle. eauto.
    eapply mult_T_linear_expr_correct. eauto.
    split; unfold RF_le, Float.neg; rewrite ?is_finite_Bopp, ?B2R_Bopp, FINf2; apply Rle_refl.
    rewrite Rmult_comm.
    unfold r_of_inum in Hxy. destruct (is_finite _ _ f1) eqn:FINf1; inv Hxy.
    pose proof Bmult_correct 53 1024 eq_refl eq_refl Float.binop_pl mode_NE f1 f2.
    rewrite FINf2, FINf1 in H.
    Rlt_bool_case H. destruct H as (? & ? & _); now eauto. eauto.
  - (* IOdivf *)
    inv H5. destruct (linearize_expr c e1); inv Hle.
    edestruct IHe1 as (y & Hxy & Hley); eauto. clear IHe1 IHe2.
    eapply get_itv_correct, itv_of_itvabs_correct in H4. 2:now eauto.
    destruct (itv_of_itvabs (get_itv c e2)) as [|[]]. discriminate.
    unfold r_of_inum in H4. destruct (is_finite _ _ f2) eqn:FINf2. 2:contradiction.
    assert (B2R 53 1024 f2 ≠ 0%R).
    { destruct f2; try discriminate.
      2:intro HH; apply F2R_eq_0_reg in HH; destruct b; discriminate.
      clear - H0 H4.
      unfold inv_itv in H0.
      destruct H4. unfold RF_le, Float.neg in H, H1. rewrite B2R_Bopp, is_finite_Bopp in H1.
      replace (Fleb (B754_finite 53 1024 false 1 (-1074) eq_refl) f) with false in H0.
      replace (Fleb f0 (B754_finite 53 1024 true 1 (-1074) eq_refl)) with false in H0.
      discriminate.
      - clear H0 H1. destruct (is_finite _ _ f0) eqn:FINf0.
        2:subst f0; reflexivity.
        rewrite Fleb_Rle by auto. rewrite Rle_bool_false. auto.
        eapply Rlt_le_trans, H. compute. Psatz.lra.
      - clear H0 H. destruct (is_finite _ _ f) eqn:FINf.
        2:destruct f as [|[]| |]; inv H1; reflexivity.
        rewrite Fleb_Rle by auto. rewrite Rle_bool_false. auto.
        eapply Rle_lt_trans. apply Ropp_le_cancel, H1. compute. Psatz.lra. }
    apply (inv_itv_correct (ITV f f0)) in H4. destruct inv_itv. discriminate.
    unfold r_of_inum in Hxy. destruct (is_finite _ _ f1) eqn:FINf1; inv Hxy.
    eapply double_round_over_T_linear_expr_correct in H0. eauto.
    eapply mult_T_linear_expr_correct; now eauto.
    pose proof Bdiv_correct 53 1024 eq_refl eq_refl Float.binop_pl mode_NE f1 f2 H.
    unfold Float.div. rewrite FINf1 in H1.
    rewrite Rmult_comm. unfold Rdiv in H1. Rlt_bool_case H1. destruct H1 as (? & ? & _).
    auto. eauto.
  - (* IOcmpf *)
    destruct itv_of_itvabs; inv Hle. destruct r_of_inum. 2:contradiction.
    unfold eval_T_linear_expr. simpl. rewrite Telements_empty. eauto.
Qed.

Definition double_min_delta : float :=
  B754_finite 53 1024 false 1 (-1074) eq_refl.

Lemma double_min_delta_correct :
  ∀ f1 f2,
    is_finite _ _ f1 = true -> is_finite _ _ f2 = true ->
    (B2R _ _ f1 < B2R _ _ f2)%R ->
    (B2R 53 1024 f1 - B2R 53 1024 f2 <= B2R _ _ (Float.neg double_min_delta))%R.
Proof.
  intros. destruct f1, f2; try discriminate.
  - simpl in H1. Psatz.lra.
  - simpl in *. apply F2R_gt_0_reg in H1. destruct b0. discriminate. clear H1.
    rewrite Rminus_0_l, F2R_Zopp with (m:=1). apply Ropp_le_contravar. simpl.
    apply Rmult_le_compat. simpl. Psatz.lra. simpl. Psatz.lra.
    apply Z2R_le. simpl. Psatz.lia.
    apply bpow_le. simpl. unfold bounded in e0.
    rewrite Bool.andb_true_iff in e0. destruct e0.
    unfold canonic_mantissa, Fcore_FLT.FLT_exp in H1.
    apply Zeq_bool_eq in H1. Psatz.lia.
  - simpl in *. apply F2R_lt_0_reg in H1. destruct b. 2:discriminate. clear H1.
    simpl. rewrite Rminus_0_r, F2R_Zopp with (m:=1), F2R_Zopp with (m:=Z.pos m).
    apply Ropp_le_contravar.
    apply Rmult_le_compat. simpl. Psatz.lra. simpl. Psatz.lra.
    apply Z2R_le. simpl. Psatz.lia.
    apply bpow_le. simpl. unfold bounded in e0.
    rewrite Bool.andb_true_iff in e0. destruct e0.
    unfold canonic_mantissa, Fcore_FLT.FLT_exp in H1.
    apply Zeq_bool_eq in H1. Psatz.lia.
  - simpl in *. unfold F2R, Fnum, Fexp in *.
    replace (bpow radix2 e) with (bpow radix2 (e-Z.min e e1+Z.min e e1))
      in H1 |- * by (f_equal; ring).
    replace (bpow radix2 e1) with (bpow radix2 (e1-Z.min e e1+Z.min e e1))
      in H1 |- * by (f_equal; ring).
    replace (bpow radix2 (-1074)) with (bpow radix2 (-1074-Z.min e e1+Z.min e e1))
      by (f_equal; ring).
    rewrite !bpow_plus, <- !Rmult_assoc, <- Rmult_minus_distr_r in *.
    apply Rmult_le_compat_r. apply bpow_ge_0.
    apply Rmult_lt_reg_r in H1. 2:apply bpow_gt_0.
    rewrite <- !Z2R_Zpower, <- !Z2R_mult, <- Z2R_minus in * by Psatz.lia.
    apply lt_Z2R in H1.
    apply Rle_trans with (Z2R (-1))%R. apply Z2R_le. Psatz.lia.
    unfold Z2R, P2R. rewrite Ropp_mult_distr_l_reverse, Rmult_1_l.
    apply Ropp_le_contravar. change 1%R with (bpow radix2 0). apply bpow_le.
    unfold bounded in e0, e2.
    rewrite Bool.andb_true_iff in e0, e2. destruct e0, e2.
    unfold canonic_mantissa, Fcore_FLT.FLT_exp in H2, H4.
    apply Zeq_bool_eq in H2. apply Zeq_bool_eq in H4. Psatz.lia.
Qed.

Definition process_msg (m:message) (ab:unit * query_chan) :=
  let c := snd ab in
  let new_msg :=
    match m with
    | Fact_msg e b =>
      match simplify_bool_expr e c with
      | (IEbinop ((IOcmp _ | IOcmpf _) as op) e1 e2, b') =>
        do_opt le1 <- linearize_expr c e1;
        do_opt le2 <- linearize_expr c e2;
        let le := add_T_linear_expr le1 (opp_T_linear_expr le2) in
        do_opt i <-
          match op, xorb b b' with
          | IOcmp Ceq, true | IOcmp Cne, false
          | IOcmpf Ceq, true | IOcmpf Cne, false =>
            Some (ITV Float.zero Float.zero)

          | (IOcmp Clt|IOcmpf Clt), false | (IOcmp Cge|IOcmpf Cge), true =>
            Some (ITV (B754_infinity _ _ true) Float.zero)
          | IOcmp Cle, false | IOcmp Cgt, true =>
            Some (ITV (B754_infinity _ _ true) (Float.neg one))
          | IOcmpf Cle, false | IOcmpf Cgt, true =>
            Some (ITV (B754_infinity _ _ true) (Float.neg double_min_delta))

          | (IOcmp Cgt|IOcmpf Cgt), false | (IOcmp Cle|IOcmpf Cle), true =>
            Some (ITV Float.zero (B754_infinity _ _ false))
          | IOcmp Cge, false | IOcmp Clt, true =>
            Some (ITV one (B754_infinity _ _ false))
          | IOcmpf Cge, false | IOcmpf Clt, true =>
            Some (ITV double_min_delta (B754_infinity _ _ false))

          | _, _ => None
          end;
          Some (add_T_linear_expr le (i, T.empty _))
      | _ => None
      end
    | Known_value_msg x v =>
      do_opt itv <-
        match v with
        | INz z => Some (itv_of_zitv (ZIntervals.ITV z z))
        | INf f => if is_finite _ _ f then Some (ITV f f) else None
        end;
      Some (opp_itv itv, T.set x (ITV one one, itv) (T.empty _))
    | Itv_msg x i =>
      match itv_of_itvabs (NotBot (Just i)) with
      | All => None
      | Just itv =>
        Some (opp_itv itv, T.set x (ITV one one, itv) (T.empty _))
      end
    | _ => None
    end
  in
  match new_msg with
  | None => NotBot (tt, m::List.nil)
  | Some m' =>
    if T_linear_expr_is_top m' then NotBot (tt, m::List.nil)
    else
      NotBot (tt, m::(Linear_zero_msg m')::List.nil)
  end.

Transparent Float.cmp.

Lemma process_msg_correct:
  forall m ρ ab,
    ρ ∈ γ ab -> ρ ∈ γ m ->
    ρ ∈ γ (process_msg m ab).
Proof.
  unfold process_msg; intros.
  match goal with
  | |- γ match ?O with _ => _ end ρ =>
    assert (forall le, O = Some le -> eval_T_linear_expr ρ le 0); [|destruct O]
  end.
  { destruct m; try discriminate.
    - simpl in H0. destruct (simplify_bool_expr i (snd ab)) eqn:EQ.
      eapply simplify_bool_expr_correct in EQ.
      2:now apply H0. 2:now apply H.
      destruct i0; try discriminate. destruct i0; try discriminate.
      + pose proof linearize_expr_correct (snd ab) i0_1.
        pose proof linearize_expr_correct (snd ab) i0_2.
        destruct linearize_expr. 2:discriminate.
        destruct linearize_expr. 2:discriminate.
        inv EQ. inv H9.
        edestruct H1 as (y & Hxy & Hy). now eauto. apply H. apply H6.
        edestruct H2 as (y' & Hxy' & Hy'). now eauto. apply H. apply H8.
        inv Hxy. inv Hxy'. clear H1 H2.
        assert (Zcmp c i1 i2 = xorb b b0)
          by (destruct b, b0, Zcmp; try discriminate; reflexivity).
        rewrite <- H1. clear H1 b b0 H0 H10 H6 H8. intros.
        match type of H0 with
        | match ?O with _ => _ end = _ =>
          assert (forall i, O = Some i -> (Z2R i2 - Z2R i1)%R ∈ γ i); [|destruct O; inv H0]
        end.
        { clear H0. intros.
          destruct c; simpl in H0; destruct (Z.compare_spec i1 i2); inv H0; split;
          try reflexivity;
          try (apply Z.le_succ_l, Z2R_le in H1;
               rewrite <- Z.add_1_r, Z2R_plus in H1; simpl in H1);
          unfold RF_le, B2R, F2R; simpl; try Psatz.lra. }
        specialize (H1 _ eq_refl).
        replace 0%R with (Z2R i1 - Z2R i2 + (Z2R i2 - Z2R i1))%R by ring.
        apply add_T_linear_expr_correct.
        apply add_T_linear_expr_correct, opp_T_linear_expr_correct; now auto.
        unfold eval_T_linear_expr. simpl. rewrite Telements_empty. apply H1.
      + pose proof linearize_expr_correct (snd ab) i0_1.
        pose proof linearize_expr_correct (snd ab) i0_2.
        destruct linearize_expr. 2:discriminate.
        destruct linearize_expr. 2:discriminate.
        inv EQ. inv H9.
        edestruct H1 as (y & Hxy & Hy). now eauto. apply H. apply H6.
        edestruct H2 as (y' & Hxy' & Hy'). now eauto. apply H. apply H8.
        simpl in Hxy, Hxy'.
        destruct (is_finite _ _ f1) eqn:FINf1. 2:discriminate.
        destruct (is_finite _ _ f2) eqn:FINf2. 2:discriminate.
        inv Hxy. inv Hxy'.
        assert (Float.cmp c f1 f2 = xorb b b0)
          by (destruct b, b0, Float.cmp; try discriminate; reflexivity).
        rewrite <- H3. clear H1 b b0 H0 H10 H3 H2. intros.
        match type of H0 with
        | match ?O with _ => _ end = _ =>
          assert (forall i, O = Some i -> (B2R _ _ f2 - B2R _ _  f1)%R ∈ γ i); [|destruct O; inv H0]
        end.
        { clear H0. intros. unfold Float.cmp in H0.
          rewrite Bcompare_correct in H0 by auto.
          destruct c; simpl cmp_of_comparison in H0;
          destruct (Rcompare_spec (B2R _ _ f1) (B2R _ _ f2));
          inv H0; split;
          try reflexivity;
          unfold RF_le, Float.zero; simpl; try Psatz.lra;
          rewrite ?Ropp_minus_distr; apply double_min_delta_correct; auto. }
        specialize (H1 _ eq_refl).
        replace 0%R with (B2R _ _ f1 - B2R _ _ f2 + (B2R _ _ f2 - B2R _ _ f1))%R by ring.
        apply add_T_linear_expr_correct.
        apply add_T_linear_expr_correct, opp_T_linear_expr_correct; now auto.
        unfold eval_T_linear_expr. simpl. rewrite Telements_empty. apply H1.
    - inv H0. intros.
      match type of H0 with
      | match ?O with _ => _ end = _ =>
        assert (∀ x, O = Some x ->
                ∃ y, r_of_inum (ρ v) = Some y /\ y ∈ γ x); [|destruct O; inv H0]
      end.
      { intros. destruct (ρ v).
        - destruct (is_finite _ _ f) eqn:FINf; inv H1. inv H0.
          simpl. rewrite FINf.
          eexists. split. eauto. unfold RF_le, Float.neg.
          rewrite B2R_Bopp, is_finite_Bopp, FINf. Psatz.lra.
        - inv H1. inv H0. eexists. split. reflexivity.
          eapply (itv_of_zitv_correct (ZIntervals.ITV z z)).
          simpl; Psatz.lia. }
      unfold eval_T_linear_expr, eval_linear_expr. simpl. rewrite Telements_singleton.
      edestruct H1 as (y & Hy & Hxy). now eauto. clear H1. unfold rρ. simpl. rewrite Hy.
      split. 2:now auto. exists 1%R. split. compute; Psatz.lra.
      replace (0-1*y)%R with (-y)%R by ring. apply opp_itv_correct, Hxy.
    - apply (itv_of_itvabs_correct (NotBot (Just a))) in H0.
      destruct itv_of_itvabs; intros; inv H1.
      unfold eval_T_linear_expr. simpl. rewrite Telements_singleton. red. red. red.
      unfold rρ. destruct r_of_inum. 2:now auto. split. 2:now auto.
      exists 1%R. split. compute; Psatz.lra.
      replace (0-1*r)%R with (-r)%R by ring. apply opp_itv_correct, H0. }
  - destruct T_linear_expr_is_top; repeat constructor; auto. apply H1. auto.
  - repeat constructor. now auto.
Qed.

Program Instance abVarExpEq_env : ab_ideal_env t t :=
  {| id_leb := fun _ _ => true;
     id_top := fun _ => NotBot (tt, List.nil);
     id_join := fun _ _ _ => NotBot (tt, List.nil);
     id_init_iter := fun _ => tt;
     id_widen := fun _ _ _ => (tt, NotBot (tt, List.nil));
     assign := fun _ _ _ _ => NotBot (tt, List.nil);
     forget := fun _ _ _ => NotBot (tt, List.nil);
     enrich_query_chan :=
       fun _ c =>
         {| get_var_ty := c.(get_var_ty);
            get_itv := c.(get_itv);
            get_congr := c.(get_congr);
            get_eq_expr := c.(get_eq_expr);
            AbIdealEnv.linearize_expr e :=
              match linearize_expr c e with
              | None => None
              | Some le =>
                if T_linear_expr_is_top le then None else Some le
              end
         |};
     process_msg := process_msg |}.
Next Obligation. split; constructor. Qed.
Next Obligation. split; constructor. Qed.
Next Obligation. split; constructor. Qed.
Next Obligation. split; constructor. Qed.
Next Obligation. split. constructor. split; constructor. Qed.
Next Obligation. split; constructor. Qed.
Next Obligation. split; constructor. Qed.
Next Obligation. apply process_msg_correct; auto. Qed.
Next Obligation.
  constructor; try apply H0.
  simpl. intros.
  pose proof linearize_expr_correct chan e. destruct linearize_expr. 2:discriminate.
  destruct T_linear_expr_is_top; inv H1. auto.
Qed.

Opaque t.
