(* *********************************************************************)
(*                                                                     *)
(*              The Verasco static analyzer                            *)
(*                                                                     *)
(*          Vincent Laporte, Université de Rennes 1                    *)
(*          Jacques-Henri Jourdan, INRIA Paris-Rocquencourt            *)
(*                                                                     *)
(*  Copyright Institut National de Recherche en Informatique et en     *)
(*  Automatique and Université de Rennes 1.  All rights reserved.      *)
(*  This file is distributed under the terms of the GNU General        *)
(*  Public License as published by the Free Software Foundatio,        *)
(*  either version 2 of the License, or (at your option) any later     *)
(*  version.  This file is also distributed under the terms of the     *)
(*  INRIA Non-Commercial License Agreement.                            *)
(*                                                                     *)
(* *********************************************************************)

Require Import Utf8.
Require Import Util.
Require Import Coqlib.
Require Import Ctypes.
Require Import Values.
Require Import Integers.
Require Import Floats.
Require Import Cminor.
Require Import AdomLib.
Require Import IntSet.
Require Import FloatIntervals.
Require Import ZIntervals.

Set Implicit Arguments.
Section nexpr.
  Variable var : Type.

  Inductive num_val : Type :=
  | NVint : int -> num_val
  | NVint64 : int64 -> num_val
  | NVfloat : float -> num_val
  | NVsingle : float32 -> num_val.

  Inductive num_val_ty : Type :=
  | NVTint | NVTint64 | NVTfloat | NVTsingle.

  Inductive num_val_ty_gamma : gamma_op num_val_ty num_val :=
  | NVTSint : ∀ i, num_val_ty_gamma NVTint (NVint i)
  | NVTSint64 : ∀ i, num_val_ty_gamma NVTint64 (NVint64 i)
  | NVTSfloat : ∀ f, num_val_ty_gamma NVTfloat (NVfloat f)
  | NVTSsingle : ∀ f, num_val_ty_gamma NVTsingle (NVsingle f).
  Global Existing Instance num_val_ty_gamma.

  Hint Constructors num_val_ty_gamma.

  Inductive nexpr : Type :=
  | NEvar : num_val_ty -> var -> nexpr
  | NEconst : num_val -> nexpr
  | NEunknown : num_val_ty -> nexpr
  | NEunop : unary_operation -> nexpr -> nexpr
  | NEbinop : binary_operation -> nexpr -> nexpr -> nexpr.

  Fixpoint ne_free_var (ne: nexpr) : ℘(var) :=
    match ne with
    | NEvar _ v => eq v
    | NEconst _ | NEunknown _ => ∅
    | NEunop _ ne' => ne_free_var ne'
    | NEbinop _ ne₁ ne₂ => ne_free_var ne₁ ∪ ne_free_var ne₂
    end.

  Section eval_nexpr.

    Variable ρ: var -> num_val.

    Inductive is_bool : num_val -> Prop :=
    | is_bool_false: is_bool (NVint Int.zero)
    | is_bool_true: is_bool (NVint Int.one).

    Definition Ntrue: num_val := NVint Int.one.
    Definition Nfalse: num_val := NVint Int.zero.

    Definition of_bool (b: bool): num_val :=
      if b then Ntrue else Nfalse.

    Inductive eval_nunop: unary_operation -> num_val -> num_val -> Prop :=
    | eval_nunop_Ocast8unsigned: ∀ i,
      eval_nunop Ocast8unsigned  (NVint i)  (NVint (Int.zero_ext 8 i))
    | eval_nunop_Ocast8signed: ∀ i,
      eval_nunop Ocast8signed  (NVint i)  (NVint (Int.sign_ext 8 i))
    | eval_nunop_Ocast16unsigned: ∀ i,
      eval_nunop Ocast16unsigned  (NVint i)  (NVint (Int.zero_ext 16 i))
    | eval_nunop_Ocast16signed: ∀ i,
      eval_nunop Ocast16signed  (NVint i)  (NVint (Int.sign_ext 16 i))
    | eval_nunop_Onegint : ∀ i,
      eval_nunop Onegint  (NVint i)  (NVint (Int.neg i))
    | eval_nunop_Onotint : ∀ i,
      eval_nunop Onotint  (NVint i)  (NVint (Int.not i))
    | eval_nunop_Onegf : ∀ f,
      eval_nunop Onegf  (NVfloat f)  (NVfloat (Float.neg f))
    | eval_nunop_Onegfs : ∀ f,
      eval_nunop Onegfs (NVsingle f)  (NVsingle (Float32.neg f))
    | eval_nunop_Oabsf : ∀ f,
      eval_nunop Oabsf  (NVfloat f)  (NVfloat (Float.abs f))
    | eval_nunop_Oabsfs : ∀ f,
      eval_nunop Oabsfs  (NVsingle f)  (NVsingle (Float32.abs f))
    | eval_nunop_Osingleoffloat : ∀ f,
      eval_nunop Osingleoffloat  (NVfloat f)  (NVsingle (Float.to_single f))
    | eval_nunop_Ofloatofsingle : ∀ f,
      eval_nunop Ofloatofsingle  (NVsingle f)  (NVfloat (Float.of_single f))
    | eval_nunop_Ointoffloat : ∀ f i,
      Float.to_int f = Some i ->
      eval_nunop Ointoffloat  (NVfloat f)  (NVint i)
    | eval_nunop_Ointofsingle : ∀ f i,
      Float32.to_int f = Some i ->
      eval_nunop Ointofsingle  (NVsingle f)  (NVint i)
    | eval_nunop_Ointuoffloat : ∀ f i,
      Float.to_intu f = Some i ->
      eval_nunop Ointuoffloat  (NVfloat f)  (NVint i)
    | eval_nunop_Ointuofsingle : ∀ f i,
      Float32.to_intu f = Some i ->
      eval_nunop Ointuofsingle  (NVsingle f)  (NVint i)
    | eval_nunop_Ofloatofint : ∀ i,
      eval_nunop Ofloatofint  (NVint i)  (NVfloat (Float.of_int i))
    | eval_nunop_Osingleofint : ∀ i,
      eval_nunop Osingleofint  (NVint i)  (NVsingle (Float32.of_int i))
    | eval_nunop_Ofloatofintu : ∀ i,
      eval_nunop Ofloatofintu  (NVint i)  (NVfloat (Float.of_intu i))
    | eval_nunop_Osingleofintu : ∀ i,
      eval_nunop Osingleofintu  (NVint i)  (NVsingle (Float32.of_intu i))
    | eval_nunop_Onegl : ∀ i,
      eval_nunop Onegl  (NVint64 i)  (NVint64 (Int64.neg i))
    | eval_nunop_Onotl : ∀ i,
      eval_nunop Onotl  (NVint64 i)  (NVint64 (Int64.not i))
    | eval_nunop_Ointoflong : ∀ i,
      eval_nunop Ointoflong  (NVint64 i)  (NVint (Int64.loword i))
    | eval_nunop_Olongofint : ∀ i,
      eval_nunop Olongofint  (NVint i)  (NVint64 (Int64.repr (Int.signed i)))
    | eval_nunop_Olongofintu : ∀ i,
      eval_nunop Olongofintu  (NVint i)  (NVint64 (Int64.repr (Int.unsigned i)))
    | eval_nunop_Olongoffloat : ∀ f i,
      Float.to_long f = Some i ->
      eval_nunop Olongoffloat  (NVfloat f)  (NVint64 i)
    | eval_nunop_Olonguoffloat : ∀ f i,
      Float.to_longu f = Some i ->
      eval_nunop Olonguoffloat  (NVfloat f)  (NVint64 i)
    | eval_nunop_Olongofsingle : ∀ f i,
      Float32.to_long f = Some i ->
      eval_nunop Olongofsingle  (NVsingle f)  (NVint64 i)
    | eval_nunop_Olonguofsingle : ∀ f i,
      Float32.to_longu f = Some i ->
      eval_nunop Olonguofsingle  (NVsingle f)  (NVint64 i)
    | eval_nunop_Ofloatoflong : ∀ i,
      eval_nunop Ofloatoflong  (NVint64 i)  (NVfloat (Float.of_long i))
    | eval_nunop_Osingleoflong : ∀ i,
      eval_nunop Osingleoflong  (NVint64 i)  (NVsingle (Float32.of_long i))
    | eval_nunop_Ofloatoflongu : ∀ i,
      eval_nunop Ofloatoflongu  (NVint64 i)  (NVfloat (Float.of_longu i))
    | eval_nunop_Osingleoflongu : ∀ i,
      eval_nunop Osingleoflongu  (NVint64 i)  (NVsingle (Float32.of_longu i)).

    Inductive eval_nbinop : binary_operation -> num_val -> num_val -> num_val -> Prop :=
    | eval_nbinop_Oadd : ∀ i1 i2,
      eval_nbinop Oadd  (NVint i1)  (NVint i2)  (NVint (Int.add i1 i2))
    | eval_nbinop_Osub : ∀ i1 i2,
      eval_nbinop Osub  (NVint i1)  (NVint i2)  (NVint (Int.sub i1 i2))
    | eval_nbinop_Omul : ∀ i1 i2,
      eval_nbinop Omul  (NVint i1)  (NVint i2)  (NVint (Int.mul i1 i2))
    | eval_nbinop_Odiv : ∀ i1 i2,
      Int.eq i2 Int.zero = false ->
      (Int.eq i1 (Int.repr Int.min_signed) = false
        \/ Int.eq i2 Int.mone = false) ->
      eval_nbinop Odiv  (NVint i1)  (NVint i2)  (NVint (Int.divs i1 i2))
    | eval_nbinop_Odivu : ∀ i1 i2,
      Int.eq i2 Int.zero = false ->
      eval_nbinop Odivu  (NVint i1)  (NVint i2)  (NVint (Int.divu i1 i2))
    | eval_nbinop_Omod : ∀ i1 i2,
      Int.eq i2 Int.zero = false ->
      (Int.eq i1 (Int.repr Int.min_signed) = false \/
        Int.eq i2 Int.mone = false) ->
      eval_nbinop Omod  (NVint i1)  (NVint i2)  (NVint (Int.mods i1 i2))
    | eval_nbinop_Omodu : ∀ i1 i2,
      Int.eq i2 Int.zero = false ->
      eval_nbinop Omodu  (NVint i1)  (NVint i2)  (NVint (Int.modu i1 i2))
    | eval_nbinop_Oand : ∀ i1 i2,
      eval_nbinop Oand  (NVint i1)  (NVint i2)  (NVint (Int.and i1 i2))
    | eval_nbinop_Oor : ∀ i1 i2,
      eval_nbinop Oor  (NVint i1)  (NVint i2)  (NVint (Int.or i1 i2))
    | eval_nbinop_Oxor : ∀ i1 i2,
      eval_nbinop Oxor  (NVint i1)  (NVint i2)  (NVint (Int.xor i1 i2))
    | eval_nbinop_Oshl : ∀ i1 i2,
      Int.ltu i2 Int.iwordsize = true ->
      eval_nbinop Oshl  (NVint i1)  (NVint i2)  (NVint (Int.shl i1 i2))
    | eval_nbinop_Oshr : ∀ i1 i2,
      Int.ltu i2 Int.iwordsize = true ->
      eval_nbinop Oshr  (NVint i1)  (NVint i2)  (NVint (Int.shr i1 i2))
    | eval_nbinop_Oshru : ∀ i1 i2,
      Int.ltu i2 Int.iwordsize = true ->
      eval_nbinop Oshru  (NVint i1)  (NVint i2)  (NVint (Int.shru i1 i2))

    | eval_nbinop_Oaddf : ∀ f1 f2,
      eval_nbinop Oaddf (NVfloat f1) (NVfloat f2) (NVfloat (Float.add f1 f2))
    | eval_nbinop_Osubf : ∀ f1 f2,
      eval_nbinop Osubf (NVfloat f1) (NVfloat f2) (NVfloat (Float.sub f1 f2))
    | eval_nbinop_Omulf : ∀ f1 f2,
      eval_nbinop Omulf (NVfloat f1) (NVfloat f2) (NVfloat (Float.mul f1 f2))
    | eval_nbinop_Odivf : ∀ f1 f2,
      eval_nbinop Odivf (NVfloat f1) (NVfloat f2) (NVfloat (Float.div f1 f2))

    | eval_nbinop_Oaddfs : ∀ f1 f2,
      eval_nbinop Oaddfs (NVsingle f1) (NVsingle f2) (NVsingle (Float32.add f1 f2))
    | eval_nbinop_Osubfs : ∀ f1 f2,
      eval_nbinop Osubfs (NVsingle f1) (NVsingle f2) (NVsingle (Float32.sub f1 f2))
    | eval_nbinop_Omulfs : ∀ f1 f2,
      eval_nbinop Omulfs (NVsingle f1) (NVsingle f2) (NVsingle (Float32.mul f1 f2))
    | eval_nbinop_Odivfs : ∀ f1 f2,
      eval_nbinop Odivfs (NVsingle f1) (NVsingle f2) (NVsingle (Float32.div f1 f2))

    | eval_nbinop_Oaddl : ∀ i1 i2,
      eval_nbinop Oaddl  (NVint64 i1)  (NVint64 i2)  (NVint64 (Int64.add i1 i2))
    | eval_nbinop_Osubl : ∀ i1 i2,
      eval_nbinop Osubl  (NVint64 i1)  (NVint64 i2)  (NVint64 (Int64.sub i1 i2))
    | eval_nbinop_Omull : ∀ i1 i2,
      eval_nbinop Omull  (NVint64 i1)  (NVint64 i2)  (NVint64 (Int64.mul i1 i2))
    | eval_nbinop_Odivl : ∀ i1 i2,
      Int64.eq i2 Int64.zero = false ->
      (Int64.eq i1 (Int64.repr Int64.min_signed) = false
        \/ Int64.eq i2 Int64.mone = false) ->
      eval_nbinop Odivl  (NVint64 i1)  (NVint64 i2)  (NVint64 (Int64.divs i1 i2))
    | eval_nbinop_Odivlu : ∀ i1 i2,
      Int64.eq i2 Int64.zero = false ->
      eval_nbinop Odivlu  (NVint64 i1)  (NVint64 i2)  (NVint64 (Int64.divu i1 i2))
    | eval_nbinop_Omodl : ∀ i1 i2,
      Int64.eq i2 Int64.zero = false ->
      (Int64.eq i1 (Int64.repr Int64.min_signed) = false \/
        Int64.eq i2 Int64.mone = false) ->
      eval_nbinop Omodl  (NVint64 i1)  (NVint64 i2)  (NVint64 (Int64.mods i1 i2))
    | eval_nbinop_Omodlu : ∀ i1 i2,
      Int64.eq i2 Int64.zero = false ->
      eval_nbinop Omodlu  (NVint64 i1)  (NVint64 i2)  (NVint64 (Int64.modu i1 i2))
    | eval_nbinop_Oandl : ∀ i1 i2,
      eval_nbinop Oandl  (NVint64 i1)  (NVint64 i2)  (NVint64 (Int64.and i1 i2))
    | eval_nbinop_Oorl : ∀ i1 i2,
      eval_nbinop Oorl  (NVint64 i1)  (NVint64 i2)  (NVint64 (Int64.or i1 i2))
    | eval_nbinop_Oxorl : ∀ i1 i2,
      eval_nbinop Oxorl  (NVint64 i1)  (NVint64 i2)  (NVint64 (Int64.xor i1 i2))
    | eval_nbinop_Oshll : ∀ i1 i2,
      Int.ltu i2 Int64.iwordsize' = true ->
      eval_nbinop Oshll  (NVint64 i1)  (NVint i2)  (NVint64 (Int64.shl' i1 i2))
    | eval_nbinop_Oshrl : ∀ i1 i2,
      Int.ltu i2 Int64.iwordsize' = true ->
      eval_nbinop Oshrl  (NVint64 i1)  (NVint i2)  (NVint64 (Int64.shr' i1 i2))
    | eval_nbinop_Oshrlu : ∀ i1 i2,
      Int.ltu i2 Int64.iwordsize' = true ->
      eval_nbinop Oshrlu  (NVint64 i1)  (NVint i2)  (NVint64 (Int64.shru' i1 i2))

    | eval_nbinop_Ocmp : ∀ c i1 i2,
      eval_nbinop (Ocmp c)  (NVint i1)  (NVint i2) (of_bool (Int.cmp c i1 i2))
    | eval_nbinop_Ocmpu : ∀ c i1 i2,
      eval_nbinop (Ocmpu c)  (NVint i1)  (NVint i2) (of_bool (Int.cmpu c i1 i2))
    | eval_nbinop_Ocmpl : ∀ c i1 i2,
      eval_nbinop (Ocmpl c)  (NVint64 i1)  (NVint64 i2) (of_bool (Int64.cmp c i1 i2))
    | eval_nbinop_Ocmplu : ∀ c i1 i2,
      eval_nbinop (Ocmplu c)  (NVint64 i1)  (NVint64 i2) (of_bool (Int64.cmpu c i1 i2))
    | eval_nbinop_Ocmpf : ∀ c f1 f2,
      eval_nbinop (Ocmpf c)  (NVfloat f1)  (NVfloat f2) (of_bool (Float.cmp c f1 f2))
    | eval_nbinop_Ocmpfs : ∀ c f1 f2,
      eval_nbinop (Ocmpfs c)  (NVsingle f1)  (NVsingle f2) (of_bool (Float32.cmp c f1 f2)).

    Inductive eval_nexpr: nexpr -> num_val -> Prop :=
    | eval_NEvar : ∀ id i ty,
      ρ id = i -> γ ty i ->
      eval_nexpr (NEvar ty id) i
    | eval_NEconst: ∀ n,
      eval_nexpr (NEconst n) n
    | eval_NEunknown: ∀ n ty,
      n ∈ γ ty ->
      eval_nexpr (NEunknown ty) n
    | eval_NEunop: ∀ op a1 n1 n,
      eval_nexpr a1 n1 ->
      eval_nunop op n1 n ->
      eval_nexpr (NEunop op a1) n
    | eval_Ebinop: ∀ op a1 a2 n1 n2 n,
      eval_nexpr a1 n1 ->
      eval_nexpr a2 n2 ->
      eval_nbinop op n1 n2 n ->
      eval_nexpr (NEbinop op a1 a2) n.

    Inductive error_nunop: unary_operation -> num_val -> Prop :=
    | error_nunop_Ointoffloat: ∀ f, Float.to_int f = None ->
      error_nunop Ointoffloat (NVfloat f)
    | error_nunop_Ointuoffloat: ∀ f, Float.to_intu f = None ->
      error_nunop Ointuoffloat (NVfloat f)
    | error_nunop_Olongoffloat: ∀ f, Float.to_long f = None ->
      error_nunop Olongoffloat (NVfloat f)
    | error_nunop_Olonguoffloat: ∀ f, Float.to_longu f = None ->
      error_nunop Olonguoffloat (NVfloat f)
    | error_nunop_Ointofsingle: ∀ f, Float32.to_int f = None ->
      error_nunop Ointofsingle (NVsingle f)
    | error_nunop_Ointuofsingle: ∀ f, Float32.to_intu f = None ->
      error_nunop Ointuofsingle (NVsingle f)
    | error_nunop_Olongofsingle: ∀ f, Float32.to_long f = None ->
      error_nunop Olongofsingle (NVsingle f)
    | error_nunop_Olonguofsingle: ∀ f, Float32.to_longu f = None ->
      error_nunop Olonguofsingle (NVsingle f).

    Inductive error_nbinop: binary_operation -> num_val -> num_val -> Prop :=
    | error_nbinop_Odiv_0: ∀ i,
      error_nbinop Odiv (NVint i) (NVint Int.zero)
    | error_nbinop_Odiv_minint:
      error_nbinop Odiv (NVint (Int.repr Int.min_signed)) (NVint Int.mone)
    | error_nbinop_Odivu_0: ∀ i,
      error_nbinop Odivu (NVint i) (NVint Int.zero)
    | error_nbinop_Omod_0: ∀ i,
      error_nbinop Omod (NVint i) (NVint Int.zero)
    | error_nbinop_Omod_minint:
      error_nbinop Omod (NVint (Int.repr Int.min_signed)) (NVint Int.mone)
    | error_nbinop_Omodu_0: ∀ i,
      error_nbinop Omodu (NVint i) (NVint Int.zero)
    | error_nbinop_Oshl: ∀ i1 i2,
      Int.ltu i2 Int.iwordsize = false ->
      error_nbinop Oshl (NVint i1) (NVint i2)
    | error_nbinop_Oshr: ∀ i1 i2,
      Int.ltu i2 Int.iwordsize = false ->
      error_nbinop Oshr (NVint i1) (NVint i2)
    | error_nbinop_Oshru: ∀ i1 i2,
      Int.ltu i2 Int.iwordsize = false ->
      error_nbinop Oshru (NVint i1) (NVint i2)
    | error_nbinop_Odivl_0: ∀ i,
      error_nbinop Odivl (NVint64 i) (NVint64 Int64.zero)
    | error_nbinop_Odivl_minint:
      error_nbinop Odivl (NVint64 (Int64.repr Int64.min_signed)) (NVint64 Int64.mone)
    | error_nbinop_Odivlu_0: ∀ i,
      error_nbinop Odivlu (NVint64 i) (NVint64 Int64.zero)
    | error_nbinop_Omodl_0: ∀ i,
      error_nbinop Omodl (NVint64 i) (NVint64 Int64.zero)
    | error_nbinop_Omodl_minint:
      error_nbinop Omodl (NVint64 (Int64.repr Int64.min_signed)) (NVint64 Int64.mone)
    | error_nbinop_Omodlu_0: ∀ i,
      error_nbinop Omodlu (NVint64 i) (NVint64 Int64.zero)
    | error_nbinop_Oshll: ∀ i1 i2,
      Int.ltu i2 Int64.iwordsize' = false ->
      error_nbinop Oshll (NVint64 i1) (NVint i2)
    | error_nbinop_Oshrl: ∀ i1 i2,
      Int.ltu i2 Int64.iwordsize' = false ->
      error_nbinop Oshrl (NVint64 i1) (NVint i2)
    | error_nbinop_Oshrlu: ∀ i1 i2,
      Int.ltu i2 Int64.iwordsize' = false ->
      error_nbinop Oshrlu (NVint64 i1) (NVint i2).

    Inductive error_nexpr: nexpr -> Prop :=
    | error_NEunop: ∀ op a1 n1,
      eval_nexpr a1 n1 ->
      error_nunop op n1 ->
      error_nexpr (NEunop op a1)
    | error_NEunop_rec: ∀ op a1,
      error_nexpr a1 ->
      error_nexpr (NEunop op a1)
    | error_NEbinop: ∀ op a1 n1 a2 n2,
      eval_nexpr a1 n1 ->
      eval_nexpr a2 n2 ->
      error_nbinop op n1 n2 ->
      error_nexpr (NEbinop op a1 a2)
    | error_NEbinop_rec1: ∀ op a1 a2,
      error_nexpr a1 ->
      error_nexpr (NEbinop op a1 a2)
    | error_NEbinop_rec2: ∀ op a1 n1 a2,
      eval_nexpr a1 n1 ->
      error_nexpr a2 ->
      error_nexpr (NEbinop op a1 a2).

  End eval_nexpr.

  Lemma eval_nexpr_fv ρ ρ' ne :
    (∀ x, x ∈ ne_free_var ne → ρ x = ρ' x) →
    eval_nexpr ρ ne ⊆ eval_nexpr ρ' ne.
  Proof.
    induction ne; simpl; intros H q Hq; inv Hq; econstructor (eauto).
    specialize (H _ eq_refl). easy.
  Qed.

  Definition nexpr_ty e : num_val_ty :=
    match e with
      | NEvar ty _ => ty
      | NEconst (NVint _) => NVTint
      | NEconst (NVint64 _) => NVTint64
      | NEconst (NVfloat _) => NVTfloat
      | NEconst (NVsingle _) => NVTsingle
      | NEunknown ty => ty
      | NEunop (Ocast8unsigned | Ocast8signed | Ocast16unsigned | Ocast16signed |
                Onegint | Onotint | Ointoffloat | Ointuoffloat |
                Ointofsingle | Ointuofsingle | Ointoflong) _ =>
        NVTint
      | NEunop (Onegf | Oabsf | Ofloatofsingle | Ofloatofint | Ofloatofintu |
                Ofloatoflong | Ofloatoflongu) _ =>
        NVTfloat
      | NEunop (Onegfs | Oabsfs | Osingleoffloat | Osingleofint | Osingleofintu |
                Osingleoflong | Osingleoflongu) _ =>
        NVTsingle
      | NEunop (Onegl | Onotl | Olongofint | Olongofintu |
                Olongoffloat | Olonguoffloat | Olongofsingle | Olonguofsingle) _ =>
        NVTint64
      | NEbinop (Oadd | Osub | Omul | Odiv | Odivu | Omod | Omodu | Oand | Oor |
                 Oxor | Oshl | Oshr | Oshru |
                 Ocmp _ | Ocmpu _ | Ocmpf _ | Ocmpfs _ | Ocmpl _ | Ocmplu _ ) _ _ =>
        NVTint
      | NEbinop (Oaddf | Osubf | Omulf | Odivf) _ _ => NVTfloat
      | NEbinop (Oaddfs | Osubfs | Omulfs | Odivfs) _ _ => NVTsingle
      | NEbinop (Oaddl | Osubl | Omull | Odivl | Odivlu |
                 Omodl | Omodlu | Oandl | Oorl | Oxorl | Oshll | Oshrl | Oshrlu) _ _ =>
        NVTint64
    end.

  Lemma nexpr_ty_correct:
    ∀ n e ρ, eval_nexpr ρ e n -> γ (nexpr_ty e) n.
  Proof.
    destruct 1.
    - destruct H0; constructor.
    - destruct n; constructor.
    - destruct H; constructor.
    - destruct H0; constructor.
    - destruct H1; try constructor; simpl;
      match goal with |- context [of_bool ?b] => destruct b end; constructor.
  Qed.

  Definition ne_const_i (i: int) : nexpr := NEconst (NVint i).
  Definition ne_const_l (i: int64) : nexpr := NEconst (NVint64 i).
  Definition ne_const_f (f: float) : nexpr := NEconst (NVfloat f).
  Definition ne_const_s (f: float32) : nexpr := NEconst (NVsingle f).

End nexpr.


Inductive num_val_itv : Type :=
| NVIint : ZIntervals.itv+⊤ -> num_val_itv
| NVIint64 : ZIntervals.itv+⊤ -> num_val_itv
| NVIfloat : FloatIntervals.itv+⊤ -> num_val_itv
| NVIsingle : FloatIntervals.itv+⊤ -> num_val_itv.

Inductive num_val_itv_gamma : gamma_op num_val_itv num_val :=
| NVIint_ok : ∀ itv i,
    i ∈ γ itv -> num_val_itv_gamma (NVIint itv) (NVint (Int.repr i))
| NVIint64_ok : ∀ itv i,
    i ∈ γ itv -> num_val_itv_gamma (NVIint64 itv) (NVint64 (Int64.repr i))
| NVIfloat_ok : ∀ itv f,
    f ∈ γ itv ->
    num_val_itv_gamma (NVIfloat itv) (NVfloat f)
| NVIsingle_ok : ∀ itv f,
    Float.of_single f ∈ γ itv ->
    num_val_itv_gamma (NVIsingle itv) (NVsingle f).
Existing Instance num_val_itv_gamma.

Unset Elimination Schemes.
(** Domain of environments using machine numbers. *)

Class ab_machine_env (var t iter_t: Type) `{EqDec var} : Type := {
  mach_top:> top_op (t+⊥);
  mach_leb:> leb_op t;
  mach_join:> join_op t (t+⊥);
  mach_widen:> widen_op iter_t (t+⊥);

  assign: var -> nexpr var -> t -> t+⊥;
  forget: var -> t -> t+⊥;
  assume: nexpr var -> bool -> t -> t+⊥;
  get_itv: nexpr var -> t -> num_val_itv+⊥;
  concretize_int: N -> nexpr var -> t -> int_set+⊤+⊥;
  noerror: nexpr var -> t -> bool;

  mach_gamma:> gamma_op t (var -> num_val);
  mach_gamma_iter:> gamma_op iter_t (var -> num_val);

  mach_top_correct:> top_op_correct (t+⊥) (var -> num_val);
  mach_leb_correct:> leb_op_correct t (var -> num_val);
  mach_join_correct:> join_op_correct t (t+⊥) (var -> num_val);
  mach_widen_correct:> widen_op_correct iter_t (t+⊥) (var -> num_val);
  assign_correct: ∀ x e ρ n ab,
    ρ ∈ γ ab ->
    n ∈ eval_nexpr ρ e ->
    ρ +[x => n] ∈ γ (assign x e ab);
  forget_correct: ∀ x ρ n ab,
    ρ ∈ γ ab ->
    ρ +[x => n] ∈ γ (forget x ab);
  assume_correct: ∀ e ρ ab b,
    ρ ∈ γ ab ->
    of_bool b ∈ eval_nexpr ρ e ->
    ρ ∈ γ (assume e b ab);
  get_itv_correct: ∀ e ρ ab,
    ρ ∈ γ ab ->
    (eval_nexpr ρ e) ⊆ γ (get_itv e ab);
  concretize_int_correct: ∀ n e ρ ab,
    ρ ∈ γ ab ->
    (eval_nexpr ρ e ∘ NVint) ⊆ γ (concretize_int n e ab);
  noerror_correct: ∀ e ρ ab,
    ρ ∈ γ ab ->
    noerror e ab = true ->
    error_nexpr ρ e ->
    False
}.

Section MORE.

Context (var t iter_t: Type) `(EQ: EqDec var) `(D: @ab_machine_env var t iter_t EQ).

Lemma forget_correct':
  ∀ (x: var) (ρ ρ': var → num_val) (ab: t),
    ρ ∈ γ(ab) →
    (∀ y, y ≠ x → ρ'(y) = ρ(y)) →
    ρ' ∈ γ (forget x ab).
Proof.
  intros x ρ ρ' ab H3 H2.
  erewrite (Axioms.extensionality ρ').
  apply forget_correct; eauto.
  intros z. unfold upd.
  destruct @eq_dec as [ -> | NE ]. reflexivity.
  apply H2, NE.
Qed.

Lemma assume_bot:
  ∀ e b (ab: t),
    is_bot (assume e b ab) = true →
    ∀ ρ, ρ ∈ γ(ab) →
         ¬ of_bool(b) ∈ eval_nexpr ρ e.
Proof.
  intros e b ab B ρ G K.
  erewrite is_bot_spec in B. discriminate.
  apply assume_correct; eauto.
Qed.

End MORE.
