(* *********************************************************************)
(*                                                                     *)
(*              The Verasco static analyzer                            *)
(*                                                                     *)
(*          Vincent Laporte, Université de Rennes 1                    *)
(*          Jacques-Henri Jourdan, INRIA Paris-Rocquencourt            *)
(*                                                                     *)
(*  Copyright Institut National de Recherche en Informatique et en     *)
(*  Automatique and Université de Rennes 1.  All rights reserved.      *)
(*  This file is distributed under the terms of the GNU General        *)
(*  Public License as published by the Free Software Foundatio,        *)
(*  either version 2 of the License, or (at your option) any later     *)
(*  version.  This file is also distributed under the terms of the     *)
(*  INRIA Non-Commercial License Agreement.                            *)
(*                                                                     *)
(* *********************************************************************)

Require Import Utf8 Coqlib Util.
Require Import String.
Require Import Maps.
Require Import AST.
Require Import Integers.
Require Import Values.
Require Import Globalenvs.
Require Import Csharpminor.
Require Import Memory.
Require Import Events.
Require Import AdomLib.
Require Import TypesDomain.
Require Import AlarmMon.

Section GE.

Variable ge : genv.

Unset Elimination Schemes.
Class mem_dom (t iter_t:Type) := {
  leb_mem :> leb_op t;
  join_mem :> join_op t (t+⊥);
  widen_mem :> widen_op iter_t (t+⊥);
  (* assign a temporary variable *)
  assign: ident → expr → t → alarm_mon (t+⊥);
  (* assign any value of a given type to a temporary variable *)
  assign_any: ident → AbTy.t → t → alarm_mon (t+⊥);
  (* assign a memory cell *)
  store: memory_chunk → expr → expr → t → alarm_mon (t+⊥);
  (* assume an expression evaluates to a non-zero value or to zero *)
  (* may return an alarm as soon as his paremeter not a comparison *)
  assume: expr → t → alarm_mon (t+⊥ * t+⊥);
  (* checks whether the expression can lead to an error  *)
  noerror: expr → t → alarm_mon unit;
  (* dereference a function pointer *)
  deref_fun: expr → t → alarm_mon (list (ident * fundef));
  (* push a new stack frame *)
  push_frame: (* function to call *) ident → function →
              (* parameter values *) list expr →
              t → alarm_mon (t+⊥);
  (* pop a stack frame *)
  pop_frame: (* return value and temporary to write to *) option (expr) → option(ident) →
             t → alarm_mon (t+⊥);
  (* allocates a new memory block *)
  ab_malloc: expr → option ident → t → alarm_mon (t+⊥);
  (* deallocate a memory block *)
  ab_free: expr → t → alarm_mon (t+⊥);
  (* memcopy : same semantics as extcall_memcpy_sem *)
  ab_memcpy: Z → Z → expr → expr → t → alarm_mon (t+⊥);
  (* initial memory state *)
  init_mem: list (ident * globdef fundef unit) → alarm_mon (t+⊥)
}.

(* A concrete execution state is made of a call stack and a memory.
 * Each frame in the call stack is a pair (function name × environments).
 *)
Definition concrete_state := (list (ident * (temp_env * env)) * mem)%type.

Inductive ExistsCState {T:Type} (E : ℘ concrete_state) Trans (cs:option T) : Prop :=
  ExistsCStateC :
    ∀ f t e s m, E ((f, (t, e)) :: s, m) → Trans f t e s m cs → ExistsCState E Trans cs.

Definition Assign (x: ident) (q: expr) (E: ℘ concrete_state) : ℘ (option concrete_state) :=
  ExistsCState E (fun f t e s m cs =>
    ∀ v, eval_expr ge e t m q v →
         cs = Some ((f, (PTree.set x v t, e)) :: s, m)).

Definition AssignAny (x: ident) (ty: AbTy.t) (E: ℘ concrete_state) : ℘ (option concrete_state) :=
  ExistsCState E (fun f t e s m cs =>
    ∃ v, v ∈ γ(ty) ∧ cs = Some ((f, (PTree.set x v t, e)) :: s, m)).

Definition Store (κ: memory_chunk) (dst src: expr) (E: ℘ concrete_state) : ℘ (option concrete_state) :=
  ExistsCState E (fun f t e s m cs =>
    ∀ v a m',
      eval_expr ge e t m src v →
      eval_expr ge e t m dst a →
      Mem.storev κ m a v = Some m' →
      cs = Some ((f, (t, e)) :: s, m')).

Definition Assume (q: expr) (E: ℘ concrete_state) : ℘ (option (bool * concrete_state)) :=
  ExistsCState E (fun f t e s m cs =>
    ∀ v b,
      Val.bool_of_val v b →
      eval_expr ge e t m q v →
      cs = Some (b, ((f, (t, e)) :: s, m))).

Local Instance boolean_partitionning t (_: gamma_op t concrete_state) :
  gamma_op (t * t) (bool * concrete_state) :=
  fun x y =>
  if fst y
  then γ (fst x) (snd y)
  else γ (snd x) (snd y).

Lemma boolean_partition t (_: gamma_op t concrete_state) (f: bool → t) :
  ∀ b cs,
    cs ∈ γ (f b) →
    Some (b, cs) ∈ γ ((f true, f false), nil).
Proof. intros () cs H; exact H. Qed.

Definition DerefFun (q: expr) (E: ℘ concrete_state) : ℘ (option (ident * fundef)) :=
  ExistsCState E (fun _ t e s m fid =>
    ∀ b fi fd, eval_expr ge e t m q (Vptr b Int.zero) →
               Genv.find_funct_ptr ge b = Some fd →
               Genv.find_symbol ge fi = Some b →
               fid = Some (fi, fd)).

Definition PopFrame (ret: option expr) (rettemp: option (ident)) (E: ℘ concrete_state) : ℘ (option concrete_state) :=
  ExistsCState E (fun _ t e s m cs =>
    ∀ m', Mem.free_list m (blocks_of_env e) = Some m' ->
    ∀ v, match ret with None => v = Vundef | Some q => eval_expr ge e t m q v end ->
         match s with
         | nil =>
           rettemp = Some 1%positive ->
           ∀ r, v = Vint r → cs = Some (s, m')
         | (f', (t', e')) :: s' =>
           cs = Some ((f', (Cminor.set_optvar rettemp v t', e')) :: s', m')
         end).

Definition PushFrame (fi: ident) (f:function) (args:list expr)
           (E: ℘ concrete_state) : ℘ (option concrete_state) :=
  fun cs =>
    ∃ s m, E (s, m) ∧
     list_norepet (List.map fst f.(fn_vars)) ∧
     (∀ vargs,
        match s with
          | nil => vargs = nil
          | (_, (t, e)) :: _ => eval_exprlist ge e t m args vargs
        end →
      ∀ m1 e1, alloc_variables empty_env m (f.(fn_vars)) e1 m1 →
      ∀ t1, bind_parameters f.(fn_params) vargs (create_undef_temps f.(fn_temps)) = Some t1 →
      cs = Some ((fi, (t1, e1))::s, m1)).

Definition Noerror (q:expr) (E: ℘ concrete_state) : ℘ (option unit) :=
  ExistsCState E (fun _ t e s m u => ∀ v, eval_expr ge e t m q v → v <> Vundef → u = Some tt).

Definition Malloc (sz_exp:expr) (rettemp:option ident) (E: ℘ concrete_state) : ℘ (option concrete_state) :=
  ExistsCState E (fun f t e s m cs =>
    ∀ sz  m' vres, eval_expr ge e t m sz_exp sz →
                   extcall_malloc_sem ge (sz::nil) m E0 vres m' →
                   cs = Some ((f, (Cminor.set_optvar rettemp vres t, e)) :: s, m')).

Definition Free (ptr_exp:expr) (E: ℘ concrete_state) : ℘ (option concrete_state) :=
  ExistsCState E (fun f t e s m cs =>
    ∀ ptr m', eval_expr ge e t m ptr_exp ptr →
              extcall_free_sem ge (ptr::nil) m E0 Vundef m' →
              cs = Some ((f, (t, e)) :: s, m')).

Definition Memcpy (sz al:Z) (dst_exp src_exp:expr)  (E: ℘ concrete_state) : ℘ (option concrete_state) :=
  ExistsCState E (fun f t e s m cs =>
    ∀ dst src  m',
      eval_expr ge e t m dst_exp dst →
      eval_expr ge e t m src_exp src →
      extcall_memcpy_sem sz al ge (dst::src::nil) m E0 Vundef m' →
      cs = Some ((f, (t, e)) :: s, m')).

Definition Init_Mem (defs:list (ident * globdef fundef unit)) : ℘ (option concrete_state) :=
  fun cs =>
    ∃ prog : program,
      ge = Genv.globalenv prog
    ∧ defs = prog.(prog_defs)
    ∧ ∀ m, Genv.alloc_globals ge Mem.empty defs = Some m →
           cs = Some (nil, m).

Local Instance ListIncl : gamma_op (list (ident * fundef)) (ident * fundef) :=
  λ l x, In x l.

Local Instance UnitGamma : gamma_op unit unit := eq.

Class mem_dom_spec {t iter_t: Type}
       (DOM: mem_dom t iter_t)
       (mem_gamma: gamma_op t concrete_state)
       (mem_gamma_iter: gamma_op iter_t concrete_state)
  : Prop := {
  leb_mem_correct:> leb_op_correct t concrete_state;
  join_mem_correct:> join_op_correct t (t+⊥) concrete_state;
  widen_mem_correct:> widen_op_correct iter_t (t+⊥) concrete_state;
  assign_sound: ∀ x e ab, Assign x e (γ ab) ⊆ γ (assign x e ab);
  assign_any_sound: ∀ x ty ab, AssignAny x ty (γ ab) ⊆ γ (assign_any x ty ab);
  store_sound: ∀ κ dst src ab, Store κ dst src (γ ab) ⊆ γ (store κ dst src ab);
  assume_sound: ∀ e ab, Assume e (γ ab) ⊆ γ(assume e ab);
  noerror_sound: ∀ e ab, Noerror e (γ ab) ⊆ γ (noerror e ab);
  deref_fun_sound: ∀ e ab, DerefFun e (γ ab) ⊆ γ (deref_fun e ab);
  pop_frame_sound: ∀ ret rettemp ab, PopFrame ret rettemp (γ ab) ⊆ γ (pop_frame ret rettemp ab);
  push_frame_sound: ∀ fi f args ab, PushFrame fi f args (γ ab) ⊆ γ (push_frame fi f args ab);
  malloc_sound: ∀ sz rettemp ab, Malloc sz rettemp (γ ab) ⊆ γ (ab_malloc sz rettemp ab);
  free_sound: ∀ ptr ab, Free ptr (γ ab) ⊆ γ (ab_free ptr ab);
  memcpy_sound: ∀ sz al dst src ab, Memcpy sz al dst src (γ ab) ⊆ γ (ab_memcpy sz al dst src ab);
  init_mem_sound: ∀ defs, Init_Mem defs ⊆ γ (init_mem defs)
}.

End GE.
