(* *********************************************************************)
(*                                                                     *)
(*              The Verasco static analyzer                            *)
(*                                                                     *)
(*          Jacques-Henri Jourdan, INRIA Paris-Rocquencourt            *)
(*                                                                     *)
(*  Copyright Institut National de Recherche en Informatique et en     *)
(*  Automatique and Université de Rennes 1.  All rights reserved.      *)
(*  This file is distributed under the terms of the GNU General        *)
(*  Public License as published by the Free Software Foundatio,        *)
(*  either version 2 of the License, or (at your option) any later     *)
(*  version.  This file is also distributed under the terms of the     *)
(*  INRIA Non-Commercial License Agreement.                            *)
(*                                                                     *)
(* *********************************************************************)

Require Import String ToString Coqlib Utf8 Maps ShareTree Sets Util.
Require Import AdomLib.
Require Import IdealExpr AbIdealEnv.

Module Dom := WPFun T.

Definition t := Dom.t ideal_num_ty.

Instance t_gamma: gamma_op t (var -> ideal_num) := Dom.gamma_tree _.

Instance eq_ty : EqDec ideal_num_ty.
Proof. intros [] []; auto; right; discriminate. Defined.

Fixpoint traverse_expr (ab:t) (e:iexpr) : t+⊥ :=
  match e with
  | IEvar ty v =>
    match T.get_set v ab with
    | (None, setter) => NotBot (setter ty)
    | (Some ty', _) => if eq_ty ty ty' then NotBot ab else Bot
    end
  | IEconst _ | IEZitv _ | IEunknown _ => NotBot ab
  | IEunop _ e => traverse_expr ab e
  | IEbinop _ e1 e2 =>
    do_bot ab <- traverse_expr ab e1;
    traverse_expr ab e2
  end.

Lemma traverse_expr_correct:
  ∀ e ab ρ n,
    ρ ∈ γ ab ->
    eval_iexpr ρ e n ->
    ρ ∈ γ (traverse_expr ab e).
Proof.
  induction e.
  - intros. inv H0. simpl. destruct (T.get_set_spec v ab), T.get_set.
    simpl in H0. subst. pose proof (H v). unfold Dom.get in H0.
    destruct T.get.
    + destruct i, i0; inv H0; inv H5; try congruence; auto.
    + rewrite H1. repeat intro. unfold Dom.get. rewrite T.gsspec.
      destruct T.elt_eq. subst. auto. apply H.
  - intros. auto.
  - intros. auto.
  - intros. auto.
  - intros. inv H0. eapply IHe; eauto.
  - intros. inv H0. eapply botbind_spec; eauto.
Qed.

Program Instance abVarExpEq_env : ab_ideal_env t (t+⊥) :=
  {| id_leb a b := Dom.leb_tree (flat_leb_op eq_dec) _ (fst a) b;
     id_top c := NotBot (Dom.top_tree _, nil);
     id_join a b c := NotBot (Dom.joinwiden (flat_join_op eq_dec) _ (fst a) (fst b), nil);
     id_init_iter := id;
     id_widen a b c :=
       match a, b with
       | Bot, Bot => (Bot, Bot)
       | Bot, NotBot (r, _) | NotBot r, Bot => (NotBot r, NotBot (r, nil))
       | NotBot (a), NotBot (b, _) =>
         let r := Dom.joinwiden (flat_join_op eq_dec) _ (a:Dom.t ideal_num_ty) b in
         (NotBot r, NotBot (r, nil))
       end;
     assign x e ab c :=
       do_bot ab <- traverse_expr (fst ab) e;
       NotBot (Dom.set ab x (Just (iexpr_ty e)), nil);
     forget x ab c := NotBot (Dom.set (fst ab) x All, nil);
     enrich_query_chan ab c :=
       {| get_var_ty := fun x => Dom.get x ab;
          get_itv := c.(get_itv);
          get_congr := c.(get_congr);
          get_eq_expr := c.(get_eq_expr);
          linearize_expr := c.(linearize_expr) |};
     process_msg m ab :=
       match m with
       | Fact_msg e _ =>
         do_bot ab <- traverse_expr (fst ab) e;
         NotBot (ab, m::nil)
       | _ => NotBot (fst ab, m::nil)
       end
  |}.
Next Obligation. destruct x; reflexivity. Qed.
Next Obligation. destruct x; reflexivity. Qed.
Next Obligation. destruct x; reflexivity. Qed.
Next Obligation. split. apply Dom.top_tree_correct. constructor. Qed.
Next Obligation.
  eapply Dom.leb_tree_correct, H0. 2:now eauto.
  apply flat_leb_op_correct. intros [] []; try discriminate; reflexivity.
Qed.
Next Obligation.
  split. 2:now constructor. apply Dom.join_tree_correct.
  apply flat_join_op_correct. intros [] []; try discriminate; reflexivity.
  destruct H as [[]|[]]; now auto.
Qed.
Next Obligation.
  destruct ab1, ab2; try contradiction. split. apply H. split. apply H. constructor.
  destruct x0.
  split; [|split;[|constructor]]; apply Dom.join_tree_correct; auto;
  apply flat_join_op_correct; intros [] []; try discriminate; reflexivity.
Qed.
Next Obligation.
  eapply botbind_spec, traverse_expr_correct, H1. 2:now apply H.
  split. 2:now constructor.
  repeat intro. simpl. unfold Dom.get. rewrite T.gsspec.
  destruct T.elt_eq. subst. rewrite upd_s. eapply iexpr_ty_correct, H1.
  rewrite upd_o by auto. apply H2.
Qed.
Next Obligation.
  split. 2:now constructor. repeat intro. simpl. unfold Dom.get. rewrite T.grspec.
  destruct T.elt_eq. constructor. rewrite upd_o by auto. apply H.
Qed.
Next Obligation.
  destruct m; try now (split;[apply H|constructor;[auto|constructor]]).
  eapply botbind_spec, traverse_expr_correct, H0. 2:now apply H.
  split. now auto. constructor. now auto. constructor.
Qed.
Next Obligation. destruct H0; constructor; simpl; auto. Qed.

Instance t_tostring : ToString t := fun _ => ""%string.
