(* *********************************************************************)
(*                                                                     *)
(*              The Verasco static analyzer                            *)
(*                                                                     *)
(*          Vincent Laporte, Université de Rennes 1                    *)
(*                                                                     *)
(*  Copyright Institut National de Recherche en Informatique et en     *)
(*  Automatique and Université de Rennes 1.  All rights reserved.      *)
(*  This file is distributed under the terms of the GNU General        *)
(*  Public License as published by the Free Software Foundatio,        *)
(*  either version 2 of the License, or (at your option) any later     *)
(*  version.  This file is also distributed under the terms of the     *)
(*  INRIA Non-Commercial License Agreement.                            *)
(*                                                                     *)
(* *********************************************************************)

Require Import
  Coqlib AST Values Maps ShareTree Globalenvs Csharpminor Memory Hash.

Require Import
  Utf8 String ToString Debug DebugCminor
  AdomLib Util TreeAl AssocList MemChunkTree Sets IntSet.

Notation fname := ident (only parsing).

Set Implicit Arguments.
Unset Elimination Schemes.

(** Type of abstract blocks. *)
Inductive ablock :=
| ABlocal (f: fname) (l: ident)
| ABglobal (b: block)
.

Instance AblockToString (ge: genv) : ToString ablock :=
  (λ b, match b with
       | ABlocal f l => "Local(" ++ string_of_ident f ++ ", " ++ string_of_ident l ++ ")"
       | ABglobal b => "Global(" ++ match Genv.invert_symbol ge b with Some s => string_of_ident s | None => "?" end ++ ")"
       end)%string.

Definition ablock_beq (x y: ablock) : bool :=
  match x, y with
  | ABlocal a b, ABlocal c d => Pos.eqb a c && Pos.eqb b d
  | ABglobal a, ABglobal b => Pos.eqb a b
  | _, _ => false
  end.

Lemma ablock_beq_correct :
  ∀ x y, ablock_beq x y = true <-> x = y.
Proof.
  intros [a b|a] [c d|c]; simpl;
  (try intuition discriminate);
  repeat
    (match goal with
     | |- appcontext[Pos.eqb ?a ?b] => generalize (Pos.eqb_spec a b); destruct (Pos.eqb a b)
     | |- appcontext[Z.eqb ?a ?b] => generalize (Z.eqb_spec a b); destruct (Z.eqb a b)
     end; simpl;
     let H := fresh in
     intros H; inversion H; clear H; subst);
  intuition congruence.
Qed.

Definition ablock_dec (x y: ablock) : { x = y } + { x ≠ y }.
  refine ((if ablock_beq x y as b return (b = true <-> x = y) → { x = y } + { x ≠ y }
           then λ H, left _
           else λ H, right _)
         (ablock_beq_correct x y)).
Proof.
  abstract intuition.
  abstract intuition.
Defined.
Instance AblockDec : EqDec ablock := ablock_dec.

Module Ablock : TYPE_EQ
with Definition s := ablock
with Definition t := (fname * ident + block)%type.

  Definition s := ablock.
  Definition t := (fname * ident + block)%type.
  Definition t_of_s (b: s) : t :=
    match b with
    | ABlocal f l => inl (f, l)
    | ABglobal b => inr b
    end.
  Definition s_of_t (x: t) : s :=
    match x with
    | inl (f, l) => ABlocal f l
    | inr b => ABglobal b
    end.
  Lemma s_of_t_of_s : forall x : s, s_of_t (t_of_s x) = x.
  Proof. intros [f l|b]; reflexivity. Qed.
  Lemma t_of_s_of_t: forall x : t, t_of_s (s_of_t x) = x.
  Proof. intros [(f,l)|b]; reflexivity. Qed.
  Definition eq: forall (x y: s), {x = y} + {x <> y} := ablock_dec.
End Ablock.

Module _MTFI : SHARETREE with Definition elt := (fname * ident)%type
  := ProdShareTree(PShareTree)(PShareTree).
Module _MTFIB : SHARETREE with Definition elt := (fname * ident + block)%type
  := SumShareTree(_MTFI)(PShareTree).
Module ABTree : SHARETREE with Definition elt := ablock := BijShareTree(Ablock)(_MTFIB).
Module BlockSet := Tree2Set ABTree.
Module BSO := SetOp BlockSet.

Definition block_of_ablock (ge: genv) (stk: list (ident * (temp_env * env)))
           (ab: ablock) : option block :=
  match ab with
  | ABlocal f l =>
    do_opt t_e <- assoc f stk;
    let (_, e) := t_e in
    do_opt b_sz <- PTree.get l e;
    let (b, _) := b_sz in
    Some b
  | ABglobal b =>
    do_opt s <- Genv.invert_symbol ge b;
    Some b
  end.

Definition size_ty_chunk (κ:typed_memory_chunk) := size_chunk (proj1_sig κ).

Inductive cell : Type :=
| AClocal (f: ident) (l: ident) (ofs: Z) (κ: typed_memory_chunk)
| ACtemp (f: ident) (r: ident)
| ACglobal (b: block) (ofs: Z) (κ: typed_memory_chunk)
.

Definition ablock_of_cell (c: cell) : ablock :=
  match c with
  | AClocal f l _ _ => ABlocal f l
  | ACglobal b _ _ => ABglobal b
  | ACtemp f r => ABglobal xH (* spurious value, should not be used *)
  end.

Definition range : Type := (Z * typed_memory_chunk)%type.

Definition range_of_cell (c: cell) : range :=
  match c with
  | AClocal _ _ o κ
  | ACglobal _ o κ => (o, κ)
  | ACtemp _ _ => (0, exist _ Mint32 I)
  end.

Definition cell_from (b: ablock) (r: range) : cell :=
  match b with
  | ABlocal f l => AClocal f l (fst r) (snd r)
  | ABglobal b => ACglobal b (fst r) (snd r)
  end.

Lemma ablock_of_cell_from b r :
  ablock_of_cell (cell_from b r) = b.
Proof. destruct b; reflexivity. Qed.

Program Definition cell_beq (x y: cell) : bool :=
  match x, y return _ with
  | AClocal f l o κ, AClocal f' l' o' κ' =>
    Pos.eqb f f' && Pos.eqb l l' && 
    Z.eqb o o' && chunk_beq κ κ'
  | ACtemp f r, ACtemp f' r' =>
    Pos.eqb f f' && Pos.eqb r r'
  | ACglobal b o κ, ACglobal b' o' κ' =>
    Pos.eqb b b' && Z.eqb o o' && chunk_beq κ κ'
  | _, _ => false
  end.

Lemma cell_beq_correct x y :
  cell_beq x y ↔ x = y.
Proof.
  unfold is_true.
  split.
- destruct x; simpl; destruct y; intros; eq_some_inv;
  try destruct κ as [[] []], κ0  as [[] []];
  InvBooleans; try discriminate;
  repeat rewrite Pos.eqb_eq in *;
  repeat rewrite Z.eqb_eq in *;
  repeat rewrite chunk_beq_correct in *;
  f_equal;
  try congruence.
- intros <-. destruct x; simpl;
  repeat rewrite (proj2 (Pos.eqb_eq _ _) eq_refl);
  repeat rewrite (proj2 (Z.eqb_eq _ _) eq_refl);
  repeat rewrite (proj2 (chunk_beq_correct _ _) eq_refl);
  reflexivity.
Qed.

Instance CellToString (ge: genv): ToString cell :=
  λ c,
  match c with
  | ACtemp f r => ("Temp(" ++ string_of_ident f ++ ", " ++ string_of_ident r ++")")%string
  | AClocal _ _ o κ
  | ACglobal _ o κ =>
    ("[" ++ to_string (ToString := AblockToString ge) (ablock_of_cell c) ++
               (if Z_zerop o then "" else ", " ++ to_string o) ++
               "]%" ++ to_string (proj1_sig κ))%string
  end.

Section OVERLAP.

Definition ranges_disjoint (a b: range) : Prop :=
  let '(i,c) := a in
  let '(j,d) := b in
  i + size_ty_chunk c <= j ∨ j + size_ty_chunk d <= i.

Definition ranges_overlap (a b: range) : Prop :=
  let '(i,c) := a in
  let '(j,d) := b in
  j < i + size_ty_chunk c ∧ i < j + size_ty_chunk d.

Remark ranges_disjoint_overlap a b :
  ranges_disjoint a b →
  ranges_overlap a b →
  ∀ P : Prop, P.
Proof.
  unfold ranges_disjoint, ranges_overlap. destruct a, b. intuition.
Qed.

Lemma ranges_disjoint_dec a b :
  { ranges_disjoint a b } + { ranges_overlap a b }.
Proof.
  unfold ranges_overlap, ranges_disjoint.
  destruct a; destruct b.
  match goal with |- context[ ?a <= ?b \/ _ ] => destruct (Z_le_dec a b) end. intuition.
  match goal with |- context[ _ \/ ?a <= ?b ] => destruct (Z_le_dec a b) end; intuition.
Defined.
Global Opaque ranges_disjoint_dec.

Lemma ranges_disjoint_not_eq r r' :
  ranges_disjoint r r' → r ≠ r'.
Proof.
  destruct r as (i,κ), r' as (i', κ').
  simpl. intros H K. eq_some_inv. subst.
  destruct κ' as [[] []]; unfold size_ty_chunk in *; simpl in *; Psatz.lia.
Qed.

Definition is_temp (c: cell) : bool :=
  match c with
  | ACtemp _ _ => true
  | _ => false
  end.

Definition cells_disjoint (a b: cell) : Prop :=
  is_temp a || is_temp b ∧ a ≠ b ∨
  ablock_of_cell a ≠ ablock_of_cell b ∨
  ranges_disjoint (range_of_cell a) (range_of_cell b).

Definition cells_overlap (a b: cell) : Prop :=
  (is_temp a || is_temp b = false ∨ a = b) ∧
  ablock_of_cell a = ablock_of_cell b ∧
  ranges_overlap (range_of_cell a) (range_of_cell b).

Remark cells_disjoint_overlap a b :
  cells_disjoint a b →
  cells_overlap a b →
  ∀ P : Prop, P.
Proof.
  unfold cells_disjoint, cells_overlap.
  unfold is_true. rewrite orb_true_iff, orb_false_iff.
  intuition. congruence. congruence.
  eapply ranges_disjoint_overlap; eauto.
  eapply ranges_disjoint_overlap; eauto.
Qed.

Lemma cells_disjoint_not_eq c c' :
  cells_disjoint c c' → c ≠ c'.
Proof.
  unfold cells_disjoint, is_true. rewrite orb_true_iff.
  intros H <-. destruct H as [(_, H) | [ H | H ] ]; try exact (H eq_refl).
  exact (ranges_disjoint_not_eq H eq_refl).
Qed.

Definition cells_disjoint_dec a b :
  { cells_disjoint a b } + { cells_overlap a b }.
  refine (
      (if (is_temp a || is_temp b) && negb (cell_beq a b) as q return _ = q → _
      then λ K, left (or_introl _)
      else
        λ K,
        match ablock_dec (ablock_of_cell a) (ablock_of_cell b) with
        | left EQ =>
          match ranges_disjoint_dec (range_of_cell a) (range_of_cell b) with
          | left D => left (or_intror (or_intror D))
          | right H => right (conj _ (conj EQ H))
          end
        | right NE => left (or_intror (or_introl NE))
        end) eq_refl).
  Proof.
    rewrite andb_true_iff, negb_true_iff in K.
    destruct K as [Htmp NE]. split. exact Htmp. intros EQ.
    generalize (proj2 (cell_beq_correct a b) EQ). rewrite NE.
    unfold is_true; intros X; eq_some_inv.
    rewrite andb_false_iff in K.
    destruct K as [K|K]. left. exact K.
    right. rewrite negb_false_iff in K. exact (proj1 (cell_beq_correct _ _) K).
  Defined.

(* We intentedly do not mention Many32 and Many64 *)
Program Definition memory_chunk_fold {A} (f: A → typed_memory_chunk → A) (a: A) : A :=
  (f (f (f (f (f (f (f (f a Mint8signed) Mint8unsigned) Mint16signed) Mint16unsigned) Mint32) Mint64) Mfloat32) Mfloat64).

Definition overlap_aux (b: ablock) (ofs: Z) (κ': typed_memory_chunk) (n: N) (acc: list cell): list cell :=
  let base := ofs - size_ty_chunk κ' + 1 in
  N.peano_rect (λ _, list cell)
               acc
               (λ n,
                let ofs' := base + Z.of_N n in
                cons (cell_from b (ofs', κ'))
                )
               n.

Lemma overlap_aux_correct b ofs κ' c:
  ∀ n acc,
    In c (overlap_aux b ofs κ' n acc)
       ↔
    In c acc ∨
    ∃ ofs', c = cell_from b (ofs', κ') ∧ ofs' < ofs - size_ty_chunk κ' + Z.of_N n + 1 ∧ ofs < ofs' + size_ty_chunk κ'.
Proof.
  unfold overlap_aux.
  induction n as [|n IH] using N.peano_ind; intros acc.
  split. exact (@or_introl _ _).
  intros [H|H]. exact H. hsplit. subst. simpl. Psatz.lia.
  rewrite N.peano_rect_succ. split.
  intros [ <- | IN ].
  right. eexists. split. reflexivity. split. rewrite N2Z.inj_succ. Psatz.lia. Psatz.lia.
  destruct (proj1 (IH _) IN). left; easy. right. hsplit. subst. eexists. split. reflexivity. split.
  rewrite N2Z.inj_succ. Psatz.lia. Psatz.lia.
  intros [H|H]. right. exact (proj2 (IH _) (or_introl H)).
  hsplit. subst.
  rewrite N2Z.inj_succ in *.
  assert (ofs' < ofs - size_ty_chunk κ' + Z.of_N n + 1 ∨ ofs' = ofs - size_ty_chunk κ' + 1 + Z.of_N n) as K by Psatz.lia.
  destruct K.
  right. apply (proj2 (IH _)). vauto.
  left. repeat f_equal. easy.
Qed.

(* All cells that overlap with and are distinct from c *)
Definition overlap (c: cell) : list cell :=
  match c with
  | ACtemp _ _ => nil
  | AClocal _ _ ofs κ
  | ACglobal _ ofs κ =>
    let b := ablock_of_cell c in
    memory_chunk_fold
      (λ l κ',
       if eq_dec (proj1_sig κ) (proj1_sig κ')
       then
         overlap_aux b ofs κ (Z.to_N (size_ty_chunk κ - 1))
        (overlap_aux b (ofs + size_ty_chunk κ) κ (Z.to_N (size_ty_chunk κ - 1)) l)
       else overlap_aux b ofs κ' (Z.to_N (size_ty_chunk κ + size_ty_chunk κ' - 1)) l)
      nil
  end.

Lemma memory_chunk_fold_rec {A} (f: A -> typed_memory_chunk -> A) :
  ∀ (P:A -> Prop), (∀ x c, P x -> P (f x c)) ->
  ∀ x, P x -> P (memory_chunk_fold f x).
Proof.
  unfold memory_chunk_fold. intros.
  repeat apply H. auto.
Qed.

Lemma in_memory_chunk_fold {A} (f: list A → typed_memory_chunk → list A) :
  (∀ x c l, In x l → In x (f l c)) →
  ∀ x c, (∀ l, In x (f l c)) →
         ∀ l, In x (memory_chunk_fold f l).
Proof.
  unfold memory_chunk_fold; intros H x c Hx l.
  destruct c as [[] []]; repeat first [apply Hx|apply H]; congruence.
Qed.

Lemma overlap_correct' : ∀ c c', cells_overlap c c' → c ≠ c' → In c' (overlap c).
Proof.
  unfold overlap, ablock_of_cell.
  intros [f l ofs κ|f r|b ofs κ] [f' l' ofs' κ'|f' r'|b' ofs' κ'];
  intros (Htmp & EQ & H);
  simpl in Htmp; unfold is_true in Htmp; eq_some_inv;
  simpl in EQ; inv EQ; intros NE;
  try (destruct Htmp as [ K | K ]; eq_some_inv; easy).
+ apply (in_memory_chunk_fold) with κ'.
  intros x q l H0. destruct (eq_dec (proj1_sig κ) (proj1_sig q)); repeat (apply overlap_aux_correct; left); easy.
  simpl in H.
  intros l. destruct (eq_dec (proj1_sig κ) (proj1_sig κ')).
  - assert (κ = κ') by (destruct κ as [[] []], κ' as [[] []]; simpl in *; congruence).
    subst κ'. apply overlap_aux_correct.
    assert (ofs' = ofs ∨ ofs' < ofs ∨ ofs' > ofs) as K by Psatz.lia.
    destruct K as [K | [K | K] ]. congruence.
    right. eexists. split. reflexivity. rewrite Z2N.id; Psatz.lia.
    left. apply overlap_aux_correct. right. eexists. split. reflexivity.
    rewrite Z2N.id; Psatz.lia.
  - apply overlap_aux_correct. right. eexists. split. reflexivity.
    simpl in H. rewrite Z2N.id; Psatz.lia.
+ (* exactly same script as above *)
  apply (in_memory_chunk_fold) with κ'.
  intros x q l H0. destruct (eq_dec (proj1_sig κ) (proj1_sig q)); repeat (apply overlap_aux_correct; left); easy.
  simpl in H.
  intros l. destruct (eq_dec (proj1_sig κ) (proj1_sig κ')).
  - assert (κ = κ') by (destruct κ as [[] []], κ' as [[] []]; simpl in *; congruence).
    subst κ'. apply overlap_aux_correct.
    assert (ofs' = ofs ∨ ofs' < ofs ∨ ofs' > ofs) as K by Psatz.lia.
    destruct K as [K | [K | K] ]. congruence.
    right. eexists. split. reflexivity. rewrite Z2N.id; Psatz.lia.
    left. apply overlap_aux_correct. right. eexists. split. reflexivity.
    rewrite Z2N.id; Psatz.lia.
  - apply overlap_aux_correct. right. eexists. split. reflexivity.
    simpl in H. rewrite Z2N.id; Psatz.lia.
Qed.

Lemma overlap_correct r r' : ¬ In r' (overlap r) → cells_disjoint r r' ∨ r = r'.
Proof.
  Hint Resolve overlap_correct'.
  intros H. destruct (cells_disjoint_dec r r'); auto.
  right. destruct (eq_dec_of_beq cell_beq cell_beq_correct r r'); auto.
  elim H. auto.
Qed.

Global Opaque overlap.

End OVERLAP.

Module Cell : TYPE_EQ
with Definition s := cell
with Definition t := (fname * ident * Z * typed_memory_chunk + fname * ident + block * Z * typed_memory_chunk)%type.

  Definition s := cell.
  Definition t := (fname * ident * Z * typed_memory_chunk + fname * ident + block * Z * typed_memory_chunk)%type.
  Definition t_of_s (b: s) : t :=
    match b with
    | AClocal f l o κ => inl (inl (f, l, o, κ))
    | ACtemp f r => inl (inr (f, r))
    | ACglobal b o κ => inr (b, o, κ)
    end.
  Definition s_of_t (x: t) : s :=
    match x with
    | inl (inl (f, l, o, κ)) => AClocal f l o κ
    | inl (inr (f, r)) => ACtemp f r
    | inr (b, o, κ) => ACglobal b o κ
    end.
  Lemma s_of_t_of_s : forall x : s, s_of_t (t_of_s x) = x.
  Proof. intros [f l|f r|b]; reflexivity. Qed.
  Lemma t_of_s_of_t: forall x : t, t_of_s (s_of_t x) = x.
  Proof. intros [[(((f,l),o), κ)|(f,r)]|((b,o),κ)]; reflexivity. Qed.
  Definition eq: forall (x y: s), {x = y} + {x <> y} :=
    eq_dec_of_beq cell_beq cell_beq_correct.
End Cell.

Module _MTFIZ : SHARETREE with Definition elt := (fname * ident * Z)%type
  := ProdShareTree(_MTFI)(ZShareTree).
Module _MTFIZC : SHARETREE with Definition elt := (fname * ident * Z * typed_memory_chunk)%type
  := ProdShareTree(_MTFIZ)(MemChunkTree).
Module _MTFIZCFI : SHARETREE with Definition elt := (fname * ident * Z * typed_memory_chunk + fname * ident)%type
  := SumShareTree(_MTFIZC)(_MTFI).
Module _MTBZ : SHARETREE with Definition elt := (block * Z)%type
  := ProdShareTree(PShareTree)(ZShareTree).
Module _MTBZC : SHARETREE with Definition elt := (block * Z * typed_memory_chunk)%type
  := ProdShareTree(_MTBZ)(MemChunkTree).
Module _MTFIZCFIBZC : SHARETREE with Definition elt := (fname * ident * Z * typed_memory_chunk + fname * ident + block * Z * typed_memory_chunk)%type
  := SumShareTree(_MTFIZCFI)(_MTBZC).

Module ACTree : SHARETREE with Definition elt := cell := BijShareTree(Cell)(_MTFIZCFIBZC).
Module CellSet  : SET with Definition elt := cell := Tree2Set ACTree.
Module CSO := SetOp CellSet.

Instance ACellDec : EqDec cell := ACTree.elt_eq.

Definition set_product (bs: BlockSet.t) κ (offs: int_set+⊤+⊥) : CellSet.t +⊤ :=
  match offs with
  | Bot => Just CellSet.empty
  | NotBot offs =>
    do_top offs <- offs;
      Just (BlockSet.fold
              (λ b cs,
               ZShareTree.fold (λ cs ofs _, CellSet.add (cell_from b (ofs, κ)) cs) offs cs)
              bs
              CellSet.empty)
  end.

Lemma ztree_fold_union_In f acc x (is: int_set) :
  ((∃i, In (i, tt) (ZShareTree.elements is) ∧ x = f i) ∨ CellSet.mem x acc)
  ↔
  CellSet.mem x (ZShareTree.fold (λ cs i _, CellSet.add (f i) cs) is acc).
Proof.
  rewrite ZShareTree.fold_spec.
  generalize (ZShareTree.elements is). clear is.
  induction l using rev_ind.
  simpl. split. now intros [(i & ()&_)|H]. intuition.
  rewrite fold_left_app.
  simpl.
  destruct IHl as [A B]. split.
  intros [(? & H & ->)|H].
  rewrite in_app in H. destruct H as [H|[->|()]].
  specialize (A (or_introl (ex_intro _ _ (conj H eq_refl)))).
  apply CellSet.mem_add; auto.
  apply CellSet.mem_add; auto.
  apply CellSet.mem_add; auto.
  intros H; apply CellSet.mem_add in H; destruct H as [H|H].
  destruct (B H) as [(i & K & ->)|K].
  left. eexists. rewrite in_app. intuition; eauto.
  auto.
  left. eexists. intuition eauto.
  rewrite in_app. right. now destruct x0 as (? & ()); left.
Qed.

Lemma set_product_correct :
  forall bs κ offs b i,
    BlockSet.mem b bs ->
    i ∈ γ offs ->
    match set_product bs κ offs with
    | All => True
    | Just res => CellSet.mem (cell_from b (proj i, κ)) res
    end.
Proof.
  unfold set_product.
  intros bs κ [|[|offs]]. intuition. intros; exact I.
  simpl. intros b i.
  rewrite BlockSet.fold_spec.
  rewrite BlockSet.elements_spec.
  generalize (BlockSet.elements bs). clear bs.

  intros l H IN.
  apply ZShareTree.elements_correct in IN.
  induction l. inversion H.
  destruct H as [->|H].
  simpl. apply ztree_fold_union_In. left. exists (proj i). intuition.
  simpl. apply ztree_fold_union_In. right. now apply IHl.
Qed.

Instance hash_cell : hashable cell := fun (h:N) (c:cell) =>
  match c with
    | AClocal x y z t =>
      hash (hash (hash (hash (MIX h 0) x) y) z) t
    | ACtemp x y =>
      hash (hash (MIX h 1) x) y
    | ACglobal x y z =>
      hash (hash (hash (MIX h 2) (Npos x)) y) z
  end.
