(* *********************************************************************)
(*                                                                     *)
(*              The Verasco static analyzer                            *)
(*                                                                     *)
(*          Jacques-Henri Jourdan, INRIA Paris-Rocquencourt            *)
(*                                                                     *)
(*  Copyright Institut National de Recherche en Informatique et en     *)
(*  Automatique and Université de Rennes 1.  All rights reserved.      *)
(*  This file is distributed under the terms of the GNU General        *)
(*  Public License as published by the Free Software Foundatio,        *)
(*  either version 2 of the License, or (at your option) any later     *)
(*  version.  This file is also distributed under the terms of the     *)
(*  INRIA Non-Commercial License Agreement.                            *)
(*                                                                     *)
(* *********************************************************************)

Require Import Coqlib.
Require Import Maps.
Require Import Integers.
Require Import Values.
Require Import Memory.
Require Import Utf8.
Require Import AST.
Require Import Events.
Require Import Globalenvs.
Require Import Smallstep.
Require Import Behaviors.
Require Import Csharpminor.
Require Import Switch.

Lemma eval_expr_det:
  ∀ ge e t m exp,
    ∀ v, eval_expr ge e t m exp v →
    ∀ v', eval_expr ge e t m exp v' →
    v = v'.
Proof.
  induction 1; intros.
  - inv H0. congruence.
  - inv H0. destruct H, H2; congruence.
  - inv H0. congruence.
  - inv H1. rewrite (IHeval_expr _ H4) in H0. congruence.
  - inv H2. rewrite (IHeval_expr1 _ H6) in H1. rewrite (IHeval_expr2 _ H8) in H1. congruence.
  - inv H1. rewrite (IHeval_expr _ H4) in H0. congruence.
Qed.

Lemma eval_exprlist_det:
  ∀ ge e t m expl,
    ∀ vl, eval_exprlist ge e t m expl vl →
    ∀ vl', eval_exprlist ge e t m expl vl' →
    vl = vl'.
Proof.
  induction 1; intros.
  inv H. auto.
  inv H1. f_equal. eapply eval_expr_det; eauto. eauto.
Qed.

Lemma alloc_variables_det:
  ∀ e1 m1 vars e2 m2 e2' m2',
    alloc_variables e1 m1 vars e2 m2 ->
    alloc_variables e1 m1 vars e2' m2' ->
    e2 = e2' /\ m2 = m2'.
Proof.
  induction 1.
  intro. inv H. auto.
  intro. inv H1.
  rewrite H in H9. inv H9. auto.
Qed.

Lemma bool_of_val_det:
  ∀ v b b',
    Val.bool_of_val v b ->
    Val.bool_of_val v b' ->
    b = b'.
Proof.
  intros. destruct H; inv H0; auto.
Qed.

Lemma switch_argument_det:
  ∀ b v n n',
    switch_argument b v n ->
    switch_argument b v n' ->
    n = n'.
Proof.
  intros. destruct H; inv H0; auto.
Qed.

Ltac determ :=
  repeat match goal with
           | H : eval_expr ?ge ?e ?t ?m ?exp ?v,
             H' : eval_expr ?ge ?e ?t ?m ?exp ?v' |- _ =>
             let EQ := fresh "EQ" in
             assert (EQ:=eval_expr_det _ _ _ _ _ _ H _ H');
               try (subst v'); clear H'
           | H : eval_exprlist ?ge ?e ?t ?m ?expl ?vl,
             H' : eval_exprlist ?ge ?e ?t ?m ?expl ?vl' |- _ =>
             let EQ := fresh "EQ" in
             assert (EQ:=eval_exprlist_det _ _ _ _ _ _ H _ H');
               subst vl'; clear H'
           | H : alloc_variables ?e1 ?m1 ?vars ?e2 ?m2,
             H' : alloc_variables ?e1 ?m1 ?vars ?e2' ?m2' |- _ =>
             let EQ := fresh "EQ" in
             let EQ1 := fresh "EQ1" in
             let EQ2 := fresh "EQ2" in
             assert (EQ:=alloc_variables_det _ _ _ _ _ _ _ H H');
               destruct EQ as [EQ1 EQ2];
               subst e2' m2'; clear H'
           | H : Val.bool_of_val ?v ?b,
             H' : Val.bool_of_val ?v ?b' |- _ =>
             let EQ := fresh "EQ" in
             assert (EQ:=bool_of_val_det _ _ _ H H');
               subst b; clear H'
           | H : switch_argument ?b ?v ?n,
             H' : switch_argument ?b ?v ?n' |- _ =>
             let EQ := fresh "EQ" in
             assert (EQ:=switch_argument_det _ _ _ _ H H');
               subst n; clear H'
           | H : ?A = Some ?x,
             H' : ?A = Some ?x' |- _ =>
             let EQ := fresh "EQ" in
             assert (EQ:x = x') by congruence;
               subst x; clear H'
         end.

Section PROG.

Variable (prog: program).
Let ge := Genv.globalenv prog.

Definition stack := list (function * env * temp_env).

CoInductive hoare_stmt:
  ∀ (f:function) (env:env) (stk:stack)
    (pre_normal:temp_env -> mem -> Prop)
    (pre_label:ident -> temp_env -> mem -> Prop)
    (stmt:stmt)
    (post_normal:temp_env -> mem -> Prop)
    (post_return:val -> mem -> Prop)
    (post_exit:nat -> temp_env -> mem -> Prop)
    (post_goto:ident -> temp_env -> mem -> Prop), Prop :=

| hoare_Sskip:
  ∀ f env stk pre_normal pre_label
    post_normal post_return post_exit post_goto,
    (∀ le mem, (pre_normal le mem:Prop) -> (post_normal le mem:Prop)) ->
    hoare_stmt f env stk
               pre_normal pre_label
               Sskip
               post_normal post_return post_exit post_goto

| hoare_Sset:
  ∀ f env stk id a pre_normal pre_label
    post_normal post_return post_exit post_goto,
    (∀ le mem, (pre_normal le mem:Prop) ->
               ∃ v, eval_expr ge env le mem a v /\
                    post_normal (PTree.set id v le) mem) ->
    hoare_stmt f env stk
               pre_normal pre_label
               (Sset id a)
               post_normal post_return post_exit post_goto

| hoare_Sstore:
  ∀ f env stk chunk addr a pre_normal pre_label
    post_normal post_return post_exit post_goto,
    (∀ le mem, (pre_normal le mem:Prop) ->
               ∃ vaddr v mem',
                 eval_expr ge env le mem addr vaddr /\
                 eval_expr ge env le mem a v /\
                 Mem.storev chunk mem vaddr v = Some mem' /\
                 post_normal le mem') ->
    hoare_stmt f env stk
               pre_normal pre_label
               (Sstore chunk addr a)
               post_normal post_return post_exit post_goto

| hoare_Scall:
  ∀ f env stk optid sig a bl pre_normal pre_label
    post_normal post_return post_exit post_goto,
    (∀ le mem, (pre_normal le mem:Prop) ->
               ∃ vf vargs fd,
                 eval_expr ge env le mem a vf /\
                 eval_exprlist ge env le mem bl vargs /\
                 Genv.find_funct ge vf = Some fd /\
                 funsig fd = sig /\
                 hoare_fun ((f, env, le)::stk)
                           (fun vargs' mem' => vargs = vargs' /\ mem = mem')
                           fd
                           (fun v mem' =>
                              post_normal (Cminor.set_optvar optid v le) mem')) ->
    hoare_stmt f env stk
               pre_normal pre_label
               (Scall optid sig a bl)
               post_normal post_return post_exit post_goto

| hoare_Sbuiltin:
  ∀ f env stk optid ef bl pre_normal pre_label
    post_normal post_return post_exit post_goto,
    (∀ le mem, (pre_normal le mem:Prop) ->
               ∃ vargs ,
                 eval_exprlist ge env le mem bl vargs /\
                 (∃ t v mem', external_call ef ge vargs mem t v mem') /\
                 ∀ t v mem', external_call ef ge vargs mem t v mem' ->
                   (post_normal (Cminor.set_optvar optid v le) mem':Prop)) ->
    hoare_stmt f env stk
               pre_normal pre_label
               (Sbuiltin optid ef bl)
               post_normal post_return post_exit post_goto

| hoare_Sseq:
  ∀ f env stk s1 s2,
  ∀ pre_normal pre_label P
    post_normal post_return post_exit post_goto,
    hoare_stmt f env stk
               pre_normal pre_label
               s1
               P post_return post_exit post_goto ->
    hoare_stmt f env stk
               P pre_label
               s2
               post_normal post_return post_exit post_goto ->
    hoare_stmt f env stk
               pre_normal pre_label
               (Sseq s1 s2)
               post_normal post_return post_exit post_goto

| hoare_Sifthenelse:
  ∀ f env stk a s1 s2,
  ∀ pre_normal1 pre_normal2 pre_normal pre_label
    post_normal post_return post_exit post_goto,
    (∀ le mem, (pre_normal le mem:Prop) ->
               ∃ v b, eval_expr ge env le mem a v /\
                      Val.bool_of_val v b /\
                      if b then pre_normal1 le mem else pre_normal2 le mem) ->
    hoare_stmt f env stk
               pre_normal1 pre_label
               s1
               post_normal post_return post_exit post_goto ->
    hoare_stmt f env stk
               pre_normal2 pre_label
               s2
               post_normal post_return post_exit post_goto ->
    hoare_stmt f env stk
               pre_normal pre_label
               (Sifthenelse a s1 s2)
               post_normal post_return post_exit post_goto

| hoare_Sloop:
  ∀ f env stk s,
  ∀ inv pre_normal pre_label
    post_normal post_return post_exit post_goto,
    (∀ le mem, (pre_normal le mem:Prop) -> (inv le mem:Prop)) ->
    hoare_stmt f env stk
               inv pre_label
               s
               inv post_return post_exit post_goto ->
    hoare_stmt f env stk
               pre_normal pre_label
               (Sloop s)
               post_normal post_return post_exit post_goto

| hoare_Sblock:
  ∀ f env stk s,
  ∀ pre_normal pre_label
    post_normal post_return post_exit post_goto,
    hoare_stmt f env stk
               pre_normal pre_label
               s
               post_normal post_return
               (fun n => match n with
                           | S n => post_exit n
                           | O => post_normal
                         end) post_goto ->
    hoare_stmt f env stk
               pre_normal pre_label
               (Sblock s)
               post_normal post_return post_exit post_goto

| hoare_Sexit:
  ∀ f env stk x,
  ∀ pre_normal pre_label
    post_normal post_return post_exit post_goto,
    (∀ le mem, (pre_normal le mem:Prop) -> (post_exit x le mem:Prop)) ->
    hoare_stmt f env stk
               pre_normal pre_label
               (Sexit x)
               post_normal post_return post_exit post_goto

| hoare_Sswitch_0:
    ∀ f env stk islong a cases,
    ∀ pre_normal pre_label
      post_normal post_return post_exit post_goto,
      (∀ le mem, (pre_normal le mem:Prop) ->
                 ∃ n v, eval_expr ge env le mem a v /\
                        switch_argument islong v n /\
                        hoare_stmt f env stk (fun le' mem' => le' = le /\ mem' = mem)
                                   (fun _ _ _ => False)
                                   (seq_of_lbl_stmt (select_switch n cases))
                                   post_normal post_return post_exit post_goto) ->
      hoare_stmt f env stk (fun _ _ => False) pre_label
                 (seq_of_lbl_stmt cases)
                 post_normal post_return post_exit post_goto ->
      hoare_stmt f env stk
               pre_normal pre_label
               (Sswitch islong a cases)
               post_normal post_return post_exit post_goto

| hoare_Sreturn_0:
  ∀ f env stk,
  ∀ pre_normal pre_label
    post_normal post_return post_exit post_goto,
    (∀ le mem, (pre_normal le mem:Prop) ->
               ∃ mem', Mem.free_list mem (blocks_of_env env) = Some mem' /\
                       post_return Vundef mem') ->
    hoare_stmt f env stk
               pre_normal pre_label
               (Sreturn None)
               post_normal post_return post_exit post_goto

| hoare_Sreturn_1:
  ∀ f env stk a,
  ∀ pre_normal pre_label
    post_normal post_return post_exit post_goto,
    (∀ le mem, (pre_normal le mem:Prop) ->
               ∃ v mem',
                 eval_expr ge env le mem a v /\
                 Mem.free_list mem (blocks_of_env env) = Some mem' /\
                 post_return v mem') ->
    hoare_stmt f env stk
               pre_normal pre_label
               (Sreturn (Some a))
               post_normal post_return post_exit post_goto

| hoare_Slabel:
  ∀ f env stk lbl s,
  ∀ pre_normal pre_label
    post_normal post_return post_exit post_goto,
    hoare_stmt f env stk
               (fun le mem => pre_label lbl le mem \/ pre_normal le mem) pre_label
               s
               post_normal post_return post_exit post_goto ->
    hoare_stmt f env stk
               pre_normal pre_label
               (Slabel lbl s)
               post_normal post_return post_exit post_goto

| hoare_Sgoto:
  ∀ f env stk lbl,
  ∀ pre_normal pre_label
    post_normal post_return post_exit post_goto,
    (∀ le mem, (pre_normal le mem:Prop) -> (post_goto lbl le mem:Prop)) ->
    hoare_stmt f env stk
               pre_normal pre_label
               (Sgoto lbl)
               post_normal post_return post_exit post_goto

with hoare_fun :
       ∀ (stk:stack)
         (pre:list val -> mem -> Prop)
         (f:fundef)
         (post:val -> mem -> Prop), Prop :=

| hoare_Internal:
  ∀ f stk pre post,
    (∀ vargs mem, (pre vargs mem:Prop) ->
     list_norepet (map fst (fn_vars f)) /\
     list_norepet (fn_params f) /\
     list_disjoint (fn_params f) (fn_temps f) /\
     ∃ mem1 env le,
       alloc_variables empty_env mem (fn_vars f) env mem1 /\
       bind_parameters (fn_params f) vargs
                       (create_undef_temps (fn_temps f)) = Some le /\
       ∃ inv,
         (∀ lbl le mem, (inv lbl le mem:Prop) -> ∀ k, find_label lbl (fn_body f) k <> None) /\
         hoare_stmt f env stk
                    (fun le' mem' => le' = le /\ mem' = mem1) inv
                    (fn_body f)
                    (fun le mem =>
                       ∃ mem',
                         Mem.free_list mem (blocks_of_env env) = Some mem' /\
                         post Vundef mem') post
                    (fun _ _ _ => False) inv) ->
    hoare_fun stk
              pre
              (Internal f)
              post
| hoare_External:
  ∀ ef stk pre post,
    (∀ vargs mem, (pre vargs mem:Prop) ->
                  (∃ t v mem', external_call ef ge vargs mem t v mem') /\
                  ∀ t v mem', external_call ef ge vargs mem t v mem' ->
                              (post v mem':Prop)) ->
    hoare_fun stk
              pre
              (External ef)
              post.

Lemma hoare_stmt_pre_impl:
  ∀ f env stk
    pre_normal pre_label post_normal post_return post_exit post_goto
    pre_normal' pre_label',
    (∀ le mem, (pre_normal' le mem:Prop) -> (pre_normal le mem:Prop)) ->
    (∀ lbl le mem, (pre_label' lbl le mem:Prop) -> (pre_label lbl le mem:Prop)) ->
  ∀ stmt,
    hoare_stmt f env stk pre_normal pre_label stmt post_normal post_return post_exit post_goto ->
    hoare_stmt f env stk pre_normal' pre_label' stmt post_normal post_return post_exit post_goto

with hoare_fun_pre_impl:
  ∀ stk
    pre pre' post,
    (∀ args mem, (pre' args mem:Prop) -> (pre args mem:Prop)) ->
  ∀ f,
    hoare_fun stk pre f post ->
    hoare_fun stk pre' f post.
Proof.
  - intros. destruct H1; try now (econstructor; eauto).
    + eapply hoare_Sloop with (inv:=inv). eauto. eauto.
    + apply hoare_Sswitch_0. intros.
      edestruct H1 as (n & Hn & HOARE). eauto. eauto.
      eapply hoare_stmt_pre_impl. contradiction. eauto. eauto.
    + eapply hoare_Slabel.
      eapply hoare_stmt_pre_impl. 3:apply H1. intuition. auto.
  - intros. destruct H0; econstructor; eauto.
Qed.

Lemma hoare_stmt_post_impl:
  ∀ f env stk
    pre_normal pre_label post_normal post_return post_exit post_goto
    post_normal' post_return' post_exit' post_goto',
    (∀ le mem, (post_normal le mem:Prop) -> (post_normal' le mem:Prop)) ->
    (∀ v mem, (post_return v mem:Prop) -> (post_return' v mem:Prop)) ->
    (∀ x le mem, (post_exit x le mem:Prop) -> (post_exit' x le mem:Prop)) ->
    (∀ lbl le mem, (post_goto lbl le mem:Prop) -> (post_goto' lbl le mem:Prop)) ->
  ∀ stmt,
    hoare_stmt f env stk pre_normal pre_label stmt post_normal post_return post_exit post_goto ->
    hoare_stmt f env stk pre_normal pre_label stmt post_normal' post_return' post_exit' post_goto'
with hoare_fun_post_impl:
  ∀ stk
    pre post post',
    (∀ val mem, (post val mem:Prop) -> (post' val mem:Prop)) ->
  ∀ f,
    hoare_fun stk pre f post ->
    hoare_fun stk pre f post'.
Proof.
  - intros. destruct H3; try now (econstructor; eauto).
    + econstructor. firstorder.
    + econstructor. intros. edestruct H3 as (vaddr & v & mem' & A & B & C & D); eauto 10.
    + econstructor. intros. edestruct H3 as (vf & vargs & fd & A & B & C & D & E); eauto.
      eexists. eexists. eexists. split. eauto. split. eauto.
      split. eauto. split. auto.
      eapply hoare_fun_post_impl; eauto. eauto.
    + econstructor. intros. edestruct H3 as (vargs & A & (t & v & mem' & B) & C); eauto.
      eauto 10.
    + econstructor. eapply hoare_stmt_post_impl; eauto.
      destruct x; eauto.
    + constructor. intros.
      edestruct H3 as (n & v & A & B & C); eauto 10.
      eapply hoare_stmt_post_impl; eauto.
    + econstructor. intros. edestruct H3 as (mem' & A & B); eauto.
    + econstructor. intros. edestruct H3 as (v & mem' & A & B & C); eauto 10.
  - intros. destruct H0.
    + econstructor; eauto. intros.
      edestruct H0 as (A & B & C & mem1 & env & le & D & E & inv & F & G); eauto.
      split. auto. split. auto. split. auto. eexists. eexists. eexists.
      split. eauto. split. eauto. exists inv. split. auto.
      eapply hoare_stmt_post_impl, G.
      intros. edestruct H2 as (mem' & I & J). eauto.
      auto. auto. auto.
    + econstructor. intros.
      edestruct H0 as ((t & v & mem' & A) & B); eauto 10.
Qed.

Scheme stmt_ind_2 := Induction for stmt Sort Prop
  with lbl_stmt_ind_2  := Induction for lbl_stmt Sort Prop.

Lemma hoare_stmt_wp:
  ∀ stmt f env stk post_normal post_return post_exit post_goto,
    hoare_stmt f env stk
               (fun le mem => ∃ P, P le mem /\
                              hoare_stmt f env stk P (fun _ _ _ => False) stmt
                                         post_normal post_return post_exit post_goto)
               (fun lbl le mem => ∃ P, P lbl le mem /\
                                       hoare_stmt f env stk  (fun _ _ => False) P stmt
                                                  post_normal post_return post_exit post_goto)
               stmt post_normal post_return post_exit post_goto.
Proof.
  induction stmt using stmt_ind_2
  with (P0 := fun l => ∀ f env stk post_normal post_return post_exit post_goto,
          hoare_stmt f env stk
                     (fun le mem =>
                        ∃ P, P le mem /\
                             hoare_stmt f env stk P (fun _ _ _ => False)
                                        (seq_of_lbl_stmt l)
                                        post_normal post_return post_exit post_goto)
                     (fun lbl le mem =>
                        ∃ P, P lbl le mem /\
                             hoare_stmt f env stk (fun _ _ => False) P
                                        (seq_of_lbl_stmt l)
                                        post_normal post_return post_exit post_goto)
                     (seq_of_lbl_stmt l) post_normal post_return post_exit post_goto).
  - constructor. intros le mem (P & HP & HPP). inv HPP. auto.
  - constructor. intros le mem (P & HP & HPP). inv HPP. auto.
  - constructor. intros le mem (P & HP & HPP). inv HPP. auto.
  - constructor. intros le mem (P & HP & HPP). inv HPP. auto.
  - constructor. intros le mem (P & HP & HPP). inv HPP. auto.
  - econstructor.
    eapply hoare_stmt_pre_impl, IHstmt1.
    3:eapply hoare_stmt_pre_impl, IHstmt2. 3:exact (fun le mem H => H).
    + intros le mem (P & HP & HPP). inv HPP.
      exists P. split. auto.
      eapply hoare_stmt_post_impl. 5:apply H6.
      eauto. auto. auto. auto.
    + simpl. intros lbl le mem (P & HP & HPP). inv HPP.
      exists P. split. auto.
      eapply hoare_stmt_post_impl. 5:apply H6.
      exists P0. split. auto.
      eapply hoare_stmt_pre_impl. 3:apply H11.
      auto. destruct 1. auto. auto. auto.
    + simpl. intros lbl le mem (P & HP & HPP). inv HPP.
      exists P. split. auto.
      eapply hoare_stmt_pre_impl. 3:apply H11.
      destruct 1. auto.
  - econstructor.
    2:eapply hoare_stmt_pre_impl, IHstmt1. 2:exact (fun le mem H => H).
    3:eapply hoare_stmt_pre_impl, IHstmt2. 3:exact (fun le mem H => H).
    + intros le mem (P & HP & HPP). inv HPP.
      specialize (H7 _ _ HP). destruct H7 as (v & b & EVAL & Hvb & Hb).
      exists v. exists b. destruct b; eauto.
    + intros lbl le mem (P & HP & HPP). inv HPP.
      exists P. split. auto. eapply hoare_stmt_pre_impl; eauto. destruct 1.
    + intros lbl le mem (P & HP & HPP). inv HPP.
      exists P. split. auto. eapply hoare_stmt_pre_impl; eauto. destruct 1.
  - intros.
    apply hoare_Sloop with
    (inv := fun le mem => ∃ P, P le mem /\
                               hoare_stmt f env stk P (fun _ _ _ => False)
                                          stmt P post_return post_exit post_goto).
    2:eapply hoare_stmt_pre_impl, IHstmt.
    + intros le mem (P & HP & HPP). inv HPP.
      exists inv. eauto.
    + intros le mem (P & HP & HPP).
      exists P. split. auto. eapply hoare_stmt_post_impl. 5:apply HPP.
      eauto. auto. auto. auto.
    + intros lbl le mem (P & HP & HPP). inv HPP.
      exists P. split. auto.
      eapply hoare_stmt_post_impl, hoare_stmt_pre_impl. 7:apply H6.
      exists inv. split. auto.
      eapply hoare_stmt_pre_impl. 3:apply H6. auto. destruct 1. auto. auto. auto.
      destruct 1. auto.
  - constructor. eapply hoare_stmt_pre_impl, IHstmt.
    + intros le mem (P & HP & HPP). inv HPP. eauto.
    + intros lbl le mem (P & HP & HPP). inv HPP. eauto.
  - constructor. intros le mem (P & HP & HPP). inv HPP. auto.
  - constructor.
    + intros le mem (P & HP & HPP). inv HPP. eauto.
    + eapply hoare_stmt_pre_impl, IHstmt. contradiction.
      intros lbl le mem (P & HP & HPP). inv HPP. eauto.
  - destruct o; constructor.
    + intros le mem (P & HP & HPP). inv HPP. eauto.
    + intros le mem (P & HP & HPP). inv HPP. eauto.
  - constructor. eapply hoare_stmt_pre_impl, IHstmt.
    + intros le mem [(P & HP & HPP)|(P & HP & HPP)]; inv HPP.
      exists (P l). split. auto.
      eapply hoare_stmt_pre_impl, H10. simpl; auto. destruct 1.
      exists P. split. auto.
      eapply hoare_stmt_pre_impl, H10. simpl; auto. destruct 1.
    + intros lbl le mem (P & HP & HPP). inv HPP.
      exists P. split. auto.
      eapply hoare_stmt_pre_impl, H10. destruct 1. auto.
  - constructor. intros le mem (P & HP & HPP). inv HPP. auto.
  - constructor. intros le mem (P & HP & HPP). inv HPP. auto.
  - simpl. econstructor. eapply hoare_stmt_pre_impl, IHstmt.
    3:eapply hoare_stmt_pre_impl, IHstmt0. 3:exact (fun le mem H => H).
    + intros le mem (P & HP & HPP). inv HPP.
      exists P. split. auto.
      eapply hoare_stmt_post_impl. 5:apply H6.
      eauto. auto. auto. auto.
    + simpl. intros lbl le mem (P & HP & HPP). inv HPP.
      exists P. split. auto.
      eapply hoare_stmt_post_impl. 5:apply H6.
      exists P0. split. auto.
      eapply hoare_stmt_pre_impl. 3:apply H11.
      auto. destruct 1. auto. auto. auto.
    + simpl. intros lbl le mem (P & HP & HPP). inv HPP.
      exists P. split. auto.
      eapply hoare_stmt_pre_impl. 3:apply H11.
      destruct 1. auto.
Qed.

Lemma hoare_fun_wp:
  ∀ f stk post,
    hoare_fun stk (fun args mem => ∃ P, P args mem /\ hoare_fun stk P f post) f post.
Proof.
  destruct f; intros.
  - constructor.
    intros vargs mem (P & HP & HPP). inv HPP.
    specialize (H2 _ _ HP).
    destruct H2 as (A & B & C & mem1 & env & le & D & E & inv & F & G).
    split. auto. split. auto. split. auto.
    exists mem1. exists env. exists le. split. auto. split. auto.
    exists inv. split. auto. auto.
  - constructor. intros vargs mem (P & HP & HPP). inv HPP. auto.
Qed.

Lemma hoare_stmt_pre_exists:
  ∀ T stmt f env stk pre_normal pre_label post_normal post_return post_exit post_goto,
    (∀ x:T, hoare_stmt f env stk (pre_normal x) (pre_label x)
                       stmt post_normal post_return post_exit post_goto) ->
    hoare_stmt f env stk (fun le mem => ∃ x, pre_normal x le mem)
               (fun lbl le mem => ∃ x, pre_label x lbl le mem)
               stmt post_normal post_return post_exit post_goto.
Proof.
  intros. eapply hoare_stmt_pre_impl, hoare_stmt_wp.
  - intros le mem (x & Hx). exists (pre_normal x).
    split. auto. eapply hoare_stmt_pre_impl, (H x). auto. destruct 1.
  - intros lbl le mem (x & Hx). exists (pre_label x).
    split. auto. eapply hoare_stmt_pre_impl, (H x). destruct 1. auto.
Qed.

Lemma hoare_stmt_pre_or:
  ∀ stmt f env stk pre_normal1 pre_normal2 pre_label1 pre_label2
    post_normal post_return post_exit post_goto,
    hoare_stmt f env stk pre_normal1 pre_label1
               stmt post_normal post_return post_exit post_goto ->
    hoare_stmt f env stk pre_normal2 pre_label2
               stmt post_normal post_return post_exit post_goto ->
    hoare_stmt f env stk (fun le mem => pre_normal1 le mem \/ pre_normal2 le mem)
               (fun lbl le mem => pre_label1 lbl le mem \/ pre_label2 lbl le mem)
               stmt post_normal post_return post_exit post_goto.
Proof.
  intros.
  eapply hoare_stmt_pre_impl, hoare_stmt_pre_exists
  with (pre_normal := fun b => if b:bool then pre_normal1 else pre_normal2)
       (pre_label := fun b => if b:bool then pre_label1 else pre_label2).
  destruct 1. exists true. auto. exists false. auto.
  destruct 1. exists true. auto. exists false. auto.
  destruct x; auto.
Qed.

Lemma hoare_fun_pre_exists:
  ∀ T f stk pre post,
    (∀ x:T, hoare_fun stk (pre x) f post) ->
    hoare_fun stk (fun vargs mem => ∃ x, pre x vargs mem) f post.
Proof.
  intros. eapply hoare_fun_pre_impl, hoare_fun_wp.
  intros args mem (x & Hx). exists (pre x). auto.
Qed.

Lemma hoare_fun_pre_or:
  ∀ f stk pre1 pre2 post,
    hoare_fun stk pre1 f post ->
    hoare_fun stk pre2 f post ->
    hoare_fun stk (fun vargs mem => pre1 vargs mem \/ pre2 vargs mem) f post.
Proof.
  intros.
  eapply hoare_fun_pre_impl, hoare_fun_pre_exists
  with (pre := fun b => if b:bool then pre1 else pre2).
  destruct 1. exists true. auto. exists false. auto.
  destruct x; auto.
Qed.

Lemma hoare_stmt_sp_aux:
  ∀ stmt f env stk pre_normal pre_label post_normal0 post_return0 post_exit0 post_goto0,
    hoare_stmt f env stk pre_normal pre_label stmt post_normal0 post_return0 post_exit0 post_goto0 ->
  ∀ post_normal post_return post_exit post_goto,
    (∀ le mem, (∀ P, hoare_stmt f env stk pre_normal pre_label stmt
                               P (fun _ _ => True)
                               (fun _ _ _ => True) (fun _ _ _ => True) ->
                     P le mem) ->
               (post_normal le mem:Prop)) ->
    (∀ v mem,  (∀ P, hoare_stmt f env stk pre_normal pre_label stmt
                               (fun _ _ => True) P
                               (fun _ _ _ => True) (fun _ _ _ => True) ->
                     P v mem) ->
               (post_return v mem:Prop)) ->
    (∀ x le mem, (∀ P, hoare_stmt f env stk pre_normal pre_label stmt
                                 (fun _ _ => True) (fun _ _ => True)
                                 P (fun _ _ _ => True) ->
                       P x le mem) ->
                 (post_exit x le mem:Prop)) ->
    (∀ lbl le mem, (∀ P, hoare_stmt f env stk pre_normal pre_label stmt
                                    (fun _ _ => True) (fun _ _ => True)
                                    (fun _ _ _ => True) P ->
                         P lbl le mem) ->
                   (post_goto lbl le mem:Prop)) ->
    hoare_stmt f env stk pre_normal pre_label stmt
               post_normal post_return post_exit post_goto
with hoare_fun_sp_aux:
  ∀ f stk pre post0, hoare_fun stk pre f post0 ->
  ∀ post,
    (∀ v mem, (∀ P, hoare_fun stk pre f P -> P v mem) ->
              (post v mem:Prop)) ->
    hoare_fun stk pre f post.
Proof.
  - destruct stmt; intros * H; inv H; intros.
    + constructor. intros le mem PRE. apply H. intros. inv H4. auto.
    + constructor. intros le mem PRE.
      edestruct H11 as (v & EVAL & POST); eauto.
      exists v. split. auto.
      apply H. intros. inv H3.
      edestruct H16 as (v' & EVAL' & POST'); eauto.
      determ. auto.
    + constructor. intros le mem PRE.
      edestruct H12 as (vaddr & v & mem0 & EVAL1 & EVAL2 & MEM & POST); eauto.
      exists vaddr. exists v. exists mem0. split. auto. split. auto. split. auto.
      apply H. intros. inv H3.
      edestruct H17 as (vaddr' & v' & mem0' & EVAL1' & EVAL2' & MEM' & POST'); eauto.
      determ. auto.
    + constructor. intros.
      edestruct H13 as (vf & vargs & fd & EVAL1 & EVAL2 & FD & SIG & POST); eauto.
      exists vf. exists vargs. exists fd. split. auto. split. auto. split. auto. split. auto.
      eapply hoare_fun_sp_aux. apply POST.
      intros. apply H. intros. inv H5.
      edestruct H20 as (vf' & vargs' & fd' & EVAL1' & EVAL2' & FD' & SIG' & POST'); eauto.
      determ.
      apply H4. apply POST'.
    + constructor. intros le mem PRE.
      edestruct H12 as (vargs & EVAL & (t & v & mem0 & EXT) & POST); eauto.
      exists vargs. split. auto. split. eauto.
      intros. apply H. intros. inv H4.
      edestruct H18 as (vargs' & EVAL' & (t' & v' & mem0' & EXT') & POST'); eauto.
      determ. eauto.
    + econstructor.
      * eapply (hoare_stmt_sp_aux _ _ _ _ _ _ _ _ _ _ H7).
        exact (fun _ _ H => H).
        intros. apply H0. intros. inv H4. apply H3. eapply hoare_stmt_post_impl, H14; eauto.
        intros. apply H1. intros. inv H4. apply H3. eapply hoare_stmt_post_impl, H14; eauto.
        intros. apply H2. intros. inv H4. apply H3. eapply hoare_stmt_post_impl, H14; eauto.
      * eapply hoare_stmt_sp_aux.
        eapply hoare_stmt_pre_impl, H12; auto.
        intros. eapply H3, hoare_stmt_post_impl, H7; auto.
        intros. apply H. intros. inv H4. apply H3. eapply hoare_stmt_pre_impl, H19; eauto.
        intros. apply H4. apply H14.
        intros. apply H0. intros. inv H4. apply H3. eapply hoare_stmt_pre_impl, H19; eauto.
        intros. apply H4. eapply hoare_stmt_post_impl, H14; auto.
        intros. apply H1. intros. inv H4. apply H3. eapply hoare_stmt_pre_impl, H19; eauto.
        intros. apply H4. eapply hoare_stmt_post_impl, H14; auto.
        intros. apply H2. intros. inv H4. apply H3. eapply hoare_stmt_pre_impl, H19; eauto.
        intros. apply H4. eapply hoare_stmt_post_impl, H14; auto.
    + apply hoare_Sifthenelse
      with (pre_normal1 := fun le mem =>
              pre_normal le mem /\ ∃ v, eval_expr ge env le mem e v ∧ Val.bool_of_val v true)
           (pre_normal2 := fun le mem =>
              pre_normal le mem /\ ∃ v, eval_expr ge env le mem e v ∧ Val.bool_of_val v false);
      [|eapply hoare_stmt_sp_aux|eapply hoare_stmt_sp_aux].
      * intros. edestruct H8 as (v & b & EVAL & BOV & PRE); eauto.
        exists v. exists b. split. auto. split. auto. destruct b; eauto.
      * eapply hoare_stmt_pre_impl, H13; eauto.
        intros le' mem' (PRE & v & EVAL & BOV).
        edestruct H8 as (v' & b' & EVAL' & BOV' & PRE'); eauto.
        determ. auto.
      * intros. apply H. intros. inv H4. apply H3. eapply hoare_stmt_pre_impl, H21; eauto.
        intros le' mem' (PRE & v & EVAL & BOV).
        edestruct H16 as (v' & b' & EVAL' & BOV' & PRE'); eauto.
        determ. auto.
      * intros. apply H0. intros. inv H4. apply H3. eapply hoare_stmt_pre_impl, H21; eauto.
        intros le' mem' (PRE & v' & EVAL' & BOV').
        edestruct H16 as (v'' & b'' & EVAL'' & BOV'' & PRE''); eauto.
        determ. auto.
      * intros. apply H1. intros. inv H4. apply H3. eapply hoare_stmt_pre_impl, H21; eauto.
        intros le' mem' (PRE & v' & EVAL' & BOV').
        edestruct H16 as (v'' & b'' & EVAL'' & BOV'' & PRE''); eauto.
        determ. auto.
      * intros. apply H2. intros. inv H4. apply H3. eapply hoare_stmt_pre_impl, H21; eauto.
        intros le' mem' (PRE & v' & EVAL' & BOV').
        edestruct H16 as (v'' & b'' & EVAL'' & BOV'' & PRE''); eauto.
        determ. auto.
      * eapply hoare_stmt_pre_impl, H14; eauto.
        intros le' mem' (PRE & v & EVAL & BOV).
        edestruct H8 as (v' & b' & EVAL' & BOV' & PRE'); eauto.
        determ. auto.
      * intros. apply H. intros. inv H4. apply H3. eapply hoare_stmt_pre_impl, H22; eauto.
        intros le' mem' (PRE & v & EVAL & BOV).
        edestruct H16 as (v' & b' & EVAL' & BOV' & PRE'); eauto.
        determ. auto.
      * intros. apply H0. intros. inv H4. apply H3. eapply hoare_stmt_pre_impl, H22; eauto.
        intros le' mem' (PRE & v' & EVAL' & BOV').
        edestruct H16 as (v'' & b'' & EVAL'' & BOV'' & PRE''); eauto.
        determ. auto.
      * intros. apply H1. intros. inv H4. apply H3. eapply hoare_stmt_pre_impl, H22; eauto.
        intros le' mem' (PRE & v' & EVAL' & BOV').
        edestruct H16 as (v'' & b'' & EVAL'' & BOV'' & PRE''); eauto.
        determ. auto.
      * intros. apply H2. intros. inv H4. apply H3. eapply hoare_stmt_pre_impl, H22; eauto.
        intros le' mem' (PRE & v' & EVAL' & BOV').
        edestruct H16 as (v'' & b'' & EVAL'' & BOV'' & PRE''); eauto.
        determ. auto.
    + apply hoare_Sloop with
      (inv := fun le mem =>
                ∀ P, hoare_stmt f env stk P pre_label stmt
                                P (fun _ _ => True) (fun _ _ _ => True) (fun _ _ _ => True) ->
                     (∀ le' mem', pre_normal le' mem' -> P le' mem')->
                     P le mem).
      eauto.
      eapply hoare_stmt_sp_aux.
      * eapply hoare_stmt_pre_impl, H7; auto.
        intros. apply H4; auto. eapply hoare_stmt_post_impl, H7; auto.
      * intros. apply H4. eapply hoare_stmt_pre_impl, H5; auto. intros. apply H8; auto.
      * intros. apply H0. intros. apply H4. inv H5.
        eapply hoare_stmt_post_impl, hoare_stmt_pre_impl, H14; auto. intros. apply H5; auto.
        eapply hoare_stmt_post_impl, H14; auto.
      * intros. apply H2. intros. apply H4. inv H5.
        eapply hoare_stmt_post_impl, hoare_stmt_pre_impl, H14; auto. intros. apply H5; auto.
        eapply hoare_stmt_post_impl, H14; auto.
      * intros. apply H3. intros. apply H4. inv H5.
        eapply hoare_stmt_post_impl, hoare_stmt_pre_impl, H14; auto. intros. apply H5; auto.
        eapply hoare_stmt_post_impl, H14; auto.
    + constructor. eapply (hoare_stmt_sp_aux _ _ _ _ _ _ _ _ _ _ H6).
      * intros. apply H. intros. apply H3. inv H4. eapply hoare_stmt_post_impl, H12; eauto.
      * intros. apply H0. intros. apply H3. inv H4. eapply hoare_stmt_post_impl, H12; eauto.
      * intros []; intros.
        apply H. intros.
        apply (H3 (λ n, match n with
                          | 0%nat => P | S _ => λ _ _, True
                        end)).
        inv H4. eapply hoare_stmt_post_impl, H12; eauto.
        apply H1. intros.
        apply (H3 (λ n, match n with
                          | 0%nat => λ _ _, True | S n => P n
                        end)).
        inv H4. eapply hoare_stmt_post_impl, H12; eauto.
      * intros. apply H2. intros. apply H3. inv H4. eapply hoare_stmt_post_impl, H12; eauto.
    + constructor. intros. apply H1. intros. inv H4. auto.
    + constructor. intros.
      { edestruct H12 as (n & v & Hv & Hnv & HOARE); eauto.
        exists n. exists v. split. auto. split. auto.
        eapply hoare_stmt_sp_aux.
        - apply HOARE.
        - intros. apply H. intros. inv H5. apply H4.
          edestruct H20 as (n' & v' & Hv' & Hn'v' & HOARE'); eauto. determ. apply HOARE'.
        - intros. apply H0. intros. inv H5. apply H4.
          edestruct H20 as (n' & v' & Hv' & Hn'v' & HOARE'); eauto. determ. apply HOARE'.
        - intros. apply H1. intros. inv H5. apply H4.
          edestruct H20 as (n' & v' & Hv' & Hn'v' & HOARE'); eauto. determ. apply HOARE'.
        - intros. apply H2. intros. inv H5. apply H4.
          edestruct H20 as (n' & v' & Hv' & Hn'v' & HOARE'); eauto. determ. apply HOARE'. }
      { eapply hoare_stmt_sp_aux.
        - apply H13.
        - intros. apply H. intros. inv H4. apply H3. apply H20.
        - intros. apply H0. intros. inv H4. apply H3. apply H20.
        - intros. apply H1. intros. inv H4. apply H3. apply H20.
        - intros. apply H2. intros. inv H4. apply H3. apply H20. }
    + constructor. intros.
      edestruct H6 as (mem' & MEM' & POST'); eauto.
      exists mem'. split. auto. intros. apply H0. intros. inv H4.
      edestruct H5 as (mem'' & MEM'' & POST''); eauto.
      determ. auto.
    + constructor. intros.
      edestruct H6 as (v & mem' & EVAL & MEM' & POST'); eauto.
      exists v. exists mem'. split. auto. split. auto. apply H0. intros. inv H4.
      edestruct H12 as (v' & mem'' & EVAL' & MEM'' & POST''); eauto.
      determ. auto.
    + constructor. eapply (hoare_stmt_sp_aux _ _ _ _ _ _ _ _ _ _ H11).
      * intros. apply H. intros. apply H3. inv H4. eapply hoare_stmt_post_impl, H17; eauto.
      * intros. apply H0. intros. apply H3. inv H4. eapply hoare_stmt_post_impl, H17; eauto.
      * intros. apply H1. intros. apply H3. inv H4. eapply hoare_stmt_post_impl, H17; eauto.
      * intros. apply H2. intros. apply H3. inv H4. eapply hoare_stmt_post_impl, H17; eauto.
    + constructor. intros. apply H2. intros. inv H4. auto.
  - destruct f; intros * H; inv H; intros.
    + constructor. intros.
      edestruct H3 as (A & B & C & mem1 & env & le & D & E & inv & F & G); eauto.
      split. auto. split. auto. split. auto. exists mem1. exists env. exists le.
      split. auto. split. auto.
      exists (fun lbl le0 mem0 =>
                (∀ k, find_label lbl (fn_body f) k ≠ None) /\
                ∀ P, hoare_stmt f env stk
                                (fun le' mem' => le' = le /\ mem' = mem1)
                                P (fn_body f)
                                (fun _ _ => True) (fun _ _ => True)
                                (fun _ _ _ => True) P ->
                P lbl le0 mem0).
      split. destruct 1. auto.
      eapply hoare_stmt_sp_aux.
      * eapply hoare_stmt_pre_impl, G. auto.
        intros. destruct H1. apply (H2 inv).
        eapply hoare_stmt_post_impl, G; auto.
      * intros.
        refine (let H1':=H1 _ _ in _).
        { eapply hoare_stmt_post_impl, hoare_stmt_pre_impl, G; auto. exact (fun _ _ H => H).
          intros. destruct H2. apply (H4 inv).
          eapply hoare_stmt_post_impl, G; auto. }
        simpl in H1'. destruct H1' as (mem' & MEM' & POST0).
        exists mem'. split. auto.
        apply H. intros. inv H2.
        edestruct H7 as (A' & B' & C' & mem1' & env' & le' & D' & E' & inv' & F' & G'); eauto.
        determ.
        refine (let H1':=H1 _ _ in _).
        { eapply hoare_stmt_pre_impl, hoare_stmt_post_impl, G'; auto.
          intros. destruct H2. apply (H4 inv').
          eapply hoare_stmt_post_impl, G'; auto. exact (fun _ _ H => H). }
        destruct H1' as (mem'' & MEM'' & HP).
        determ. auto.
      * intros. apply H. intros. inv H2.
        edestruct H7 as (A' & B' & C' & mem1' & env' & le' & D' & E' & inv' & F' & G'); eauto.
        determ.
        apply H1.
        eapply hoare_stmt_pre_impl, hoare_stmt_post_impl, G'; auto.
        intros. destruct H2. apply (H4 inv').
        eapply hoare_stmt_pre_impl, hoare_stmt_post_impl, G'; auto.
      * intros. apply (H1 (fun _ _ _  => False)).
        eapply hoare_stmt_pre_impl, hoare_stmt_post_impl, G; auto.
        intros. destruct H2. apply (H4 inv).
        eapply hoare_stmt_pre_impl, hoare_stmt_post_impl, G; auto.
      * intros. split.
        eapply F. apply (H1 inv).
        eapply hoare_stmt_pre_impl, hoare_stmt_post_impl, G; auto.
        intros. destruct H2. apply (H4 inv).
        eapply hoare_stmt_pre_impl, hoare_stmt_post_impl, G; auto.
        intros. apply (H1 P).
        eapply hoare_stmt_pre_impl, H2; auto.
        intros. destruct H4. apply H5. apply H2.
    + constructor.
      intros vargs mem PRE.
      edestruct H3 as ((t & v & mem' & EXT) & A); eauto.
      split. eauto.
      intros. apply H.
      intros. inv H1.
      edestruct H6 as ((t' & v' & mem'' & EXT') & A'); eauto.
Qed.

Lemma hoare_stmt_sp:
  ∀ stmt f env stk pre_normal pre_label post_normal post_return post_exit post_goto,
    hoare_stmt f env stk pre_normal pre_label stmt post_normal post_return post_exit post_goto ->
    hoare_stmt f env stk pre_normal pre_label stmt
               (fun le mem =>
                  ∀ P, hoare_stmt f env stk pre_normal pre_label stmt
                                  P (fun _ _ => True)
                                  (fun _ _ _ => True) (fun _ _ _ => True) ->
                       P le mem)
               (fun v mem =>
                  ∀ P, hoare_stmt f env stk pre_normal pre_label stmt
                                  (fun _ _ => True) P
                                  (fun _ _ _ => True) (fun _ _ _ => True) ->
                       P v mem)
               (fun x le mem =>
                  ∀ P, hoare_stmt f env stk pre_normal pre_label stmt
                                  (fun _ _ => True) (fun _ _ => True)
                                  P (fun _ _ _ => True) ->
                       P x le mem)
               (fun lbl le mem =>
                  ∀ P, hoare_stmt f env stk pre_normal pre_label stmt
                                  (fun _ _ => True) (fun _ _ => True)
                                  (fun _ _ _ => True) P ->
                       P lbl le mem).
Proof.
  intros. eapply hoare_stmt_sp_aux; eauto.
Qed.

Lemma hoare_fun_sp:
  ∀ f stk pre post, hoare_fun stk pre f post ->
    hoare_fun stk pre f (fun v mem => ∀ P, hoare_fun stk pre f P -> P v mem).
Proof.
  intros. eapply hoare_fun_sp_aux; eauto.
Qed.

Lemma hoare_stmt_post_forall:
  ∀ T stmt f env stk pre_normal pre_label post_normal post_return post_exit post_goto,
    inhabited T ->
    (∀ x:T, hoare_stmt f env stk pre_normal pre_label
                       stmt (post_normal x) (post_return x) (post_exit x) (post_goto x)) ->
    hoare_stmt f env stk pre_normal pre_label stmt
               (fun le mem => ∀ x, post_normal x le mem)
               (fun v mem => ∀ x, post_return x v mem)
               (fun xx le mem => ∀ x, post_exit x xx le mem)
               (fun lbl le mem => ∀ x, post_goto x lbl le mem).
Proof.
  intros. destruct H.
  eapply hoare_stmt_post_impl, hoare_stmt_sp, (H0 X); instantiate; simpl.
  - intros. apply H. eapply hoare_stmt_post_impl, H0; eauto.
  - intros. apply H. eapply hoare_stmt_post_impl, H0; eauto.
  - intros. apply H. eapply hoare_stmt_post_impl, H0; eauto.
  - intros. apply H. eapply hoare_stmt_post_impl, H0; eauto.
Qed.

Lemma hoare_stmt_post_and:
  ∀ stmt f env stk pre_normal pre_label
    post_normal1 post_normal2 post_return1 post_return2 post_exit1 post_exit2 post_goto1 post_goto2,
    hoare_stmt f env stk pre_normal pre_label
               stmt post_normal1 post_return1 post_exit1 post_goto1 ->
    hoare_stmt f env stk pre_normal pre_label
               stmt post_normal2 post_return2 post_exit2 post_goto2 ->
    hoare_stmt f env stk pre_normal pre_label stmt
               (fun le mem => post_normal1 le mem /\ post_normal2 le mem)
               (fun v mem => post_return1 v mem /\ post_return2 v mem)
               (fun x le mem => post_exit1 x le mem /\ post_exit2 x le mem)
               (fun lbl le mem => post_goto1 lbl le mem /\ post_goto2 lbl le mem).
Proof.
  intros.
  eapply hoare_stmt_post_impl, hoare_stmt_post_forall
  with (post_normal := fun b => if b:bool then post_normal1 else post_normal2)
       (post_return := fun b => if b:bool then post_return1 else post_return2)
       (post_exit := fun b => if b:bool then post_exit1 else post_exit2)
       (post_goto := fun b => if b:bool then post_goto1 else post_goto2).
  split. apply (H1 true). apply (H1 false).
  split. apply (H1 true). apply (H1 false).
  split. apply (H1 true). apply (H1 false).
  split. apply (H1 true). apply (H1 false).
  exists. exact true.
  destruct x; auto.
Qed.

Lemma hoare_fun_post_forall:
  ∀ T f stk pre post,
    inhabited T ->
    (∀ x:T, hoare_fun stk pre f (post x)) ->
    hoare_fun stk pre f (fun mem v => ∀ x, post x mem v).
Proof.
  intros. destruct H.
  eapply hoare_fun_post_impl, hoare_fun_sp, (H0 X).
  intros. apply H. eapply hoare_fun_post_impl, H0; eauto.
Qed.

Lemma hoare_fun_post_and:
  ∀ f stk pre post1 post2,
    hoare_fun stk pre f post1 ->
    hoare_fun stk pre f post2 ->
    hoare_fun stk pre f (fun mem v => post1 mem v /\ post2 mem v).
Proof.
  intros.
  eapply hoare_fun_post_impl, hoare_fun_post_forall
  with (post := fun b => if b:bool then post1 else post2).
  split. apply (H1 true). apply (H1 false).
  exists. exact true.
  destruct x; auto.
Qed.

Definition safe k s :=
  ∀ tr s' k',
    starN step ge k' s tr s' ->
    Nostep (semantics prog) s' ->
    (∀ r, ¬final_state s' r) ->
    (k <= k')%nat.

Lemma safe_step:
  ∀ s k s' t,
    step ge s t s' ->
    (∀ s' t, step ge s t s' -> safe k s') ->
    safe (S k) s.
Proof.
  unfold safe. repeat intro. destruct H1.
  edestruct H2; eauto.
  eapply le_n_S; eauto.
Qed.

Ltac safe_step :=
  lazymatch goal with
    | |- safe (S ?k) ?st =>
      eapply safe_step; [now (econstructor; eauto) | ];
      let s' := fresh "s'" in
      let tr := fresh "tr" in
      let Hs' := fresh "Hs'" in
      intros s' tr Hs'; inv Hs'; try contradiction; [];
      determ
    | |- safe ?k ?st =>
      destruct k; [repeat intro; apply le_0_n|safe_step]
  end.

Inductive rel_cont_stack: cont -> stack -> Prop :=
| RCS_Kseq :
    forall k st s, rel_cont_stack k s -> rel_cont_stack (Kseq st k) s
| RCS_Kblock :
    forall k s, rel_cont_stack k s -> rel_cont_stack (Kblock k) s
| RCS_Kstop : rel_cont_stack Kstop nil
| RCS_Kcall :
    forall k s tmp f e le, rel_cont_stack k s ->
    rel_cont_stack (Kcall tmp f e le k) ((f, e, le)::s).

Definition hoare_stmt_sem (f:function) (env:env) (stk:stack)
           (pre_normal:temp_env -> mem -> Prop)
           (pre_label:ident -> temp_env -> mem -> Prop)
           (stmt:stmt)
           (post_normal:temp_env -> mem -> Prop)
           (post_return:val -> mem -> Prop)
           (post_exit:nat -> temp_env -> mem -> Prop)
           (post_goto:ident -> temp_env -> mem -> Prop) : Prop :=
  ∀ n k,
    rel_cont_stack k stk ->

    (∀ n', (n' <= n)%nat ->
     ∀ le' mem', post_normal le' mem' ->
                 safe n' (State f Sskip k env le' mem')) ->

    (∀ n', (n' < n)%nat ->
     ∀ mem' v, post_return v mem' ->
               safe n' (Returnstate v (call_cont k) mem')) ->

    (∀ n', (n' <= n)%nat ->
     ∀ le' mem' x, post_exit x le' mem' ->
                   safe n' (State f (Sexit x) k env le' mem')) ->

    (∀ n', (n' < n)%nat ->
     ∀ le' mem' lbl, post_goto lbl le' mem' ->
     ∃ s' k', find_label lbl f.(fn_body) (call_cont k) = Some (s', k') /\
              safe n' (State f s' k' env le' mem')) ->

    ∀ le mem,
      (pre_normal le mem -> safe n (State f stmt k env le mem)) /\
      (∀ lbl s' k', find_label lbl stmt k = Some (s', k') ->
                    pre_label lbl le mem ->
                    safe n (State f s' k' env le mem)).

Definition hoare_fun_sem (stk:stack)
           (pre:list val -> mem -> Prop)
           (f:fundef)
           (post:val -> mem -> Prop) :=
  ∀ n k,
    rel_cont_stack k stk ->
    is_call_cont k ->

    (∀ n', (n' <= n)%nat ->
     ∀ mem' v, post v mem' ->
               safe n' (Returnstate v k mem')) ->

    ∀ args mem,
      pre args mem -> safe n (Callstate f args k mem).

Lemma hoare_correct_aux:
  (∀ f env stk stmt pre_normal pre_label post_normal post_return post_exit post_goto,
  hoare_stmt f env stk pre_normal pre_label stmt
             post_normal post_return post_exit post_goto ->
  hoare_stmt_sem f env stk pre_normal pre_label stmt
             post_normal post_return post_exit post_goto) /\
  (∀ stk pre f post,
  hoare_fun stk pre f post ->
  hoare_fun_sem stk pre f post).
Proof.
  unfold hoare_stmt_sem, hoare_fun_sem.
  refine (let H := (_: (∀ n:nat,
                        (∀ f env stk stmt pre_normal pre_label post_normal post_return post_exit post_goto, _:Prop) /\
                        (∀ stk pre f post, _:Prop))) in
          conj (fun f env stk stmt pre_normal pre_label
                    post_normal post_return post_exit post_goto H0 n =>
                  match H n with
                    | conj H _ => H f env stk stmt pre_normal pre_label
                                    post_normal post_return post_exit post_goto H0
                  end)
               (fun stk pre f post H0 n =>
                  match H n with
                    | conj _ H => H stk pre f post H0
                  end)).
  induction n using lt_wf_ind. rename H into IND.
  split; [intros f env stk; induction stmt using stmt_ind_2
                                       with (P0 := fun l =>
    ∀ pre_label post_normal post_return post_exit post_goto,
      hoare_stmt f env stk (fun _ _ => False) pre_label (seq_of_lbl_stmt l)
                 post_normal post_return post_exit post_goto ->
      ∀ k, rel_cont_stack k stk ->
        (∀ n', n' ≤ n ->
         ∀ le' mem', post_normal le' mem' -> safe n' (State f Sskip k env le' mem')) ->
        (∀ n', (n' < n)%nat ->
         ∀ mem' v, post_return v mem' -> safe n' (Returnstate v (call_cont k) mem')) ->
        (∀ n', n' ≤ n ->
         ∀ le' mem' x, post_exit x le' mem' -> safe n' (State f (Sexit x) k env le' mem')) ->
        (∀ n', (n' < n)%nat ->
         ∀ le' mem' lbl, post_goto lbl le' mem' ->
           ∃ s' k', find_label lbl (fn_body f) (call_cont k) =  Some (s', k') /\
                    safe n' (State f s' k' env le' mem')) ->
        ∀ le mem lbl s' k', find_label_ls lbl l k = Some (s', k') ->
                            pre_label lbl le mem ->
                            safe n (State f s' k' env le mem)); intros |
          intros; destruct x].
  - (* Sskip *)
    inv x. split; intros. 2:discriminate. auto.
  - (* Sset *)
    inv x. split; intros. 2:discriminate.
    edestruct H15 as (v & Hv & POST); eauto. safe_step. auto.
  - (* Sstore *)
    inv x. split; intros. 2:discriminate.
    edestruct H16 as (vaddr & v & mem' & Hvaddr & Hv & MEM' & POST); eauto. safe_step. auto.
  - (* Scall *)
    inv x. split; intros. 2:discriminate.
    edestruct H17 as (vf & vargs & fd & Hvf & Hvargs & Hfd & SIG & POST); eauto. safe_step.
    eapply IND. auto. apply POST. constructor. auto. exact I. 2:simpl; auto.
    simpl. intros. safe_step. apply H0. omega. auto.
  - (* Sbuitin *)
    inv x. split; intros. 2:discriminate.
    edestruct H16 as (vargs & Hvargs & (t & v & mem' & EXT) & POST); eauto. safe_step. eauto.
  - (* Sseq *)
    inv x. split; intros.
    + safe_step. edestruct (proj1 (IND n (lt_n_Sn n))) as [IND' _]; [..|apply IND']; eauto.
      constructor. auto.
      intros. safe_step.
      assert (n' < S n)%nat by omega.
      edestruct (proj1 (IND n' H7)) as [IND' _]; [..|apply IND']; eauto.
      intros. apply H0. omega. auto.
      intros. apply H1. omega. auto.
      intros. apply H2. omega. auto.
      intros. apply H3. omega. auto.
      intros. safe_step. apply H2. omega. auto.
    + simpl in H4. destruct (find_label lbl stmt1 (Kseq stmt2 k)) eqn:?.
      * inv H4.
        eapply IHstmt1. 7:apply Heqo. eauto. constructor. auto.
        intros. safe_step.
        assert (n' < n)%nat by omega.
        edestruct (proj1 (IND n' H7)) as [IND' _]; [..|apply IND']; eauto.
        intros. apply H0. omega. auto.
        intros. apply H1. omega. auto.
        intros. apply H2. omega. auto.
        intros. apply H3. omega. auto.
        eauto.
        intros. safe_step. apply H2. omega. auto.
        auto.
        auto.
      * eapply IHstmt2; eauto.
  - (* Sifthenelse *)
    inv x. split; intros.
    + edestruct H12 as (v & b & Hv & Hvb & PRE); eauto. safe_step.
      destruct b0; (edestruct (proj1 (IND n (lt_n_Sn n))) as [IND' _]; [..|apply IND']); eauto.
    + simpl in H4. destruct (find_label lbl stmt1 k) eqn:?.
      * inv H4. eapply IHstmt1; eauto.
      * eapply IHstmt2; eauto.
  - (* Sloop *)
    inv x. split; intros.
    + safe_step. edestruct (proj1 (IND n (lt_n_Sn n))) as [IND' _]; [..|apply IND']; eauto.
      constructor. auto.
      intros. safe_step.
      assert (n' < S n)%nat by omega.
      edestruct (proj1 (IND n' H8)) as [IND' _]; [..|apply IND']; eauto.
      econstructor; eauto.
      instantiate (1:=fun _ _ => False). contradiction.
      intros. eapply H1. omega. auto.
      intros. eapply H2. omega. auto.
      intros. eapply H3. omega. auto.
      intros. safe_step. eapply H2. omega. auto.
      auto.
    + simpl in H4.
      eapply IHstmt. 7:apply H4. eauto. constructor. auto.
      intros. safe_step.
      assert (n' < n)%nat by omega.
      edestruct (proj1 (IND n' H9)) as [IND' _]; [..|apply IND']; eauto.
      econstructor; eauto.
      instantiate (1:=fun _ _ => False). contradiction.
      intros. eapply H1. omega. auto.
      intros. eapply H2. omega. auto.
      intros. eapply H3. omega. auto.
      intros. eauto.
      intros. safe_step. eapply H2. omega. auto.
      auto.
      auto.
  - (* Sblock *)
    inv x. split; intros.
    + safe_step. edestruct (proj1 (IND n (lt_n_Sn n))) as [IND' _]; [..|apply IND']; eauto.
      constructor. auto.
      intros. safe_step. apply H0. omega. auto.
      simpl. destruct x.
      intros. safe_step. apply H0. omega. auto.
      intros. safe_step. apply H2. omega. auto.
    + eapply IHstmt. 7:eauto. eauto. constructor. auto.
      intros. safe_step. apply H0. omega. auto.
      intros. apply H1. auto. auto.
      simpl. destruct x.
      intros. safe_step. apply H0. omega. auto.
      intros. safe_step. apply H2. omega. auto.
      intros. apply H3. omega. auto.
      auto.
  - (* Sexit *)
    inv x. split; intros. 2:discriminate. auto.
  - (* Sswitch *)
    inv x. split; intros.
    + edestruct H16 as (n' & v' & Hv' & Hn'v' & HOARE). eauto.
      safe_step. edestruct (proj1 (IND n (lt_n_Sn n))) as [IND' _]; [..|apply IND']; eauto.
      simpl. auto.
    + eapply IHstmt; eauto.
  - (* Sreturn *)
    inv x.
    + (* None *)
      split; intros. 2:discriminate.
      edestruct H10 as (mem' & MEM' & POST). eauto.
      safe_step. auto.
    + (* Some *)
      split; intros. 2:discriminate.
      edestruct H10 as (v & mem' & Hv & MEM' & POST). eauto.
      safe_step. auto.
  - (* Slabel *)
    inv x. split; intros.
    + safe_step. edestruct (proj1 (IND n (lt_n_Sn n))) as [IND' _]; [..|apply IND']; eauto.
      simpl; auto.
    + simpl in H4. destruct ident_eq.
      * subst. inv H4.
        edestruct IHstmt as [IND' _]; [..|apply IND']; eauto.
        simpl; auto.
      * eapply IHstmt; eauto.
  - (* Sgoto *)
    inv x. split; intros. 2:discriminate.
    destruct n. repeat intro; apply le_0_n.
    edestruct H3 as (s' & k' & Hs'k' & SAFE); eauto.
    safe_step. rewrite H14 in Hs'k'. inv Hs'k'. auto.
  - (* LSnil *)
    inv H5.
  - (* LScons *)
    simpl in H5, H. inv H.
    destruct (find_label lbl stmt (Kseq (seq_of_lbl_stmt l) k)) eqn:?.
    + inv H5. eapply IHstmt. 7:eauto. eauto. constructor. auto.
      intros. safe_step.
      assert (n' < n)%nat by omega.
      edestruct (proj1 (IND n' H7)) as [IND' _]; [..|apply IND']; eauto.
      intros. apply H1. omega. auto.
      intros. apply H2. omega. auto.
      intros. apply H3. omega. auto.
      intros. apply H4. omega. auto.
      auto.
      intros. safe_step. apply H3. omega. auto.
      auto.
      auto.
    + eapply IHstmt0; eauto.
      eapply hoare_stmt_pre_impl; eauto. destruct 1.
  - (* Internal *)
    edestruct H3 as (A & B & C & mem1 & env & le & ALLOC & BIND & inv & FIND & HOARE). eauto.
    safe_step.
    edestruct (proj1 (IND n (lt_n_Sn n))) as [IND' _]; [..|apply IND']; eauto.
    + intros. destruct H5 as (mem'0 & MEM'0 & POST). safe_step. apply H1. omega. auto.
    + intros. replace (call_cont k) with k by (destruct k; contradiction || reflexivity).
      apply H1. omega. auto.
    + contradiction.
    + intros n' ?.
      induction n' using lt_wf_ind.
      replace (call_cont k) with k by (destruct k; contradiction || reflexivity).
      intros.
      destruct (find_label lbl (fn_body f) k) as [[s' k']|] eqn:?. 2:now (edestruct FIND; eauto).
      exists s'. exists k'. split. auto.
      eapply IND; eauto.
      * intros. destruct H10 as (mem2 & FREE & POST). safe_step. apply H1. omega. auto.
      * intros.
        replace (call_cont k) with k by (destruct k; contradiction || reflexivity).
        apply H1. omega. auto.
      * contradiction.
      * intros. apply H5; eauto. omega.
    + simpl. auto.
  - (* External *)
    edestruct H3 as ((t & v & mem' & EXT) & POST). eauto.
    safe_step. eauto.
Qed.

Lemma hoare_stmt_correct:
  ∀ f env stk stmt pre_normal pre_label post_normal post_return post_exit post_goto,
    hoare_stmt f env stk pre_normal pre_label stmt
               post_normal post_return post_exit post_goto ->
    hoare_stmt_sem f env stk pre_normal pre_label stmt
                   post_normal post_return post_exit post_goto.
Proof.
  intros. apply hoare_correct_aux. auto.
Qed.

Lemma hoare_fun_correct:
  ∀ stk pre f post,
    hoare_fun stk pre f post ->
    hoare_fun_sem stk pre f post.
Proof.
  intros. apply hoare_correct_aux. auto.
Qed.

Lemma hoare_prog:
  ∀ f b mem,
    Genv.find_symbol ge (prog_main prog) = Some b ->
    Genv.find_funct_ptr ge b = Some f ->
    funsig f = signature_main ->
    Genv.init_mem prog = Some mem ->
    hoare_fun nil (fun vargs mem' => vargs = nil /\ mem' = mem)
              f
              (fun v mem => ∃ r, v = Vint r) ->
    ∀ tr, program_behaves (semantics prog) (Goes_wrong tr) ->
          False.
Proof.
  intros. apply hoare_fun_correct in H3.
  inv H4.
  - assert (∀ n, safe n s).
    { intros. inv H5.
      change ge0 with ge in *. clear ge0.
      assert (mem = m0) by congruence. subst m0.
      assert (b = b0) by congruence. subst b0.
      assert (f = f0) by congruence. subst f0.
      apply H3. constructor. constructor.
      intros. destruct H10 as [r Hr]. subst v.
      repeat intro. inv H10. edestruct H12. constructor. inv H13.
      auto. }
    inv H6.
    apply star_starN in H8. destruct H8 as [n Hn].
    pose proof (H4 (S n) _ _ _ Hn H9 H10).
    omega.
  - eapply H6. econstructor; eauto.
Qed.

(** Alternative rules for labelled statements *)
Definition hoare_lbl_stmt
    (f:function) (env:env) (stk:stack)
    (pre_case:Z -> temp_env -> mem -> Prop)
    (pre_default:temp_env -> mem -> Prop)
    (pre_fallthrough:temp_env -> mem -> Prop)
    (pre_label:ident -> temp_env -> mem -> Prop)
    (lbl_stmt:lbl_stmt)
    (post_normal:temp_env -> mem -> Prop)
    (post_return:val -> mem -> Prop)
    (post_exit:nat -> temp_env -> mem -> Prop)
    (post_goto:ident -> temp_env -> mem -> Prop) : Prop :=

  (∀ n s, select_switch_case n lbl_stmt = Some s ->
          hoare_stmt f env stk
                     (pre_case n) (fun _ _ _ => False)
                     (seq_of_lbl_stmt s)
                     post_normal post_return post_exit post_goto) /\
  hoare_stmt f env stk
             pre_default (fun _ _ _ => False)
             (seq_of_lbl_stmt (select_switch_default lbl_stmt))
             post_normal post_return post_exit post_goto /\
  hoare_stmt f env stk
             pre_fallthrough (fun _ _ _ => False)
             (seq_of_lbl_stmt lbl_stmt)
             post_normal post_return post_exit post_goto /\
  hoare_stmt f env stk
             (fun _ _ => False) pre_label
             (seq_of_lbl_stmt lbl_stmt)
             post_normal post_return post_exit post_goto.

Lemma hoare_lbl_stmt_post_impl:
  ∀ f env stk
    pre_case pre_default pre_fallthrough pre_label post_normal post_return post_exit post_goto
    post_normal' post_return' post_exit' post_goto',
    (∀ le mem, (post_normal le mem:Prop) -> (post_normal' le mem:Prop)) ->
    (∀ v mem, (post_return v mem:Prop) -> (post_return' v mem:Prop)) ->
    (∀ x le mem, (post_exit x le mem:Prop) -> (post_exit' x le mem:Prop)) ->
    (∀ lbl le mem, (post_goto lbl le mem:Prop) -> (post_goto' lbl le mem:Prop)) ->
  ∀ stmt,
    hoare_lbl_stmt f env stk pre_case pre_default pre_fallthrough pre_label stmt post_normal post_return post_exit post_goto ->
    hoare_lbl_stmt f env stk pre_case pre_default pre_fallthrough pre_label stmt post_normal' post_return' post_exit' post_goto'.
Proof.
  unfold hoare_lbl_stmt.
  intuition; eapply hoare_stmt_post_impl; eauto.
Qed.

Lemma hoare_lbl_stmt_pre_impl:
  ∀ f env stk
    pre_case pre_default pre_fallthrough pre_label post_normal post_return post_exit post_goto
    pre_case' pre_default' pre_fallthrough' pre_label',
    (∀ i le mem, (pre_case' i le mem:Prop) -> (pre_case i le mem:Prop)) ->
    (∀ le mem, (pre_default' le mem:Prop) -> (pre_default le mem:Prop)) ->
    (∀ le mem, (pre_fallthrough' le mem:Prop) -> (pre_fallthrough le mem:Prop)) ->
    (∀ lbl le mem, (pre_label' lbl le mem:Prop) -> (pre_label lbl le mem:Prop)) ->
  ∀ stmt,
    hoare_lbl_stmt f env stk pre_case pre_default pre_fallthrough pre_label stmt post_normal post_return post_exit post_goto ->
    hoare_lbl_stmt f env stk pre_case' pre_default' pre_fallthrough' pre_label' stmt post_normal post_return post_exit post_goto.
Proof.
  unfold hoare_lbl_stmt.
  intuition; eapply hoare_stmt_pre_impl; try eassumption.
  3:eapply H4;eauto. auto. auto. auto. auto. auto.
Qed.

Lemma hoare_LSnil :
  ∀ f env stk pre_case pre_default pre_fallthrough pre_label
    post_normal post_return post_exit post_goto,
    (∀ le mem, (pre_default le mem:Prop) -> (post_normal le mem:Prop)) ->
    (∀ le mem, (pre_fallthrough le mem:Prop) -> (post_normal le mem:Prop)) ->
    hoare_lbl_stmt f env stk
                   pre_case pre_default pre_fallthrough pre_label
                   LSnil
                   post_normal post_return post_exit post_goto.
Proof.
  split; [|split; [|split]].
  discriminate.
  constructor. auto.
  constructor. auto.
  constructor. destruct 1.
Qed.

Lemma hoare_LScons_case :
  ∀ n s ls f env stk pre_case pre_default pre_fallthrough pre_label
    P post_normal post_return post_exit post_goto,

    hoare_stmt f env stk
               (fun le mem => pre_case n le mem \/ pre_fallthrough le mem) pre_label
               s
               P post_return post_exit post_goto ->

    hoare_lbl_stmt f env stk
               pre_case pre_default P pre_label
               ls
               post_normal post_return post_exit post_goto ->

    hoare_lbl_stmt f env stk
                   pre_case pre_default pre_fallthrough pre_label
                   (LScons (Some n) s ls)
                   post_normal post_return post_exit post_goto.
Proof.
  intros. destruct H0 as (A & B & C & D).
  split; [|split; [|split]].
  - simpl. intros. destruct zeq.
    + inv H0. econstructor. eapply hoare_stmt_pre_impl, H. left. auto. destruct 1. eauto.
    + auto.
  - auto.
  - econstructor. eapply hoare_stmt_pre_impl, H. right. auto. destruct 1. eauto.
  - econstructor. eapply hoare_stmt_pre_impl, H. destruct 1. auto.
    eapply hoare_stmt_pre_impl, hoare_stmt_pre_or. 3:apply C. 3:apply D.
    now intuition. now intuition.
Qed.

Lemma hoare_LScons_default :
  ∀ s ls f env stk pre_case pre_default pre_fallthrough pre_label
    P post_normal post_return post_exit post_goto,

    hoare_stmt f env stk
               (fun le mem => pre_default le mem \/ pre_fallthrough le mem) pre_label
               s
               P post_return post_exit post_goto ->

    hoare_lbl_stmt f env stk
               pre_case (fun _ _ => False) P pre_label
               ls
               post_normal post_return post_exit post_goto ->

    hoare_lbl_stmt f env stk
                   pre_case pre_default pre_fallthrough pre_label
                   (LScons None s ls)
                   post_normal post_return post_exit post_goto.
Proof.
  intros. destruct H0 as (A & B & C & D).
  split; [|split; [|split]].
  - auto.
  - econstructor. eapply hoare_stmt_pre_impl, H. left. auto. destruct 1. auto.
  - econstructor. eapply hoare_stmt_pre_impl, H. right. auto. destruct 1. eauto.
  - econstructor. eapply hoare_stmt_pre_impl, H. destruct 1. auto.
    eapply hoare_stmt_pre_impl, hoare_stmt_pre_or. 3:apply C. 3:apply D.
    now intuition. now intuition.
Qed.

Lemma hoare_Sswitch:
  ∀ f env stk islong a cases,
  ∀ pre_normal pre_label pre_case pre_default
    post_normal post_return post_exit post_goto,
    (∀ le mem, (pre_normal le mem:Prop) ->
       ∃ v n,
         switch_argument islong v n /\
         eval_expr ge env le mem a v /\
         match select_switch_case n cases with
           | None => pre_default le mem
           | Some _ => pre_case n le mem
         end) ->

    hoare_lbl_stmt f env stk
                   pre_case pre_default (fun _ _ => False) pre_label
                   cases
                   post_normal post_return post_exit post_goto ->

    hoare_stmt f env stk
               pre_normal pre_label
               (Sswitch islong a cases)
               post_normal post_return post_exit post_goto.
Proof.
  intros. destruct H0 as (? & ? & _ & ?).
  constructor.
  - intros. edestruct H as (v & n & SA & EVAL & PRE). eauto.
    eexists. eexists. split. eauto. split. eauto. unfold select_switch.
    specialize (H0 n). destruct select_switch_case.
    eapply hoare_stmt_pre_impl, H0; auto. intros ? ? [-> ->]. auto.
    eapply hoare_stmt_pre_impl, H1; auto. intros ? ? [-> ->]. auto.
  - auto.
Qed.

End PROG.
