(* *********************************************************************)
(*                                                                     *)
(*              The Verasco static analyzer                            *)
(*                                                                     *)
(*          Jacques-Henri Jourdan, INRIA Paris-Rocquencourt            *)
(*                                                                     *)
(*  Copyright Institut National de Recherche en Informatique et en     *)
(*  Automatique and Université de Rennes 1.  All rights reserved.      *)
(*  This file is distributed under the terms of the GNU General        *)
(*  Public License as published by the Free Software Foundatio,        *)
(*  either version 2 of the License, or (at your option) any later     *)
(*  version.  This file is also distributed under the terms of the     *)
(*  INRIA Non-Commercial License Agreement.                            *)
(*                                                                     *)
(* *********************************************************************)

Require Import Coqlib Utf8.
Require Import Maps ShareTree Util.
Require Import AdomLib.
Require Import IdealExpr.
Require Import AbIdealEnv.
Require Import AbIdealNonrel.
Require Import IdealIntervals.
Require Import FactMsgHelpers.

Module Dom := WPFun T.

Section S.

  Context {Val: Type} {Dval: ab_ideal_nonrel Val}.
  Context (G:gamma_op Val ideal_num) (Dval_cor: ab_ideal_nonrel_correct Dval G).

  Definition t := T.t Val.

  Inductive tagged_iexpr : iexpr -> Type :=
  | IEvar_tag : Val+⊤+⊥ -> ∀ ty x, tagged_iexpr (IEvar ty x)
  | IEconst_tag : Val+⊤+⊥ -> ∀ n, tagged_iexpr (IEconst n)
  | IEZitv_tag : Val+⊤+⊥ -> ∀ i, tagged_iexpr (IEZitv i)
  | IEunknown_tag : Val+⊤+⊥ -> ∀ ty, tagged_iexpr (IEunknown ty)
  | IEunop_tag : Val+⊤+⊥ -> ∀ op {e}, tagged_iexpr e -> tagged_iexpr (IEunop op e)
  | IEbinop_tag : Val+⊤+⊥ -> ∀ op {e1 e2}, tagged_iexpr e1 -> tagged_iexpr e2 ->
                                       tagged_iexpr (IEbinop op e1 e2).

  Inductive tagged_iexpr_gamma:
    ∀ {e}, gamma_op (tagged_iexpr e) (var -> ideal_num) :=
  | IEvar_tag_ok: ∀ ab x ty ρ,
    ρ x ∈ γ ab ->
    tagged_iexpr_gamma (IEvar_tag ab ty x) ρ
  | IEZconst_tag_ok: ∀ ab n ρ,
    n ∈ γ ab ->
    tagged_iexpr_gamma (IEconst_tag ab n) ρ
  | IEZitv_tag_ok: ∀ ab itv ρ,
    γ itv  ⊆ (γ ab ∘ INz) ->
    tagged_iexpr_gamma (IEZitv_tag ab itv) ρ
  | IEunknown_tag_ok: ∀ ab ty ρ,
    γ ty ⊆ γ ab ->
    tagged_iexpr_gamma (IEunknown_tag ab ty) ρ
  | IEunop_tag_ok: ∀ ab op e (te:tagged_iexpr e) ρ,
    (∀ n r, eval_iexpr ρ e n -> eval_iunop op n r -> r ∈ γ ab) ->
    tagged_iexpr_gamma te ρ ->
    tagged_iexpr_gamma (IEunop_tag ab op te) ρ
  | IEbinop_tag_ok: ∀ ab op e1 (te1:tagged_iexpr e1) e2 (te2:tagged_iexpr e2) ρ,
    (∀ n1 n2 r, eval_iexpr ρ e1 n1 -> eval_iexpr ρ e2 n2 ->
                eval_ibinop op n1 n2 r -> r ∈ γ ab) ->
    tagged_iexpr_gamma te1 ρ ->
    tagged_iexpr_gamma te2 ρ ->
    tagged_iexpr_gamma (IEbinop_tag ab op te1 te2) ρ.

  Existing Instance tagged_iexpr_gamma.

  Definition iexpr_get_tag {e:iexpr} (te:tagged_iexpr e) : Val+⊤+⊥ :=
    match te with
      | IEvar_tag ab _ _ => ab
      | IEconst_tag ab _ => ab
      | IEZitv_tag ab _ => ab
      | IEunknown_tag ab _ => ab
      | IEunop_tag ab _ _ _ => ab
      | IEbinop_tag ab _ _ _ _ _ => ab
    end.

  Lemma iexpr_get_tag_correct:
    ∀ ρ e (te:tagged_iexpr e) n,
      ρ ∈ γ te ->
      eval_iexpr ρ e n ->
      n ∈ γ (iexpr_get_tag te).
  Proof.
    destruct 1; intros EVAL; inv EVAL; eauto.
  Qed.

  Fixpoint eval_expr (fuel:nat) :
    ∀ e:iexpr, t * query_chan -> tagged_iexpr e :=
    fix eval_expr' (e:iexpr) (ρ:t * query_chan) : tagged_iexpr e :=
      match e with
        | IEvar _ x =>
          let ab := Dom.get x (fst ρ) in
          let ab :=
            match fuel with
              | O => NotBot ab
              | S fuel =>
                forward_var (fun e => iexpr_get_tag (eval_expr fuel e ρ))
                            x ab (snd ρ)
            end
          in
          IEvar_tag ab _ _
        | IEconst n => IEconst_tag (NotBot (const n)) _
        | IEZitv itv => IEZitv_tag (NotBot (z_itv itv)) _
        | IEunknown _ => IEunknown_tag (NotBot All) _
        | IEunop op e =>
          let te := eval_expr' e ρ in
          let ab :=
            do_bot ab <- iexpr_get_tag te;
            forward_unop op ab
          in
          IEunop_tag ab _ te
        | IEbinop op e1 e2 =>
          let te1 := eval_expr' e1 ρ in
          let te2 := eval_expr' e2 ρ in
          let ab :=
            do_bot ab1 <- iexpr_get_tag te1;
            do_bot ab2 <- iexpr_get_tag te2;
            forward_binop op ab1 ab2
          in
          IEbinop_tag ab _ te1 te2
      end.
  Definition eval_expr_fuel := 5%nat.
  Opaque eval_expr_fuel.

  Lemma eval_expr_correct:
    ∀ fuel e ρ ab,
      ρ ∈ γ ab ->
      ρ ∈ γ (eval_expr fuel e ab).
  Proof.
    fix 1.
    intros fuel e ρ ab Hρ. induction e.
    - destruct fuel; constructor. apply Hρ.
      eapply forward_var_sound; try apply Hρ; auto.
      repeat intro. eapply iexpr_get_tag_correct; eauto.
    - destruct fuel; constructor; try apply const_correct; auto.
    - destruct fuel; constructor; try constructor; apply z_itv_correct.
    - destruct fuel; constructor; constructor.
    - destruct fuel; constructor; eauto; intros;
      eapply botbind_spec; eauto using iexpr_get_tag_correct, forward_unop_sound.
    - destruct fuel; constructor; eauto; intros;
      eapply botbind2_spec; eauto using iexpr_get_tag_correct, forward_binop_sound.
  Qed.

  Definition assign (x:var) (e:iexpr) (ρ:t * query_chan) (chan:query_chan) : (t*messages_chan)+⊥ :=
    do_bot abv <- iexpr_get_tag (eval_expr eval_expr_fuel e ρ);
    NotBot (Dom.set (fst ρ) x abv,
            match abv with
              | Just abv => assign_msgs x abv
              | All => nil
            end).

  Lemma assign_correct:
    ∀ x e ρ, ∀ n ab c,
      ρ ∈ γ ab -> (upd ρ x n) ∈ γ c ->
      n ∈ eval_iexpr ρ e ->
      (upd ρ x n) ∈ γ (assign x e ab c).
  Proof.
    intros x e ρ n ab c Hρ Hc Hn.
    eapply botbind_spec, iexpr_get_tag_correct; eauto. 2:apply eval_expr_correct; auto.
    split.
    - intro. simpl. rewrite Dom.gsspec.
      unfold upd. unfold eq_dec, var_eq. destruct T.elt_eq. auto. apply Hρ.
    - destruct a. constructor. apply assign_msgs_sound.
      rewrite upd_s. auto.
  Qed.

  Fixpoint backward_eval (fuel:nat) :
    ∀ {e:iexpr}, tagged_iexpr e -> t * query_chan -> Val -> t+⊥ :=
    fix backward_eval' {e:iexpr} (te:tagged_iexpr e) (ρ:t * query_chan) (v:Val) : t+⊥ :=
    do_bot abres <- meet (NotBot (Just v)) (iexpr_get_tag te);
    match te with
      | IEvar_tag abx _ x =>
        let newρ := Dom.set (fst ρ) x abres in
        match abres with
          | All => NotBot newρ
          | Just abres =>
            match fuel with
              | O => NotBot newρ
              | S fuel =>
                backward_var
                  (fun e => iexpr_get_tag (eval_expr fuel e ρ))
                  (fun e abe abenv => backward_eval fuel (eval_expr fuel e ρ) abenv abe)
                  (fun abx abenv =>
                     match T.get_set x (fst abenv) with
                     | (None, setter) => NotBot (setter abx)
                     | (Some abx', setter) =>
                       do_bot abx <- meet abx abx';
                       NotBot (setter abx)
                     end)
                  x abres (newρ, snd ρ)
            end
         end
      | IEconst_tag _ _ | IEZitv_tag _ _ | IEunknown_tag _ _ =>
        NotBot (fst ρ)
      | IEunop_tag _ op _ te' =>
        do_bot ve' <- iexpr_get_tag te';
        do_bot ve' <- backward_unop op abres ve';
        match ve' with
        | Just ve' => backward_eval' te' ρ ve'
        | All => NotBot (fst ρ)
        end
      | IEbinop_tag _ op _ _ te1 te2 =>
        do_bot ve1 <- iexpr_get_tag te1;
        do_bot ve2 <- iexpr_get_tag te2;
        do_bot (ve1, ve2) <- backward_binop op abres ve1 ve2;
        do_bot ρ' <-
          match ve1 with
          | Just ve1 => backward_eval' te1 ρ ve1
          | All => NotBot (fst ρ)
          end;
        match ve2 with
        | Just ve2 => backward_eval' te2 (ρ', snd ρ) ve2
        | All => NotBot ρ'
        end
    end.
  Definition backward_eval_fuel := 5%nat.

  Lemma backward_eval_correct:
    ∀ fuel e (te:tagged_iexpr e) res res_ab,
      res ∈ γ res_ab ->
    ∀ ρ ρ_ab,
      ρ ∈ γ ρ_ab ->
      ρ ∈ γ te ->
      eval_iexpr ρ e res ->
      ρ ∈ γ (backward_eval fuel te ρ_ab res_ab).
  Proof.
    fix 1.
    induction te; intros res res_ab Hres ρ ρ_ab Hρ Hte EVAL; inversion EVAL; subst; clear EVAL.
    - assert (Hρ':=Hρ). destruct Hρ' as [Hρ0 Hρ1].
      destruct fuel; (eapply botbind_spec; [|apply meet_correct; eauto using iexpr_get_tag_correct]);
      (intros [|abx] Habx; try (eapply Dom.gamma_set; now eauto)).
      apply backward_var_sound; eauto.
      + repeat intro. eapply iexpr_get_tag_correct, H. apply eval_expr_correct; eauto.
      + repeat intro. eapply backward_eval_correct; eauto. apply eval_expr_correct; eauto.
      + repeat intro. destruct (T.get_set_spec x (fst ab)). destruct T.get_set as [[]?].
        eapply botbind_spec with (ρ x). intros. rewrite H2.
        eapply Dom.gamma_set with (v:=Just _). apply H0. apply H4. auto.
        apply meet_correct. split. apply H.
        destruct H0. specialize (H0 x). unfold Dom.get in H0. rewrite <- H1 in H0. apply H0.
        rewrite H2. eapply Dom.gamma_set with (v:=Just _). apply H0. apply H. auto.
      + split. eapply Dom.gamma_set; eauto. auto.
    - assert (Hρ':=Hρ). destruct Hρ' as [Hρ0 Hρ1].
      destruct fuel; auto; (eapply botlift_fun1_spec; [auto|]);
      apply meet_correct; eauto using iexpr_get_tag_correct, const_correct.
    - assert (Hρ':=Hρ). destruct Hρ' as [Hρ0 Hρ1].
      destruct fuel; auto; eapply botlift_fun1_spec; auto;
      apply meet_correct; eauto using iexpr_get_tag_correct.
    - assert (Hρ':=Hρ). destruct Hρ' as [Hρ0 Hρ1].
      destruct fuel; auto; eapply botlift_fun1_spec; auto;
      apply meet_correct; eauto using iexpr_get_tag_correct.
    - pose proof Hte. inv Hte. apply Classical_Prop.EqdepTheory.inj_pair2 in H6. subst te0.
      destruct fuel;
      (repeat (eapply botbind_spec; [intros|]);
        [destruct a1; [apply Hρ|eapply IHte; eauto]|
         eapply backward_unop_sound; eauto |
         eapply iexpr_get_tag_correct; eauto |
         eapply meet_correct; eauto |
         eapply iexpr_get_tag_correct; eauto ]).
    - pose proof Hte. inv Hte.
      apply Classical_Prop.EqdepTheory.inj_pair2 in H8. subst te0.
      apply Classical_Prop.EqdepTheory.inj_pair2 in H9. subst te3.
      destruct fuel;
      (repeat (eapply botbind_spec; [intros; try destruct a2|]);
      [| |eapply backward_binop_sound; eauto|
       eapply iexpr_get_tag_correct; eauto|
       eapply iexpr_get_tag_correct; eauto|
       eapply meet_correct; eauto|
       eapply iexpr_get_tag_correct; eauto];
      destruct H6; [destruct t1; [|eapply IHte2]|destruct t0; [|eapply IHte1]];
      eauto);
      try split; auto; apply Hρ.
  Qed.

  Program Definition accumulate_assume_msgs (ρold ρnew:t) (c:messages_chan): messages_chan :=
    let messages := T.shcombine_diff
                      (fun x vOld vNew =>
                         match vOld, vNew return _ with
                           | _, None => None
                           | None, Some vNew =>
                             match assume_msgs x All vNew return _ with
                               | nil => None
                               | x => Some x
                             end
                           | Some vOld, Some vNew =>
                             if vOld ⊑ vNew then None
                             else
                               match assume_msgs x (Just vOld) vNew return _ with
                                 | nil => None
                                 | x => Some x
                               end
                         end)
                      _ ρold ρnew
    in
    T.fold (fun acc _ l => app l acc) messages c.
  Next Obligation. destruct v; auto. rewrite leb_refl; auto. Qed.

  Lemma accumulate_assume_msgs_correct:
    forall ρold ρnew c ρ,
      ρ ∈ γ ρold ->
      ρ ∈ γ ρnew ->
      ρ ∈ γ c ->
      ρ ∈ γ (accumulate_assume_msgs ρold ρnew c).
  Proof.
    intros.
    unfold accumulate_assume_msgs.
    apply Dom.P.fold_rec; auto.
    intros.
    assert (γ v ρ).
    { rewrite T.gshcombine_diff in H3.
      specialize (H k). specialize (H0 k). unfold Dom.get in H, H0.
      destruct T.get, T.get; try discriminate.
      - destruct (v0 ⊑ v1). discriminate.
        pose proof assume_msgs_sound k (Just v0) v1 ρ.
        destruct (assume_msgs k (Just v0) v1); inv H3. auto.
      - pose proof assume_msgs_sound k All v0 ρ.
        destruct (assume_msgs k All v0); inv H3. auto. }
    clear H3. induction v.
    auto. inv H5. constructor. auto. apply IHv. auto.
  Qed.

  Definition assume (e:iexpr) (b:bool) (ρ:t * query_chan): t+⊥ :=
    let n := INz (if b then 1 else 0) in
    match const n with
    | All => NotBot (fst ρ)
    | Just ab =>
      backward_eval backward_eval_fuel (eval_expr eval_expr_fuel e ρ)
                    ρ ab
    end.

  Lemma assume_correct:
    ∀ e ρ ab b, ρ ∈ γ ab ->
                (INz (if b:bool then 1 else 0)) ∈ eval_iexpr ρ e ->
                ρ ∈ γ (assume e b ab).
  Proof.
    intros. unfold assume.
    destruct (const (INz (if b then 1 else 0))) eqn:EQ.
    apply H.
    eapply backward_eval_correct; eauto.
    change (γ (Just v) (INz (if b:bool then 1 else 0))). rewrite <- EQ.
    apply const_correct. apply eval_expr_correct; auto.
  Qed.

  Definition forget (x:var) (ρ:t * query_chan) (c:query_chan): t+⊥ :=
    NotBot (Dom.set (fst ρ) x All).

  Lemma forget_correct:
    ∀ x ρ, ∀ n ab c,
      ρ ∈ γ ab -> (upd ρ x n) ∈ γ c ->
      (upd ρ x n) ∈ γ (forget x ab c).
  Proof.
    intros x ρ n ab c Hρ Hc y. unfold upd.
    rewrite Dom.gsspec. unfold eq_dec, var_eq. destruct T.elt_eq; auto. exact I.
    destruct Hρ. auto.
  Qed.

  Instance leb_tree : leb_op t := Dom.leb_tree _ leb_refl.
  Instance join_tree : join_op t t := Dom.join_tree _ join_idem.
  Definition widen_tree : t -> t -> t := Dom.widen_tree _ widen_idem.

  Instance ideal_box_domain : ab_ideal_env t (t+⊥) := {
    id_top c := NotBot (top, nil);
    id_leb x y := leb (fst x) y;
    id_join x y c := NotBot (join (fst x) (fst y), nil);
    id_init_iter := id;
    id_widen x y c :=
      let w :=
        match x with
        | Bot => do_bot (y, _) <- y; NotBot y
        | NotBot x =>
          match y with
          | Bot => NotBot x
          | NotBot (y, _) => NotBot (widen_tree x y)
          end
        end
      in
      (w, do_bot w <- w; NotBot (w, nil));
    assign x e ab c := assign x e ab c;
    forget x ab c := botlift_fun1 (fun x => (x, nil)) (forget x ab c);
    process_msg m ab :=
      match m with
        | Fact_msg e b =>
          do_bot res <- assume_unfold (fun e b ab => assume e b ab) e b ab;
          let c := accumulate_assume_msgs (fst ab) res (m::nil) in
          NotBot (res, c)
        | Known_value_msg x v =>
          do_bot v_ab <- meet (Dom.get x (fst ab)) (const v);
          NotBot (Dom.set (fst ab) x v_ab, m::nil)
        | _ =>
          process_msg
            (AB:=t)
            (fun e => iexpr_get_tag (eval_expr 0 e ab))
            (fun e ab' abenv => backward_eval 0 (eval_expr 0 e abenv) abenv ab')
                      m ab
      end;
    enrich_query_chan ab c :=
      enrich_query_chan (fun e => iexpr_get_tag (eval_expr eval_expr_fuel e (ab, c))) c
  }.
  Proof.
    - split. intro. unfold Dom.get. rewrite T.gempty. constructor. constructor.
    - intros. intro. destruct H0. eapply leb_correct in H0. 2:apply H.
      specialize (H0 e). unfold flat_top_leb in H. auto.
    - intros. split. apply join_correct. destruct H as [[]|[]]; auto. constructor.
    - auto.
    - intros. destruct ab1. contradiction.
      destruct ab2. split. auto. eapply botbind_spec with ρ.
      intros. split. auto. constructor. auto.
      destruct x0 as [y ?].
      unfold widen_tree.
      pose proof Dom.widen_tree_incr widen widen_idem AbIdealNonrel.widen_incr.
      specialize (H1 x y _ H). split. apply H1. split. apply H1. constructor.
    - eapply assign_correct.
    - intros. eapply botlift_fun1_spec, forget_correct; eauto. split. auto. constructor.
    - intros.
      pose proof H. destruct H1.
      destruct m;
        (try now (apply process_msg_sound;
          [repeat intro; eapply iexpr_get_tag_correct, H3; apply eval_expr_correct; eauto|
           repeat intro; eapply backward_eval_correct; eauto; apply eval_expr_correct; eauto|
           auto|auto])).
      + eapply botbind_spec, assume_unfold_correct; eauto.
        split. eauto. apply accumulate_assume_msgs_correct; auto. constructor. auto. constructor.
        intros. apply join_correct. auto.
        intros. apply assume_correct; auto.
      + eapply botbind_spec.
        * intros. split; simpl. eapply Dom.gamma_set; eauto. repeat constructor. auto.
        * apply meet_correct. split. apply H. inv H0. apply const_correct.
    - intros. apply enrich_query_chan_sound; auto.
      intros e ve He. eapply iexpr_get_tag_correct; eauto. eapply eval_expr_correct; split; eauto.
  Defined.

End S.
