(* *********************************************************************)
(*                                                                     *)
(*              The Verasco static analyzer                            *)
(*                                                                     *)
(*          Jacques-Henri Jourdan, INRIA Paris-Rocquencourt            *)
(*                                                                     *)
(*  Copyright Institut National de Recherche en Informatique et en     *)
(*  Automatique and Université de Rennes 1.  All rights reserved.      *)
(*  This file is distributed under the terms of the GNU General        *)
(*  Public License as published by the Free Software Foundatio,        *)
(*  either version 2 of the License, or (at your option) any later     *)
(*  version.  This file is also distributed under the terms of the     *)
(*  INRIA Non-Commercial License Agreement.                            *)
(*                                                                     *)
(* *********************************************************************)

Require Import Utf8 String.
Require Import Coqlib.
Require Import AbIdealEnv IdealExpr.
Require Import AbMachineEnv.
Require Import Integers Floats.
Require Import FloatLib FloatIntervals.
Require Import Reals.
Require Import Fappli_IEEE.
Require Import ZArith.
Require Import Cminor Ctypes.
Require Import AdomLib.
Require Import Maps ShareTree TreeAl Util Hash ToString.
Require Import FactMsgHelpers.

Open Scope Z_scope.

Section Functor.
  Context (abt abt_iter:Type) {IdD : ab_ideal_env_nochan abt abt_iter}.

  Inductive related : num_val -> ideal_num -> Prop :=
  | rel_int : forall x, related (NVint (Int.repr x)) (INz x)
  | rel_long : forall x, related (NVint64 (Int64.repr x)) (INz x)
  | rel_float : forall x, related (NVfloat x) (INf x)
  | rel_single : forall x, related (NVsingle x) (INf (Float.of_single x)).

  Hint Constructors related.

  Lemma related_unsigned:
    forall x, related (NVint x) (INz (Int.unsigned x)).
  Proof.
    intros. rewrite <- Int.repr_unsigned with (i:=x) at 1. constructor.
  Qed.

  Lemma related_signed:
    forall x, related (NVint x) (INz (Int.signed x)).
  Proof.
    intros. rewrite <- Int.repr_signed with (i:=x) at 1. constructor.
  Qed.

  Lemma related_unsigned_long:
    forall x, related (NVint64 x) (INz (Int64.unsigned x)).
  Proof.
    intros. rewrite <- Int64.repr_unsigned with (i:=x) at 1. constructor.
  Qed.

  Lemma related_signed_long:
    forall x, related (NVint64 x) (INz (Int64.signed x)).
  Proof.
    intros. rewrite <- Int64.repr_signed with (i:=x) at 1. constructor.
  Qed.

  Instance gamma_mach : gamma_op abt (Cells.cell -> num_val) :=
    fun v ρ =>
      ∃ ρ':var -> ideal_num,
        ρ' ∈ γ v ∧ ∀ id, related (ρ id) (ρ' (Real id)).

  Instance gamma_mach_iter : gamma_op abt_iter (Cells.cell -> num_val) :=
    fun v ρ =>
      ∃ ρ':var -> ideal_num,
        ρ' ∈ γ v ∧ ∀ id, related (ρ id) (ρ' (Real id)).

  Definition typesmap := Cells.ACTree.t (option signedness * num_val_ty).

  Instance gamma_typs : gamma_op typesmap (Cells.cell -> num_val) :=
    fun typs ρ =>
      ∀ v t, Cells.ACTree.get v typs = Some t ->
             match snd t with NVTfloat | NVTsingle => fst t = None | _ => True end /\
             ρ v ∈ γ (snd t).

  Record ghosts_info : Type := {
    next_g: positive;
    typs_g: PTree.t ideal_num_ty
  }.

  Definition init_ghost_info := {| next_g := 1; typs_g := PTree.empty _ |}.

  Definition compat_gh (gi:ghosts_info) (ρ1 ρ2:var -> ideal_num) :=
    (∀ id, ρ1 (Real id) = ρ2 (Real id)) /\
    (∀ g, (g < gi.(next_g))%positive -> ρ1 (Ghost g) = ρ2 (Ghost g)).

  Lemma compat_gh_refl:
    ∀ next_g ρ, compat_gh next_g ρ ρ.
  Proof.
    intros. split; auto.
  Qed.
  Hint Resolve compat_gh_refl.

  Lemma compat_gh_trans:
    ∀ gi1 gi2 ρ1 ρ2 ρ3,
      (gi1.(next_g) <= gi2.(next_g))%positive ->
      compat_gh gi1 ρ1 ρ2 ->
      compat_gh gi2 ρ2 ρ3 ->
      compat_gh gi1 ρ1 ρ3.
  Proof.
    split; intros; (etransitivity; [apply H0|apply H1]).
    auto. Psatz.lia.
  Qed.
  Hint Resolve compat_gh_trans.

  Instance gamma_gh : gamma_op ghosts_info (var -> ideal_num) :=
    fun gi ρ =>
      ∀ p ty, PTree.get p gi.(typs_g) = Some ty -> ρ (Ghost p) ∈ γ ty.

  Lemma gamma_gh_init:
    ∀ ρ, ρ ∈ γ init_ghost_info.
  Proof.
    repeat intro. rewrite PTree.gempty in H. discriminate.
  Qed.
  Hint Resolve gamma_gh_init.

  Definition signed_itv :=
    ZIntervals.ITV Int.min_signed Int.max_signed.

  Lemma signed_itv_correct:
    ∀ x, Int.signed x ∈ γ signed_itv.
  Proof.
    simpl. apply Int.signed_range.
  Qed.

  Definition unsigned_itv :=
    ZIntervals.ITV 0 Int.max_unsigned.

  Lemma unsigned_itv_correct:
    ∀ x, Int.unsigned x ∈ γ unsigned_itv.
  Proof.
    intros. simpl. apply Int.unsigned_range_2.
  Qed.

  Definition signed64_itv :=
    ZIntervals.ITV Int64.min_signed Int64.max_signed.

  Lemma signed64_itv_correct:
    ∀ x, Int64.signed x ∈ γ signed64_itv.
  Proof.
    simpl. apply Int64.signed_range.
  Qed.

  Definition unsigned64_itv :=
    ZIntervals.ITV 0 Int64.max_unsigned.

  Lemma unsigned64_itv_correct:
    ∀ x, Int64.unsigned x ∈ γ unsigned64_itv.
  Proof.
    intros. simpl. apply Int64.unsigned_range_2.
  Qed.

  Definition alias_sz : Z := 1.
  Opaque alias_sz.

  Definition simplify_expr_alias (v:abt+⊥) (e:iexpr) (gi:ghosts_info):
    iexpr*abt+⊥*ghosts_info :=
    let e := simplify_expr e in
    if Z.leb (iexpr_sz e) alias_sz then (e, v, gi)
    else
      (IEvar (iexpr_ty e) (Ghost gi.(next_g)),
       do_bot v <- v;
       AbIdealEnv.idnc_assign (Ghost gi.(next_g)) e v,
       {| next_g := Pos.succ gi.(next_g);
          typs_g := PTree.set gi.(next_g) (iexpr_ty e) gi.(typs_g) |}).

  Lemma simplify_expr_alias_correct:
    ∀ v e gi,
    ∀ e' v' gi', simplify_expr_alias v e gi = (e', v', gi') ->
    ∀ ρ x, ρ ∈ γ v -> ρ ∈ γ gi ->
          (∀ ρ', compat_gh gi ρ ρ' -> eval_iexpr ρ' e x ) ->
    ∃ ρ', compat_gh gi ρ ρ' /\ ρ' ∈ γ v' /\ ρ' ∈ γ gi' /\
          (∀ ρ'', compat_gh gi' ρ' ρ'' -> eval_iexpr ρ'' e' x).
  Proof.
    unfold simplify_expr_alias. intros.
    pose proof simplify_expr_correct e.
    destruct Z.leb; inv H; eauto 10.
    eexists (ρ+[Ghost gi.(next_g) => x]). split; [|split;[|split]].
    - split. symmetry. apply upd_o. discriminate.
      intros. unfold upd. destruct eq_dec. inv e0. Psatz.lia. auto.
    - eapply botbind_spec with ρ. 2:now auto. intros.
      eapply AbIdealEnv.idnc_assign_correct; now auto.
    - repeat intro. unfold upd. simpl in H. rewrite PTree.gsspec in H.
      destruct eq_dec, peq; try congruence. 2:now apply H1, H.
      inv e0; inv H. eapply iexpr_ty_correct, H3; eauto.
    - constructor. rewrite <- (proj2 H) by (simpl; Psatz.lia). apply upd_s.
      eapply iexpr_ty_correct, H3, H2. eauto.
  Qed.

  Lemma simplify_expr_alias_monotone:
    ∀ v e gi,
    ∀ e' v' gi',
      simplify_expr_alias v e gi = (e', v', gi') ->
      (gi.(next_g) <= gi'.(next_g))%positive.
  Proof.
    unfold simplify_expr_alias. intros.
    destruct Z.leb; inv H; simpl in *; Psatz.lia.
  Qed.

  Definition itv_normalizer (i:ZIntervals.itv+⊤+⊥) (min max:Z) (logdiv:Z): option Z :=
    match i with
      | NotBot (Just (ZIntervals.ITV a b)) =>
        if Z.eqb (Z.shiftr (b-a) logdiv) 0 then
          let q := Z.shiftr (a-min) logdiv in
          if Zle_bool (b-Z.shiftl q logdiv) max
          then Some q
          else None
        else None
      | _ => None
    end.

  Lemma itv_normalizer_correct:
    ∀ i min max logdiv, 0 < logdiv ->
    ∀ q, itv_normalizer i min max logdiv = Some q ->
    ∀ n, n ∈ γ i -> min <= n - q * 2 ^ logdiv <= max.
  Proof.
    intros [|[|[a b]]] min max logdiv Hlogdiv q Hq n Hn; try discriminate Hq; simpl in *.
    rewrite Z.shiftl_mul_pow2, !Z.shiftr_div_pow2 in Hq by omega.
    destruct (Z.eqb_spec ((b-a) / (2 ^ logdiv)) 0). 2:discriminate.
    match goal with
      | H:context [Zle_bool ?x ?y] |- _ =>
        assert (HZle:= Zle_cases x y); destruct (Zle_bool x y)
    end; inv Hq.

    assert (Hlt:2^logdiv > 0). apply Z.lt_gt, Z.pow_pos_nonneg; omega.
    assert (Hmodeq:=Zmod_eq (a-min) _ Hlt).
    assert (Hmodlt:=Z_mod_lt (a-min) _ Hlt).
    destruct Hn. omega.
  Qed.

  Definition normalize_expr (v:abt+⊥) (e:iexpr) (min max:Z) (logdiv:Z) (gi:ghosts_info):
    iexpr*abt+⊥*ghosts_info*bool :=
    (* This trick makes things actually slower. *)
    (* let '(e, v, gi) := simplify_expr_alias v e gi in *)
    match v with
    | NotBot v =>
      let itv :=
          do_bot x <- (idnc_get_query_chan v).(AbIdealEnv.get_itv) e;
            IdealIntervals.in_z x
      in
      match itv_normalizer itv min max logdiv with
      | Some q =>
        let e :=
            if Z_eq_dec q 0 then e
            else
              let shift := Z.shiftl q logdiv in
              IEbinop IOsub e (IEconst (INz shift))
        in
        (e, NotBot v, gi, true)
      | None => (e, NotBot v, gi, false)
      end
    | Bot => (e, v, gi, false)
    end.

  Lemma normalize_expr_correct:
    ∀ v e min max logdiv gi, 0 < logdiv ->
    ∀ e' v' gi' b, normalize_expr v e min max logdiv gi = (e', v', gi', b) ->
    ∀ ρ x, ρ ∈ γ v -> ρ ∈ γ gi ->
      (∀ ρ', compat_gh gi ρ ρ' -> eval_iexpr ρ' e (INz x)) ->
      ∃ ρ', compat_gh gi ρ ρ' /\ ρ' ∈ γ v' /\ ρ' ∈ γ gi' /\
        if b then
          ∃ x', min <= x' <= max ∧ Int.eqmod (two_p logdiv) x x' /\
                (∀ ρ'', compat_gh gi' ρ' ρ'' -> eval_iexpr ρ'' e' (INz x'))
        else
          ∀ ρ'', compat_gh gi' ρ' ρ'' -> eval_iexpr ρ'' e' (INz x).
  Proof.
    unfold normalize_expr.
    intros v e min max logdiv gi Hlogdiv e' v' gi' b Heq ρ x Hγ Hgi EVAL.

    (* pose proof simplify_expr_alias_correct v e gi. *)
    (* destruct simplify_expr_alias as [[? [|]] ?]. inv Heq. now auto. *)
    (* edestruct H as (ρ' & Hρρ' & Hρ' & Hgi' & EVAL'); eauto. clear H. *)
    (* assert (Hnorm:=itv_normalizer_correct (do_bot x <- (idnc_get_query_chan x0).(AbIdealEnv.get_itv) i; *)
    (*          IdealIntervals.in_z x) min max logdiv Hlogdiv). *)

    destruct v. contradiction.

    assert (Hnorm:=itv_normalizer_correct
                (do_bot x <- (idnc_get_query_chan x0).(AbIdealEnv.get_itv) e;
             IdealIntervals.in_z x) min max logdiv Hlogdiv).
    destruct itv_normalizer; inv Heq. 2:now eauto.
    eexists. split. eauto. split. auto. split. auto.
    eexists. split. apply Hnorm. auto.
    eapply botbind_spec; eauto. eapply AbIdealEnv.get_itv_correct; eauto.
    apply idnc_get_query_chan_correct; eauto.
    split. exists z. rewrite two_p_equiv. ring.
    destruct (Z_eq_dec z 0). subst.
    replace (x-0*2^logdiv) with x by ring. auto.
    intros. econstructor; eauto.
    rewrite Z.shiftl_mul_pow2 by (zify; omega). constructor.
  Qed.

  Lemma normalize_expr_next_g_monotone:
    ∀ v e min max logdiv gi e' v' gi' b,
      normalize_expr v e min max logdiv gi = (e', v', gi', b) ->
      (gi.(next_g) <= gi'.(next_g))%positive.
  Proof.
    intros. unfold normalize_expr in H.
    (* pose proof (simplify_expr_alias_monotone v e gi). *)
    (* destruct simplify_expr_alias as [[? []]?]. inv H. eauto. *)
    destruct v. inv H. Psatz.lia.
    destruct itv_normalizer; inv H.
    Psatz.lia. Psatz.lia.
  Qed.

  Definition normalize_expr_full (v:abt+⊥) (e:iexpr) (f:signedness) (gi:ghosts_info):
    iexpr * abt+⊥ * ghosts_info :=
    let '(min, max) :=
      match f with Signed => (Int.min_signed, Int.max_signed) | Unsigned => (0, Int.max_unsigned) end
    in
    match normalize_expr v e min max Int.zwordsize gi with
      | (e, v, ng, true) => (e, v, ng)
      | _ => (IEZitv (ZIntervals.ITV min max), v, gi)
    end.

  Lemma normalize_expr_full_correct:
    ∀ v e f,
    ∀ ρ x gi, ρ ∈ γ v -> ρ ∈ γ gi ->
    (∀ ρ', compat_gh gi ρ ρ' -> eval_iexpr ρ' e (INz x)) ->
    ∀ e' v' gi', normalize_expr_full v e f gi = (e', v', gi') ->
    ∃ ρ', compat_gh gi ρ ρ' /\ ρ' ∈ γ v' /\ ρ' ∈ γ gi' /\
    ∃ x', x' ∈ γ (match f with Signed => signed_itv | Unsigned => unsigned_itv end) ∧ 
          Int.eqm x x' ∧
          (∀ ρ'', compat_gh gi' ρ' ρ'' -> eval_iexpr ρ'' e' (INz x')).
  Proof.
    intros v e [] ρ x gi Hρ Hgi EVAL e' v' gi' EQ; unfold normalize_expr_full in EQ.
    - destruct (normalize_expr v e Int.min_signed Int.max_signed Int.zwordsize gi) as [[[? ?]?]?]eqn:NORM.
      eapply normalize_expr_correct in NORM; eauto. 2:reflexivity.
      destruct b0; inv EQ. unfold Int.eqm. rewrite Int.modulus_power. apply NORM.
      eexists. split. eauto. split. eauto. split. auto.
      eexists. split. apply signed_itv_correct. split.
      eapply Int.eqm_trans, Int.eqm_sym, Int.eqm_signed_unsigned. apply Int.eqm_unsigned_repr.
      econstructor. apply signed_itv_correct.
    - destruct (normalize_expr v e 0 Int.max_unsigned Int.zwordsize gi) as [[[? ?]?]?]eqn:NORM.
      eapply normalize_expr_correct in NORM; eauto. 2:reflexivity.
      destruct b0; inv EQ. unfold Int.eqm. rewrite Int.modulus_power. apply NORM.
      eexists. split. eauto. split. eauto. split. auto.
      eexists. split. apply unsigned_itv_correct. split.
      apply Int.eqm_unsigned_repr.
      econstructor. apply unsigned_itv_correct.
  Qed.

  Lemma normalize_expr_full_next_g_monotone:
    ∀ v e f gi e' v' gi',
      normalize_expr_full v e f gi = (e', v', gi') ->
      (gi.(next_g) <= gi'.(next_g))%positive.
  Proof.
    intros. unfold normalize_expr_full in H.
    match type of H with
    | (let '(_, _) := ?P in _) = _ => destruct P
    end.
    pose proof normalize_expr_next_g_monotone v e z z0 Int.zwordsize gi.
    destruct normalize_expr as [[[? ?]?][]]; inv H. eauto. Psatz.lia.
  Qed.

  Definition normalize_expr_full64 (v:abt+⊥) (e:iexpr) (f:signedness) (gi:ghosts_info):
    iexpr * abt+⊥ * ghosts_info :=
    let '(min, max) :=
      match f with Signed => (Int64.min_signed, Int64.max_signed) | Unsigned => (0, Int64.max_unsigned) end
    in
    match normalize_expr v e min max Int64.zwordsize gi with
      | (e, v, ng, true) => (e, v, ng)
      | _ => (IEZitv (ZIntervals.ITV min max), v, gi)
    end.

  Lemma normalize_expr_full64_correct:
    ∀ v e f,
    ∀ ρ x gi, ρ ∈ γ v -> ρ ∈ γ gi ->
              (∀ ρ', compat_gh gi ρ ρ' -> eval_iexpr ρ' e (INz x)) ->
    ∀ e' v' gi', normalize_expr_full64 v e f gi = (e', v', gi') ->
    ∃ ρ', compat_gh gi ρ ρ' /\ ρ' ∈ γ v' /\ ρ' ∈ γ gi' /\
    ∃ x', x' ∈ γ (match f with Signed => signed64_itv | Unsigned => unsigned64_itv end) ∧ 
          Int64.eqm x x' ∧
          (∀ ρ'', compat_gh gi' ρ' ρ'' -> eval_iexpr ρ'' e' (INz x')).
  Proof.
    intros v e [] ρ x gi Hρ Hgi EVAL e' v' gi' EQ; unfold normalize_expr_full64 in EQ.
    - destruct (normalize_expr v e Int64.min_signed Int64.max_signed Int64.zwordsize gi) as [[[? ?]?]?]eqn:NORM.
      eapply normalize_expr_correct in NORM; eauto. 2:reflexivity.
      destruct b0; inv EQ. unfold Int64.eqm. rewrite Int64.modulus_power. apply NORM.
      eexists. split. eauto. split. eauto. split. auto.
      eexists. split. apply signed64_itv_correct. split.
      eapply Int64.eqm_trans, Int64.eqm_sym, Int64.eqm_signed_unsigned. apply Int64.eqm_unsigned_repr.
      econstructor. apply signed64_itv_correct.
    - destruct (normalize_expr v e 0 Int64.max_unsigned Int64.zwordsize gi) as [[[? ?]?]?]eqn:NORM.
      eapply normalize_expr_correct in NORM; eauto. 2:reflexivity.
      destruct b0; inv EQ. unfold Int64.eqm. rewrite Int64.modulus_power. apply NORM.
      eexists. split. eauto. split. eauto. split. auto.
      eexists. split. apply unsigned64_itv_correct. split.
      apply Int64.eqm_unsigned_repr.
      econstructor. apply unsigned64_itv_correct.
  Qed.

  Lemma normalize_expr_full64_next_g_monotone:
    ∀ v e f gi e' v' gi',
      normalize_expr_full64 v e f gi = (e', v', gi') ->
      (gi.(next_g) <= gi'.(next_g))%positive.
  Proof.
    intros. unfold normalize_expr_full64 in H.
    match type of H with
    | (let '(_, _) := ?P in _) = _ => destruct P
    end.
    pose proof normalize_expr_next_g_monotone v e z z0 Int64.zwordsize gi.
    destruct normalize_expr as [[[? ?]?][]]; inv H. eauto. Psatz.lia.
  Qed.

  Definition assume_cmp (c:comparison) (e:iexpr) (z:Z) (v:abt+⊥): abt+⊥ :=
    do_bot v <- v;
    idnc_assume (IEbinop (IOcmp c) e (IEconst (INz z))) true v.

  Lemma assume_cmp_correct:
    ∀ c e x z ρ ab,
      ρ ∈ γ ab ->
      INz x ∈ eval_iexpr ρ e ->
      ZIntervals.Zcmp c x z = true ->
      ρ ∈ γ (assume_cmp c e z ab).
  Proof.
    intros. apply botbind_spec with ρ; auto.
    intros. apply idnc_assume_correct. auto.
    econstructor. eauto. constructor.
    replace 1 with (if ZIntervals.Zcmp c x z then 1 else 0). constructor.
    rewrite H1. auto.
  Qed.

  Definition assume_cmpf (c:comparison) (e:iexpr) (f:float) (b:bool) (v:abt+⊥): abt+⊥ :=
    do_bot v <- v;
    idnc_assume (IEbinop (IOcmpf c) e (IEconst (INf f))) b v.

  Lemma assume_cmpf_correct:
    ∀ c e x f ρ b ab,
      ρ ∈ γ ab ->
      INf x ∈ eval_iexpr ρ e ->
      Float.cmp c x f = b ->
      ρ ∈ γ (assume_cmpf c e f b ab).
  Proof.
    intros. apply botbind_spec with ρ; auto.
    intros. apply idnc_assume_correct. auto.
    econstructor. eauto. constructor. subst. econstructor.
  Qed.

  Fixpoint convert_expr (v:abt+⊥) (e:nexpr Cells.cell) (gi:ghosts_info):
    (iexpr * abt+⊥ * bool * ghosts_info) :=
    match e with
      | NEvar (NVTint | NVTint64) x => (IEvar INTz (Real x), v, true, gi)
      | NEvar (NVTfloat | NVTsingle) x => (IEvar INTf (Real x), v, true, gi)
      | NEconst (NVint i) => (IEconst (INz (Int.signed i)), v, true, gi)
      | NEconst (NVint64 i) => (IEconst (INz (Int64.signed i)), v, true, gi)
      | NEconst (NVfloat f) => (IEconst (INf f), v, true, gi)
      | NEconst (NVsingle f) => (IEconst (INf (Float.of_single f)), v, true, gi)
      | NEunknown (NVTint | NVTint64) => (IEunknown INTz, v, true, gi)
      | NEunknown (NVTfloat | NVTsingle) => (IEunknown INTf, v, true, gi)


      | NEunop op e =>
        let '(e, v, nerr, gi) := convert_expr v e gi in
        match op with
          | Ocast8unsigned =>
            match normalize_expr v e 0 255 8 gi with
              | (e, v, gi, true) => (e, v, nerr, gi)
              | _ => (IEZitv (ZIntervals.ITV 0 255), v, nerr, gi)
            end
          | Ocast8signed =>
            match normalize_expr v e (-128) 127 8 gi with
              | (e, v, gi, true) => (e, v, nerr, gi)
              | _ => (IEZitv (ZIntervals.ITV (-128) 127), v, nerr, gi)
            end
          | Ocast16unsigned =>
            match normalize_expr v e 0 65535 16 gi with
              | (e, v, gi, true) => (e, v, nerr, gi)
              | _ => (IEZitv (ZIntervals.ITV 0 65535), v, nerr, gi)
            end
          | Ocast16signed =>
            match normalize_expr v e (-32768) 32767 16 gi with
              | (e, v, gi, true) => (e, v, nerr, gi)
              | _ => (IEZitv (ZIntervals.ITV (-32768) 32767), v, nerr, gi)
            end
          | Olongofint =>
            let '(e, v, gi) := normalize_expr_full v e Signed gi in
            (e, v, nerr, gi)
          | Olongofintu =>
            let '(e, v, gi) := normalize_expr_full v e Unsigned gi in
            (e, v, nerr, gi)
          | Ointoflong => (e, v, nerr, gi)

          | Onegint => (IEunop IOneg e, v, nerr, gi)
          | Onotint => (IEunop IOnot e, v, nerr, gi)

          | Onegl => (IEunop IOneg e, v, nerr, gi)
          | Onotl => (IEunop IOnot e, v, nerr, gi)

          | Onegf | Onegfs => (IEunop IOnegf e, v, nerr, gi)
          | Oabsf | Oabsfs => (IEunop IOabsf e, v, nerr, gi)
          | Ofloatofsingle => (e, v, nerr, gi)
          | Osingleoffloat => (IEunop IOsingleoffloat e, v, nerr, gi)
          | Ointoffloat =>
            let '(e, v, gi) := simplify_expr_alias v e gi in
            let nerr := nerr &&
              is_bot (assume_cmpf Cge e smallest_intoffloat false v) &&
              is_bot (assume_cmpf Cle e largest_intoffloat false v)
            in
            (IEunop IOzoffloat e, v, nerr, gi)
          | Ointofsingle =>
            let '(e, v, gi) := simplify_expr_alias v e gi in
            let nerr := nerr &&
              is_bot (assume_cmpf Cge e (Float.of_single smallest_intofsingle) false v) &&
              is_bot (assume_cmpf Cle e (Float.of_single largest_intofsingle) false v)
            in
            (IEunop IOzoffloat e, v, nerr, gi)
          | Ointuoffloat =>
            let '(e, v, gi) := simplify_expr_alias v e gi in
            let nerr := nerr &&
              is_bot (assume_cmpf Cge e smallest_intuoffloat false v) &&
              is_bot (assume_cmpf Cle e largest_intuoffloat false v)
            in
            (IEunop IOzoffloat e, v, nerr, gi)
          | Ointuofsingle =>
            let '(e, v, gi) := simplify_expr_alias v e gi in
            let nerr := nerr &&
              is_bot (assume_cmpf Cge e (Float.of_single smallest_intuofsingle) false v) &&
              is_bot (assume_cmpf Cle e (Float.of_single largest_intuofsingle) false v)
            in
            (IEunop IOzoffloat e, v, nerr, gi)
          | Ofloatofint =>
            let '(e, v, gi) := normalize_expr_full v e Signed gi in
            (IEunop IOfloatofz e, v, nerr, gi)
          | Osingleofint =>
            let '(e, v, gi) := normalize_expr_full v e Signed gi in
            (IEunop IOsingleofz e, v, nerr, gi)
          | Ofloatofintu =>
            let '(e, v, gi) := normalize_expr_full v e Unsigned gi in
            (IEunop IOfloatofz e, v, nerr, gi)
          | Osingleofintu =>
            let '(e, v, gi) := normalize_expr_full v e Unsigned gi in
            (IEunop IOsingleofz e, v, nerr, gi)
          | Olongoffloat =>
            let '(e, v, gi) := simplify_expr_alias v e gi in
            let nerr := nerr &&
              is_bot (assume_cmpf Cge e smallest_longoffloat false v) &&
              is_bot (assume_cmpf Cle e largest_longoffloat false v)
            in
            (IEunop IOzoffloat e, v, nerr, gi)
          | Olongofsingle =>
            let '(e, v, gi) := simplify_expr_alias v e gi in
            let nerr := nerr &&
              is_bot (assume_cmpf Cge e (Float.of_single smallest_longofsingle) false v) &&
              is_bot (assume_cmpf Cle e (Float.of_single largest_longofsingle) false v)
            in
            (IEunop IOzoffloat e, v, nerr, gi)
          | Olonguoffloat =>
            let '(e, v, gi) := simplify_expr_alias v e gi in
            let nerr := nerr &&
              is_bot (assume_cmpf Cge e smallest_longuoffloat false v) &&
              is_bot (assume_cmpf Cle e largest_longuoffloat false v)
            in
            (IEunop IOzoffloat e, v, nerr, gi)
          | Olonguofsingle =>
            let '(e, v, gi) := simplify_expr_alias v e gi in
            let nerr := nerr &&
              is_bot (assume_cmpf Cge e (Float.of_single smallest_longuofsingle) false v) &&
              is_bot (assume_cmpf Cle e (Float.of_single largest_longuofsingle) false v)
            in
            (IEunop IOzoffloat e, v, nerr, gi)
          | Ofloatoflong =>
            let '(e, v, gi) := normalize_expr_full64 v e Signed gi in
            (IEunop IOfloatofz e, v, nerr, gi)
          | Ofloatoflongu =>
            let '(e, v, gi) := normalize_expr_full64 v e Unsigned gi in
            (IEunop IOfloatofz e, v, nerr, gi)
          | Osingleoflong =>
            let '(e, v, gi) := normalize_expr_full64 v e Signed gi in
            (IEunop IOsingleofz e, v, nerr, gi)
          | Osingleoflongu =>
            let '(e, v, gi) := normalize_expr_full64 v e Unsigned gi in
            (IEunop IOsingleofz e, v, nerr, gi)
        end

      | NEbinop op e1 e2 =>
        let '(e1, v, nerr1, gi) := convert_expr v e1 gi in
        let '(e2, v, nerr2, gi) := convert_expr v e2 gi in
        match op with
          | Ocmp c =>
            let '(e1, v, gi) := normalize_expr_full v e1 Signed gi in
            let '(e2, v, gi) := normalize_expr_full v e2 Signed gi in
            (IEbinop (IOcmp c) e1 e2, v, (nerr1 && nerr2)%bool, gi)
          | Ocmpu c =>
            let '(e1, v, gi) := normalize_expr_full v e1 Unsigned gi in
            let '(e2, v, gi) := normalize_expr_full v e2 Unsigned gi in
            (IEbinop (IOcmp c) e1 e2, v, (nerr1 && nerr2)%bool, gi)
          | Ocmpl c =>
            let '(e1, v, gi) := normalize_expr_full64 v e1 Signed gi in
            let '(e2, v, gi) := normalize_expr_full64 v e2 Signed gi in
            (IEbinop (IOcmp c) e1 e2, v, (nerr1 && nerr2)%bool, gi)
          | Ocmplu c =>
            let '(e1, v, gi) := normalize_expr_full64 v e1 Unsigned gi in
            let '(e2, v, gi) := normalize_expr_full64 v e2 Unsigned gi in
            (IEbinop (IOcmp c) e1 e2, v, (nerr1 && nerr2)%bool, gi)

          | Oaddf => (IEbinop IOaddf e1 e2, v, (nerr1 && nerr2)%bool, gi)
          | Oaddfs => (IEunop IOsingleoffloat (IEbinop IOaddf e1 e2), v, (nerr1 && nerr2)%bool, gi)
          | Osubf => (IEbinop IOsubf e1 e2, v, (nerr1 && nerr2)%bool, gi)
          | Osubfs => (IEunop IOsingleoffloat (IEbinop IOsubf e1 e2), v, (nerr1 && nerr2)%bool, gi)
          | Omulf => (IEbinop IOmulf e1 e2, v, (nerr1 && nerr2)%bool, gi)
          | Omulfs => (IEunop IOsingleoffloat (IEbinop IOmulf e1 e2), v, (nerr1 && nerr2)%bool, gi)
          | Odivf => (IEbinop IOdivf e1 e2, v, (nerr1 && nerr2)%bool, gi)
          | Odivfs => (IEunop IOsingleoffloat (IEbinop IOdivf e1 e2), v, (nerr1 && nerr2)%bool, gi)
          | Ocmpf c | Ocmpfs c => (IEbinop (IOcmpf c) e1 e2, v, (nerr1 && nerr2)%bool, gi)
          | Oadd | Oaddl => (IEbinop IOadd e1 e2, v, (nerr1 && nerr2)%bool, gi)
          | Osub | Osubl => (IEbinop IOsub e1 e2, v, (nerr1 && nerr2)%bool, gi)
          | Omul | Omull => (IEbinop IOmul e1 e2, v, (nerr1 && nerr2)%bool, gi)

          | Odiv =>
            let '(e1, v, gi) := normalize_expr_full v e1 Signed gi in
            let '(e2, v, gi) := normalize_expr_full v e2 Signed gi in
            if nerr1 && nerr2 && is_bot (assume_cmp Ceq e2 0 v) &&
               is_bot (assume_cmp Cle e1 Int.min_signed (assume_cmp Ceq e2 (-1) v))
            then (IEbinop IOdiv e1 e2, v, true, gi)
            else
              let v := assume_cmp Cne e2 0 v in
              let v := (assume_cmp Cgt e1 Int.min_signed v) ⊔
                       (assume_cmp Cne e2 (-1) v) in
              (IEbinop IOdiv e1 e2, v, false, gi)
          | Odivl =>
            let '(e1, v, gi) := normalize_expr_full64 v e1 Signed gi in
            let '(e2, v, gi) := normalize_expr_full64 v e2 Signed gi in
            if nerr1 && nerr2 && is_bot (assume_cmp Ceq e2 0 v) &&
               is_bot (assume_cmp Cle e1 Int64.min_signed (assume_cmp Ceq e2 (-1) v))
            then (IEbinop IOdiv e1 e2, v, true, gi)
            else
              let v := assume_cmp Cne e2 0 v in
              let v := (assume_cmp Cgt e1 Int64.min_signed v) ⊔
                       (assume_cmp Cne e2 (-1) v) in
              (IEbinop IOdiv e1 e2, v, false, gi)
          | Odivu =>
            let '(e1, v, gi) := normalize_expr_full v e1 Unsigned gi in
            let '(e2, v, gi) := normalize_expr_full v e2 Unsigned gi in
            if nerr1 && nerr2 && is_bot (assume_cmp Ceq e2 0 v) then
              (IEbinop IOdiv e1 e2, v, true, gi)
            else
              let v := assume_cmp Cne e2 0 v in
              (IEbinop IOdiv e1 e2, v, false, gi)
          | Odivlu =>
            let '(e1, v, gi) := normalize_expr_full64 v e1 Unsigned gi in
            let '(e2, v, gi) := normalize_expr_full64 v e2 Unsigned gi in
            if nerr1 && nerr2 && is_bot (assume_cmp Ceq e2 0 v) then
              (IEbinop IOdiv e1 e2, v, true, gi)
            else
              let v := assume_cmp Cne e2 0 v in
              (IEbinop IOdiv e1 e2, v, false, gi)
          | Omod =>
            let '(e1, v, gi) := normalize_expr_full v e1 Signed gi in
            let '(e2, v, gi) := normalize_expr_full v e2 Signed gi in
            if nerr1 && nerr2 && is_bot (assume_cmp Ceq e2 0 v) &&
               is_bot (assume_cmp Cle e1 Int.min_signed (assume_cmp Ceq e2 (-1) v))
            then (IEbinop IOmod e1 e2, v, true, gi)
            else
              let v := assume_cmp Cne e2 0 v in
              let v := (assume_cmp Cgt e1 Int.min_signed v) ⊔
                       (assume_cmp Cne e2 (-1) v) in
              (IEbinop IOmod e1 e2, v, false, gi)
          | Omodl =>
            let '(e1, v, gi) := normalize_expr_full64 v e1 Signed gi in
            let '(e2, v, gi) := normalize_expr_full64 v e2 Signed gi in
            if nerr1 && nerr2 && is_bot (assume_cmp Ceq e2 0 v) &&
               is_bot (assume_cmp Cle e1 Int64.min_signed (assume_cmp Ceq e2 (-1) v))
            then (IEbinop IOmod e1 e2, v, true, gi)
            else
              let v := assume_cmp Cne e2 0 v in
              let v := (assume_cmp Cgt e1 Int64.min_signed v) ⊔
                       (assume_cmp Cne e2 (-1) v) in
              (IEbinop IOmod e1 e2, v, false, gi)
          | Omodu =>
            let '(e1, v, gi) := normalize_expr_full v e1 Unsigned gi in
            let '(e2, v, gi) := normalize_expr_full v e2 Unsigned gi in
            if nerr1 && nerr2 && is_bot (assume_cmp Ceq e2 0 v) then
              (IEbinop IOmod e1 e2, v, true, gi)
            else
              let v := assume_cmp Cne e2 0 v in
              (IEbinop IOmod e1 e2, v, false, gi)
          | Omodlu =>
            let '(e1, v, gi) := normalize_expr_full64 v e1 Unsigned gi in
            let '(e2, v, gi) := normalize_expr_full64 v e2 Unsigned gi in
            if nerr1 && nerr2 && is_bot (assume_cmp Ceq e2 0 v) then
              (IEbinop IOmod e1 e2, v, true, gi)
            else
              let v := assume_cmp Cne e2 0 v in
              (IEbinop IOmod e1 e2, v, false, gi)


          | Oand | Oandl => (IEbinop IOand e1 e2, v, (nerr1 && nerr2)%bool, gi)
          | Oor | Oorl => (IEbinop IOor e1 e2, v, (nerr1 && nerr2)%bool, gi)
          | Oxor | Oxorl => (IEbinop IOxor e1 e2, v, (nerr1 && nerr2)%bool, gi)
          | Oshl =>
            let '(e2, v, gi) := normalize_expr_full v e2 Unsigned gi in
            if nerr1 && nerr2 && is_bot (assume_cmp Cge e2 Int.zwordsize v) then
              (IEbinop IOshl e1 e2, v, true, gi)
            else
              let v := assume_cmp Clt e2 Int.zwordsize v in
              (IEbinop IOshl e1 e2, v, false, gi)
          | Oshll =>
            let '(e2, v, gi) := normalize_expr_full v e2 Unsigned gi in
            if nerr1 && nerr2 && is_bot (assume_cmp Cge e2 Int64.zwordsize v) then
              (IEbinop IOshl e1 e2, v, true, gi)
            else
              let v := assume_cmp Clt e2 Int64.zwordsize v in
              (IEbinop IOshl e1 e2, v, false, gi)
          | Oshr =>
            let '(e1, v, gi) := normalize_expr_full v e1 Signed gi in
            let '(e2, v, gi) := normalize_expr_full v e2 Unsigned gi in
            if nerr1 && nerr2 && is_bot (assume_cmp Cge e2 Int.zwordsize v) then
              (IEbinop IOshr e1 e2, v, true, gi)
            else
              let v := assume_cmp Clt e2 Int.zwordsize v in
              (IEbinop IOshr e1 e2, v, false, gi)
          | Oshrl =>
            let '(e1, v, gi) := normalize_expr_full64 v e1 Signed gi in
            let '(e2, v, gi) := normalize_expr_full v e2 Unsigned gi in
            if nerr1 && nerr2 && is_bot (assume_cmp Cge e2 Int64.zwordsize v) then
              (IEbinop IOshr e1 e2, v, true, gi)
            else
              let v := assume_cmp Clt e2 Int64.zwordsize v in
              (IEbinop IOshr e1 e2, v, false, gi)
          | Oshru =>
            let '(e1, v, gi) := normalize_expr_full v e1 Unsigned gi in
            let '(e2, v, gi) := normalize_expr_full v e2 Unsigned gi in
            if nerr1 && nerr2 && is_bot (assume_cmp Cge e2 Int.zwordsize v) then
              (IEbinop IOshr e1 e2, v, true, gi)
            else
              let v := assume_cmp Clt e2 Int.zwordsize v in
              (IEbinop IOshr e1 e2, v, false, gi)
          | Oshrlu =>
            let '(e1, v, gi) := normalize_expr_full64 v e1 Unsigned gi in
            let '(e2, v, gi) := normalize_expr_full v e2 Unsigned gi in
            if nerr1 && nerr2 && is_bot (assume_cmp Cge e2 Int64.zwordsize v) then
              (IEbinop IOshr e1 e2, v, true, gi)
            else
              let v := assume_cmp Clt e2 Int64.zwordsize v in
              (IEbinop IOshr e1 e2, v, false, gi)
        end
    end.

  Lemma convert_expr_next_g_monotone:
    ∀ e v gi e' v' nerr' gi',
      convert_expr v e gi = (e', v', nerr', gi') ->
      (gi.(next_g) <= gi'.(next_g))%positive.
  Proof.
    induction e; simpl; intros.
    - destruct n; inv H; Psatz.lia.
    - destruct n; try destruct t; inv H; Psatz.lia.
    - destruct n; inv H; Psatz.lia.
    - specialize (IHe v gi).
      destruct (convert_expr v e gi) as [[[e'0 v'0] nerr'0] next_g'0].
      specialize (IHe _ _ _ _ eq_refl).
      destruct u;
      repeat match type of H with
      | context [simplify_expr_alias ?A ?B ?C] =>
        let SIMPL_MONO := fresh "SIMPL_MONO" in
        assert (SIMPL_MONO := simplify_expr_alias_monotone A B C);
        destruct (simplify_expr_alias A B C) as [[? ?] ?];
        specialize (SIMPL_MONO _ _ _ eq_refl)
      | context [normalize_expr ?A ?B ?C ?D ?E ?F] =>
        let NORM_MONO := fresh "NORM_MONO" in
        assert (NORM_MONO:=normalize_expr_next_g_monotone A B C D E F);
        destruct (normalize_expr A B C D E F) as [[[? ?] ?][]];
        specialize (NORM_MONO _ _ _ _ eq_refl)
      | context [normalize_expr_full ?A ?B ?C ?D] =>
        let NORM_MONO := fresh "NORM_MONO" in
        assert (NORM_MONO:=normalize_expr_full_next_g_monotone A B C D);
        destruct (normalize_expr_full A B C D) as [[? ?] ?];
        specialize (NORM_MONO _ _ _ eq_refl)
      | context [normalize_expr_full64 ?A ?B ?C ?D] =>
        let NORM_MONO := fresh "NORM_MONO" in
        assert (NORM_MONO:=normalize_expr_full64_next_g_monotone A B C D);
        destruct (normalize_expr_full64 A B C D) as [[? ?] ?];
        specialize (NORM_MONO _ _ _ eq_refl)
      end;
      inv H;
      Psatz.lia.
    - specialize (IHe1 v gi).
      destruct (convert_expr v e1 gi) as [[[e'0 v'0] nerr'0] gi'0].
      specialize (IHe1 _ _ _ _ eq_refl).
      specialize (IHe2 v'0 gi'0).
      destruct (convert_expr v'0 e2 gi'0) as [[[e'1 v'1] nerr'1] gi'1].
      specialize (IHe2 _ _ _ _ eq_refl).
      destruct b;
      repeat match type of H with
      | (if ?B then _ else _) = _ => destruct B as [|]
      | context [normalize_expr ?A ?B ?C ?D ?E ?F] =>
        let NORM_MONO := fresh "NORM_MONO" in
        assert (NORM_MONO:=normalize_expr_next_g_monotone A B C D E F);
        destruct (normalize_expr A B C D E F) as [[[? ?] ?][]];
        specialize (NORM_MONO _ _ _ _ eq_refl); simpl in H
      | context [normalize_expr_full ?A ?B ?C ?D] =>
        let NORM_MONO := fresh "NORM_MONO" in
        assert (NORM_MONO:=normalize_expr_full_next_g_monotone A B C D);
        destruct (normalize_expr_full A B C D) as [[? ?] ?];
        specialize (NORM_MONO _ _ _ eq_refl)
      | context [normalize_expr_full64 ?A ?B ?C ?D] =>
        let NORM_MONO := fresh "NORM_MONO" in
        assert (NORM_MONO:=normalize_expr_full64_next_g_monotone A B C D);
        destruct (normalize_expr_full64 A B C D) as [[? ?] ?];
        specialize (NORM_MONO _ _ _ eq_refl)
      end;
      inv H; Psatz.lia.
  Qed.

  Lemma convert_expr_ok:
    ∀ e ρ x, eval_nexpr ρ e x ->
    ∀ v ρ'1 gi, ρ'1 ∈ γ v -> ρ'1 ∈ γ gi -> (∀ id, related (ρ id) (ρ'1 (Real id))) ->
    ∀ e' v' nerr' gi', convert_expr v e gi = (e', v', nerr', gi') ->
    ∃ ρ'2,
      ρ'2 ∈ γ v' /\ ρ'2 ∈ γ gi' /\
      compat_gh gi ρ'1 ρ'2 /\
    ∃ x',
      (∀ ρ'3, compat_gh gi' ρ'2 ρ'3 -> eval_iexpr ρ'3 e' x') ∧
      related x x'.
  Proof.
    induction 1; simpl convert_expr;
    intros ab ρ'1 gi Hρ'1 Hgi Hρρ'1 e' v' nerr gi' EQ.

    - specialize (Hρρ'1 id). exists ρ'1.
      split; [|split;[|split;[|exists (ρ'1 (Real id)); split]]].
      + destruct ty; congruence.
      + destruct ty; inv EQ; auto.
      + auto.
      + destruct ty; inv EQ; inv H0; inv Hρρ'1; try congruence; repeat constructor;
        rewrite H2; symmetry; apply H0.
      + congruence.

    - exists ρ'1. split. destruct n; congruence. split.
      destruct n; inv EQ; auto. split. auto.
      destruct n; inv EQ; eexists;
      (split; [|auto using related_signed, related_signed_long]);
      constructor; constructor.

    - exists ρ'1. split. destruct ty; congruence. split.
      destruct ty; inv EQ; auto. split. auto.
      destruct H; inv EQ; eexists; repeat constructor.
      apply related_signed. apply related_signed_long.

    - specialize (IHeval_nexpr _ _ _ Hρ'1 Hgi Hρρ'1).
      assert (MONOe1:=convert_expr_next_g_monotone a1 ab gi).
      destruct (convert_expr ab a1 gi) as [[[e2 ab2] nerr2] gi''].
      specialize (MONOe1 _ _ _ _ eq_refl).
      specialize (IHeval_nexpr _ _ _ _ eq_refl).
      destruct IHeval_nexpr as (ρ'2 & Hab2 & Hgi'' & Hρ'1ρ'2 & x' & EVAL & Hx').
      destruct H0; inv Hx';
      repeat (match type of EQ with
      | context [simplify_expr_alias ?ab ?e ?gi] =>
        let SIMPL := fresh "SIMPL" in let SIMPL_MONO := fresh "SIMPL_MONO" in
        destruct (simplify_expr_alias ab e gi) as [[] ?] eqn:SIMPL;
        assert (SIMPL_MONO:=SIMPL);
        apply simplify_expr_alias_monotone in SIMPL_MONO;
        (eapply simplify_expr_alias_correct in SIMPL;[|now eauto|now eauto|now eauto]);
        destruct SIMPL as (? & ? & ? & ? & ?)
      | context [normalize_expr ?ab ?e ?min ?max ?logmod ?gi] =>
        let NORM := fresh "NORM" in let NORM_MONO := fresh "NORM_MONO" in
        destruct (normalize_expr ab e min max logmod gi) as [[[] ?][]] eqn:NORM;
        assert (NORM_MONO:=NORM);
        apply normalize_expr_next_g_monotone in NORM_MONO;
        (eapply normalize_expr_correct in NORM;[|reflexivity|now eauto|now eauto|now eauto]);
        [destruct NORM as (? & ? & ? & ? & ? & ? & ? & ?)|destruct NORM as (? & ? & ? & ?)]
      | context [normalize_expr_full ?ab ?e ?sgn ?gi] =>
        let NORM := fresh "NORM" in let NORM_MONO := fresh "NORM_MONO" in
        destruct (normalize_expr_full ab e sgn gi) as [[] ?] eqn:NORM;
        assert (NORM_MONO:=NORM);
        apply normalize_expr_full_next_g_monotone in NORM_MONO;
        eapply normalize_expr_full_correct in NORM;
        [destruct NORM as (? & ? & ? & ? & ? & ? & ? & ?)|now eauto|now eauto|now eauto]
      | context [normalize_expr_full64 ?ab ?e ?sgn ?gi] =>
        let NORM := fresh "NORM" in let NORM_MONO := fresh "NORM_MONO" in
        destruct (normalize_expr_full64 ab e sgn gi) as [[] ?] eqn:NORM;
        assert (NORM_MONO:=NORM);
        apply normalize_expr_full64_next_g_monotone in NORM_MONO;
        eapply normalize_expr_full64_correct in NORM;
        [destruct NORM as (? & ? & ? & ? & ? & ? & ? & ?)|now eauto|now eauto|now eauto]
      end);
      inv EQ.

      + eexists. split. eauto. split. eauto. split. eauto. eexists. split. eauto.
        replace (Int.zero_ext 8 (Int.repr x)) with (Int.repr x1). constructor.
        rewrite <- Int.repr_unsigned with (Int.zero_ext 8 (Int.repr x)). f_equal.
        assert (8 < Int.zwordsize) by reflexivity.
        apply Int.eqmod_small_eq with 256. 2:omega. 2:apply Int.zero_ext_range; omega.
        eapply Int.eqmod_trans. eapply Int.eqmod_sym, H4.
        eapply Int.eqmod_trans, Int.eqmod_sym, Int.eqmod_zero_ext. 2:omega.
        eapply Int.eqmod_divides. eapply Int.eqm_unsigned_repr.
        apply Z.mod_divide. discriminate. reflexivity.
      + eexists. split. eauto. split. auto. split. auto. eexists.
        split. 2:apply related_unsigned. constructor.
        destruct (Int.zero_ext_range 8 (Int.repr x)). compute; split; congruence.
        change (two_p 8) with 256 in H5. split; omega.

      + eexists. split. eauto. split. eauto. split. eauto. eexists. split. eauto.
        replace (Int.sign_ext 8 (Int.repr x)) with (Int.repr x1). constructor.
        rewrite <- Int.repr_signed with (Int.sign_ext 8 (Int.repr x)). f_equal.
        assert (8 < Int.zwordsize) by reflexivity.
        apply Z.add_cancel_r with (p:=128).
        apply Int.eqmod_small_eq with 256. 2:omega.
        apply Int.eqmod_add. 2:apply Int.eqmod_refl.
        eapply Int.eqmod_trans. eapply Int.eqmod_sym, H4.
        eapply Int.eqmod_trans, Int.eqmod_sym, Int.eqmod_sign_ext. 2:omega.
        eapply Int.eqmod_divides. eapply Int.eqm_unsigned_repr.
        apply Z.mod_divide. discriminate. reflexivity.
        destruct (Int.sign_ext_range 8 (Int.repr x)). omega.
        change (two_p (8-1)) with 128 in H7, H8. omega.
      + eexists. split. eauto. split. auto. split. auto. eexists.
        split. 2:apply related_signed. constructor.
        destruct (Int.sign_ext_range 8 (Int.repr x)). compute; split; congruence.
        change (two_p (8-1)) with 128 in H4, H5. split; omega.

      + eexists. split. eauto. split. eauto. split. eauto. eexists. split. eauto.
        replace (Int.zero_ext 16 (Int.repr x)) with (Int.repr x1). constructor.
        rewrite <- Int.repr_unsigned with (Int.zero_ext 16 (Int.repr x)). f_equal.
        assert (16 < Int.zwordsize) by reflexivity.
        apply Int.eqmod_small_eq with 65536. 2:omega. 2:apply Int.zero_ext_range; omega.
        eapply Int.eqmod_trans. eapply Int.eqmod_sym, H4.
        eapply Int.eqmod_trans, Int.eqmod_sym, Int.eqmod_zero_ext. 2:omega.
        eapply Int.eqmod_divides. eapply Int.eqm_unsigned_repr.
        apply Z.mod_divide. discriminate. reflexivity.
      + eexists. split. eauto. split. auto. split. auto. eexists.
        split. 2:apply related_unsigned. constructor.
        destruct (Int.zero_ext_range 16 (Int.repr x)). compute; split; congruence.
        change (two_p 16) with 65536 in H5. split; omega.

      + eexists. split. eauto. split. eauto. split. eauto. eexists. split. eauto.
        replace (Int.sign_ext 16 (Int.repr x)) with (Int.repr x1). constructor.
        rewrite <- Int.repr_signed with (Int.sign_ext 16 (Int.repr x)). f_equal.
        assert (16 < Int.zwordsize) by reflexivity.
        apply Z.add_cancel_r with (p:=32768).
        apply Int.eqmod_small_eq with 65536. 2:omega.
        apply Int.eqmod_add. 2:apply Int.eqmod_refl.
        eapply Int.eqmod_trans. eapply Int.eqmod_sym, H4.
        eapply Int.eqmod_trans, Int.eqmod_sym, Int.eqmod_sign_ext. 2:omega.
        eapply Int.eqmod_divides. eapply Int.eqm_unsigned_repr.
        apply Z.mod_divide. discriminate. reflexivity.
        destruct (Int.sign_ext_range 16 (Int.repr x)). omega.
        change (two_p (16-1)) with 32768 in H7, H8. omega.
      + eexists. split. eauto. split. auto. split. auto. eexists.
        split. 2:apply related_signed. constructor.
        destruct (Int.sign_ext_range 16 (Int.repr x)). compute; split; congruence.
        change (two_p (16-1)) with 32768 in H4, H5. split; omega.

      + eexists. split. eauto. split. auto. split. auto. eexists. split. eauto.
        rewrite Int.neg_repr. constructor.

      + eexists. split. eauto. split. auto. split. auto. eexists. split. eauto.
        replace (Int.not (Int.repr x)) with (Int.repr (Z.lnot x)). constructor.
        apply Int.same_bits_eq. intros i Hi.
        rewrite Int.bits_not, !Int.testbit_repr, Z.lnot_spec; intuition.

      + eauto 10.

      + eexists. split. eauto. split. auto. split. auto. eexists. split. 2:constructor.
        Transparent Float.neg Float32.neg Float.of_single.
        replace (Float.of_single (Float32.neg f))
        with (Float.neg (Float.of_single f))
          by (apply Bconv_Bopp; auto).
        econstructor; eauto.

      + eauto 10.

      + eexists. split. eauto. split. auto. split. auto. eexists. split. 2:constructor.
        Transparent Float.abs Float32.abs Float.of_single.
        replace (Float.of_single (Float32.abs f))
        with (Float.abs (Float.of_single f))
          by (apply Bconv_Babs; auto).
        econstructor; eauto.

      + eauto 10.

      + eauto 10.

      + Transparent Float.to_int.
        eexists. split. eauto. split. auto. split. eauto. exists (INz (Int.signed i)).
        split. 2:apply related_signed. econstructor. eauto. constructor.
        unfold Float.to_int, Fappli_IEEE_extra.ZofB_range in H0.
        destruct (Fappli_IEEE_extra.ZofB _ _ f) eqn:?; [|discriminate].
        destruct ((Int.min_signed <=? z) && (z <=? Int.max_signed))%bool eqn:?; inv H0.
        rewrite Bool.andb_true_iff, <- !Zle_is_le_bool in Heqb.
        rewrite Int.signed_repr by auto. auto.

      + Transparent Float32.to_int.
        eexists. split. eauto. split. auto. split. eauto. exists (INz (Int.signed i)).
        split. 2:apply related_signed. econstructor. eauto. constructor.
        unfold Float32.to_int, Fappli_IEEE_extra.ZofB_range in H0.
        destruct (Fappli_IEEE_extra.ZofB _ _ f) eqn:?; [|discriminate].
        destruct ((Int.min_signed <=? z) && (z <=? Int.max_signed))%bool eqn:?; inv H0.
        rewrite Bool.andb_true_iff, <- !Zle_is_le_bool in Heqb.
        rewrite Int.signed_repr by auto.
        rewrite Fappli_IEEE_extra.ZofB_correct in *.
        destruct (is_finite _ _ f) eqn:?.
        rewrite (proj2 (floatofsingle_exact _)), (proj1 (floatofsingle_exact _)); auto.
        discriminate.

      + Transparent Float.to_intu.
        eexists. split. eauto. split. auto. split. eauto. exists (INz (Int.unsigned i)).
        split. 2:apply related_unsigned. econstructor. eauto. constructor.
        unfold Float.to_intu, Fappli_IEEE_extra.ZofB_range in H0.
        destruct (Fappli_IEEE_extra.ZofB _ _ f) eqn:?; [|discriminate].
        destruct ((0 <=? z) && (z <=? Int.max_unsigned))%bool eqn:?; inv H0.
        rewrite Bool.andb_true_iff, <- !Zle_is_le_bool in Heqb.
        rewrite Int.unsigned_repr by auto. auto.

      + Transparent Float32.to_intu.
        eexists. split. eauto. split. eauto. split. eauto. exists (INz (Int.unsigned i)).
        split. 2:apply related_unsigned. econstructor. eauto. constructor.
        unfold Float32.to_intu, Fappli_IEEE_extra.ZofB_range in H0.
        destruct (Fappli_IEEE_extra.ZofB _ _ f) eqn:?; [|discriminate].
        destruct ((0 <=? z) && (z <=? Int.max_unsigned))%bool eqn:?; inv H0.
        rewrite Bool.andb_true_iff, <- !Zle_is_le_bool in Heqb.
        rewrite Int.unsigned_repr by auto.
        rewrite Fappli_IEEE_extra.ZofB_correct in *.
        destruct (is_finite _ _ f) eqn:?.
        rewrite (proj2 (floatofsingle_exact _)), (proj1 (floatofsingle_exact _)); auto.
        discriminate.

      + Transparent Float.of_int.
        eexists. split. eauto. split. eauto. split. eauto. eexists. split. 2:constructor.
        rewrite !Int.eqm_samerepr with (x:=x) (y:=x1) by trivial.
        unfold Float.of_int.
        rewrite Int.signed_repr by auto.
        econstructor. eauto. constructor.

      + Transparent Float32.of_int.
        eexists. split. eauto. split. eauto. split. eauto. eexists. split. 2:constructor.
        rewrite !Int.eqm_samerepr with (x:=x) (y:=x1) by trivial.
        unfold Float32.of_int.
        rewrite Int.signed_repr by auto.
        econstructor. eauto. constructor.

      + Transparent Float.of_intu.
        eexists. split. eauto. split. eauto. split. eauto. eexists. split. 2:constructor.
        rewrite !Int.eqm_samerepr with (x:=x) (y:=x1) by trivial.
        unfold Float.of_intu.
        rewrite Int.unsigned_repr by auto.
        econstructor. eauto. constructor.

      + Transparent Float32.of_intu.
        eexists. split. eauto. split. eauto. split. eauto. eexists. split. 2:constructor.
        rewrite !Int.eqm_samerepr with (x:=x) (y:=x1) by trivial.
        unfold Float32.of_intu.
        rewrite Int.unsigned_repr by auto.
        econstructor. eauto. constructor.

      + eexists. split. eauto. split. eauto. split. eauto. eexists.
        split; eauto. rewrite Int64.neg_repr. constructor.

      + eexists. split. eauto. split. eauto. split. eauto. eexists. split; eauto.
        replace (Int64.not (Int64.repr x)) with (Int64.repr (Z.lnot x)). constructor.
        apply Int64.same_bits_eq. intros i Hi.
        rewrite Int64.bits_not, !Int64.testbit_repr, Z.lnot_spec; intuition.

      + eexists. split. eauto. split. eauto. split. eauto. eexists. split; eauto.
        replace (Int64.loword (Int64.repr x)) with (Int.repr x). constructor.
        eapply Int.eqm_samerepr, Int.eqmod_divides. apply Int64.eqm_unsigned_repr.
        exists Int.modulus. reflexivity.

      + eexists. split. eauto. split. eauto. split. eauto. eexists. split; eauto.
        rewrite !Int.eqm_samerepr with (x:=x) (y:=x1) by trivial.
        rewrite Int.signed_repr by auto.
        auto.

      + eexists. split. eauto. split. eauto. split. eauto. eexists. split; eauto.
        rewrite !Int.eqm_samerepr with (x:=x) (y:=x1) by trivial.
        rewrite Int.unsigned_repr by auto.
        auto.

      + Transparent Float.to_long.
        eexists. split. eauto. split. eauto. split. eauto. exists (INz (Int64.signed i)).
        split. 2:apply related_signed_long. econstructor. eauto. constructor.
        unfold Float.to_long, Fappli_IEEE_extra.ZofB_range in H0.
        destruct (Fappli_IEEE_extra.ZofB _ _ f) eqn:?; [|discriminate].
        destruct ((Int64.min_signed <=? z) && (z <=? Int64.max_signed))%bool eqn:?; inv H0.
        rewrite Bool.andb_true_iff, <- !Zle_is_le_bool in Heqb.
        rewrite Int64.signed_repr by auto. auto.

      + Transparent Float.to_longu.
        eexists. split. eauto. split. eauto. split. eauto. exists (INz (Int64.unsigned i)).
        split. 2:apply related_unsigned_long. econstructor. eauto. constructor.
        unfold Float.to_longu, Fappli_IEEE_extra.ZofB_range in H0.
        destruct (Fappli_IEEE_extra.ZofB _ _ f) eqn:?; [|discriminate].
        destruct ((0 <=? z) && (z <=? Int64.max_unsigned))%bool eqn:?; inv H0.
        rewrite Bool.andb_true_iff, <- !Zle_is_le_bool in Heqb.
        rewrite Int64.unsigned_repr by auto. auto.

      + Transparent Float32.to_long.
        eexists. split. eauto. split. eauto. split. eauto. exists (INz (Int64.signed i)).
        split. 2:apply related_signed_long. econstructor. eauto. constructor.
        unfold Float32.to_long, Fappli_IEEE_extra.ZofB_range in H0.
        destruct (Fappli_IEEE_extra.ZofB _ _ f) eqn:?; [|discriminate].
        destruct ((Int64.min_signed <=? z) && (z <=? Int64.max_signed))%bool eqn:?; inv H0.
        rewrite Bool.andb_true_iff, <- !Zle_is_le_bool in Heqb.
        rewrite Int64.signed_repr by auto.
        rewrite Fappli_IEEE_extra.ZofB_correct in *.
        destruct (is_finite _ _ f) eqn:?.
        rewrite (proj2 (floatofsingle_exact _)), (proj1 (floatofsingle_exact _)); auto.
        discriminate.

      + Transparent Float32.to_longu.
        eexists. split. eauto. split. eauto. split. eauto. exists (INz (Int64.unsigned i)).
        split. 2:apply related_unsigned_long. econstructor. eauto. constructor.
        unfold Float32.to_longu, Fappli_IEEE_extra.ZofB_range in H0.
        destruct (Fappli_IEEE_extra.ZofB _ _ f) eqn:?; [|discriminate].
        destruct ((0 <=? z) && (z <=? Int64.max_unsigned))%bool eqn:?; inv H0.
        rewrite Bool.andb_true_iff, <- !Zle_is_le_bool in Heqb.
        rewrite Int64.unsigned_repr by auto.
        rewrite Fappli_IEEE_extra.ZofB_correct in *.
        destruct (is_finite _ _ f) eqn:?.
        rewrite (proj2 (floatofsingle_exact _)), (proj1 (floatofsingle_exact _)); auto.
        discriminate.

      + Transparent Float.of_long.
        eexists. split. eauto. split. eauto. split. eauto. eexists. split; eauto.
        rewrite !Int64.eqm_samerepr with (x:=x) (y:=x1) by trivial.
        unfold Float.of_long. rewrite Int64.signed_repr by auto. econstructor.

      + Transparent Float32.of_long.
        eexists. split. eauto. split. eauto. split. eauto. eexists. split; eauto.
        rewrite !Int64.eqm_samerepr with (x:=x) (y:=x1) by trivial.
        unfold Float32.of_long. rewrite Int64.signed_repr by auto. econstructor.

      + Transparent Float.of_longu.
        eexists. split. eauto. split. eauto. split. eauto. eexists. split; eauto.
        rewrite !Int64.eqm_samerepr with (x:=x) (y:=x1) by trivial.
        unfold Float.of_longu. rewrite Int64.unsigned_repr by auto. econstructor.

      + Transparent Float32.of_longu.
        eexists. split. eauto. split. eauto. split. eauto. eexists. split; eauto.
        rewrite !Int64.eqm_samerepr with (x:=x) (y:=x1) by trivial.
        unfold Float32.of_longu. rewrite Int64.unsigned_repr by auto. econstructor.

    - specialize (IHeval_nexpr1 _ _ _ Hρ'1 Hgi Hρρ'1).
      assert (MONO1:=convert_expr_next_g_monotone a1 ab gi).
      destruct (convert_expr ab a1 gi) as [[[e1' ab'] nerr1] gi''].
      specialize (IHeval_nexpr1 _ _ _ _ eq_refl). specialize (MONO1 _ _ _ _ eq_refl).
      edestruct IHeval_nexpr1 as (ρ'e1 & Hρ'e1 & Hgi'' & Hρ'1ρ'e1 & x'e1 & EVAL1 & Hx'e1).
      assert (Hρρ'e1 : ∀ id : Cells.cell, related (ρ id) (ρ'e1 (Real id))).
      { intros. rewrite <- (proj1 Hρ'1ρ'e1). auto. }
      specialize (IHeval_nexpr2 _ _ _ Hρ'e1 Hgi'' Hρρ'e1).
      assert (MONO2:=convert_expr_next_g_monotone a2 ab' gi'').
      destruct (convert_expr ab' a2 gi'') as [[[e2' ab''] nerr2] gi'''].
      specialize (IHeval_nexpr2 _ _ _ _ eq_refl). specialize (MONO2 _ _ _ _ eq_refl).
      edestruct IHeval_nexpr2 as  (ρ'e2 & Hρ'e2 & Hgi'''& Hρ'e1ρ'e2 & x'e2 & EVAL2 & Hx'e2).

      destruct H1; inv Hx'e1; inv Hx'e2;
      repeat (match type of EQ with
        | context [normalize_expr ?ab ?e ?min ?max ?logmod ?gi] =>
          let NORM := fresh "NORM" in let NORM_MONO := fresh "NORM_MONO" in
          destruct (normalize_expr ab e min max logmod gi) as [[[] ?][]] eqn:NORM;
          assert (NORM_MONO:=NORM);
          apply normalize_expr_next_g_monotone in NORM_MONO;
          (eapply normalize_expr_correct in NORM; [|reflexivity|now eauto|now eauto|now eauto]);
          [destruct NORM as (? & ? & ? & ? & ? & ? & ? & ?)|destruct NORM as (? & ? & ? & ?)];
          simpl in EQ
        | context [normalize_expr_full ?ab ?e ?sgn ?gi] =>
          let NORM := fresh "NORM" in let NORM_MONO := fresh "NORM_MONO" in
          destruct (normalize_expr_full ab e sgn gi) as [[] ?] eqn:NORM;
          assert (NORM_MONO:=NORM);
          apply normalize_expr_full_next_g_monotone in NORM_MONO;
          eapply normalize_expr_full_correct in NORM;
          [destruct NORM as (? & ? & ? & ? & ? & ? & ? & ?)|now eauto|now eauto|now eauto]
        | context [normalize_expr_full64 ?ab ?e ?sgn ?gi] =>
          let NORM := fresh "NORM" in let NORM_MONO := fresh "NORM_MONO" in
          destruct (normalize_expr_full64 ab e sgn gi) as [[] ?] eqn:NORM;
          assert (NORM_MONO:=NORM);
          apply normalize_expr_full64_next_g_monotone in NORM_MONO;
          eapply normalize_expr_full64_correct in NORM;
          [destruct NORM as (? & ? & ? & ? & ? & ? & ? & ?)|now eauto|now eauto|now eauto]
      end).

      + inv EQ. eexists. split. eauto. split. eauto. split. eauto. eexists. split. eauto.
        rewrite Int.add_unsigned.
        erewrite Int.eqm_samerepr. constructor. apply Int.eqm_add; eauto with ints.

      + inv EQ. eexists. split. eauto. split. eauto. split. eauto. eexists. split. eauto.
        unfold Int.sub.
        erewrite Int.eqm_samerepr. constructor. apply Int.eqm_sub; eauto with ints.

      + inv EQ. eexists. split. eauto. split. eauto. split. eauto. eexists. split. eauto.
        unfold Int.mul. erewrite Int.eqm_samerepr. constructor. apply Int.eqm_mult; eauto with ints.

      + assert (x4 ≠ 0).
        { assert (Heqspec:=Int.eq_spec (Int.repr x0) Int.zero).
          rewrite H1 in Heqspec.
          contradict Heqspec. subst x4. apply Int.eqm_samerepr; trivial. }
        match type of EQ with context [if ?c then _ else _] => destruct c end; inv EQ.
        * eexists. split. eauto. split. eauto. split. eauto. eexists. split. eauto 10.
          rewrite !Int.eqm_samerepr with (x:=x) (y:=x2) by trivial.
          rewrite !Int.eqm_samerepr with (x:=x0) (y:=x4) by trivial.
          unfold Int.divs. rewrite !Int.signed_repr; eauto.
        * exists x3. split. eapply join_correct.
          { destruct H2; [left|right];
            (eapply assume_cmp_correct; [eapply assume_cmp_correct| |]);
            eauto 3; simpl.
            * destruct (Z.compare_spec x4 0); congruence.
            * assert (Heqspec:=Int.eq_spec (Int.repr x) (Int.repr Int.min_signed)).
              rewrite H2 in Heqspec.
              erewrite Int.eqm_samerepr in Heqspec by eauto.
              assert (x2 ≠ Int.min_signed) by congruence.
              destruct (Z.compare_spec x2 Int.min_signed); simpl in *; try congruence. omega.
            * destruct (Z.compare_spec x4 0); congruence.
            * assert (Heqspec:=Int.eq_spec (Int.repr x0) Int.mone).
              rewrite H2 in Heqspec.
              erewrite Int.eqm_samerepr in Heqspec by eauto.
              assert (x4 ≠ -1) by (unfold Int.mone in Heqspec; congruence).
              destruct (Z.compare_spec x4 (-1)); simpl in *; congruence. }
          split. eauto. split. eauto. eexists; split; econstructor; eauto.
          erewrite Int.eqm_samerepr with (x:=x), Int.eqm_samerepr with (x:=x0); eauto.
          rewrite !Int.signed_repr by auto. auto.

      + assert (x4 ≠ 0).
        { assert (Heqspec:=Int.eq_spec (Int.repr x0) Int.zero).
          rewrite H1 in Heqspec.
          contradict Heqspec. subst x4. apply Int.eqm_samerepr; trivial. }
        match type of EQ with context [if ?c then _ else _] => destruct c end; inv EQ.
        * eexists. split. eauto. split. eauto. split. eauto. eexists. split. eauto 10.
          rewrite !Int.eqm_samerepr with (x:=x) (y:=x2) by trivial.
          rewrite !Int.eqm_samerepr with (x:=x0) (y:=x4) by trivial.
          unfold Int.divu. rewrite !Int.unsigned_repr; eauto.
          simpl in *. rewrite Int.Zquot_Zdiv by omega. destruct zlt. omega. constructor.
        * exists x3. split. eapply assume_cmp_correct; eauto 3.
          simpl. destruct (Z.compare_spec x4 0); try congruence; omega.
          split. eauto. split. eauto. eexists; split; econstructor; eauto.
          erewrite Int.eqm_samerepr with (x:=x), Int.eqm_samerepr with (x:=x0); eauto.
          rewrite !Int.unsigned_repr by auto.
          replace (x2/x4) with (x2÷x4). auto.
          simpl in *. rewrite Int.Zquot_Zdiv by omega. destruct zlt; omega.

      + assert (x4 ≠ 0).
        { assert (Heqspec:=Int.eq_spec (Int.repr x0) Int.zero).
          rewrite H1 in Heqspec.
          contradict Heqspec. subst x4. apply Int.eqm_samerepr; trivial. }
        match type of EQ with context [if ?c then _ else _] => destruct c end; inv EQ.
        * eexists. split. eauto. split. eauto. split. eauto. eexists. split. eauto 10.
          rewrite !Int.eqm_samerepr with (x:=x) (y:=x2) by trivial.
          rewrite !Int.eqm_samerepr with (x:=x0) (y:=x4) by trivial.
          unfold Int.mods. rewrite !Int.signed_repr; eauto.
        * exists x3. split. eapply join_correct.
          { simpl in *. destruct H2; [left|right];
            (eapply assume_cmp_correct; [eapply assume_cmp_correct| |]);
            eauto 3; simpl.
            * destruct (Z.compare_spec x4 0); congruence.
            * assert (Heqspec:=Int.eq_spec (Int.repr x) (Int.repr Int.min_signed)).
              rewrite H2 in Heqspec.
              erewrite Int.eqm_samerepr in Heqspec by eauto.
              assert (x2 ≠ Int.min_signed) by congruence.
              destruct (Z.compare_spec x2 Int.min_signed); simpl in *; try congruence. omega.
            * destruct (Z.compare_spec x4 0); congruence.
            * assert (Heqspec:=Int.eq_spec (Int.repr x0) Int.mone).
              rewrite H2 in Heqspec.
              erewrite Int.eqm_samerepr in Heqspec by eauto.
              assert (x4 ≠ -1) by (unfold Int.mone in Heqspec; congruence).
              destruct (Z.compare_spec x4 (-1)); simpl in *; congruence. }
          split. eauto. split. eauto. eexists; split; econstructor; eauto.
          erewrite Int.eqm_samerepr with (x:=x), Int.eqm_samerepr with (x:=x0); eauto.
          rewrite !Int.signed_repr by auto. auto.

      + assert (x4 ≠ 0).
        { assert (Heqspec:=Int.eq_spec (Int.repr x0) Int.zero).
          rewrite H1 in Heqspec.
          contradict Heqspec. subst x4. apply Int.eqm_samerepr; trivial. }
        match type of EQ with context [if ?c then _ else _] => destruct c end; inv EQ.
        * eexists. split. eauto. split. eauto. split. eauto. eexists. split. eauto 10.
          rewrite !Int.eqm_samerepr with (x:=x) (y:=x2) by trivial.
          rewrite !Int.eqm_samerepr with (x:=x0) (y:=x4) by trivial.
          unfold Int.modu. rewrite !Int.unsigned_repr; eauto.
          simpl in *. rewrite Z.rem_mod_nonneg by omega. constructor.
        * exists x3. split. eapply assume_cmp_correct; eauto 3.
          simpl. destruct (Z.compare_spec x4 0); try congruence; omega.
          split. eauto. split. eauto. eexists; split; econstructor; eauto.
          erewrite Int.eqm_samerepr with (x:=x), Int.eqm_samerepr with (x:=x0); eauto.
          rewrite !Int.unsigned_repr by auto.
          simpl in *. rewrite <- Z.rem_mod_nonneg by omega. auto.

      + inv EQ. eexists. split. eauto. split. eauto. split. eauto. eexists. split. eauto.
        replace (Int.and (Int.repr x) (Int.repr x0)) with (Int.repr (Z.land x x0)).
        constructor.
        apply Int.same_bits_eq. intros.
        rewrite Int.bits_and, !Int.testbit_repr, Z.land_spec by trivial.
        reflexivity.

      + inv EQ. eexists. split. eauto. split. eauto. split. eauto. eexists. split. eauto.
        replace (Int.or (Int.repr x) (Int.repr x0)) with (Int.repr (Z.lor x x0)).
        constructor.
        apply Int.same_bits_eq. intros.
        rewrite Int.bits_or, !Int.testbit_repr, Z.lor_spec by trivial.
        reflexivity.

      + inv EQ. eexists. split. eauto. split. eauto. split. eauto. eexists. split. eauto.
        replace (Int.xor (Int.repr x) (Int.repr x0)) with (Int.repr (Z.lxor x x0)).
        constructor.
        apply Int.same_bits_eq. intros.
        rewrite Int.bits_xor, !Int.testbit_repr, Z.lxor_spec by trivial.
        reflexivity.

      + assert (∃ x',
          (∀ ρ', compat_gh g x1 ρ' ->
                 eval_iexpr ρ' (IEbinop IOshl e1' i) x') /\
          related (NVint (Int.shl (Int.repr x) (Int.repr x0))) x').
        { simpl in *. eexists; split. eauto 10.
          rewrite !Int.eqm_samerepr with (x:=x0) (y:=x2) by trivial.
          replace (Int.shl (Int.repr x) (Int.repr x2)) with (Int.repr (Z.shiftl x x2)).
          constructor.
          apply Int.same_bits_eq. intros.
          rewrite Int.bits_shl by trivial.
          rewrite Int.unsigned_repr by trivial.
          destruct Coqlib.zlt; rewrite !Int.testbit_repr by omega.
          rewrite Z.shiftl_spec by omega.
          apply Z.testbit_neg_r. omega.
          apply Z.shiftl_spec. omega. }
        match type of EQ with context [if ?c then _ else _] => destruct c end; inv EQ.
        * eauto 10.
        * exists x1. split. eapply assume_cmp_correct; eauto 3. 2:eauto.
          simpl. unfold Int.ltu in H1. erewrite Int.eqm_samerepr in H1; eauto 2. rewrite Int.unsigned_repr in H1; eauto 2.
          change (Int.unsigned Int.iwordsize) with Int.zwordsize in H1.
          destruct Coqlib.zlt in H1; [|discriminate].
          destruct (Z.compare_spec x2 Int.zwordsize); try omega; auto.

      + assert (∃ x',
          (∀ ρ', compat_gh g0 x3 ρ' ->
                 eval_iexpr ρ' (IEbinop IOshr i i0) x') /\
          related (NVint (Int.shr (Int.repr x) (Int.repr x0))) x').
        { simpl in *. eexists; split. eauto.
          rewrite !Int.eqm_samerepr with (x:=x0) (y:=x4) by trivial.
          rewrite !Int.eqm_samerepr with (x:=x) (y:=x2) by trivial.
          replace (Int.shr (Int.repr x2) (Int.repr x4)) with (Int.repr (Z.shiftr x2 x4)).
          constructor.
          apply Int.same_bits_eq. intros.
          rewrite Int.bits_shr by trivial.
          rewrite Int.unsigned_repr by trivial.
          destruct Coqlib.zlt; rewrite !Int.testbit_repr by omega.
          apply Z.shiftr_spec. omega.
          rewrite Z.shiftr_spec by omega.
          rewrite <- Int.signed_repr with x2 by auto.
          rewrite !Int.bits_signed by omega.
          destruct Coqlib.zlt. omega.
          destruct Coqlib.zlt; reflexivity. }
        match type of EQ with context [if ?c then _ else _] => destruct c end; inv EQ.
        * eauto 10.
        * exists x3. split. eapply assume_cmp_correct; eauto 3. 2:now eauto 10. simpl.
          unfold Int.ltu in H1. erewrite Int.eqm_samerepr in H1; eauto 2.
          rewrite Int.unsigned_repr in H1; eauto 2.
          change (Int.unsigned Int.iwordsize) with Int.zwordsize in H1.
          destruct Coqlib.zlt in H1; [|discriminate].
          destruct (Z.compare_spec x4 Int.zwordsize); try omega; auto.

      + assert (∃ x',
          (∀ ρ', compat_gh g0 x3 ρ' ->
                 eval_iexpr ρ' (IEbinop IOshr i i0) x') /\
          related (NVint (Int.shru (Int.repr x) (Int.repr x0))) x').
        { eexists; split. eauto.
          rewrite !Int.eqm_samerepr with (x:=x0) (y:=x4) by trivial.
          rewrite !Int.eqm_samerepr with (x:=x) (y:=x2) by trivial.
          replace (Int.shru (Int.repr x2) (Int.repr x4)) with (Int.repr (Z.shiftr x2 x4)).
          constructor.
          apply Int.same_bits_eq. intros.
          rewrite Int.bits_shru by trivial.
          rewrite Int.unsigned_repr by trivial.
          simpl in *. destruct Coqlib.zlt; rewrite !Int.testbit_repr by omega.
          apply Z.shiftr_spec. omega.
          rewrite Z.shiftr_spec by omega.
          apply Z.testbit_false. omega.
          rewrite Z.div_small. reflexivity. split.
          simpl in *; omega.
          eapply Z.lt_le_trans with (2^Int.zwordsize).
          change (0 <= x2 <= 2 ^ Int.zwordsize - 1) in H5. omega.
          apply Z.pow_le_mono_r; omega. }
        match type of EQ with context [if ?c then _ else _] => destruct c end; inv EQ.
        * eauto 10.
        * exists x3. split. eauto. eapply assume_cmp_correct; eauto 3. simpl. 2:now eauto 10.
          unfold Int.ltu in H1. erewrite Int.eqm_samerepr in H1; eauto 3. rewrite Int.unsigned_repr in H1; eauto 3.
          change (Int.unsigned Int.iwordsize) with Int.zwordsize in H1.
          destruct Coqlib.zlt in H1; [|discriminate].
          destruct (Z.compare_spec x4 Int.zwordsize); try omega; auto.

      + inv EQ. eauto 15.

      + inv EQ. eauto 15.

      + inv EQ. eauto 15.

      + inv EQ. eauto 15.

      + inv EQ. eexists. split. eauto. split. eauto. split. eauto. eexists. split. eauto 10.
        rewrite plus_double_round. auto.

      + inv EQ. eexists. split. eauto. split. eauto. split. eauto. eexists. split. eauto 10.
        rewrite sub_double_round. auto.

      + inv EQ. eexists. split. eauto. split. eauto. split. eauto. eexists. split. eauto 10.
        rewrite mult_double_round. auto.

      + inv EQ. eexists. split. eauto. split. eauto. split. eauto. eexists. split. eauto 10.
        rewrite div_double_round. auto.

      + inv EQ. eexists. split. eauto. split. eauto. split. eauto. eexists. split. eauto 10.
        rewrite Int64.add_unsigned.
        erewrite Int64.eqm_samerepr. constructor. apply Int64.eqm_add; eauto with ints.

      + inv EQ. eexists. split. eauto. split. eauto. split. eauto. eexists. split. eauto 10.
        unfold Int64.sub.
        erewrite Int64.eqm_samerepr. constructor. apply Int64.eqm_add; eauto with ints.

      + inv EQ. eexists. split. eauto. split. eauto. split. eauto. eexists. split. eauto. unfold Int64.mul.
        erewrite Int64.eqm_samerepr. econstructor. apply Int64.eqm_mult; eauto with ints.

      + assert (x4 ≠ 0).
        { assert (Heqspec:=Int64.eq_spec (Int64.repr x0) Int64.zero).
          rewrite H1 in Heqspec.
          contradict Heqspec. subst x4. apply Int64.eqm_samerepr; trivial. }
        match type of EQ with context [if ?c then _ else _] => destruct c end; inv EQ.
        * eexists. split. eauto. split. eauto. split. eauto. eexists. split. eauto 10.
          rewrite !Int64.eqm_samerepr with (x:=x) (y:=x2) by trivial.
          rewrite !Int64.eqm_samerepr with (x:=x0) (y:=x4) by trivial.
          unfold Int64.divs. rewrite !Int64.signed_repr; eauto.
        * eexists. split. apply join_correct.
          { destruct H2; [left|right];
            (eapply assume_cmp_correct; [eapply assume_cmp_correct| |]);
            eauto 3; simpl.
            * destruct (Z.compare_spec x4 0); congruence.
            * assert (Heqspec:=Int64.eq_spec (Int64.repr x) (Int64.repr Int64.min_signed)).
              rewrite H2 in Heqspec.
              erewrite Int64.eqm_samerepr in Heqspec by eauto.
              assert (x2 ≠ Int64.min_signed) by congruence.
              destruct (Z.compare_spec x2 Int64.min_signed); simpl in *; try congruence. omega.
            * destruct (Z.compare_spec x4 0); congruence.
            * assert (Heqspec:=Int64.eq_spec (Int64.repr x0) Int64.mone).
              rewrite H2 in Heqspec.
              erewrite Int64.eqm_samerepr in Heqspec by eauto.
              assert (x4 ≠ -1) by (unfold Int64.mone in Heqspec; congruence).
              destruct (Z.compare_spec x4 (-1)); simpl in *; congruence. }
          split. eauto. split. eauto. eexists; split; econstructor; eauto.
          erewrite Int64.eqm_samerepr with (x:=x), Int64.eqm_samerepr with (x:=x0); eauto.
          rewrite !Int64.signed_repr by auto. auto.

      + assert (x4 ≠ 0).
        { assert (Heqspec:=Int64.eq_spec (Int64.repr x0) Int64.zero).
          rewrite H1 in Heqspec.
          contradict Heqspec. subst x4. apply Int64.eqm_samerepr; trivial. }
        match type of EQ with context [if ?c then _ else _] => destruct c end; inv EQ.
        * eexists. split. eauto. split. eauto. split. eauto. eexists. split. eauto 10.
          rewrite !Int64.eqm_samerepr with (x:=x) (y:=x2) by trivial.
          rewrite !Int64.eqm_samerepr with (x:=x0) (y:=x4) by trivial.
          unfold Int64.divu. rewrite !Int64.unsigned_repr; eauto.
          simpl in *. rewrite Int64.Zquot_Zdiv by omega. destruct zlt. omega. constructor.
        * eexists. split. eapply assume_cmp_correct; eauto 3.
          simpl. destruct (Z.compare_spec x4 0); try congruence; omega.
          split. eauto. split. eauto. eexists; split; econstructor; eauto.
          erewrite Int64.eqm_samerepr with (x:=x), Int64.eqm_samerepr with (x:=x0); eauto.
          rewrite !Int64.unsigned_repr by auto.
          replace (x2/x4) with (x2÷x4). auto.
          simpl in *. rewrite Int64.Zquot_Zdiv by omega. destruct zlt; omega.

      + assert (x4 ≠ 0).
        { assert (Heqspec:=Int64.eq_spec (Int64.repr x0) Int64.zero).
          rewrite H1 in Heqspec.
          contradict Heqspec. subst x4. apply Int64.eqm_samerepr; trivial. }
        match type of EQ with context [if ?c then _ else _] => destruct c end; inv EQ.
        * eexists. split. eauto. split. eauto. split. eauto. eexists. split. eauto 10.
          rewrite !Int64.eqm_samerepr with (x:=x) (y:=x2) by trivial.
          rewrite !Int64.eqm_samerepr with (x:=x0) (y:=x4) by trivial.
          unfold Int64.mods. rewrite !Int64.signed_repr; eauto.
        * eexists. split. apply join_correct.
          { destruct H2; [left|right];
            (eapply assume_cmp_correct; [eapply assume_cmp_correct| |]);
            eauto 3; simpl.
            * destruct (Z.compare_spec x4 0); congruence.
            * assert (Heqspec:=Int64.eq_spec (Int64.repr x) (Int64.repr Int64.min_signed)).
              rewrite H2 in Heqspec.
              erewrite Int64.eqm_samerepr in Heqspec by eauto.
              assert (x2 ≠ Int64.min_signed) by congruence.
              destruct (Z.compare_spec x2 Int64.min_signed); simpl in *; try congruence. omega.
            * destruct (Z.compare_spec x4 0); congruence.
            * assert (Heqspec:=Int64.eq_spec (Int64.repr x0) Int64.mone).
              rewrite H2 in Heqspec.
              erewrite Int64.eqm_samerepr in Heqspec by eauto.
              assert (x4 ≠ -1) by (unfold Int64.mone in Heqspec; congruence).
              destruct (Z.compare_spec x4 (-1)); simpl in *; congruence. }
          split. eauto. split. eauto. eexists; split; econstructor; eauto.
          erewrite Int64.eqm_samerepr with (x:=x), Int64.eqm_samerepr with (x:=x0); eauto.
          rewrite !Int64.signed_repr by auto. auto.

      + assert (x4 ≠ 0).
        { assert (Heqspec:=Int64.eq_spec (Int64.repr x0) Int64.zero).
          rewrite H1 in Heqspec.
          contradict Heqspec. subst x4. apply Int64.eqm_samerepr; trivial. }
        match type of EQ with context [if ?c then _ else _] => destruct c end; inv EQ.
        * eexists. split. eauto. split. eauto. split. eauto. eexists. split. eauto 10.
          rewrite !Int64.eqm_samerepr with (x:=x) (y:=x2) by trivial.
          rewrite !Int64.eqm_samerepr with (x:=x0) (y:=x4) by trivial.
          unfold Int64.modu. rewrite !Int64.unsigned_repr; eauto.
          simpl in *. rewrite Z.rem_mod_nonneg by omega. constructor.
        * eexists. split. eauto. eapply assume_cmp_correct; eauto 3.
          simpl. destruct (Z.compare_spec x4 0); try congruence; omega.
          split. eauto. split. eauto. eexists; split; econstructor; eauto.
          erewrite Int64.eqm_samerepr with (x:=x), Int64.eqm_samerepr with (x:=x0); eauto.
          rewrite !Int64.unsigned_repr by auto.
          simpl in *. rewrite <- Z.rem_mod_nonneg by omega. auto.

      + inv EQ. eexists. split. eauto. split. eauto. split. eauto. eexists. split. eauto.
        replace (Int64.and (Int64.repr x) (Int64.repr x0)) with (Int64.repr (Z.land x x0)).
        constructor.
        apply Int64.same_bits_eq. intros.
        rewrite Int64.bits_and, !Int64.testbit_repr, Z.land_spec by trivial.
        reflexivity.

      + inv EQ. eexists. split. eauto. split. eauto. split. eauto. eexists. split. eauto.
        replace (Int64.or (Int64.repr x) (Int64.repr x0)) with (Int64.repr (Z.lor x x0)).
        constructor.
        apply Int64.same_bits_eq. intros.
        rewrite Int64.bits_or, !Int64.testbit_repr, Z.lor_spec by trivial.
        reflexivity.

      + inv EQ. eexists. split. eauto. split. eauto. split. eauto. eexists. split. eauto.
        replace (Int64.xor (Int64.repr x) (Int64.repr x0)) with (Int64.repr (Z.lxor x x0)).
        constructor.
        apply Int64.same_bits_eq. intros.
        rewrite Int64.bits_xor, !Int64.testbit_repr, Z.lxor_spec by trivial.
        reflexivity.

      + assert (∃ x',
          (∀ ρ', compat_gh g x1 ρ' -> eval_iexpr ρ' (IEbinop IOshl e1' i) x') /\
          related (NVint64 (Int64.shl' (Int64.repr x) (Int.repr x0))) x').
        { eexists. split. eauto 10.
          rewrite !Int.eqm_samerepr with (x:=x0) (y:=x2) by trivial.
          replace (Int64.shl' (Int64.repr x) (Int.repr x2)) with (Int64.repr (Z.shiftl x x2)).
          constructor.
          apply Int64.same_bits_eq. intros.
          rewrite Int64.bits_shl' by trivial.
          rewrite Int.unsigned_repr by trivial.
          simpl in *. destruct Coqlib.zlt; rewrite !Int64.testbit_repr by omega.
          rewrite Z.shiftl_spec by omega.
          apply Z.testbit_neg_r. omega.
          apply Z.shiftl_spec. omega. }
        match type of EQ with context [if ?c then _ else _] => destruct c end; inv EQ.
        * eauto 10.
        * eexists. split. eapply assume_cmp_correct; eauto 3. simpl. 2:eauto.
          unfold Int.ltu in H1. erewrite Int.eqm_samerepr in H1; eauto 3. rewrite Int.unsigned_repr in H1; eauto 3.
          change (Int.unsigned Int64.iwordsize') with Int64.zwordsize in H1.
          destruct Coqlib.zlt in H1; [|discriminate].
          destruct (Z.compare_spec x2 Int64.zwordsize); try omega; auto.

      + assert (∃ x',
          (∀ ρ', compat_gh g0 x3 ρ' -> eval_iexpr ρ' (IEbinop IOshr i i0) x') /\
          related (NVint64 (Int64.shr' (Int64.repr x) (Int.repr x0))) x').
        { eexists; split. eauto 10.
          rewrite !Int.eqm_samerepr with (x:=x0) (y:=x4) by trivial.
          rewrite !Int64.eqm_samerepr with (x:=x) (y:=x2) by trivial.
          replace (Int64.shr' (Int64.repr x2) (Int.repr x4)) with (Int64.repr (Z.shiftr x2 x4)).
          constructor.
          apply Int64.same_bits_eq. intros.
          rewrite Int64.bits_shr' by trivial.
          rewrite Int.unsigned_repr by trivial.
          simpl in *. destruct Coqlib.zlt; rewrite !Int64.testbit_repr by omega.
          apply Z.shiftr_spec. omega.
          rewrite Z.shiftr_spec by omega.
          rewrite <- Int64.signed_repr with x2 by auto.
          rewrite !Int64.bits_signed by omega.
          destruct Coqlib.zlt. omega.
          rewrite Int64.repr_signed. reflexivity. }
        match type of EQ with context [if ?c then _ else _] => destruct c end; inv EQ.
        * eauto 10.
        * eexists. split. eapply assume_cmp_correct; eauto 3. simpl. 2:now eauto 10.
          unfold Int.ltu in H1. erewrite Int.eqm_samerepr in H1; eauto 3. rewrite Int.unsigned_repr in H1; eauto 3.
          change (Int.unsigned Int64.iwordsize') with Int64.zwordsize in H1.
          destruct Coqlib.zlt in H1; [|discriminate].
          destruct (Z.compare_spec x4 Int64.zwordsize); try omega; auto.

      + assert (∃ x',
          (∀ ρ', compat_gh g0 x3 ρ' -> eval_iexpr ρ' (IEbinop IOshr i i0) x') /\
          related (NVint64 (Int64.shru' (Int64.repr x) (Int.repr x0))) x').
        { eexists; split. eauto 10.
          rewrite !Int.eqm_samerepr with (x:=x0) (y:=x4) by trivial.
          rewrite !Int64.eqm_samerepr with (x:=x) (y:=x2) by trivial.
          replace (Int64.shru' (Int64.repr x2) (Int.repr x4)) with (Int64.repr (Z.shiftr x2 x4)).
          constructor.
          apply Int64.same_bits_eq. intros.
          rewrite Int64.bits_shru' by trivial.
          rewrite Int.unsigned_repr by trivial.
          simpl in *. destruct Coqlib.zlt; rewrite !Int64.testbit_repr by omega.
          apply Z.shiftr_spec. omega.
          rewrite Z.shiftr_spec by omega.
          apply Z.testbit_false. omega.
          rewrite Z.div_small. reflexivity. split. omega.
          eapply Z.lt_le_trans with (2^Int64.zwordsize).
          change (0 <= x2 <= 2 ^ Int64.zwordsize - 1) in H5. omega.
          apply Z.pow_le_mono_r; omega. }
        match type of EQ with context [if ?c then _ else _] => destruct c end; inv EQ.
        * eauto 10.
        * eexists. split. eapply assume_cmp_correct; eauto 3. 2:now eauto 10. simpl.
          unfold Int.ltu in H1. erewrite Int.eqm_samerepr in H1; eauto 3. rewrite Int.unsigned_repr in H1; eauto 3.
          change (Int.unsigned Int64.iwordsize') with Int64.zwordsize in H1.
          destruct Coqlib.zlt in H1; [|discriminate].
          destruct (Z.compare_spec x4 Int64.zwordsize); try omega; auto.

      + inv EQ. eexists. split. eauto. split. eauto. split. eauto.
        exists (INz (if Int.cmp c (Int.repr x) (Int.repr x0) then 1 else 0)).
        split. econstructor; eauto. 2:destruct Int.cmp; constructor.
        rewrite !Int.eqm_samerepr with (x:=x) (y:=x2) by trivial.
        rewrite !Int.eqm_samerepr with (x:=x0) (y:=x4) by trivial.
        replace (Int.cmp c (Int.repr x2) (Int.repr x4)) with (ZIntervals.Zcmp c x2 x4).
        constructor.
        unfold Int.cmp, Int.lt.
        rewrite !Int.signed_repr by eauto.
        pose proof (Int.eq_spec (Int.repr x2) (Int.repr x4)).
        destruct c; simpl.
        * destruct Int.eq; destruct (Z.compare_spec x2 x4); try constructor; exfalso.
          assert (Int.signed (Int.repr x2) = Int.signed (Int.repr x4)) by congruence.
          rewrite !Int.signed_repr in H16 by trivial. omega.
          assert (Int.signed (Int.repr x2) = Int.signed (Int.repr x4)) by congruence.
          rewrite !Int.signed_repr in H16 by trivial. omega.
          congruence.
        * destruct Int.eq; destruct (Z.compare_spec x2 x4); try constructor; exfalso.
          assert (Int.signed (Int.repr x2) = Int.signed (Int.repr x4)) by congruence.
          rewrite !Int.signed_repr in H16 by trivial. omega.
          assert (Int.signed (Int.repr x2) = Int.signed (Int.repr x4)) by congruence.
          rewrite !Int.signed_repr in H16 by trivial. omega.
          congruence.
        * destruct Coqlib.zlt; destruct (Z.compare_spec x2 x4); try constructor; exfalso; omega.
        * destruct Coqlib.zlt; destruct (Z.compare_spec x2 x4); try constructor; exfalso; omega.
        * destruct Coqlib.zlt; destruct (Z.compare_spec x2 x4); try constructor; exfalso; omega.
        * destruct Coqlib.zlt; destruct (Z.compare_spec x2 x4); try constructor; exfalso; omega.

      + inv EQ. eexists. split. eauto. split. eauto. split. eauto.
        exists (INz (if Int.cmpu c (Int.repr x) (Int.repr x0) then 1 else 0)).
        split. econstructor; eauto. 2:destruct Int.cmpu; constructor.
        rewrite !Int.eqm_samerepr with (x:=x) (y:=x2) by trivial.
        rewrite !Int.eqm_samerepr with (x:=x0) (y:=x4) by trivial.
        replace (Int.cmpu c (Int.repr x2) (Int.repr x4)) with (ZIntervals.Zcmp c x2 x4).
        constructor.
        unfold Int.cmpu, Int.ltu.
        rewrite !Int.unsigned_repr by auto.
        pose proof (Int.eq_spec (Int.repr x2) (Int.repr x4)).
        destruct c; simpl.
        * destruct Int.eq; destruct (Z.compare_spec x2 x4); try constructor; exfalso.
          assert (Int.unsigned (Int.repr x2) = Int.unsigned (Int.repr x4)) by congruence.
          rewrite !Int.unsigned_repr in H16 by trivial. omega.
          assert (Int.unsigned (Int.repr x2) = Int.unsigned (Int.repr x4)) by congruence.
          rewrite !Int.unsigned_repr in H16 by trivial. omega.
          congruence.
        * destruct Int.eq; destruct (Z.compare_spec x2 x4); try constructor; exfalso.
          assert (Int.unsigned (Int.repr x2) = Int.unsigned (Int.repr x4)) by congruence.
          rewrite !Int.unsigned_repr in H16 by trivial. omega.
          assert (Int.unsigned (Int.repr x2) = Int.unsigned (Int.repr x4)) by congruence.
          rewrite !Int.unsigned_repr in H16 by trivial. omega.
          congruence.
        * destruct Coqlib.zlt; destruct (Z.compare_spec x2 x4); try constructor; exfalso; omega.
        * destruct Coqlib.zlt; destruct (Z.compare_spec x2 x4); try constructor; exfalso; omega.
        * destruct Coqlib.zlt; destruct (Z.compare_spec x2 x4); try constructor; exfalso; omega.
        * destruct Coqlib.zlt; destruct (Z.compare_spec x2 x4); try constructor; exfalso; omega.

      + inv EQ. eexists. split. eauto. split. eauto. split. eauto. 
        exists (INz (if Int64.cmp c (Int64.repr x) (Int64.repr x0) then 1 else 0)).
        split. econstructor; eauto. 2:destruct Int64.cmp; constructor.
        rewrite !Int64.eqm_samerepr with (x:=x) (y:=x2) by trivial.
        rewrite !Int64.eqm_samerepr with (x:=x0) (y:=x4) by trivial.
        replace (Int64.cmp c (Int64.repr x2) (Int64.repr x4)) with (ZIntervals.Zcmp c x2 x4).
        constructor.
        unfold Int64.cmp, Int64.lt.
        rewrite !Int64.signed_repr by eauto.
        pose proof (Int64.eq_spec (Int64.repr x2) (Int64.repr x4)).
        destruct c; simpl.
        * destruct Int64.eq; destruct (Z.compare_spec x2 x4); try constructor; exfalso.
          assert (Int64.signed (Int64.repr x2) = Int64.signed (Int64.repr x4)) by congruence.
          rewrite !Int64.signed_repr in H16 by trivial. omega.
          assert (Int64.signed (Int64.repr x2) = Int64.signed (Int64.repr x4)) by congruence.
          rewrite !Int64.signed_repr in H16 by trivial. omega.
          congruence.
        * destruct Int64.eq; destruct (Z.compare_spec x2 x4); try constructor; exfalso.
          assert (Int64.signed (Int64.repr x2) = Int64.signed (Int64.repr x4)) by congruence.
          rewrite !Int64.signed_repr in H16 by trivial. omega.
          assert (Int64.signed (Int64.repr x2) = Int64.signed (Int64.repr x4)) by congruence.
          rewrite !Int64.signed_repr in H16 by trivial. omega.
          congruence.
        * destruct Coqlib.zlt; destruct (Z.compare_spec x2 x4); try constructor; exfalso; omega.
        * destruct Coqlib.zlt; destruct (Z.compare_spec x2 x4); try constructor; exfalso; omega.
        * destruct Coqlib.zlt; destruct (Z.compare_spec x2 x4); try constructor; exfalso; omega.
        * destruct Coqlib.zlt; destruct (Z.compare_spec x2 x4); try constructor; exfalso; omega.

      + inv EQ. eexists. split. eauto. split. eauto. split. eauto. 
        exists (INz (if Int64.cmpu c (Int64.repr x) (Int64.repr x0) then 1 else 0)).
        split. econstructor; eauto. 2:destruct Int64.cmpu; constructor.
        rewrite !Int64.eqm_samerepr with (x:=x) (y:=x2) by trivial.
        rewrite !Int64.eqm_samerepr with (x:=x0) (y:=x4) by trivial.
        replace (Int64.cmpu c (Int64.repr x2) (Int64.repr x4)) with (ZIntervals.Zcmp c x2 x4).
        constructor.
        unfold Int64.cmpu, Int64.ltu.
        rewrite !Int64.unsigned_repr by auto.
        pose proof (Int64.eq_spec (Int64.repr x2) (Int64.repr x4)).
        destruct c; simpl.
        * destruct Int64.eq; destruct (Z.compare_spec x2 x4); try constructor; exfalso.
          assert (Int64.unsigned (Int64.repr x2) = Int64.unsigned (Int64.repr x4)) by congruence.
          rewrite !Int64.unsigned_repr in H16 by trivial. omega.
          assert (Int64.unsigned (Int64.repr x2) = Int64.unsigned (Int64.repr x4)) by congruence.
          rewrite !Int64.unsigned_repr in H16 by trivial. omega.
          congruence.
        * destruct Int64.eq; destruct (Z.compare_spec x2 x4); try constructor; exfalso.
          assert (Int64.unsigned (Int64.repr x2) = Int64.unsigned (Int64.repr x4)) by congruence.
          rewrite !Int64.unsigned_repr in H16 by trivial. omega.
          assert (Int64.unsigned (Int64.repr x2) = Int64.unsigned (Int64.repr x4)) by congruence.
          rewrite !Int64.unsigned_repr in H16 by trivial. omega.
          congruence.
        * destruct Coqlib.zlt; destruct (Z.compare_spec x2 x4); try constructor; exfalso; omega.
        * destruct Coqlib.zlt; destruct (Z.compare_spec x2 x4); try constructor; exfalso; omega.
        * destruct Coqlib.zlt; destruct (Z.compare_spec x2 x4); try constructor; exfalso; omega.
        * destruct Coqlib.zlt; destruct (Z.compare_spec x2 x4); try constructor; exfalso; omega.

      + inv EQ. eexists. split. eauto. split. eauto. split. eauto. eexists. split. eauto.
        destruct Float.cmp; constructor.

      + inv EQ. eexists. split. eauto. split. eauto. split. eauto. eexists. split. eauto.
        rewrite Float32.cmp_double. unfold Float32.to_double.
        destruct Float.cmp; constructor.
  Qed.

  Lemma convert_expr_noerror:
    ∀ ρ e v ρ'1 gi, ρ'1 ∈ γ v -> ρ'1 ∈ γ gi ->
      (∀ id, related (ρ id) (ρ'1 (Real id))) ->
    ∀ e' v' gi', convert_expr v e gi = (e', v', true, gi') ->
      error_nexpr ρ e ->
      False.
  Proof.
    induction e; intros ab ρ'1 gi Hρ'1 Hgi Hρρ'1 e' ab' gi' EQCVT ERR;
    inv ERR; simpl convert_expr in EQCVT.
    - destruct (convert_expr ab e gi) as [[[e0 ab0] nerr0] gi''] eqn:EQCVT'.
      eapply convert_expr_ok in EQCVT'; eauto.
      destruct EQCVT' as (ρ'2 & Hρ'2 & Hgi'' & Hρ'1ρ'2 & x' & EVAL & REL).
      destruct H2; inv REL;
      match type of EQCVT with
      | context [simplify_expr_alias ?ab ?e ?gi] =>
        let SIMPL := fresh "SIMPL" in let SIMPL_MONO := fresh "SIMPL_MONO" in
        destruct (simplify_expr_alias ab e gi) as [[] ?] eqn:SIMPL;
        assert (SIMPL_MONO:=SIMPL);
        apply simplify_expr_alias_monotone in SIMPL_MONO;
        (eapply simplify_expr_alias_correct in SIMPL;[|now eauto|now eauto|now eauto]);
        destruct SIMPL as (? & ? & ? & ? & ?)
      end;
      inv EQCVT;
      rewrite !Bool.andb_true_iff in H8; destruct H8 as [[? ?] ?]; subst.
      + eapply intoffloat_intv; [| |now eauto].
        * destruct (Fleb smallest_intoffloat f) eqn:EQ. reflexivity.
          erewrite @is_bot_spec in H6. assumption.
          eapply assume_cmpf_correct; eauto 3.
          rewrite <- Float.cmp_swap. apply EQ.
        * destruct (Fleb f largest_intoffloat) eqn:EQ. reflexivity.
          erewrite @is_bot_spec in H7. assumption.
          eapply assume_cmpf_correct; eauto.
      + eapply intuoffloat_intv; [| |now eauto].
        * destruct (Fleb smallest_intuoffloat f) eqn:EQ. reflexivity.
          erewrite @is_bot_spec in H6. assumption.
          eapply assume_cmpf_correct; eauto 3.
          rewrite <- Float.cmp_swap. apply EQ.
        * destruct (Fleb f largest_intuoffloat) eqn:EQ. reflexivity.
          erewrite @is_bot_spec in H7. assumption.
          eapply assume_cmpf_correct; eauto.
      + eapply longoffloat_intv; [| |now eauto].
        * destruct (Fleb smallest_longoffloat f) eqn:EQ. reflexivity.
          erewrite @is_bot_spec in H6. assumption.
          eapply assume_cmpf_correct; eauto 3.
          rewrite <- Float.cmp_swap. apply EQ.
        * destruct (Fleb f largest_longoffloat) eqn:EQ. reflexivity.
          erewrite @is_bot_spec in H7. assumption.
          eapply assume_cmpf_correct; eauto.
      + eapply longuoffloat_intv; [| |now eauto].
        * destruct (Fleb smallest_longuoffloat f) eqn:EQ. reflexivity.
          erewrite @is_bot_spec in H6. assumption.
          eapply assume_cmpf_correct; eauto 3.
          rewrite <- Float.cmp_swap. apply EQ.
        * destruct (Fleb f largest_longuoffloat) eqn:EQ. reflexivity.
          erewrite @is_bot_spec in H7. assumption.
          eapply assume_cmpf_correct; eauto.
      + eapply intofsingle_intv; [| |now eauto].
        * destruct (Fleb smallest_intofsingle f) eqn:EQ. reflexivity.
          erewrite @is_bot_spec in H6. assumption.
          eapply assume_cmpf_correct; eauto 3.
          rewrite <- Float.cmp_swap. rewrite <- Float32.cmp_double. apply EQ.
        * destruct (Fleb f largest_intofsingle) eqn:EQ. reflexivity.
          erewrite @is_bot_spec in H7. assumption.
          eapply assume_cmpf_correct; eauto 3. rewrite <- Float32.cmp_double. apply EQ.
      + eapply intuofsingle_intv; [| |now eauto].
        * destruct (Fleb smallest_intuofsingle f) eqn:EQ. reflexivity.
          erewrite @is_bot_spec in H6. assumption.
          eapply assume_cmpf_correct; eauto 3.
          rewrite <- Float.cmp_swap. rewrite <- Float32.cmp_double. apply EQ.
        * destruct (Fleb f largest_intuofsingle) eqn:EQ. reflexivity.
          erewrite @is_bot_spec in H7. assumption.
          eapply assume_cmpf_correct; eauto 3. rewrite <- Float32.cmp_double. apply EQ.
      + eapply longofsingle_intv; [| |now eauto].
        * destruct (Fleb smallest_longofsingle f) eqn:EQ. reflexivity.
          erewrite @is_bot_spec in H6. assumption.
          eapply assume_cmpf_correct; eauto 3.
          rewrite <- Float.cmp_swap. rewrite <- Float32.cmp_double. apply EQ.
        * destruct (Fleb f largest_longofsingle) eqn:EQ. reflexivity.
          erewrite @is_bot_spec in H7. assumption.
          eapply assume_cmpf_correct; eauto 3. rewrite <- Float32.cmp_double. apply EQ.
      + eapply longuofsingle_intv; [| |now eauto].
        * destruct (Fleb smallest_longuofsingle f) eqn:EQ. reflexivity.
          erewrite @is_bot_spec in H6. assumption.
          eapply assume_cmpf_correct; eauto 3.
          rewrite <- Float.cmp_swap. rewrite <- Float32.cmp_double. apply EQ.
        * destruct (Fleb f largest_longuofsingle) eqn:EQ. reflexivity.
          erewrite @is_bot_spec in H7. assumption.
          eapply assume_cmpf_correct; eauto 3. rewrite <- Float32.cmp_double. apply EQ.
    - destruct (convert_expr ab e gi) as [[[e0 ab0] nerr0] gi''] eqn:EQCVT'.
      assert (nerr0 = true).
      { destruct u; try congruence;
        try (destruct normalize_expr as [[[? ?] ?][]]; congruence);
        try (destruct normalize_expr_full as [[? ?] ?]; congruence);
        try (destruct normalize_expr_full64 as [[? ?] ?]; congruence);
        destruct (simplify_expr_alias ab0 e0 gi'') as [[? ?] ?];
        injection EQCVT; destruct nerr0; simpl; congruence. }
      subst. eapply IHe; eauto.
    - destruct (convert_expr ab e1 gi) as [[[e01 ab01] nerr01] gi''] eqn:EQCVT'1.
      eapply convert_expr_ok in EQCVT'1; eauto.
      destruct EQCVT'1 as (ρ'e1 & Hρ'e1 & Hgi'' & Hρ'1ρ'e1 & x'e1 & EVAL1 & REL1).
      assert (MONO2:=convert_expr_next_g_monotone e2 ab01 gi'').
      destruct (convert_expr ab01 e2 gi'') as [[[e02 ab02] nerr02] gi'''] eqn:EQCVT'2.
      specialize (MONO2 _ _ _ _ eq_refl).
      assert (Hρρ'e1 : ∀ id, related (ρ id) (ρ'e1 (Real id))).
      { intros. rewrite <- (proj1 (Hρ'1ρ'e1)). auto. }
      eapply convert_expr_ok in EQCVT'2; eauto.
      destruct EQCVT'2 as (ρ'e2 & Hρ'e2 & Hgi''' & Hρ'e1ρ'e2 & x'e2 & EVAL2 & REL2).
      assert (EVAL1':∀ ρ'', compat_gh gi''' ρ'e2 ρ'' -> eval_iexpr ρ'' e01 x'e1).
      { intros. apply EVAL1. split; intros; etransitivity.
        apply Hρ'e1ρ'e2. apply H.
        apply Hρ'e1ρ'e2. auto. apply H. Psatz.lia. }
      destruct H4; inv REL1; inv REL2;
      repeat (match type of EQCVT with
      | context [simplify_expr_alias ?ab ?e ?gi] =>
        let SIMPL := fresh "SIMPL" in let SIMPL_MONO := fresh "SIMPL_MONO" in
        destruct (simplify_expr_alias ab e gi) as [[] ?] eqn:SIMPL;
        assert (SIMPL_MONO:=SIMPL);
        apply simplify_expr_alias_monotone in SIMPL_MONO;
        (eapply simplify_expr_alias_correct in SIMPL;[|now eauto|now eauto|now eauto]);
        destruct SIMPL as (? & ? & ? & ? & ?)
      | context [normalize_expr ?ab ?e ?min ?max ?logmod ?gi] =>
        let NORM := fresh "NORM" in let NORM_MONO := fresh "NORM_MONO" in
        destruct (normalize_expr ab e min max logmod gi) as [[[] ?][]] eqn:NORM;
        assert (NORM_MONO:=NORM);
        apply normalize_expr_next_g_monotone in NORM_MONO;
        (eapply normalize_expr_correct in NORM; [|reflexivity|now eauto|now eauto|now eauto]);
        [destruct NORM as (? & ? & ? & ? & ? & ? & ? & ?)|destruct NORM as (? & ? & ? & ?)];
        simpl in EQCVT
      | context [normalize_expr_full ?ab ?e ?sgn ?gi] =>
        let NORM := fresh "NORM" in let NORM_MONO := fresh "NORM_MONO" in
        destruct (normalize_expr_full ab e sgn gi) as [[] ?] eqn:NORM;
        assert (NORM_MONO:=NORM);
        apply normalize_expr_full_next_g_monotone in NORM_MONO;
        eapply normalize_expr_full_correct in NORM;
        [destruct NORM as (? & ? & ? & ? & ? & ? & ? & ?)|now eauto|now eauto|now eauto]
      | context [normalize_expr_full64 ?ab ?e ?sgn ?gi] =>
        let NORM := fresh "NORM" in let NORM_MONO := fresh "NORM_MONO" in
        destruct (normalize_expr_full64 ab e sgn gi) as [[] ?] eqn:NORM;
        assert (NORM_MONO:=NORM);
        apply normalize_expr_full64_next_g_monotone in NORM_MONO;
        eapply normalize_expr_full64_correct in NORM;
        [destruct NORM as (? & ? & ? & ? & ? & ? & ? & ?)|now eauto|now eauto|now eauto]
      end);
      match type of EQCVT with context [if ?c then _ else _] => destruct c eqn:EQCOND end;
        inv EQCVT; rewrite !Bool.andb_true_iff in EQCOND;
      repeat match type of EQCOND with | _ /\ _ => destruct EQCOND as [EQCOND ?] end; subst.
      + change (Int.unsigned (Int.repr x0) = 0) in H0.
        rewrite <- Int.signed_eq_unsigned in H0 by (rewrite H0; compute; congruence).
        erewrite Int.eqm_samerepr, Int.signed_repr in H0 by eauto. subst.
        erewrite (is_bot_spec x3) in H15. discriminate.
        eauto using assume_cmp_correct.
      + change (Int.unsigned (Int.repr x) = Int.unsigned (Int.repr (Int.min_signed))) in H0.
        change (Int.unsigned (Int.repr x0) = Int.unsigned (Int.repr (-1))) in H1.
        eapply f_equal with (f:=Int.signed ∘ Int.repr) in H0.
        eapply f_equal with (f:=Int.signed ∘ Int.repr) in H1.
        rewrite !Int.repr_unsigned in H0, H1.
        erewrite Int.eqm_samerepr, Int.signed_repr in H0, H1 by eauto. subst.
        erewrite (is_bot_spec x3) in H15. discriminate.
        eauto using assume_cmp_correct.
      + change (Int.unsigned (Int.repr x0) = 0) in H0.
        erewrite Int.eqm_samerepr, Int.unsigned_repr in H0 by eauto. subst.
        erewrite (is_bot_spec x3) in H14. discriminate.
        eauto using assume_cmp_correct.
      + change (Int.unsigned (Int.repr x0) = 0) in H0.
        rewrite <- Int.signed_eq_unsigned in H0 by (rewrite H0; compute; congruence).
        erewrite Int.eqm_samerepr, Int.signed_repr in H0 by eauto. subst.
        erewrite (is_bot_spec x3) in H15. discriminate.
        eauto using assume_cmp_correct.
      + change (Int.unsigned (Int.repr x) = Int.unsigned (Int.repr (Int.min_signed))) in H0.
        change (Int.unsigned (Int.repr x0) = Int.unsigned (Int.repr (-1))) in H1.
        eapply f_equal with (f:=Int.signed ∘ Int.repr) in H0.
        eapply f_equal with (f:=Int.signed ∘ Int.repr) in H1.
        rewrite !Int.repr_unsigned in H0, H1.
        erewrite Int.eqm_samerepr, Int.signed_repr in H0, H1 by eauto. subst.
        erewrite (is_bot_spec x3) in H15. discriminate.
        eauto using assume_cmp_correct.
      + change (Int.unsigned (Int.repr x0) = 0) in H0.
        erewrite Int.eqm_samerepr, Int.unsigned_repr in H0 by eauto. subst.
        erewrite (is_bot_spec x3) in H14. discriminate.
        eauto using assume_cmp_correct.
      + rewrite (is_bot_spec x1) in H8. discriminate.
        eapply assume_cmp_correct; eauto 2.
        erewrite Int.eqm_samerepr in H by eauto.
        unfold Int.ltu in H. rewrite Int.unsigned_repr in H by auto. destruct zlt. discriminate.
        change (x2 >= Int.zwordsize) in g.
        simpl. destruct (Z.compare_spec x2 Int.zwordsize); try omega; auto.
      + rewrite (is_bot_spec x3) in H14. discriminate.
        eapply assume_cmp_correct; eauto 2.
        erewrite Int.eqm_samerepr in H by eauto.
        unfold Int.ltu in H. rewrite Int.unsigned_repr in H by auto. destruct zlt. discriminate.
        change (x4 >= Int.zwordsize) in g0.
        simpl. destruct (Z.compare_spec x4 Int.zwordsize); try omega; auto.
      + rewrite (is_bot_spec x3) in H14. discriminate.
        eapply assume_cmp_correct; eauto 2.
        erewrite Int.eqm_samerepr in H by eauto.
        unfold Int.ltu in H. rewrite Int.unsigned_repr in H by auto. destruct zlt. discriminate.
        change (x4 >= Int.zwordsize) in g0.
        simpl. destruct (Z.compare_spec x4 Int.zwordsize); try omega; auto.
      + change (Int64.unsigned (Int64.repr x0) = 0) in H0.
        rewrite <- Int64.signed_eq_unsigned in H0 by (rewrite H0; compute; congruence).
        erewrite Int64.eqm_samerepr, Int64.signed_repr in H0 by eauto. subst.
        rewrite (is_bot_spec x3) in H15. discriminate.
        eauto using assume_cmp_correct.
      + change (Int64.unsigned (Int64.repr x) = Int64.unsigned (Int64.repr (Int64.min_signed))) in H0.
        change (Int64.unsigned (Int64.repr x0) = Int64.unsigned (Int64.repr (-1))) in H1.
        eapply f_equal with (f:=Int64.signed ∘ Int64.repr) in H0.
        eapply f_equal with (f:=Int64.signed ∘ Int64.repr) in H1.
        rewrite !Int64.repr_unsigned in H0, H1.
        erewrite Int64.eqm_samerepr, Int64.signed_repr in H0, H1 by eauto. subst.
        rewrite (is_bot_spec x3) in H15. discriminate.
        eauto using assume_cmp_correct.
      + change (Int64.unsigned (Int64.repr x0) = 0) in H0.
        erewrite Int64.eqm_samerepr, Int64.unsigned_repr in H0 by eauto. subst.
        rewrite (is_bot_spec x3) in H14. discriminate.
        eauto using assume_cmp_correct.
      + change (Int64.unsigned (Int64.repr x0) = 0) in H0.
        rewrite <- Int64.signed_eq_unsigned in H0 by (rewrite H0; compute; congruence).
        erewrite Int64.eqm_samerepr, Int64.signed_repr in H0 by eauto. subst.
        rewrite (is_bot_spec x3) in H15. discriminate.
        eauto using assume_cmp_correct.
      + change (Int64.unsigned (Int64.repr x) = Int64.unsigned (Int64.repr (Int64.min_signed))) in H0.
        change (Int64.unsigned (Int64.repr x0) = Int64.unsigned (Int64.repr (-1))) in H1.
        eapply f_equal with (f:=Int64.signed ∘ Int64.repr) in H0.
        eapply f_equal with (f:=Int64.signed ∘ Int64.repr) in H1.
        rewrite !Int64.repr_unsigned in H0, H1.
        erewrite Int64.eqm_samerepr, Int64.signed_repr in H0, H1 by eauto. subst.
        rewrite (is_bot_spec x3) in H15. discriminate.
        eauto using assume_cmp_correct.
      + change (Int64.unsigned (Int64.repr x0) = 0) in H0.
        erewrite Int64.eqm_samerepr, Int64.unsigned_repr in H0 by eauto. subst.
        rewrite (is_bot_spec x3) in H14. discriminate.
        eauto using assume_cmp_correct.
      + rewrite (is_bot_spec x1) in H8. discriminate.
        eapply assume_cmp_correct; eauto 2.
        erewrite Int.eqm_samerepr in H by eauto.
        unfold Int.ltu in H. rewrite Int.unsigned_repr in H by auto. destruct zlt. discriminate.
        change (x2 >= Int64.zwordsize) in g.
        simpl. destruct (Z.compare_spec x2 Int64.zwordsize); try omega; auto.
      + rewrite (is_bot_spec x3) in H14. discriminate.
        eapply assume_cmp_correct; eauto 2.
        erewrite Int.eqm_samerepr in H by eauto.
        unfold Int.ltu in H. rewrite Int.unsigned_repr in H by auto. destruct zlt. discriminate.
        change (x4 >= Int64.zwordsize) in g0.
        simpl. destruct (Z.compare_spec x4 Int64.zwordsize); try omega; auto.
      + rewrite (is_bot_spec x3) in H14. discriminate.
        eapply assume_cmp_correct; eauto 2.
        erewrite Int.eqm_samerepr in H by eauto.
        unfold Int.ltu in H. rewrite Int.unsigned_repr in H by auto. destruct zlt. discriminate.
        change (x4 >= Int64.zwordsize) in g0.
        simpl. destruct (Z.compare_spec x4 Int64.zwordsize); try omega; auto.

    - destruct (convert_expr ab e1 gi) as [[[e01 ab01] nerr01] gi''] eqn:EQCVT'1.
      destruct (convert_expr ab01 e2 gi'') as [[[e02 ab02] nerr02] g'''] eqn:EQCVT'2.
      assert (nerr01 = true).
      { destruct b, nerr01, nerr02; simpl in EQCVT;
        repeat (match type of EQCVT with
          | context [normalize_expr ?ab ?e ?min ?max ?logmod ?gi] =>
            destruct (normalize_expr ab e min max logmod gi) as [[[] ?][]]; simpl in EQCVT
          | context [normalize_expr_full ?ab ?e ?sgn ?gi] =>
            destruct (normalize_expr_full ab e sgn gi) as [[] ?]
          | context [normalize_expr_full64 ?ab ?e ?sgn ?gi] =>
            destruct (normalize_expr_full64 ab e sgn gi) as [[] ?]
        end);
        congruence. }
      subst. eapply IHe1; eauto.
    - destruct (convert_expr ab e1 gi) as [[[e01 ab01] nerr01] gi''] eqn:EQCVT'1.
      edestruct convert_expr_ok as (ρ'e1 & Hρ'e1 & Hgi'' & Hρ'1ρ'e1 & x'e1 & EVAL & Hx'e1); eauto.
      destruct (convert_expr ab01 e2 gi'') as [[[e02 ab02] nerr02] gi'''] eqn:EQCVT'2.
      assert (nerr02 = true).
      { destruct b, nerr01, nerr02; simpl in EQCVT;
        repeat (match type of EQCVT with
          | context [normalize_expr ?ab ?e ?min ?max ?logmod ?gi] =>
            destruct (normalize_expr ab e min max logmod gi) as [[[] ?][]]; simpl in EQCVT
          | context [normalize_expr_full ?ab ?e ?sgn ?gi] =>
            destruct (normalize_expr_full ab e sgn gi) as [[] ?]
          | context [normalize_expr_full64 ?ab ?e ?sgn ?gi] =>
            destruct (normalize_expr_full64 ab e sgn gi) as [[] ?]
        end);
        congruence. }
      subst. eapply IHe2; eauto.
      intros. rewrite <- (proj1 Hρ'1ρ'e1). auto.
  Qed.

  Definition forget_temps (v:abt) (gi:ghosts_info) : abt+⊥ :=
    positive_rec _ (NotBot v)
      (fun x v => do_bot v <- v; idnc_forget (Ghost x) v)
      gi.(next_g).

  Lemma forget_temps_correct:
    ∀ ab (ρ:var -> ideal_num) gi, ρ ∈ γ ab ->
    ρ ∈ γ (forget_temps ab gi).
  Proof.
    unfold forget_temps. intros ab ρ gi.
    change (ρ ∈ γ ab) with (ρ ∈ γ (NotBot ab)).
    revert ρ. generalize (NotBot ab). clear ab.
    induction gi.(next_g) using Pos.peano_ind; intros.
    + rewrite positive_rec_base. auto.
    + rewrite positive_rec_succ.
      apply botbind_spec with ρ. intros. 2:now auto.
      replace ρ with (ρ+[Ghost p => ρ (Ghost p)]).
      eapply idnc_forget_correct, H0.
      apply Axioms.extensionality. intros. unfold upd. destruct eq_dec; congruence.
  Qed.

  Definition get_itv (e:nexpr Cells.cell) (v:abt * typesmap) : num_val_itv+⊥ :=
    let '(ie, v, _, gi) := convert_expr (NotBot (fst v)) e init_ghost_info in
    (* No need to forget the tmps here, because we will discard v *)
    do_bot v <- v;
    do_bot itv <- (idnc_get_query_chan v).(AbIdealEnv.get_itv) (simplify_expr ie);
      match itv, nexpr_ty e with
        | Just (IdealIntervals.Az ((ZIntervals.ITV a b) as itv')), NVTint =>
          NotBot (NVIint
            (match itv_normalizer (NotBot (Just itv')) Int.min_signed Int.max_unsigned Int.zwordsize with
               | Some q => Just (ZIntervals.ITV (a-Int.modulus*q) (b-Int.modulus*q))
               | None => All
             end))
        | All, NVTint => NotBot (NVIint All)
        | Just (IdealIntervals.Az ((ZIntervals.ITV a b) as itv')), NVTint64 =>
          NotBot (NVIint64
            (match itv_normalizer (NotBot (Just itv')) Int64.min_signed Int64.max_unsigned Int64.zwordsize with
               | Some q => Just (ZIntervals.ITV (a-Int64.modulus*q) (b-Int64.modulus*q))
               | None => All
             end))
        | All, NVTint64 => NotBot (NVIint64 All)
        | Just (IdealIntervals.Af itv), NVTfloat => NotBot (NVIfloat (Just itv))
        | Just (IdealIntervals.Af itv), NVTsingle => NotBot (NVIsingle (Just itv))
        | All, NVTfloat => NotBot (NVIfloat All)
        | All, NVTsingle => NotBot (NVIsingle All)
        | _, _ => Bot
      end.

  Lemma get_itv_correct:
    ∀ e ρ ab, ρ ∈ γ ab -> eval_nexpr ρ e ⊆ γ (get_itv e ab).
  Proof.
    intros e ρ [ab typs] [[ρ' [Hγρ' Hρρ']] Htyps] x EVAL. unfold get_itv.
    pose proof (convert_expr_ok _ _ _ EVAL (NotBot ab) _ _ Hγρ' (gamma_gh_init _ ) Hρρ'). unfold fst.
    destruct convert_expr as [[[e' ab'] nerr'] gi].
    edestruct H as (ρ'2 & Hab' & Hgi & Hρ'ρ'2 & x' & EVAL' & REL); eauto. clear H.
    assert (TY:=nexpr_ty_correct EVAL).
    revert TY. generalize (nexpr_ty e) as ty. intros.
    clear Hγρ' EVAL e ab typs Htyps.
    eapply botbind_spec with ρ'2; eauto. intros.
    eapply botbind_spec with x';
      [|eapply AbIdealEnv.get_itv_correct, simplify_expr_correct]; eauto.
    2:apply idnc_get_query_chan_correct; auto.
    clear EVAL' e' ρ ρ'2 ρ' Hρρ' Hρ'ρ'2 Hab' H Hgi.
    intros. destruct a0 as [|[]], x'; try contradiction; destruct TY; inv REL;
    try (constructor; auto).
    - destruct i.
      destruct itv_normalizer; [|now repeat constructor].
      rewrite Int.eqm_samerepr with (y:=z-Int.modulus * z2). destruct H; repeat constructor; omega.
      exists z2. ring.
    - destruct i.
      destruct itv_normalizer; [|now repeat constructor].
      rewrite Int64.eqm_samerepr with (y:=z-Int64.modulus * z2). destruct H; repeat constructor; omega.
      exists z2. ring.
  Qed.

  Definition forget (x:Cells.cell) (v:abt * typesmap) : (abt * typesmap)+⊥ :=
    do_bot fstres <- idnc_forget (Real x) (fst v);
    NotBot (fstres, Cells.ACTree.remove x (snd v)).

  Lemma forget_correct:
    ∀ x ρ, ∀ n ab,
      ρ ∈ γ ab ->
      (upd ρ x n) ∈ γ (forget x ab).
  Proof.
    intros x ρ n ab Hρ.
    destruct Hρ as [[ρ' [Hρ' Hρρ']] Htyps].
    eapply botbind_spec with (upd ρ' (Real x) (match n with NVint _ => _ | _ => _ end)).
    2:apply idnc_forget_correct; now auto. intros. split.
    - eexists. split. apply H.
      intro. unfold upd. destruct eq_dec. inv e. rewrite eq_dec_true.
      destruct n. apply related_signed. apply related_signed_long. constructor. constructor.
      destruct (eq_dec id x). congruence. auto.
    - simpl. intros y s Hx. erewrite Cells.ACTree.grspec in Hx; eauto.
      unfold upd. destruct (Cells.ACTree.elt_eq y x), (@eq_dec _ Cells.ACellDec y x); try congruence. auto.
  Qed.

  Fixpoint infer_var_types (e:nexpr Cells.cell) (ctx:option signedness) (typs:typesmap)
  : typesmap :=
    match e with
      | NEvar nvt x =>
        match ctx with
          | None =>
            match Cells.ACTree.get x typs with
              | None =>
                match nvt with
                  | NVTfloat | NVTsingle => typs
                  | _ => Cells.ACTree.set x (None, nvt) typs
                end
              | Some _ => typs
            end
          | Some _ => Cells.ACTree.set x (ctx, nvt) typs
        end
      | NEconst _ | NEunknown _ => typs
      | NEunop op e =>
        let ctx :=
            match op with
              | (Ofloatofint | Ofloatoflong | Olongofint | Osingleofint | Osingleoflong |
                 Ocast16signed | Ocast8signed) => Some Signed
              | (Ofloatofintu | Ofloatoflongu | Olongofintu | Osingleofintu | Osingleoflongu |
                 Ocast16unsigned | Ocast8unsigned) => Some Unsigned
              | _ => None
            end
        in
        infer_var_types e ctx typs
      | NEbinop op e1 e2 =>
        let ctx1 :=
          match op with
            | (Odiv | Omod | Odivl | Omodl | Ocmp _ | Ocmpl _ | Oshr | Oshrl) =>
              Some Signed
            | (Odivu | Omodu | Odivlu | Omodlu | Ocmpu _ | Ocmplu _ | Oshru | Oshrlu) =>
              Some Unsigned
            | _ => None
          end
        in
        let ctx2 :=
          match op with
            | (Odiv | Omod | Odivl | Omodl | Ocmp _ | Ocmpl _) => Some Signed
            | (Odivu | Omodu | Odivlu | Omodlu | Ocmpu _ | Ocmplu _ |
               Oshl | Oshr | Oshru | Oshll | Oshrl | Oshrlu) => Some Unsigned
            | _ => None
          end
        in
        (infer_var_types e2 ctx2 (infer_var_types e1 ctx1 typs))
    end.

  Lemma infer_var_types_consistent :
    ∀ ρ e nv ctx typs,
      eval_nexpr ρ e nv ->
      ρ ∈ γ typs ->
      (match nv with NVfloat _ | NVsingle _ => ctx = None | _ => True end) ->
      ρ ∈ γ (infer_var_types e ctx typs).
  Proof.
    induction e; simpl; intros nv ctx typs H Htyps Hfl v' [s ty] Hmap; inv H; simpl.
    - destruct ctx.
      + erewrite Cells.ACTree.gsspec in Hmap; eauto.
        destruct Cells.ACTree.elt_eq.
        inv Hmap. split; auto. inv H4; try easy.
        rewrite <- H1 in Hfl. destruct Hfl. congruence.
        rewrite <- H1 in Hfl. destruct Hfl. congruence.
        apply Htyps in Hmap. auto.
      + destruct (Cells.ACTree.get v typs) eqn:?. apply Htyps in Hmap. auto.
        destruct n.
        * erewrite Cells.ACTree.gsspec in Hmap; eauto.
          destruct Cells.ACTree.elt_eq.
          inv Hmap. auto. apply Htyps in Hmap. auto.
        * erewrite Cells.ACTree.gsspec in Hmap; eauto.
          destruct Cells.ACTree.elt_eq.
          inv Hmap. auto. apply Htyps in Hmap. auto.
        * apply Htyps in Hmap. auto.
        * apply Htyps in Hmap. auto.
    - apply Htyps in Hmap. auto.
    - apply Htyps in Hmap. auto.
    - match goal with Hmap : context [infer_var_types e ?ctx' typs] |- _ => set ctx' in Hmap end.
      edestruct (IHe _ o typs H2); eauto. destruct H4; auto.
    - match goal with Hmap : context [infer_var_types e1 ?ctx' typs] |- _ => set ctx' in Hmap end.
      match goal with Hmap : context [infer_var_types e2 ?ctx' _] |- _ => set ctx' in Hmap end.
      edestruct (fun typs => IHe2 _ o0 typs H5). 3:eauto. 2:destruct H6; auto. 2:auto.
      intros v [s' ty'] Hmap'.
      edestruct (IHe1 _ o typs H3); eauto. destruct H6; auto.
  Qed.

  (* If e is a variable, and we cannot get a better approximation for this variable than 
     top then we assign it to the interval corresponding to this sign, before assume.

     This is usefull in the following case (for example): if we have an assume of the form x < e,
     where e is well approximated but where we do not know anything about x, we need to force x
     to be in [-min_int; +max_int] before forwarding the assume. *)
  Definition assign_vars_to_top (e:nexpr Cells.cell) (v:abt * typesmap) : (abt * typesmap)+⊥ :=
    let types := infer_var_types e None (Cells.ACTree.empty _) in
    Cells.ACTree.fold (fun ab v sty =>
      do_bot (ab, typs) <- ab;
      match sty with
      | (Some s, ty) =>
        let typs := Cells.ACTree.set v sty typs in
        match get_itv (NEvar ty v) (ab, typs) with
        | NotBot (NVIint All) | NotBot (NVIint64 All) =>
          do_bot ab <- (idnc_assign (Real v) (IEZitv
                 (match s, ty with
                  | Signed, NVTint => signed_itv
                  | Unsigned, NVTint => unsigned_itv
                  | Signed, NVTint64 => signed64_itv
                  | Unsigned, NVTint64 => unsigned64_itv
                  | _, _ => signed_itv (* Dummy *)
                  end)) ab);
          NotBot (ab, typs)
        | _ => NotBot (ab, typs)
        end
      | (None, _) =>
        let typs :=
          match Cells.ACTree.get v typs with
          | None => Cells.ACTree.set v sty typs
          | Some _ => typs
          end
        in
        NotBot (ab, typs)
      end)
     types (NotBot v).

  Lemma assign_vars_to_top_correct:
    forall ab e ρ v,
      eval_nexpr ρ e v ->
      γ ab ρ -> γ (assign_vars_to_top e ab) ρ.
  Proof.
    unfold assign_vars_to_top. intros.
    pose proof infer_var_types_consistent ρ e _ None (Cells.ACTree.empty _) H.
    lapply H1. 2:intro; rewrite Cells.ACTree.gempty; discriminate.
    clear H1. intro. lapply H1. 2:destruct v; now auto. clear H1.
    change (γ (NotBot ab) ρ) in H0. revert H0. generalize (NotBot ab). clear ab.
    generalize (infer_var_types e None). clear e v H.
    intros. revert ρ H0 H.
    assert (∀ v s0, Cells.ACTree.get v (t (Cells.ACTree.empty _)) = Some s0 <->
                    List.In (v, s0) (Cells.ACTree.elements (t (Cells.ACTree.empty _)))).
    { split. apply Cells.ACTree.elements_correct. apply Cells.ACTree.elements_complete. }
    simpl. unfold gamma_typs. setoid_rewrite H. clear H.
    rewrite Cells.ACTree.fold_spec, <- fold_left_rev_right. setoid_rewrite in_rev.
    induction (rev (Cells.ACTree.elements (t (Cells.ACTree.empty _)))). auto. simpl. intros.
    eapply botbind_spec with ρ; [|apply IHl]; eauto.
    intros. destruct a. specialize (H0 _ _ (or_introl eq_refl)). destruct H0, p.
    destruct a0, H1 as [[? []] ?].
    destruct o.
    2:simpl; destruct (Cells.ACTree.get e t0); (split; [eexists; split; now eauto|]); auto;
      simpl; intro; rewrite Cells.ACTree.gsspec; destruct Cells.ACTree.elt_eq; [|now apply H4];
      intros s0 Hs0; inv Hs0; auto. simpl.
    assert (eval_nexpr ρ (NEvar n e) (ρ e)) by (constructor; auto).
    assert (γ (Cells.ACTree.set e (Some s, n) t0) ρ).
    { intro. simpl. rewrite Cells.ACTree.gsspec. destruct Cells.ACTree.elt_eq.
      - intros. inv H6. auto.
      - apply H4. }
    unfold snd in H2. inv H2; try discriminate.
    + assert (ASS:=idnc_assign_correct (Real e)
                (e:=IEZitv (match s with
                       | Signed => signed_itv
                       | Unsigned => unsigned_itv
                     end))
                (n:=INz (match s with
                           | Signed => Int.signed i
                           | Unsigned => Int.unsigned i
                         end)) a H1).
      lapply ASS.
      2:constructor; destruct s;
      [apply signed_itv_correct|apply unsigned_itv_correct].
      clear ASS; intro.
      change Cells.ACTree.elt with Cells.cell in *.
      destruct idnc_assign. contradiction.
      assert (γ (NotBot(x0, Cells.ACTree.set e (Some s, NVTint) t0)) ρ).
      { split. eexists. split. now eauto.
        intro. unfold upd. destruct eq_dec; auto. inv e0. rewrite <- H9.
        destruct s. apply related_signed. apply related_unsigned.
        intro. simpl. rewrite Cells.ACTree.gsspec. destruct Cells.ACTree.elt_eq.
        intros. inv H7. split. auto. rewrite <- H9. constructor. auto. }
      destruct get_itv as [|[[]|[]| |]]; (try now (econstructor; eauto)); eauto;
      (split; [eexists; eauto |auto]).
    + assert (ASS:=idnc_assign_correct (Real e)
                (e:=IEZitv (match s with
                       | Signed => signed64_itv
                       | Unsigned => unsigned64_itv
                     end))
                (n:=INz (match s with
                           | Signed => Int64.signed i
                           | Unsigned => Int64.unsigned i
                         end)) a H1).
      lapply ASS.
      2:constructor; destruct s;
      [apply signed64_itv_correct|apply unsigned64_itv_correct].
      clear ASS; intro.
      change Cells.ACTree.elt with Cells.cell in *.
      destruct idnc_assign. contradiction.
      assert (γ (NotBot(x0, Cells.ACTree.set e (Some s, NVTint64) t0)) ρ).
      { split. eexists. split. now eauto.
        intro. unfold upd. destruct eq_dec; auto. inv e0. rewrite <- H9.
        destruct s. apply related_signed_long. apply related_unsigned_long.
        intro. simpl. rewrite Cells.ACTree.gsspec. destruct Cells.ACTree.elt_eq.
        intros. inv H7. split. auto. rewrite <- H9. constructor. auto. }
      destruct get_itv as [|[[]|[]| |]]; (try now (econstructor; eauto)); eauto;
      (split; [eexists; eauto |auto]).
  Qed.

  Definition assign (x:Cells.cell) (e:nexpr Cells.cell) (v:abt*typesmap) : (abt*typesmap)+⊥ :=
    let v := assign_vars_to_top e v in
    do_bot (v, typs) <- v;
    let '(e', v, _, gi) := convert_expr (NotBot v) e init_ghost_info in
    do_bot v <- v;
    let ty := nexpr_ty e in
    let typs := Cells.ACTree.set x (None, ty) typs in
    let '(v, gi) :=
    match ty with
    | NVTint =>
      match normalize_expr (NotBot v) e'
              Int.min_signed Int.max_unsigned Int.zwordsize gi with
      | (e, v, gi, true) =>
        (do_bot v <- v; idnc_assign (Real x) e v, gi)
      | _ =>
        (idnc_forget (Real x) v, gi)
      end
    | NVTint64 =>
      match normalize_expr (NotBot v) e'
              Int64.min_signed Int64.max_unsigned Int64.zwordsize gi with
      | (e, v, gi, true) =>
        (do_bot v <- v; idnc_assign (Real x) e v, gi)
      | _ =>
        (idnc_forget (Real x) v, gi)
      end
    | NVTfloat | NVTsingle =>
      (let e := simplify_expr e' in idnc_assign (Real x) e v, gi)
    end
    in
    do_bot v <- v;
    do_bot v <- forget_temps v gi;
    NotBot (v, typs).

  Lemma assign_correct:
    ∀ x e ρ, ∀ n ab,
      ρ ∈ γ ab ->
      n ∈ eval_nexpr ρ e ->
      (upd ρ x n) ∈ γ (assign x e ab).
  Proof.
    intros x e ρ n ab0 Hρ EVAL.
    unfold assign. apply botbind_spec with ρ.
    2:eauto using assign_vars_to_top_correct; eauto. clear Hρ.
    intros [ab typs] [[ρ' [Hρ' Hρρ']] Htyps].
    pose proof (convert_expr_ok _ _ _ EVAL (NotBot ab) _ _ Hρ' (gamma_gh_init _) Hρρ').
    destruct convert_expr as [[[e' v'] nerr] gi].
    edestruct H as (ρ'2 & Hρ'2 & Hgi & Hρ'ρ'2 & x' & Hx' & Hnx'); auto. clear H.
    apply botbind_spec with ρ'2. 2:now eauto. intros.
    pose proof (nexpr_ty_correct EVAL).
    destruct H0; inv Hnx'.
    - destruct (normalize_expr (NotBot a) e' Int.min_signed Int.max_unsigned
                               Int.zwordsize gi) as [[[e1 ab1] gi'][]] eqn:EQNORM.
      edestruct normalize_expr_correct with (b:=true)as
          (ρ'3 & Hρ'2ρ'3 & Hρ'3 & Hgi' & x'0  & Hx'0bounded &Hx'x'0 & EVALx'0); eauto.
      reflexivity.
      { eapply botbind_spec with (upd ρ'3 (Real x) (INz x'0)).
        - intros. eapply botbind_spec with (upd ρ'3 (Real x) (INz x'0)).
          2:now apply forget_temps_correct, H0. split.
          + eexists. split. apply H1. intro. unfold upd.
            destruct (eq_dec id x), (eq_dec (Real id) (Real x)); try congruence.
            erewrite Int.eqm_samerepr. econstructor. auto.
            rewrite <- (proj1 Hρ'2ρ'3), <- (proj1 Hρ'ρ'2). auto.
          + intros y typsy. simpl. rewrite Cells.ACTree.gsspec. unfold upd, eq_dec, Cells.ACellDec.
            destruct Cells.ACTree.elt_eq. intros eq. inv eq. split; constructor. auto.
        - eapply botbind_spec with ρ'3, Hρ'3.
          intros. apply idnc_assign_correct; auto. }
      { eapply botbind_spec with (upd ρ'2 (Real x) (INz x0)).
        - intros. eapply botbind_spec with (upd ρ'2 (Real x) (INz x0)).
          2:now apply forget_temps_correct, H0. split.
          + eexists. split. apply H1. intro. unfold upd.
            destruct (eq_dec id x), (eq_dec (Real id) (Real x)); try congruence. auto.
            rewrite <- (proj1 Hρ'ρ'2). auto.
          + intros y typsy. simpl. rewrite Cells.ACTree.gsspec.
            unfold upd, eq_dec, Cells.ACellDec. destruct Cells.ACTree.elt_eq.
            intros eq. inv eq. split; constructor. auto.
        - apply idnc_forget_correct; auto. }
    - destruct (normalize_expr (NotBot a) e' Int64.min_signed Int64.max_unsigned
                               Int64.zwordsize gi) as [[[e1 ab1] gi'][]] eqn:EQNORM.
      edestruct normalize_expr_correct with (b:=true)as
          (ρ'3 & Hρ'2ρ'3 & Hρ'3 & Hgi' & x'0  & Hx'0bounded &Hx'x'0 & EVALx'0); eauto. reflexivity.
      { eapply botbind_spec with (upd ρ'3 (Real x) (INz x'0)).
        - intros. eapply botbind_spec with (upd ρ'3 (Real x) (INz x'0)).
          2:now apply forget_temps_correct, H0. split.
          + eexists. split. apply H1. intro. unfold upd.
            destruct (eq_dec id x), (eq_dec (Real id) (Real x)); try congruence.
            erewrite Int64.eqm_samerepr. econstructor. auto.
            rewrite <- (proj1 Hρ'2ρ'3), <- (proj1 Hρ'ρ'2). auto.
          + intros y typsy. simpl. rewrite Cells.ACTree.gsspec.
            unfold upd, eq_dec, Cells.ACellDec. destruct Cells.ACTree.elt_eq.
             intros eq. inv eq. split; constructor. auto.
        - eapply botbind_spec with ρ'3, Hρ'3.
          intros. apply idnc_assign_correct; auto. }
      { eapply botbind_spec with (upd ρ'2 (Real x) (INz x0)).
        - intros. eapply botbind_spec with (upd ρ'2 (Real x) (INz x0)).
          2:now apply forget_temps_correct, H0. split.
          + eexists. split. apply H1. intro. unfold upd.
            destruct (eq_dec id x), (eq_dec (Real id) (Real x)); try congruence. auto.
            rewrite <- (proj1 Hρ'ρ'2). auto.
          + intros y typsy. simpl. rewrite Cells.ACTree.gsspec.
            unfold upd, eq_dec, Cells.ACellDec. destruct Cells.ACTree.elt_eq.
            intros eq. inv eq. split; constructor. auto.
        - apply idnc_forget_correct; auto. }
    - eapply botbind_spec with (upd ρ'2 (Real x) (INf f)).
      + intros. eapply botbind_spec with (upd ρ'2 (Real x) (INf f)).
        2:now apply forget_temps_correct, H0. split.
        * eexists. split. apply H1. intro. unfold upd.
          destruct (eq_dec id x), (eq_dec (Real id) (Real x)); try congruence. auto.
          rewrite <- (proj1 Hρ'ρ'2). auto.
        * intros y typsy. simpl. rewrite Cells.ACTree.gsspec.
          unfold upd, eq_dec, Cells.ACellDec. destruct Cells.ACTree.elt_eq.
          intro. inv H2. split. reflexivity. constructor. auto.
      + apply idnc_assign_correct, simplify_expr_correct; auto.
    - eapply botbind_spec with (upd ρ'2 (Real x) (INf (Float.of_single f))).
      + intros. eapply botbind_spec with (upd ρ'2 (Real x) (INf (Float.of_single f))).
        2:now apply forget_temps_correct, H0. split.
        * eexists. split. apply H1. intro. unfold upd.
          destruct (eq_dec id x), (eq_dec (Real id) (Real x)); try congruence. auto.
          rewrite <- (proj1 Hρ'ρ'2). auto.
        * intros y typsy. simpl. rewrite Cells.ACTree.gsspec.
          unfold upd, eq_dec, Cells.ACellDec. destruct Cells.ACTree.elt_eq.
          intro. inv H2. split. reflexivity. constructor. auto.
      + apply idnc_assign_correct, simplify_expr_correct; auto.
  Qed.

  Definition assume (e:nexpr Cells.cell) (b:bool) (v:abt * typesmap) : (abt*typesmap)+⊥ :=
    let v := assign_vars_to_top e v in
    do_bot (v, typs) <- v;
    let '(e, v, _, gi) := convert_expr (NotBot v) e init_ghost_info in
    do_bot v <- v;
    match normalize_expr (NotBot v) e Int.min_signed Int.max_unsigned Int.zwordsize gi with
      | (e, v, gi, true) =>
        do_bot v <- v;
        do_bot v <- idnc_assume e b v;
        do_bot v <- forget_temps v gi;
        NotBot (v, typs)
      | _ =>
        do_bot v <- forget_temps v gi;
        NotBot (v, typs)
    end.

  Lemma assume_correct:
    ∀ e ρ ab b,
      ρ ∈ γ ab ->
      of_bool b ∈ eval_nexpr ρ e ->
      ρ ∈ γ (assume e b ab).
  Proof.
    intros e ρ ab b Hρ EVAL. unfold assume.
    eapply botbind_spec with ρ. 2:eauto using assign_vars_to_top_correct.
    intros. destruct a, H as [[ρ' [Hρ' Hρρ']] Htyps].
    pose proof (convert_expr_ok _ _ _ EVAL (NotBot a) _ _ Hρ' (gamma_gh_init _) Hρρ').
    destruct convert_expr as [[[e' v'] nerr'] gi].
    edestruct H as (ρ'2 & Hρ'2 & Hgi & Hρ'ρ'2 & x' & EVAL2 & REL); auto. clear H.
    eapply botbind_spec with ρ'2; eauto.
    assert (∃ z, x' = INz z) by (destruct b; inv REL; eauto).
    destruct H; subst. intros.
    destruct (normalize_expr (NotBot a0) e' Int.min_signed Int.max_unsigned
                               Int.zwordsize gi) as [[[e1 ab1] gi'][]] eqn:EQNORM.
    edestruct normalize_expr_correct with (b:=true) as (ρ'3 & Hρ'2ρ'3 & Hρ'3 & Hgi' & x'0  & Hx'0bounded &Hx'x'0 & EVALx'0); eauto. reflexivity.
    - eapply botbind_spec with ρ'3, Hρ'3. intros. eapply botbind_spec with ρ'3. intros.
      + eapply botbind_spec with ρ'3, forget_temps_correct, H1.
        split. exists ρ'3. split. auto.
        intros. rewrite <- (proj1 Hρ'2ρ'3), <- (proj1 Hρ'ρ'2). auto. auto.
      + apply idnc_assume_correct; eauto.
        replace (if b then 1 else 0) with x'0. auto.
        inv REL. 2:destruct b; discriminate.
        apply Int.eqm_samerepr in Hx'x'0. rewrite Hx'x'0 in H2.
        assert (Int.repr x'0 = if b then Int.one else Int.zero) by
            (destruct b; compute in H2; unfold Int.one, Int.zero; congruence).
        destruct (Coqlib.zlt x'0 0).
        * exfalso. apply (f_equal Int.signed) in H1.
          rewrite Int.signed_repr in H1. subst x'0. destruct b; discriminate.
          split. now intuition. assert (0 < Int.max_signed) by reflexivity. omega.
        * apply (f_equal Int.unsigned) in H1.
          rewrite Int.unsigned_repr in H1 by intuition. destruct b; apply H1.
    - eapply botbind_spec with ρ'2, forget_temps_correct, H. split.
      exists ρ'2. split. eauto. intros. rewrite <- (proj1 Hρ'ρ'2). auto. auto.
  Qed.

  Definition noerror (e:nexpr Cells.cell) (v:abt * typesmap) : bool :=
    let '(_, _, nerr, _) := convert_expr (NotBot (fst v)) e init_ghost_info in
    nerr.

  Lemma noerror_correct:
    ∀ e ρ ab,
      ρ ∈ γ ab ->
      noerror e ab = true ->
      error_nexpr ρ e ->
      False.
  Proof.
    unfold noerror. intros.
    destruct H as [[? [? ?]] ?].
    destruct (convert_expr (NotBot (fst ab)) e init_ghost_info) as [[[] ?] ?] eqn:EQ. subst.
    eapply convert_expr_noerror with (v:=NotBot (fst ab)); eauto.
  Qed.

  Function zitv_concretize_int_aux (from: Z) (nb: N)
           (acc: IntSet.int_set) {measure N.to_nat nb} : TreeAl.ZShareTree.t unit :=
    match nb with
    | N0 => acc
    | Npos p =>
      let nb' := Pos.pred_N p in
      zitv_concretize_int_aux (Z.succ from) nb' (TreeAl.ZShareTree.set from tt acc)
    end.
  Proof.
    intros from nb acc p ->. rewrite N.pos_pred_spec, Nnat.N2Nat.inj_pred. zify. intuition.
  Defined.

  Definition zitv_concretize_int (max: N) (i: ZIntervals.itv) : TreeAl.ZShareTree.t unit+⊤+⊥ :=
    match i with
    | ZIntervals.ITV m M =>
      if zlt M m then Bot
      else
        let n := Z.to_N (M-m) in
        if N.leb n max
        then NotBot (Just (zitv_concretize_int_aux m (N.succ n) (TreeAl.ZShareTree.empty _)))
        else NotBot All
    end.

  Lemma zitv_concretize_int_correct:
    ∀ itv max z, γ itv z ->
      match zitv_concretize_int max itv with
        | NotBot (Just ab) => TreeAl.ZShareTree.get z ab = Some tt
        | NotBot All => True
        | Bot => False
      end.
  Proof.
    intros [a b] max z ITV.
    unfold zitv_concretize_int.
    destruct ITV as [A B].
    destruct zlt. omega.
    destruct N.leb. 2:exact I.
    remember (TreeAl.ZShareTree.empty unit) as init.
    assert (TreeAl.ZShareTree.get z init = Some tt \/ a <= z) by auto.
    clear Heqinit max A.
    remember (N.succ (Z.to_N (b-a))).
    assert (b-a+1= Z.of_N n).
    { subst n. rewrite N2Z.inj_succ, Z2N.id. Psatz.lia. omega. }
    clear Heqn g.
    functional induction (zitv_concretize_int_aux a n init).
    - destruct H. apply H. zify; omega.
    - eapply IHt.
      rewrite TreeAl.ZShareTree.gsspec. destruct (TreeAl.ZShareTree.elt_eq z from). auto.
      destruct H. auto. right. unfold TreeAl.ZShareTree.elt, TreeAl.Z_EQ_POS.s in *. omega.
      rewrite N.pos_pred_spec, N2Z.inj_pred_max. zify; omega.
  Qed.

  Definition concretize_int (max: N) (e: nexpr Cells.cell) (ab: abt*typesmap) : IntSet.int_set+⊤+⊥ :=
    let ab := fst ab in
    let '(e, ab, _, gi) := convert_expr (NotBot ab) e init_ghost_info in
    do_bot ab  <- ab;
    let e := simplify_expr e in
    let c := idnc_get_query_chan ab in
    do_bot cong <- c.(AbIdealEnv.get_congr) e;
    let '{| ZCongruences_defs.zc_rem := rem; ZCongruences_defs.zc_mod := modn |} := cong in
    let modz := Z.of_N modn in
    match modn with
      | N0 => NotBot (Just (IntSet.singleton (Int.repr rem)))
      | Npos modn =>
        do_bot itv <- c.(AbIdealEnv.get_itv) e;
        do_bot itv <- IdealIntervals.in_z itv;
        match itv with
          | All => NotBot All
          | Just (ZIntervals.ITV a b) =>
            let itvdiv := ZIntervals.ITV (-((rem-a)/(Zpos modn))) ((b-rem)/(Zpos modn)) in
            do_bot set <- zitv_concretize_int max itvdiv;
            NotBot (toplift_lift (IntSet.map (fun x => IntSet.proj (Int.repr (x*Zpos modn+rem)))) set)
        end
    end.

  Lemma concretize_int_correct :
    ∀ max e ρ ab, ρ ∈ γ ab ->
      (eval_nexpr ρ e ∘ NVint) ⊆ γ (concretize_int max e ab).
  Proof.
    intros. unfold concretize_int.
    simpl in H0. destruct H as [[ρ' [Hρ' REL]] _].
    destruct (convert_expr (NotBot (fst ab)) e init_ghost_info)  as [[[e' ab'] ?] ?]  eqn:EQ.
    edestruct convert_expr_ok with (v:=NotBot (fst ab)) as (ρ'2 & Hρ'2 & Hgi & Hρ'ρ'2 & z & EVAL & RELz); eauto.
    inv RELz.
    eapply botbind_spec with ρ'2, Hρ'2. intros.
    eapply botbind_spec.
    2:eapply get_congr_correct, simplify_expr_correct; eauto;
    apply idnc_get_query_chan_correct; auto.
    intros [] (q & Hq). simpl in Hq.
    destruct zc_mod.
    - simpl. unfold IntSet.singleton. red.
      rewrite TreeAl.ZShareTree.gsspec. destruct TreeAl.ZShareTree.elt_eq. auto.
      destruct n. f_equal. f_equal. subst zc_rem. ring.
    - eapply botbind_spec.
      2:eapply AbIdealEnv.get_itv_correct, simplify_expr_correct;
        [apply idnc_get_query_chan_correct; eauto|eauto].
      intros. eapply botbind_spec; [|apply IdealIntervals.in_z_correct; eauto].
      intros. destruct a1 as [|[]]. exact I.
      assert ((-q) ∈ γ (ZIntervals.ITV (- ((zc_rem - z) / Z.pos p))
                                       ((z0 - zc_rem) / Z.pos p))).
      { destruct H2. split; [apply -> Z.opp_le_mono|]; apply Z.div_le_lower_bound; Psatz.lia. }
      pose proof (zitv_concretize_int_correct _ max _ H3).
      destruct zitv_concretize_int as [|[|set]]. auto. exact I. simpl.
      replace x with ((-q) * Z.pos p + zc_rem).
      apply (IntSet.map_in (fun _ =>  _)). auto.
      rewrite Hq. simpl. ring.
  Qed.

  Program Definition transfer_types (v2:abt) (typs1 typs2:typesmap) : abt :=
    let typs := Cells.ACTree.shcombine_diff (fun _ s1 s2 =>
      match s1, s2 return _ with
        | Some (Some s, NVTint), Some (None, NVTint) =>
          Some (Some s, NVTint)
        | Some (Some s, NVTint64), Some (None, NVTint64) =>
          Some (Some s, NVTint64)
        | _, _ => None
      end) _ typs1 typs2
    in
    Cells.ACTree.fold (fun ab x sty =>
              match sty return _ with
                | (Some s, ty) =>
                  match get_itv (NEvar ty x) (ab, typs2) return _ with
                    | (NotBot (NVIint All)) | NotBot (NVIint64 All) =>
                      match idnc_assign (Real x) (IEZitv
                                   (match s, ty return _ with
                                    | Signed, NVTint => signed_itv
                                    | Unsigned, NVTint => unsigned_itv
                                    | Signed, NVTint64 => signed64_itv
                                    | Unsigned, NVTint64 => unsigned64_itv
                                    | _, _ => signed_itv (* Dummy *)
                                    end)) ab
                            return _
                      with Bot => ab (* I see no reason why this would happend *)
                        | NotBot ab => ab
                      end
                    | _ => ab
                  end
                | _ => ab
              end)
           typs v2.
  Next Obligation. destruct v as [[[] []]|]; auto. Qed.

  Lemma transfer_types_correct:
    ∀ typs1 typs2 v2 (ρ:Cells.cell -> num_val),
      ρ ∈ γ (v2, typs2) ->
      ρ ∈ γ (transfer_types v2 typs1 typs2).
  Proof.
    intros. unfold transfer_types.
    match goal with |- context [Cells.ACTree.shcombine_diff ?f' ?Hf' typs1 typs2] => set (f:=f'); set (Hf:=Hf') end.
    assert (ρ ∈ γ (Cells.ACTree.shcombine_diff f Hf typs1 typs2)).
    { intros x s. rewrite Cells.ACTree.gshcombine_diff by auto.
      destruct H. red in H0.
      destruct (Cells.ACTree.get x typs1) as [[[] []]|] eqn:?, (Cells.ACTree.get x typs2) as [[[] []]|] eqn:?; simpl;
        intro eq; inv eq; (split; [constructor|apply (H0 _ _ Heqo0)]). }
    revert H0.
    assert (∀ v s0, Cells.ACTree.get v (Cells.ACTree.shcombine_diff f Hf typs1 typs2) = Some s0 <->
                    List.In (v, s0) (Cells.ACTree.elements (Cells.ACTree.shcombine_diff f Hf typs1 typs2))).
    { split. apply Cells.ACTree.elements_correct. apply Cells.ACTree.elements_complete. }
    simpl. unfold gamma_typs. setoid_rewrite H0. clear H0.
    rewrite Cells.ACTree.fold_spec, <- fold_left_rev_right. setoid_rewrite in_rev.
    match goal with
      | |- appcontext [fold_right ?f] => set (F:=f)
    end.
    induction (rev (Cells.ACTree.elements (Cells.ACTree.shcombine_diff f Hf typs1 typs2))); intros. apply H.
    lapply IHl. 2:intros; apply H0; simpl; auto.
    clear IHl. intro IHl.
    destruct a as [? [[] ?]]. 2:now auto. simpl. unfold F at 1. simpl.
    edestruct H0. left. eauto.
    pose proof (get_itv_correct (NEvar n e) ρ (_, typs2) (conj IHl (proj2 H)) (ρ e)
                                (eval_NEvar _ _ eq_refl H2)).
    pose proof IHl. destruct IHl as (ρ' & Hρ' & Hρρ').
    change Cells.ACTree.elt with Cells.cell in *.
    destruct (get_itv (NEvar n e) (fold_right F v2 l, typs2)) as [|[[]|[]| |]]; auto;
    simpl in H2; inv H2; try discriminate; simpl;
    match goal with |- context [idnc_assign (Real e) ?e' ?ab] =>
      refine (let ASS' :=
        let val := match s with Signed => _ | _ => _ end in
        @idnc_assign_correct _ _ _ (Real e) e' ρ' val ab Hρ'  _
              in (fun ASS => _) ASS'); [
                        (destruct s; constructor;
                         [apply signed_itv_correct|apply unsigned_itv_correct]) ||
                        (destruct s; constructor;
                         [apply signed64_itv_correct|apply unsigned64_itv_correct])
                       |clear ASS'; intros; destruct (idnc_assign (Real e) e' ab)]
    end; auto;
    eexists; (split; [apply ASS|]);
    intro y; unfold upd; (destruct eq_dec; [|now auto]); inv e0; destruct s;
    rewrite <- H7;
    apply related_signed || apply related_unsigned ||
          apply related_signed_long || apply related_unsigned_long.
  Qed.

  Program Definition join_typs (typs1 typs2:typesmap) : typesmap :=
    Cells.ACTree.shcombine (fun _ s1 s2 =>
      match s1, s2 return _ with
        | Some (Some Signed, NVTint),     Some (Some Signed | None,   NVTint)
        | Some (Some Signed, NVTint64),   Some (Some Signed | None,   NVTint64)
        | Some (Some Unsigned, NVTint),   Some (Some Unsigned | None, NVTint)
        | Some (Some Unsigned, NVTint64), Some (Some Unsigned | None, NVTint64)
          => s1
        | Some (None, NVTint),   Some (Some Signed,   NVTint)
        | Some (None, NVTint64), Some (Some Signed,   NVTint64)
        | Some (None, NVTint),   Some (Some Unsigned, NVTint)
        | Some (None, NVTint64), Some (Some Unsigned, NVTint64)
          => s2
        | Some (Some _, NVTfloat), Some (Some _, NVTfloat)
        | Some (Some _, NVTsingle), Some (Some _, NVTsingle)
        | Some (None, NVTint), Some (None, NVTint)
        | Some (None, NVTint64), Some (None, NVTint64)
        | Some (None, NVTfloat), Some (None, NVTfloat)
        | Some (None, NVTsingle), Some (None, NVTsingle)
          => s1
        | _, _ => None
      end) _ typs1 typs2.
  Next Obligation. destruct v as [[[[]|] []]|]; auto. Qed.

  Lemma join_typs_correct:
    ∀ a b, γ a ∪ γ b ⊆ γ (join_typs a b).
  Proof.
    unfold join_typs. intros a b ρ Hρ k. rewrite Cells.ACTree.gshcombine.
    destruct Hρ as [Hρ|Hρ]; specialize (Hρ k);
    destruct (Cells.ACTree.get k a) as [[[[]|] []]|],
             (Cells.ACTree.get k b) as [[[[]|] []]|];
      try discriminate;
      destruct (Hρ _ eq_refl) as [? ?];
      try discriminate;
      intros ? eq; inv eq; simpl; auto.
  Qed.

  Instance top_ietme : top_op ((abt * typesmap)+⊥) :=
    do_bot t <- top;
    NotBot (t, Cells.ACTree.empty _).

  Program Instance leb_ietme : leb_op (abt * typesmap) := fun ab1 ab2 =>
    (fst ab1) ⊑ (fst ab2) &&
    Cells.ACTree.shforall2 (fun _ s1 s2 =>
                   match s1, s2 return _ with
                     | _, None
                     | Some (_, NVTint), Some (_, NVTint)
                     | Some (_, NVTint64), Some (_, NVTint64)
                     | Some (_, NVTsingle), Some (None, NVTsingle)
                     | Some (_, NVTfloat), Some (None, NVTfloat)
                     | Some (Some _, NVTsingle), Some (Some _, NVTsingle)
                     | Some (Some _, NVTfloat), Some (Some _, NVTfloat) => true
                     | _, _ => false
                   end) _
      (snd ab1) (snd ab2).
  Next Obligation. destruct v as [[[] []]|]; auto. Qed.

  Instance join_ietme : join_op (abt * typesmap) ((abt * typesmap)+⊥) := fun ab1 ab2 =>
    let ab1' := transfer_types (fst ab1) (snd ab2) (snd ab1) in
    let ab2' := transfer_types (fst ab2) (snd ab1) (snd ab2) in
    do_bot j <- ab1' ⊔ ab2';
    NotBot (j, join_typs (snd ab1) (snd ab2)).

  (* TODO :
     - should we use transfer_signedness in widen_ietme ? *)
  Instance widen_ietme : widen_op (abt_iter * typesmap+⊥) ((abt * typesmap)+⊥) :=
    {| widen ab1 ab2 :=
         let '(w1, w2) := fst ab1 ▽ do_bot (ab2, _) <- ab2; NotBot (ab2) in
         let typs :=
           match ab1 with
           | (_, Bot) => do_bot (_, typs2) <- ab2; NotBot typs2
           | (_, NotBot typs1) =>
             match ab2 with
             | Bot => NotBot typs1
             | NotBot (_, typs2) => NotBot (join_typs typs1 typs2)
             end
           end
         in
         ((w1, typs),
          do_bot w <- w2;
          do_bot typs <- typs;
          NotBot (w, typs));
       init_iter x :=
         match x with
         | Bot => (init_iter Bot, Bot)
         | NotBot x => (init_iter (NotBot (fst x)), NotBot (snd x))
        end |}.

  Instance top_ietme_correct : top_op_correct ((abt * typesmap)+⊥) (Cells.cell->num_val).
  Proof.
    intro ρ. unfold top, top_ietme. eapply botbind_spec with (x:=_:var->ideal_num).
    2:apply top_correct.
    - split.
      exists (fun id =>
                match id with
                | Real id =>
                  match ρ id with
                  | NVint i => INz (Int.unsigned i)
                  | NVint64 i => INz (Int64.unsigned i)
                  | NVfloat f => INf f
                  | NVsingle f => INf (Float.of_single f)
                  end
                | Ghost _ => INz 0
                end).
      split. apply H.
      intro. destruct (ρ id). apply related_unsigned. apply related_unsigned_long. constructor. constructor.
      intros x s. simpl. erewrite Cells.ACTree.gempty. discriminate.
  Qed.

  Instance leb_ietme_correct : leb_op_correct (abt * typesmap) (Cells.cell->num_val).
  Proof.
    unfold leb_op_correct. intros. unfold leb, leb_ietme in H.
    rewrite Bool.andb_true_iff in H. destruct H.
    destruct H0 as [[ρ []] ?]. split.
    - eexists ρ. split. eapply leb_correct; eauto. auto.
    - intros v ab Hvab. rewrite Cells.ACTree.shforall2_correct in H1. specialize (H1 v).
      unfold typesmap in *. rewrite Hvab in H1.
      specialize (H3 v).
      destruct ab as [? []], (Cells.ACTree.get v (snd a1)) as [[[] []]|]; try easy;
      destruct (H3 _ eq_refl), o; inv H4; inv H1; simpl in *; subst; auto.
  Qed.

  Instance join_ietme_correct : join_op_correct (abt * typesmap) ((abt * typesmap)+⊥) (Cells.cell->num_val).
  Proof.
    intros [x typsx] [y typsy] ρ [Hρ|Hρ]; assert (Hρtr:=Hρ);
    eapply transfer_types_correct in Hρtr;
    destruct Hρtr as (ρ' & Hρ' & Hρρ');
    eapply (botbind_spec _ (_:var->ideal_num)).
    - eexists. eexists ρ'. split. apply H. auto. apply join_typs_correct. left. apply Hρ.
    - apply join_correct. eauto.
    - eexists. eexists ρ'. split. apply H. auto. apply join_typs_correct. right. apply Hρ.
    - apply join_correct. eauto.
  Qed.

  Instance widen_ietme_correct :
    widen_op_correct (abt_iter * typesmap+⊥) ((abt * typesmap)+⊥) (Cells.cell->num_val).
  Proof.
    split.
    - destruct x. contradiction. destruct 1. split.
      destruct H as (? & ? & ?). eexists. split. apply init_iter_correct. eauto. auto. auto.
    - simpl. intros [x typsx] ytypsy ρ [(ρ' & Hρ' & Hρρ') Htyps].
      match goal with
      | |- context [?A ▽ ?B] =>
        pose proof (widen_incr A B ρ' Hρ'); destruct (A ▽ B)
      end.
      destruct typsx. contradiction.
      split. split. eexists. split. apply H. auto.
      + destruct ytypsy. apply Htyps. destruct x1. apply join_typs_correct. left. apply Htyps.
      + apply botbind_spec with ρ'. 2:apply H. intros.
        apply botbind_spec with ρ. intros. split. eexists. split. apply H0.  auto. apply H1.
        destruct ytypsy. apply Htyps. destruct x1. apply join_typs_correct. left. apply Htyps.
  Qed.

  Instance idealenv_to_machineenv:
    ab_machine_env (var:=Cells.cell) (abt * typesmap) (abt_iter * typesmap+⊥) :=
    { assign := assign
    ; forget := forget
    ; assume := assume
    ; get_itv := fun e v => do_bot v <- assign_vars_to_top e v; get_itv e v
    ; concretize_int := concretize_int
    ; noerror := noerror }.
  Proof.
  - apply assign_correct.
  - apply forget_correct.
  - apply assume_correct.
  - intros e ρ ab H a E.
    generalize (assign_vars_to_top_correct _ _ _ _ E H).
    destruct (assign_vars_to_top _ _) as [ | v ]. intros ().
    intros K. exact (get_itv_correct _ _ _ K _ E).
  - apply concretize_int_correct.
  - apply noerror_correct.
  Defined.

End Functor.

Instance typesmap_TS : ToString typesmap :=
  fun _ => ""%string.

Existing Instances idealenv_to_machineenv.
