(* *********************************************************************)
(*                                                                     *)
(*              The Verasco static analyzer                            *)
(*                                                                     *)
(*          Vincent Laporte, Université de Rennes 1                    *)
(*          Jacques-Henri Jourdan, INRIA Paris-Rocquencourt            *)
(*                                                                     *)
(*  Copyright Institut National de Recherche en Informatique et en     *)
(*  Automatique and Université de Rennes 1.  All rights reserved.      *)
(*  This file is distributed under the terms of the GNU General        *)
(*  Public License as published by the Free Software Foundatio,        *)
(*  either version 2 of the License, or (at your option) any later     *)
(*  version.  This file is also distributed under the terms of the     *)
(*  INRIA Non-Commercial License Agreement.                            *)
(*                                                                     *)
(* *********************************************************************)

Require Import Utf8.
Require Import IdealExpr.
Require Import AbIdealNonrel.
Require Import ZArith.
Require Import Coqlib.
Require Import ZIntervals FloatIntervals FloatIntervalsBackward IdealIntervals.
Require Import ZCongruences_defs.
Require Import Integers.
Require Import AdomLib.
Require Import AbIdealEnv IdealBoxDomain ShareTree Util.

Existing Instance FloatIntervals.gamma.

Definition forward_binop (op:i_binary_operation) (x y:abs+⊤) : abs+⊤+⊥ :=
  match op with
    | IOadd =>
      do_bot x <- in_z x;
      do_bot y <- in_z y;
      match x, y with
        | Just (ZIntervals.ITV mx Mx), Just (ZIntervals.ITV my My) =>
          NotBot (Just (Az (ZIntervals.ITV (mx+my) (Mx+My))))
        | _, _ => NotBot All
      end
    | IOsub =>
      do_bot x <- in_z x;
      do_bot y <- in_z y;
      match x, y with
        | Just (ZIntervals.ITV mx Mx), Just (ZIntervals.ITV my My) =>
          NotBot (Just (Az (ZIntervals.ITV (mx-My) (Mx-my))))
        | _, _ => NotBot All
      end
    | IOmul => botlift_fun1 (toplift_lift Az) (botbind2 forward_mult (in_z x) (in_z y))
    | IOdiv => botlift_fun1 (toplift_lift Az) (botbind2 forward_div (in_z x) (in_z y))
    | IOmod => botlift_fun1 (toplift_lift Az) (botbind2 forward_mod (in_z x) (in_z y))

    | IOand => botlift_fun1 (toplift_lift Az) (botbind2 forward_land (in_z x) (in_z y))
    | IOor => botlift_fun1 (toplift_lift Az) (botbind2 forward_lor (in_z x) (in_z y))
    | IOxor => botlift_fun1 (toplift_lift Az) (botbind2 forward_lxor (in_z x) (in_z y))
    | IOshl => botlift_fun1 (toplift_lift Az) (botbind2 forward_shl (in_z x) (in_z y))
    | IOshr => botlift_fun1 (toplift_lift Az) (botbind2 forward_shr (in_z x) (in_z y))

    | IOcmp c =>
      do_bot x <- in_z x;
      do_bot y <- in_z y;
      let hasOne := is_bot (backward_cmp_l c x y) in
      let hasZero := is_bot (backward_cmp_l (negate_comparison c) x y) in
      match hasOne, hasZero with
        | false, false => NotBot (Just (Az (ZIntervals.ITV 0 1)))
        | true, false => NotBot (Just (Az (ZIntervals.ITV 0 0)))
        | false, true => NotBot (Just (Az (ZIntervals.ITV 1 1)))
        | true, true => Bot
      end
    | IOcmpf c =>
      do_bot x <- in_f x;
      do_bot y <- in_f y;
      let hasOne := is_bot (backward_cmpf_l c x y true) in
      let hasZero := is_bot (backward_cmpf_l c x y false) in
      match hasOne, hasZero with
        | false, false => NotBot (Just (Az (ZIntervals.ITV 0 1)))
        | true, false => NotBot (Just (Az (ZIntervals.ITV 0 0)))
        | false, true => NotBot (Just (Az (ZIntervals.ITV 1 1)))
        | true, true => Bot
      end

    | IOaddf => botlift_fun1 (toplift_lift Af) (botbind2 forward_fadd (in_f x) (in_f y))
    | IOsubf => botlift_fun1 (toplift_lift Af) (botbind2 forward_fsub (in_f x) (in_f y))
    | IOmulf => botlift_fun1 (toplift_lift Af) (botbind2 forward_fmult (in_f x) (in_f y))
    | IOdivf => botlift_fun1 (toplift_lift Af) (botbind2 forward_fdiv (in_f x) (in_f y))
  end.

Lemma forward_binop_correct:
  ∀ op,
  ∀ x x_ab, x ∈ γ x_ab ->
  ∀ y y_ab, y ∈ γ y_ab ->
  eval_ibinop op x y ⊆ γ (forward_binop op x_ab y_ab).
Proof.
  intros [] x x_ab Hx y y_ab Hy z EVAL; try reflexivity; inversion EVAL; clear EVAL; subst;
    try eapply botlift_fun1_spec; eauto; intros;
    try eapply toplift_lift_spec; eauto;
    try eapply botbind2_spec; eauto; intros.

  - destruct x_ab0 as [|[]], y_ab0 as [|[]]; try easy.
    destruct H, H0; split; omega.
  - destruct x_ab0 as [|[]], y_ab0 as [|[]]; try easy.
    destruct H, H0; split; omega.
  - auto using forward_mult_correct.
  - auto using forward_div_correct.
  - auto using forward_mod_correct.
  - auto using forward_land_correct.
  - auto using forward_lor_correct.
  - auto using forward_lxor_correct.
  - auto using forward_shl_correct.
  - auto using forward_shr_correct.
  - destruct (Zcmp c i1 i2) eqn:?.
    erewrite (is_bot_spec _ (backward_cmp_l c x_ab0 y_ab0)).
    destruct @is_bot; split; omega.
    eapply backward_cmp_l_correct; eauto.
    erewrite (is_bot_spec _ (backward_cmp_l (negate_comparison c) x_ab0 y_ab0)).
    destruct @is_bot; split; omega.
    eapply backward_cmp_l_correct; eauto.
    destruct c; simpl in Heqb |- *; destruct Z.compare; congruence.
  - auto using forward_fadd_correct.
  - auto using forward_fsub_correct.
  - auto using forward_fmult_correct.
  - auto using forward_fdiv_correct.
  - destruct (Floats.Float.cmp c f1 f2) eqn:?.
    erewrite (is_bot_spec _ (backward_cmpf_l c x_ab0 y_ab0 true)).
    destruct @is_bot; split; omega.
    eapply backward_cmpf_l_correct; eauto.
    erewrite (is_bot_spec _ (backward_cmpf_l c x_ab0 y_ab0 false)).
    destruct @is_bot; split; omega.
    eapply backward_cmpf_l_correct; eauto.
Qed.

Definition forward_unop (op:i_unary_operation) (x:abs+⊤) : abs+⊤+⊥ :=
  match op with
    | IOneg => botlift_fun1 (toplift_lift Az) (botlift_fun1 forward_neg (in_z x))
    | IOnot => botlift_fun1 (toplift_lift Az) (botlift_fun1 forward_lnot (in_z x))
    | IOnegf => botlift_fun1 (toplift_lift Af) (botlift_fun1 forward_fneg (in_f x))
    | IOabsf => botlift_fun1 (toplift_lift Af) (botlift_fun1 forward_fabs (in_f x))
    | IOsingleoffloat => botlift_fun1 (toplift_lift Af) (botlift_fun1 forward_singleoffloat (in_f x))
    | IOfloatofz => botlift_fun1 (toplift_lift Af) (botlift_fun1 forward_floatofz (in_z x))
    | IOsingleofz => botlift_fun1 (toplift_lift Af) (botlift_fun1 forward_singleofz (in_z x))
    | IOzoffloat => botlift_fun1 (toplift_lift Az) (botlift_fun1 forward_zoffloat (in_f x))
  end.

Lemma forward_unop_correct:
  ∀ op,
  ∀ x x_ab, x ∈ γ x_ab ->
  eval_iunop op x ⊆ γ (forward_unop op x_ab).
Proof.
  intros [] x x_ab Hx z EVAL; inversion EVAL; clear EVAL; subst;
    eapply botlift_fun1_spec; eauto; intros;
  try eapply toplift_lift_spec; eauto; intros;
  eapply botlift_fun1_spec; eauto; intros.
  - apply forward_neg_correct; auto.
  - apply forward_lnot_correct; auto.
  - apply forward_fneg_correct; auto.
  - apply forward_fabs_correct; auto.
  - apply forward_singleoffloat_correct; auto.
  - apply forward_floatofz_correct; auto.
  - apply forward_singleofz_correct; auto.
  - intros. eapply forward_zoffloat_correct; eauto.
Qed.

Definition backward_binop (op:i_binary_operation) (res x y:abs+⊤) : (abs+⊤*abs+⊤)+⊥ :=
  match op with
    | IOadd =>
      do_bot res <- in_z res;
      do_bot x <- in_z x;
      do_bot y <- in_z y;
      do_bot x' <-
        match y, res with
          | Just (ZIntervals.ITV my My), Just (ZIntervals.ITV mres Mres) =>
            x ⊓ (Just (ZIntervals.ITV (mres - My) (Mres - my)))
          | _, _ => NotBot All
        end;
      do_bot y' <-
        match x, res with
          | Just (ZIntervals.ITV mx Mx), Just (ZIntervals.ITV mres Mres) =>
            y ⊓ (Just (ZIntervals.ITV (mres - Mx) (Mres - mx)))
          | _, _ => NotBot All
        end;
      NotBot (toplift_lift Az x', toplift_lift Az y')
    | IOsub =>
      do_bot res <- in_z res;
      do_bot x <- in_z x;
      do_bot y <- in_z y;
      do_bot x' <-
        match y, res with
          | Just (ZIntervals.ITV my My), Just (ZIntervals.ITV mres Mres) =>
            x ⊓ (Just (ZIntervals.ITV (mres + my) (Mres + My)))
          | _, _ => NotBot All
        end;
      do_bot y' <-
        match x, res with
          | Just (ZIntervals.ITV mx Mx), Just (ZIntervals.ITV mres Mres) =>
            y ⊓ (Just (ZIntervals.ITV (mx - Mres) (Mx - mres)))
          | _, _ => NotBot All
        end;
      NotBot (toplift_lift Az x', toplift_lift Az y')
    | IOmul =>
      do_bot res <- in_z res;
      do_bot x <- in_z x;
      do_bot y <- in_z y;
      do_bot x' <- backward_mul_l res x y;
      do_bot y' <- backward_mul_l res y x;
      NotBot (toplift_lift Az x', toplift_lift Az y')
    | IOcmp c =>
      do_bot res <- in_z res;
      do_bot x <- in_z x;
      do_bot y <- in_z y;
      match res with
        | Just (ZIntervals.ITV m M) =>
          match (Zlt_bool 1 m || Zlt_bool M 1)%bool,
                (Zlt_bool 0 m || Zlt_bool M 0)%bool with
            | false, false => NotBot (toplift_lift Az x, toplift_lift Az y)
            | false, true =>
              do_bot x' <- backward_cmp_l c x y;
              do_bot y' <- backward_cmp_l (swap_comparison c) y x;
              NotBot (toplift_lift Az x', toplift_lift Az y')
            | true, false =>
              do_bot x' <- backward_cmp_l (negate_comparison c) x y;
              do_bot y' <- backward_cmp_l (swap_comparison (negate_comparison c)) y x;
              NotBot (toplift_lift Az x', toplift_lift Az y')
            | true, true => Bot
          end
        | All =>
          NotBot (toplift_lift Az x, toplift_lift Az y)
      end
    | IOdiv | IOmod | IOand | IOor | IOxor | IOshl | IOshr =>
      (* TODO *)
      do_bot res <- in_z res;
      do_bot x <- in_z x;
      do_bot y <- in_z y;
      NotBot (toplift_lift Az x, toplift_lift Az y)
    | IOaddf =>
      do_bot res <- in_f res;
      do_bot x <- in_f x;
      do_bot y <- in_f y;
      do_bot x' <- backward_addf_l res x y;
      do_bot y' <- backward_addf_r res x y;
      NotBot (toplift_lift Af x', toplift_lift Af y')
    | IOsubf =>
      do_bot res <- in_f res;
      do_bot x <- in_f x;
      do_bot y <- in_f y;
      do_bot x' <- backward_subf_l res x y;
      do_bot y' <- backward_subf_r res x y;
      NotBot (toplift_lift Af x', toplift_lift Af y')
    | IOmulf | IOdivf =>
      (* TODO *)
      do_bot res <- in_f res;
      do_bot x <- in_f x;
      do_bot y <- in_f y;
      NotBot (toplift_lift Af x, toplift_lift Af y)
    | IOcmpf c =>
      do_bot res <- in_z res;
      do_bot x <- in_f x;
      do_bot y <- in_f y;
      match res with
        | Just (ZIntervals.ITV m M) =>
          match (Zlt_bool 1 m || Zlt_bool M 1)%bool,
                (Zlt_bool 0 m || Zlt_bool M 0)%bool with
            | false, false => NotBot (toplift_lift Af x, toplift_lift Af y)
            | false, true =>
              do_bot x' <- backward_cmpf_l c x y true;
              do_bot y' <- backward_cmpf_l (swap_comparison c) y x true;
              NotBot (toplift_lift Af x', toplift_lift Af y')
            | true, false =>
              do_bot x' <- backward_cmpf_l c x y false;
              do_bot y' <- backward_cmpf_l (swap_comparison c) y x false;
              NotBot (toplift_lift Af x', toplift_lift Af y')
            | true, true => Bot
          end
        | All =>
          NotBot (toplift_lift Af x, toplift_lift Af y)
      end
  end.

Lemma backward_binop_correct:
  ∀ op,
    ∀ x x_ab, x ∈ γ x_ab ->
    ∀ y y_ab, y ∈ γ y_ab ->
    ∀ res res_ab, res ∈ γ res_ab ->
      eval_ibinop op x y res ->
      (x, y) ∈ γ (backward_binop op res_ab x_ab y_ab).
Proof.
Opaque backward_subf_r.
  intros op x x_ab Hx y y_ab Hy res res_ab Hres EVAL.
  destruct op;
    inv EVAL;
    try (eapply botbind_spec; eauto; intros);
    try (eapply botbind2_spec; eauto; intros);
    try (eapply botbind2_spec; eauto; intros);
    try (split; eapply toplift_lift_spec; simpl in *; now eauto).
  - destruct y_ab0 as [|[]], a as [|[]]; try easy.
    eapply meet_correct. split. auto. simpl in *. omega.
  - destruct x_ab0 as [|[]], a as [|[]]; try easy.
    eapply meet_correct. split. auto. simpl in *. omega.
  - destruct y_ab0 as [|[]], a as [|[]]; try easy.
    eapply meet_correct. split. auto. simpl in *. omega.
  - destruct x_ab0 as [|[]], a as [|[]]; try easy.
    eapply meet_correct. split. auto. simpl in *. omega.
  - eapply backward_mul_l_correct; eauto.
  - eapply backward_mul_l_correct; eauto.
    rewrite Zmult_comm. auto.
  - destruct a as [|[]]. split; eapply toplift_lift_spec; simpl in *; eauto.
    assert (Zlt_bool 1 z || Zlt_bool z0 1 = true -> Zcmp c i1 i2 = false)%bool.
    { rewrite Bool.orb_true_iff, <- !Zlt_is_lt_bool. simpl in H. destruct Zcmp. omega. auto. }
    assert (Zlt_bool 0 z || Zlt_bool z0 0 = true -> Zcmp c i1 i2 = true)%bool.
    { rewrite Bool.orb_true_iff, <- !Zlt_is_lt_bool. simpl in H. destruct Zcmp. auto. omega. }
    destruct (Zlt_bool 1 z || Zlt_bool z0 1)%bool eqn:Hres1,
             (Zlt_bool 0 z || Zlt_bool z0 0)%bool eqn:Hres0;
      (specialize (H2 eq_refl) || clear H2); (specialize (H3 eq_refl) || clear H3).
    + congruence.
    + eapply botbind2_spec. split; eapply toplift_lift_spec; simpl in *; eauto.
      eapply backward_cmp_l_correct; eauto.
      destruct c; simpl in *; destruct (i1 ?= i2); congruence.
      eapply backward_cmp_l_correct; eauto.
      destruct c; simpl in *; destruct (Zcompare_spec i1 i2), (Zcompare_spec i2 i1);
        try congruence; exfalso; omega.
    + eapply botbind2_spec. split; eapply toplift_lift_spec; simpl in *; eauto.
      eapply backward_cmp_l_correct; eauto.
      eapply backward_cmp_l_correct; eauto.
      destruct c; simpl in *; destruct (Zcompare_spec i1 i2), (Zcompare_spec i2 i1);
        try congruence; exfalso; omega.
    + split; eapply toplift_lift_spec; simpl in *; eauto.
  - eapply backward_addf_l_correct; eauto.
  - eapply backward_addf_r_correct; eauto.
  - eapply backward_subf_l_correct; eauto.
  - eapply backward_subf_r_correct; eauto.
  - destruct a as [|[]]. split; eapply toplift_lift_spec; simpl in *; eauto.
    assert (Zlt_bool 1 z || Zlt_bool z0 1 = true -> Floats.Float.cmp c f1 f2 = false)%bool.
    { rewrite Bool.orb_true_iff, <- !Zlt_is_lt_bool. simpl in H. destruct Floats.Float.cmp. omega. auto. }
    assert (Zlt_bool 0 z || Zlt_bool z0 0 = true -> Floats.Float.cmp c f1 f2 = true)%bool.
    { rewrite Bool.orb_true_iff, <- !Zlt_is_lt_bool. simpl in H. destruct Floats.Float.cmp. auto. omega. }
    destruct (Zlt_bool 1 z || Zlt_bool z0 1)%bool eqn:Hres1,
             (Zlt_bool 0 z || Zlt_bool z0 0)%bool eqn:Hres0;
      (specialize (H2 eq_refl) || clear H2); (specialize (H3 eq_refl) || clear H3).
    + congruence.
    + eapply botbind2_spec. split; eapply toplift_lift_spec; simpl in *; eauto.
      eapply backward_cmpf_l_correct; eauto.
      eapply backward_cmpf_l_correct; eauto.
      rewrite Floats.Float.cmp_swap. auto.
    + eapply botbind2_spec. split; eapply toplift_lift_spec; simpl in *; eauto.
      eapply backward_cmpf_l_correct; eauto.
      eapply backward_cmpf_l_correct; eauto.
      rewrite Floats.Float.cmp_swap. auto.
    + split; eapply toplift_lift_spec; simpl in *; eauto.
Qed.

Definition backward_unop (op:i_unary_operation) (res x:abs+⊤) : abs+⊤+⊥ :=
  match op with
    | IOneg =>
      botbind2 (fun x res =>
        botlift_fun1 (toplift_lift Az) (x ⊓ (forward_neg res)))
        (in_z x) (in_z res)
    | IOnot =>
      botbind2 (fun x res =>
        botlift_fun1 (toplift_lift Az) (x ⊓ (forward_lnot res)))
        (in_z x) (in_z res)
    | IOnegf =>
      botbind2 (fun x res =>
        botlift_fun1 (toplift_lift Af) (x ⊓ (forward_fneg res)))
        (in_f x) (in_f res)

    | IOabsf =>
      botlift_fun1 (toplift_lift Af) (botbind2 backward_absf (in_f res) (in_f x))
    | IOsingleoffloat =>
      botlift_fun1 (toplift_lift Af) (botbind2 backward_singleoffloat (in_f res) (in_f x))
    | IOzoffloat =>
      botlift_fun1 (toplift_lift Af) (botbind2 backward_zoffloat (in_z res) (in_f x))
    | IOfloatofz =>
      botlift_fun1 (toplift_lift Az) (botbind2 backward_floatofz (in_f res) (in_z x))
    | IOsingleofz =>
      botlift_fun1 (toplift_lift Az) (botbind2 backward_singleofz (in_f res) (in_z x))
  end.

Lemma backward_unop_correct:
  ∀ op,
    ∀ x x_ab, x ∈ γ x_ab ->
    ∀ res res_ab, res ∈ γ res_ab ->
      eval_iunop op x res ->
      x ∈ γ (backward_unop op res_ab x_ab).
Proof.
  intros op x x_ab Hx res res_ab Hres EVAL.
  destruct op; inv EVAL;
    try (eapply botbind2_spec; eauto; intros);
    try (eapply botlift_fun1_spec; eauto; intros);
    try (eapply toplift_lift_spec; eauto).
  - eapply meet_correct. split. auto.
    rewrite <- Z.opp_involutive. apply forward_neg_correct; auto.
  - eapply meet_correct. split. auto.
    rewrite <- Z.lnot_involutive. apply forward_lnot_correct; auto.
  - eapply meet_correct. split. auto.
    apply forward_fneg_correct in H0.
    destruct f as [[]|[]|[]|[]]; easy.
  - eapply botbind2_spec; eauto.
    intros.
    eapply backward_absf_correct; auto.
  - eapply botbind2_spec; eauto.
    intros.
    eapply backward_singleoffloat_correct; auto.
  - eapply botbind2_spec; eauto.
    intros.
    eapply backward_floatofz_correct; auto.
  - eapply botbind2_spec; eauto.
    intros.
    eapply backward_singleofz_correct; auto.
  - eapply botbind2_spec; eauto.
    intros.
    eapply backward_zoffloat_correct; eauto.
Qed.

Definition extract_msgs (x:var) v : messages_chan :=
  match singleton v with
  | Some v => Broadcasted_msg (Known_value_msg x v)::nil
  | None => Itv_msg x v::nil
  end.

Lemma extract_msgs_correct:
  ∀ x v ρ,
    γ v (ρ x) → γ (extract_msgs x v) ρ.
Proof.
  intros. unfold extract_msgs. pose proof (singleton_correct v).
  destruct singleton; (constructor; [|constructor]).
  - apply H0; auto.
  - apply H.
Qed.

Definition enrich_query_chan fev (chan:query_chan) : query_chan :=
  {| get_var_ty := chan.(get_var_ty);
     get_itv expr :=
       do_bot ab <- fev expr;
       do_bot ab' <- chan.(get_itv) expr;
       match ab return (abs+⊤)+⊥ with
         | All => NotBot ab'
         | Just (Az i1) =>
           do_bot i2 <- in_z ab';
           botlift_fun1 (toplift_lift Az) (Just i1 ⊓ i2)
         | Just (Af i1) =>
           do_bot i2 <- in_f ab';
           botlift_fun1 (toplift_lift Af) (Just i1 ⊓ i2)
       end;
     get_congr := chan.(get_congr);
     get_eq_expr := chan.(get_eq_expr);
     linearize_expr := chan.(linearize_expr) |}.

Definition reduce_itv_congr (zc:zcongr) (itv:ZIntervals.itv) : ZIntervals.itv+⊥ :=
  let 'ZIntervals.ITV a b := itv in
  match zc.(zc_mod) with
    | 0%N =>
      let x := zc.(zc_rem) in
      meet itv (ZIntervals.ITV x x)
    | 1%N => NotBot itv
    | m =>
      let b := b-(b-zc.(zc_rem)) mod (Z_of_N m) in
      let a := a+(zc.(zc_rem)-a) mod (Z_of_N m) in
      if Z.leb a b then NotBot (ZIntervals.ITV a b)
      else Bot
  end.

Lemma reduce_itv_congr_correct:
  ∀ zc itv v,
    v ∈ γ zc ->
    v ∈ γ itv ->
    v ∈ γ (reduce_itv_congr zc itv).
Proof.
  intros. destruct itv as [a b]. simpl.
  destruct (N.eq_dec zc.(zc_mod) 0), (N.eq_dec zc.(zc_mod) 1);
    try congruence; try rewrite e.
  - apply meet_correct. split. auto.
    destruct H. rewrite H, e. simpl. omega.
  - auto.
  - destruct H0, H.
    assert (v = zc.(zc_rem) - x * Z.of_N zc.(zc_mod)) by omega. subst. clear H.
    assert (x * Z.of_N zc.(zc_mod) <= zc.(zc_rem) - a) by omega.
    assert (-x * Z.of_N zc.(zc_mod) <= b - zc.(zc_rem)) by (rewrite <- Zopp_mult_distr_l; omega).
    apply Zdiv_le_lower_bound, Zmult_le_compat_l with (p:=Z.of_N zc.(zc_mod)) in H; try (zify; omega).
    apply Zdiv_le_lower_bound, Zmult_le_compat_l with (p:=Z.of_N zc.(zc_mod)) in H2; try (zify; omega).
    rewrite Z.mul_opp_r in H2.
    destruct (zc_mod zc) as [|[]]; try congruence;
    rewrite !Z.mod_eq by (zify; omega);
    match goal with
      | |- context [Z.leb ?x ?y] => destruct (Z.leb_spec x y)
    end; try split; Psatz.lia.
Qed.

Definition process_msg fev {AB} bev (msg:message) (abenv:AB*query_chan)
           : (AB*messages_chan)+⊥ :=
  match msg with
    | Congr_msg x zc =>
      do_bot ab <- fev (IEvar INTz x);
      match ab with
        | Just (Az itv0) =>
          do_bot itv <- reduce_itv_congr zc itv0;
          let 'ZIntervals.ITV a b := itv in
          let msgs :=
            if Z.eqb a b then Broadcasted_msg (Known_value_msg x (INz a))::nil
            else if itv0 ⊑ itv then msg::nil
                 else msg::Itv_msg x (Az itv)::nil
          in
          do_bot abenv <- bev (IEvar INTz x) (Az itv) abenv;
          NotBot (abenv, msgs)
        | _ => NotBot (fst abenv, msg::nil)
      end
    | Itv_msg x itv =>
      let ty := match itv with Az _ => INTz | Af _ => INTf end in
      do_bot abenv <- bev (IEvar ty x) itv abenv;
      NotBot (abenv, msg::nil)
    | _ => NotBot (fst abenv, msg::nil)
  end.

Lemma process_msg_correct :
  ∀ fev AB (GAB:gamma_op AB (var -> ideal_num)) bev (msg:message) (abenv:AB * query_chan) ρ,

    (∀ e v, eval_iexpr ρ e v -> v ∈ γ (fev e)) ->
    (∀ e v abv ab, eval_iexpr ρ e v -> v ∈ γ abv -> ρ ∈ γ ab -> ρ ∈ γ (bev e abv ab)) ->

    γ msg ρ → γ abenv ρ → γ (process_msg fev bev msg abenv) ρ.
Proof.
  intros * Hfev Hbev Hmsg Habenv. unfold process_msg.
  destruct msg; try now (repeat constructor; simpl; auto; apply Habenv).
  - eapply botbind_spec. constructor. eauto. constructor. auto. constructor.
    eapply Hbev; eauto. simpl in Hmsg. constructor. auto.
    destruct a, (ρ v); try contradiction; constructor.
  - do 2 red in Hmsg. destruct (ρ v) eqn:?. contradiction.
    eapply botbind_spec, Hfev. 2:constructor; eauto. 2:constructor. intros ab Hab.
    destruct ab as [|[itv|]]; try now (repeat constructor; simpl; try rewrite Heqi; auto; apply Habenv).
    eapply botbind_spec, reduce_itv_congr_correct; eauto.
    intros [a' b'] Ha'b'. eapply botbind_spec, Hbev; eauto.
    2:constructor; eauto. 2:constructor. 2:now auto. split. now auto.
    destruct (Z.eqb_spec a' b'); [|destruct @leb]; simpl; repeat constructor;
    simpl; rewrite Heqi; auto.
    f_equal. simpl in Ha'b'. Psatz.lia.
Qed.

Instance itv_dom : ab_ideal_nonrel abs :=
  { widen := iditv_widen
  ; const n :=
      match n with
      | INz z => Just (Az (ZIntervals.ITV z z))
      | INf f =>
        if Fappli_IEEE.is_nan _ _ f then All else Just (Af (FloatIntervals.ITV f f))
      end
  ; z_itv := @Just _ ∘ Az
  ; forward_unop := forward_unop
  ; forward_binop := forward_binop
  ; backward_unop := backward_unop
  ; backward_binop := backward_binop

  ; forward_var fev x ab c := NotBot ab
  ; backward_var fev AB bev bvae x ab abenv :=
      match ab with
        | Az itv =>
          do_bot zc <- (snd abenv).(get_congr) (IEvar INTz x);
          do_bot itv' <- reduce_itv_congr zc itv;
          do_bot ab <-
            if itv ⊑ itv' then NotBot (fst abenv)
            else bvae (Az itv') abenv;
          match (snd abenv).(get_eq_expr) x with
            | Some (Leaf e) => bev e (Az itv') (ab, snd abenv)
            | _ => NotBot ab
          end
        | _ => NotBot (fst abenv)
      end
  ; enrich_query_chan := enrich_query_chan
  ; process_msg := process_msg
  ; assign_msgs := extract_msgs
  ; assume_msgs x v_old v := extract_msgs x v
}.

Instance itv_dom_correct : ab_ideal_nonrel_correct itv_dom _ := {}.
Proof.
  - destruct x; unfold join, iditv_join; simpl; unfold physEq; f_equal; f_equal;
    (apply ZIntervals.join_eq || apply FloatIntervals.join_eq).
  - destruct x; unfold widen, iditv_widen; simpl; unfold physEq;
    rewrite ?ZIntervals.widen_eq, ?FloatIntervals.widen_eq; auto.
  - destruct x; unfold leb, iditv_leb; simpl;
    rewrite ?ZIntervals.leb_eq, ?FloatIntervals.leb_eq; auto.
  - apply IdealIntervals.widen_incr.
  - destruct n; simpl.
    + destruct (Fappli_IEEE.is_nan _ _ f) eqn:ISNAN. constructor.
      split; apply FloatLib.Fleb_refl; auto.
    + omega.
  - auto.
  - apply forward_unop_correct.
  - apply forward_binop_correct.
  - apply backward_unop_correct.
  - apply backward_binop_correct.
  - intros. simpl. auto.
  - intros. destruct ab. 2:apply H3. simpl.
    destruct (ρ x) eqn:?. contradiction.
    eapply botbind_spec, H3. 2:repeat constructor; eauto.
    intros. eapply botbind_spec, reduce_itv_congr_correct; eauto.
    intros i' Hi'. eapply botbind_spec with ρ.
    intros.
    pose proof get_eq_expr_correct (proj2 H3) x.
    destruct @get_eq_expr as [[]|]; try apply H5.
    specialize (H6 _ eq_refl). inv H6. eapply H0. eauto. rewrite Heqi0. auto.
    split. auto. apply H3.
    destruct (i ⊑ i'). apply H3. auto.
  - constructor.
    + intros. apply H0.
    + intros.
      apply botbind_spec with (x:=a). intros.
      eapply botbind_spec. intros. 2:eapply get_itv_correct; eauto.
      destruct a0 as [|[]], a; eauto; try contradiction.
      eapply botbind_spec. intros. 2:apply in_z_correct, H3.
      eapply botlift_fun1_spec. intros. eapply toplift_lift_spec; eauto.
      eapply meet_correct. auto.
      eapply botbind_spec. intros. 2:apply in_f_correct, H3.
      eapply botlift_fun1_spec. intros. eapply toplift_lift_spec; eauto.
      eapply meet_correct. auto. auto.
    + intros. apply H0. auto.
    + intros. apply H0. auto.
    + intros. eapply H0; eauto.
  - apply process_msg_correct.
  - intros. apply extract_msgs_correct. auto.
  - intros. simpl. apply extract_msgs_correct. auto.
Qed.

Definition t := @IdealBoxDomain.t abs.

Instance intervals_ideal_env : ab_ideal_env t (t+⊥) :=
  IdealBoxDomain.ideal_box_domain _ itv_dom_correct.
