(* *********************************************************************)
(*                                                                     *)
(*              The Verasco static analyzer                            *)
(*                                                                     *)
(*          Jacques-Henri Jourdan, INRIA Paris-Rocquencourt            *)
(*                                                                     *)
(*  Copyright Institut National de Recherche en Informatique et en     *)
(*  Automatique and Université de Rennes 1.  All rights reserved.      *)
(*  This file is distributed under the terms of the GNU General        *)
(*  Public License as published by the Free Software Foundatio,        *)
(*  either version 2 of the License, or (at your option) any later     *)
(*  version.  This file is also distributed under the terms of the     *)
(*  INRIA Non-Commercial License Agreement.                            *)
(*                                                                     *)
(* *********************************************************************)

Require Import Utf8 Coqlib Util.
Require Import AdomLib.
Require Import IdealExpr IdealIntervals.
Require Import FloatIntervals FloatLib.
Require Import Fcore_Raux.
Require Import Fcore_defs.
Require Import Fcore_generic_fmt.
Require Import Fcore_float_prop.
Require Import Fappli_IEEE.
Require Import Fappli_IEEE_bits.
Require Import Fappli_IEEE_extra.
Require Import Reals Floats.

Transparent Float.neg Float.zero.

Definition r_of_inum (x:ideal_num) : option R :=
  match x with
  | INz z => Some (Z2R z)
  | INf f => if is_finite _ _ f then Some (B2R _ _ f) else None
  end.

Definition RF_le (x:R) (y:float) :=
  if is_finite _ _ y then (x <= B2R _ _ y)%R
  else y = B754_infinity _ _ false.

Local Instance itv_gamma_R : gamma_op itv R :=
  fun itv x =>
    let 'ITV a b := itv in
    RF_le x b /\ RF_le (-x) (Float.neg a).

Local Instance itv_gamma_oR : gamma_op itv (option R) :=
  fun itv x => match x with None => False | Some x => x ∈ γ itv end.

Instance igR_meet_op_correct : meet_op_correct itv (itv+⊥) R.
Proof.
  repeat intro. destruct x, y, H as [[][]]. unfold meet, itv_meet.
  assert (forall f b, is_finite 53 1024 f = true -> Fleb (B754_infinity _ _ b) f = b).
  { intros [] []; try discriminate; reflexivity. }
  assert (forall f b, is_finite 53 1024 f = true -> Fleb f (B754_infinity _ _ b) = negb b).
  { intros [|[]| |[]] []; try discriminate; try reflexivity. }
  assert (forall f, Bopp 53 1024 Float.neg_pl f = B754_infinity 53 1024 false ->
                    f = B754_infinity _ _ true).
  { intros [|[]| |]; try discriminate. auto. }
  unfold RF_le, Float.neg in H, H0, H1, H2.
  repeat match goal with
         | H:if ?FIN then _ else _ |- _ => destruct FIN eqn:?
         end;
  repeat match goal with
         | H:Bopp _ _ _ ?f = B754_infinity _ _ false |- _ => apply H5 in H
         end;
  subst; rewrite is_finite_Bopp in *; rewrite ?H3, ?H4 by auto; simpl;
  repeat match goal with
  | |- context [Fleb ?f1 ?f2] =>
    let EQ := fresh in
    destruct (Fleb f1 f2) eqn:EQ;
    rewrite Fleb_Rle in EQ by auto;
    Rle_bool_case EQ; try discriminate
  end; simpl;
  repeat match goal with
  | |- context [Fleb (B754_infinity _ _ _) ?f] => rewrite (H3 f) by auto
  | |- context [Fleb ?f (B754_infinity _ _ _)] => rewrite (H4 f) by auto; simpl
  end;
  simpl; unfold RF_le, Float.neg;
  rewrite ?is_finite_Bopp, ?B2R_Bopp in * by auto; simpl;
  repeat match goal with
         | H:is_finite _ _ ?f = true |- _ => rewrite H
         end;
  try split; try reflexivity; try Psatz.lra.
Qed.

Instance igoR_meet_op_correct : meet_op_correct itv (itv+⊥) (option R).
Proof.
  repeat intro. destruct a. 2:destruct H as [[]]. apply igR_meet_op_correct in H.
  destruct (x⊓y); auto.
Qed.

Definition itv_of_zitv (i:ZIntervals.itv) : itv :=
  let 'ZIntervals.ITV a b := i in
  ITV (binary_normalize 53 1024 eq_refl eq_refl mode_DN a 0 false)
      (binary_normalize 53 1024 eq_refl eq_refl mode_UP b 0 false).

Lemma itv_of_zitv_correct:
  ∀ i z, z ∈ γ i -> Z2R z ∈ γ (itv_of_zitv i).
Proof.
  intros. destruct i. simpl. unfold RF_le. split.
  - pose proof binary_normalize_correct 53 1024 eq_refl eq_refl mode_UP z1 0 false.
    simpl in H0. destruct H. apply Z2R_le in H. apply Z2R_le in H1.
    Rlt_bool_case H0. 2:Rlt_bool_case H0.
    + destruct H0 as (? & ? & ?). rewrite H3, H0.
      eapply Rle_trans. apply round_UP_pt. apply (fexp_correct 53 1024). reflexivity.
      apply round_le. apply (fexp_correct 53 1024), eq_refl.
      apply (valid_rnd_round_mode mode_UP).
      unfold F2R. simpl. Psatz.lra.
    + destruct binary_normalize as [| |? []|]; inv H0.
      eapply Rlt_le, round_le in H3. rewrite round_0 in H3.
      rewrite Rabs_left1 in H2 by apply H3.
      eapply Rle_trans. apply H1.
      eapply Rle_trans. apply (round_UP_pt radix2), (fexp_correct 53 1024), eq_refl.
      unfold F2R in H2. simpl in H2. rewrite Rmult_1_r in H2.
      apply Ropp_le_cancel. eapply Rle_trans, H2. compute; Psatz.lra.
      apply (valid_rnd_round_mode mode_UP). apply (fexp_correct 53 1024), eq_refl.
      apply (valid_rnd_round_mode mode_UP).
    + destruct binary_normalize as [| |? []|]; inv H0. reflexivity.
  - pose proof binary_normalize_correct 53 1024 eq_refl eq_refl mode_DN z0 0 false.
    simpl in H0. destruct H. apply Z2R_le in H. apply Z2R_le in H1.
    unfold Float.neg. rewrite is_finite_Bopp, B2R_Bopp.
    Rlt_bool_case H0. 2:Rlt_bool_case H0.
    + destruct H0 as (? & ? & ?). rewrite H3, H0.
      apply Ropp_le_contravar. eapply Rle_trans, H.
      unfold F2R. simpl. rewrite Rmult_1_r. apply round_DN_pt.
      apply (fexp_correct 53 1024), eq_refl.
    + destruct binary_normalize as [| |? []|]; inv H0. reflexivity.
    + destruct binary_normalize as [| |? []|]; inv H0.
      eapply round_le in H3. rewrite round_0 in H3.
      rewrite Rabs_right in H2 by apply Rle_ge, H3.
      eapply Ropp_le_contravar, Rle_trans, H.
      unfold F2R in H2. simpl in H2. rewrite Rmult_1_r in H2.
      eapply Rle_trans, round_DN_pt. eapply Rle_trans, H2.
      compute; Psatz.lra.
      apply (fexp_correct 53 1024), eq_refl. apply (valid_rnd_round_mode mode_DN).
      apply (fexp_correct 53 1024), eq_refl. apply (valid_rnd_round_mode mode_DN).
Qed.

Definition itv_of_z (z:Z) : itv :=
  itv_of_zitv (ZIntervals.ITV z z).

Lemma itv_of_z_correct:
  ∀ z, Z2R z ∈ γ (itv_of_z z).
Proof.
  intros. apply itv_of_zitv_correct. simpl. Psatz.lia.
Qed.

Definition itv_of_itvabs (i:IdealIntervals.abs+⊤+⊥) : itv+⊤ :=
  match i with
  | Bot | NotBot All => All
  | NotBot (Just (IdealIntervals.Az i)) => Just (itv_of_zitv i)
  | NotBot (Just (IdealIntervals.Af (ITV a b as i))) =>
    if is_finite _ _ a && is_finite _ _ b then Just i else All
  end.

Lemma itv_of_itvabs_correct:
  ∀ i x, x ∈ γ i -> r_of_inum x ∈ γ (itv_of_itvabs i).
Proof.
  unfold itv_of_itvabs. intros. destruct i as [|[|[|[]]]]; try exact I.
  - destruct x. destruct H. apply itv_of_zitv_correct, H.
  - destruct x. 2:destruct H.
    destruct (is_finite 53 1024 f) eqn:FINf, (is_finite 53 1024 f0) eqn:FINf0;
      try exact I. destruct H.
    assert (FINf1:is_finite _ _ f1 = true).
    { destruct f as [| | |[]], f1 as [|[]| |], f0; try discriminate; reflexivity. }
    simpl. rewrite FINf1. split; unfold RF_le, Float.neg.
    + rewrite FINf0. rewrite Fleb_Rle in H0 by auto. Rle_bool_case H0. auto. discriminate.
    + rewrite is_finite_Bopp, FINf, B2R_Bopp.
      rewrite Fleb_Rle in H by auto. Rle_bool_case H. Psatz.lra. discriminate.
Qed.

Definition itvabs_of_itv (ty:IdealExpr.ideal_num_ty) (i:itv) : IdealIntervals.abs+⊤ :=
  match ty with
  | INTf => Just (Af i)
  | INTz =>
    let 'ITV a b := i in
    match Zoffloat_DN (Float.neg a), Zoffloat_DN b with
    | Some a, Some b => Just (Az (ZIntervals.ITV (-a) b))
    | _, _ => All
    end
  end.

Lemma itvabs_of_itv_correct:
  ∀ i x ty, r_of_inum x ∈ γ i -> x ∈ γ ty -> x ∈ γ (itvabs_of_itv ty i).
Proof.
  destruct 2, i; intros; simpl.
  - pose proof Zoffloat_DN_correct (Float.neg f).
    pose proof Zoffloat_DN_correct f0.
    destruct (Zoffloat_DN (Float.neg f)), (Zoffloat_DN f0), H0, H1; try exact I.
    simpl in H. unfold RF_le, Float.neg in *. rewrite H0, H1, B2R_Bopp in H.
    destruct H. apply Ropp_le_cancel in H4.
    split; apply le_Z2R.
    + apply Ropp_le_cancel. rewrite Z2R_opp, H2, Ropp_involutive, B2R_Bopp.
      apply round_DN_pt. apply Fcore_FIX.FIX_exp_valid.
      apply generic_format_opp. 2:Psatz.lra.
      apply Fcore_FIX.generic_format_FIX. exists (Float radix2 i0 0); unfold F2R; simpl.
      split. ring. auto.
    + rewrite H3. apply round_DN_pt, H. apply Fcore_FIX.FIX_exp_valid.
      apply Fcore_FIX.generic_format_FIX. exists (Float radix2 i0 0); unfold F2R; simpl.
      split. ring. auto.
  - red in H. simpl in H. red in H. destruct (is_finite _ _ f) eqn:FINf. 2:contradiction.
    simpl in H. unfold RF_le, Float.neg in H. rewrite B2R_Bopp, is_finite_Bopp in H.
    destruct H. split.
    + destruct (is_finite _ _ f0) eqn:FINf0. rewrite Fleb_Rle by auto.
      apply Rle_bool_true. Psatz.lra.
      destruct f0 as [|[]| |], f; try discriminate; reflexivity.
    + destruct (is_finite _ _ f1) eqn:FINf1. rewrite Fleb_Rle by auto.
      apply Rle_bool_true. Psatz.lra. subst.
      destruct f as [| | |[]]; try discriminate; reflexivity.
Qed.

Definition opp_itv (x:itv) : itv :=
  match x with
  | ITV m M => ITV (Float.neg M) (Float.neg m)
  end.

Lemma opp_itv_correct:
  ∀ ix x,
    x ∈ γ ix ->
    (-x)%R ∈ γ (opp_itv ix).
Proof.
  intros. unfold opp_itv. destruct ix, H; split. auto. clear H0.
  unfold RF_le, Float.neg in *. rewrite !is_finite_Bopp, !B2R_Bopp, !Ropp_involutive.
  destruct is_finite. auto. subst. auto.
Qed.

Definition add_itv (x y:itv) : itv :=
  match x, y with
  | ITV mx Mx, ITV my My =>
    let m := Bplus 53 1024 eq_refl eq_refl Float.binop_pl mode_DN mx my in
    let m := if is_nan _ _ m then B754_infinity _ _ true else m in
    let M := Bplus 53 1024 eq_refl eq_refl Float.binop_pl mode_UP Mx My in
    let M := if is_nan _ _ M then B754_infinity _ _ false else M in
    ITV m M
  end.

Lemma add_itv_correct:
  ∀ ix iy x y,
    x ∈ γ ix -> y ∈ γ iy ->
    (x+y)%R ∈ γ (add_itv ix iy).
Proof.
  intros [mx Mx] [my My] x y [HMx Hmx] [HMy Hmy]. split; unfold add_itv, RF_le in *.
  - clear Hmx Hmy.
    destruct (is_finite _ _ Mx) eqn:FINx, (is_finite _ _ My) eqn:FINy.
    + pose proof Bplus_correct 53 1024 eq_refl eq_refl Float.binop_pl mode_UP Mx My FINx FINy.
      pose proof @round_UP_pt radix2 _ (fexp_correct 53 1024 eq_refl)
           (B2R 53 1024 Mx + B2R 53 1024 My).
      destruct H0 as (? & ? & ?).
      unfold round_mode in H. Rlt_bool_case H.
      * destruct H as (? & ? & ?). rewrite is_finite_not_is_nan, H4 by auto. Psatz.lra.
      * destruct H, (Bsign 53 1024 Mx) eqn:signMx,
                 (Bplus 53 1024 eq_refl eq_refl Float.binop_pl mode_UP Mx My) as [| |? []|];
        inv H. 2:now auto. simpl.
        apply Rabs_ge_inv in H3. destruct H3.
        apply Rle_trans with (r2:=(- bpow radix2 1024)%R). Psatz.lra. compute. Psatz.lra.
        assert (B2R _ _ Mx <= 0)%R.
        { destruct Mx as [[]|[]|[]|[]]; inv signMx; simpl; try Psatz.lra.
          apply F2R_le_0_compat. compute; discriminate. }
        assert (B2R _ _ My <= 0)%R.
        { destruct My as [[]|[]|[]|[]]; inv H4; simpl; try Psatz.lra.
          apply F2R_le_0_compat. compute; discriminate. }
        specialize (H2 _ (generic_format_0 _ _)). lapply H2. 2:Psatz.lra.
        assert (bpow radix2 1024 <= 0)%R by Psatz.lra.
        compute in H6. Psatz.lra.
    + subst. destruct Mx as [[]|[]|[]|[]]; try discriminate; reflexivity.
    + subst. destruct My as [[]|[]|[]|[]]; try discriminate; reflexivity.
    + subst. reflexivity.
  - clear HMx HMy.
    destruct (is_finite _ _ (Float.neg mx)) eqn:FINx,
             (is_finite _ _ (Float.neg my)) eqn:FINy;
      unfold Float.neg in FINx, FINy; rewrite is_finite_Bopp in FINx, FINy.
    + pose proof Bplus_correct 53 1024 eq_refl eq_refl Float.binop_pl mode_DN mx my FINx FINy.
      pose proof @round_DN_pt radix2 _ (fexp_correct 53 1024 eq_refl)
           (B2R 53 1024 mx + B2R 53 1024 my).
      destruct H0 as (? & ? & ?).
      unfold round_mode in H. Rlt_bool_case H.
      * destruct H as (? & ? & ?). unfold Float.neg.
        rewrite is_finite_Bopp, is_finite_not_is_nan, H4, B2R_Bopp by auto.
        unfold Float.neg in Hmx, Hmy. rewrite B2R_Bopp in Hmx, Hmy. Psatz.lra.
      * destruct H, (Bsign 53 1024 mx) eqn:signmx,
                 (Bplus 53 1024 eq_refl eq_refl Float.binop_pl mode_DN mx my) as [| |? []|];
        inv H. now auto. simpl.
        apply Rabs_ge_inv in H3. destruct H3.
        assert (0 <= B2R _ _ mx)%R.
        { destruct mx as [[]|[]|[]|[]]; inv signmx; simpl; try Psatz.lra.
          apply F2R_ge_0_compat. compute; discriminate. }
        assert (0 <= B2R _ _ my)%R.
        { destruct my as [[]|[]|[]|[]]; inv H4; simpl; try Psatz.lra.
          apply F2R_ge_0_compat. compute; discriminate. }
        specialize (H2 _ (generic_format_0 _ _)). lapply H2. 2:Psatz.lra.
        assert (bpow radix2 1024 <= 0)%R by Psatz.lra.
        compute in H6. Psatz.lra.
        apply Rle_trans with (r2:=(-bpow radix2 1024)%R).
        unfold Float.neg in Hmx, Hmy. rewrite B2R_Bopp in Hmx, Hmy. Psatz.lra.
        compute. Psatz.lra.
    + destruct mx as [[]|[]|[]|[]], my as [[]|[]|[]|[]]; try discriminate; reflexivity.
    + destruct mx as [[]|[]|[]|[]], my as [[]|[]|[]|[]]; try discriminate; reflexivity.
    + destruct mx as [[]|[]|[]|[]], my as [[]|[]|[]|[]]; try discriminate; reflexivity.
Qed.

Definition mult_itv (x y:itv) : itv :=
  match x, y with
  | ITV mx Mx, ITV my My =>
    let m1 :=
      if Bsign _ _ mx then Bmult 53 1024 eq_refl eq_refl Float.binop_pl mode_DN mx My
      else Bmult 53 1024 eq_refl eq_refl Float.binop_pl mode_DN mx my
    in
    let m1 := if is_nan _ _ m1 then Float.zero else m1 in
    let m2 :=
      if Bsign _ _ Mx then Bmult 53 1024 eq_refl eq_refl Float.binop_pl mode_DN Mx My
      else Bmult 53 1024 eq_refl eq_refl Float.binop_pl mode_DN Mx my
    in
    let m2 := if is_nan _ _ m2 then Float.zero else m2 in
    let m := if Fleb m1 m2 then m1 else m2 in
    let M1 :=
      if Bsign _ _ mx then Bmult 53 1024 eq_refl eq_refl Float.binop_pl mode_UP mx my
      else Bmult 53 1024 eq_refl eq_refl Float.binop_pl mode_UP mx My
    in
    let M1 := if is_nan _ _ M1 then Float.zero else M1 in
    let M2 :=
      if Bsign _ _ Mx then Bmult 53 1024 eq_refl eq_refl Float.binop_pl mode_UP Mx my
      else Bmult 53 1024 eq_refl eq_refl Float.binop_pl mode_UP Mx My
    in
    let M2 := if is_nan _ _ M2 then Float.zero else M2 in
    let M := if Fleb M1 M2 then M2 else M1 in
    ITV m M
  end.

Lemma mult_itv_correct:
  ∀ ix iy x y,
    x ∈ γ ix -> y ∈ γ iy ->
    (x*y)%R ∈ γ (mult_itv ix iy).
Proof.
  intros [mx Mx] [my My] x y [HMx Hmx] [HMy Hmy]. unfold mult_itv.
  set (m1 :=
      if Bsign _ _ mx then Bmult 53 1024 eq_refl eq_refl Float.binop_pl mode_DN mx My
      else Bmult 53 1024 eq_refl eq_refl Float.binop_pl mode_DN mx my).
  set (m1':=if is_nan _ _ m1 then Float.zero else m1).
  set (m2 :=
      if Bsign _ _ Mx then Bmult 53 1024 eq_refl eq_refl Float.binop_pl mode_DN Mx My
      else Bmult 53 1024 eq_refl eq_refl Float.binop_pl mode_DN Mx my).
  set (m2':=if is_nan _ _ m2 then Float.zero else m2).
  set (M1 :=
      if Bsign _ _ mx then Bmult 53 1024 eq_refl eq_refl Float.binop_pl mode_UP mx my
      else Bmult 53 1024 eq_refl eq_refl Float.binop_pl mode_UP mx My).
  set (M1':=if is_nan _ _ M1 then Float.zero else M1).
  set (M2 :=
      if Bsign _ _ Mx then Bmult 53 1024 eq_refl eq_refl Float.binop_pl mode_UP Mx my
      else Bmult 53 1024 eq_refl eq_refl Float.binop_pl mode_UP Mx My).
  set (M2':=if is_nan _ _ M2 then Float.zero else M2).
  unfold mult_itv, Float.neg, RF_le in *.
  rewrite is_finite_Bopp, B2R_Bopp in Hmx, Hmy.
  split; unfold mult_itv, Float.neg, RF_le in *; destruct (Rle_lt_dec 0 y) as [ySgn|ySgn].
  - assert (RF_le (x*y) M2').
    { subst M1 M1' M2 M2' m1 m1' m2 m2'. clear mx Hmx.
      destruct (is_finite _ _ Mx) eqn:FINMx.
      - assert (x*y <= B2R _ _ Mx*y)%R by (apply Rmult_le_compat_r; auto).
        destruct (Bsign _ _ Mx) eqn:SGNMx.
        + pose proof Bsign_true_le_0 _ _ _ SGNMx.
          destruct (is_finite _ _ my) eqn:FINmy.
          * assert (x*y <= B2R _ _ Mx * B2R _ _ my)%R.
            { eapply Rle_trans. eauto. apply Rmult_le_compat_neg_l; Psatz.lra. }
            clear HMx My HMy Hmy ySgn H.
            { pose proof Bmult_correct 53 1024 eq_refl eq_refl Float.binop_pl mode_UP Mx my.
              rewrite FINMx, FINmy, SGNMx in H. Rlt_bool_case H.
              - destruct H as (? & ? & _).
                unfold RF_le. rewrite is_finite_not_is_nan, H3, H by auto.
                eapply Rle_trans. eauto. apply round_UP_pt. apply fexp_correct. reflexivity.
              - destruct (Bsign _ _ my) eqn:SGNmy, Bmult as [| |? []|]; inv H.
                reflexivity.
                eapply Rle_trans. eauto. clear x y H1. apply Bsign_false_ge_0 in SGNmy.
                assert (B2R 53 1024 Mx * B2R 53 1024 my <= 0)%R.
                { apply Ropp_le_cancel. rewrite Ropp_0, <- Ropp_mult_distr_l_reverse.
                  apply Rmult_le_pos; Psatz.lra. }
                apply Rabs_ge_inv in H2. destruct H2.
                + eapply Rle_trans. apply round_UP_pt with (beta:=radix2).
                  apply (fexp_correct 53). reflexivity.
                  eapply Rle_trans. apply H1. compute; Psatz.lra.
                + eapply Rle_trans with (r3:=0%R) in H1. compute in H1. Psatz.lra.
                  eapply Rle_trans. apply round_UP_pt with (beta:=radix2).
                  apply fexp_correct. reflexivity. apply generic_format_0. auto.
                  Psatz.lra. }
          * destruct Mx as [[]| | |[]], my as [|[]| |]; try discriminate.
            unfold RF_le. simpl in *. Psatz.lra. reflexivity.
        + pose proof Bsign_false_ge_0 _ _ _ SGNMx.
          destruct (is_finite _ _ My) eqn:FINMy.
          * assert (x*y <= B2R _ _ Mx * B2R _ _ My)%R.
            { eapply Rle_trans, Ropp_le_cancel. eauto.
              rewrite <- !Ropp_mult_distr_l_reverse.
              apply Rmult_le_compat_neg_l; Psatz.lra. }
            clear HMx my HMy Hmy ySgn H.
            { pose proof Bmult_correct 53 1024 eq_refl eq_refl Float.binop_pl mode_UP Mx My.
              rewrite FINMx, FINMy, SGNMx in H. Rlt_bool_case H.
              - destruct H as (? & ? & _).
                unfold RF_le. rewrite is_finite_not_is_nan, H3, H by auto.
                eapply Rle_trans. eauto. apply round_UP_pt. apply fexp_correct. reflexivity.
              - destruct (Bsign _ _ My) eqn:SGNMy, Bmult as [| |? []|]; inv H.
                2:reflexivity.
                eapply Rle_trans. eauto. clear x y H1. apply Bsign_true_le_0 in SGNMy.
                assert (B2R 53 1024 Mx * B2R 53 1024 My <= 0)%R.
                { apply Ropp_le_cancel. rewrite Ropp_0, <- Ropp_mult_distr_r_reverse.
                  apply Rmult_le_pos; Psatz.lra. }
                apply Rabs_ge_inv in H2. destruct H2.
                + eapply Rle_trans. apply round_UP_pt with (beta:=radix2).
                  apply (fexp_correct 53). reflexivity.
                  eapply Rle_trans. apply H1. compute; Psatz.lra.
                + eapply Rle_trans with (r3:=0%R) in H1. compute in H1. Psatz.lra.
                  eapply Rle_trans. apply round_UP_pt with (beta:=radix2).
                  apply fexp_correct. reflexivity. apply generic_format_0. auto.
                  Psatz.lra. }
          * destruct Mx as [[]| | |[]], My as [|[]| |]; try discriminate.
            unfold RF_le. simpl in *. Psatz.lra. reflexivity.
      - clear FINMx. subst Mx. destruct My as [|[]| |]; try discriminate.
        simpl in *. replace y with 0%R by Psatz.lra. rewrite Rmult_0_r. apply Rle_refl.
        reflexivity.
        simpl in HMy. eapply Rle_trans in HMy.
        apply F2R_ge_0_reg in HMy. destruct b.
        destruct HMy. auto. reflexivity. auto. }
    clear - H. destruct (Fleb M1' M2') eqn:EQFleb. apply H.
    unfold RF_le in H.
    destruct (is_finite _ _ M1') eqn:FINM1', (is_finite _ _ M2') eqn:FINM2'.
    + rewrite Fleb_Rle in EQFleb by auto. Rle_bool_case EQFleb. discriminate. Psatz.lra.
    + rewrite H in EQFleb. destruct M1 as [| | |[]]; discriminate.
    + destruct M1 as [|[]| |]; try discriminate. 2:now auto.
      destruct M2'; discriminate.
    + destruct M1 as [|[]| |]; try discriminate. 2:now auto.
      destruct M2' as [|[]| |]; discriminate.
  - assert (RF_le (x*y) M1').
    { subst M1 M1' M2 M2' m1 m1' m2 m2'. clear Mx HMx.
      destruct (is_finite _ _ mx) eqn:FINmx.
      - assert (x*y <= B2R _ _ mx*y)%R.
        { apply Ropp_le_cancel. rewrite <- !Ropp_mult_distr_r_reverse.
          apply Rmult_le_compat_r; Psatz.lra. }
        destruct (Bsign _ _ mx) eqn:SGNmx.
        + pose proof Bsign_true_le_0 _ _ _ SGNmx.
          destruct (is_finite _ _ my) eqn:FINmy.
          * assert (x*y <= B2R _ _ mx * B2R _ _ my)%R.
            { eapply Rle_trans, Ropp_le_cancel. eauto.
              rewrite <- !Ropp_mult_distr_l_reverse.
              apply Rmult_le_compat_l; Psatz.lra. }
            clear Hmx My HMy Hmy ySgn H.
            { pose proof Bmult_correct 53 1024 eq_refl eq_refl Float.binop_pl mode_UP mx my.
              rewrite FINmx, FINmy, SGNmx in H. Rlt_bool_case H.
              - destruct H as (? & ? & _).
                unfold RF_le. rewrite is_finite_not_is_nan, H3, H by auto.
                eapply Rle_trans. eauto. apply round_UP_pt. apply fexp_correct. reflexivity.
              - destruct (Bsign _ _ my) eqn:SGNmy, Bmult as [| |? []|]; inv H.
                reflexivity.
                eapply Rle_trans. eauto. clear x y H1. apply Bsign_false_ge_0 in SGNmy.
                assert (B2R 53 1024 mx * B2R 53 1024 my <= 0)%R.
                { apply Ropp_le_cancel. rewrite Ropp_0, <- Ropp_mult_distr_l_reverse.
                  apply Rmult_le_pos; Psatz.lra. }
                apply Rabs_ge_inv in H2. destruct H2.
                + eapply Rle_trans. apply round_UP_pt with (beta:=radix2).
                  apply (fexp_correct 53). reflexivity.
                  eapply Rle_trans. apply H1. compute; Psatz.lra.
                + eapply Rle_trans with (r3:=0%R) in H1. compute in H1. Psatz.lra.
                  eapply Rle_trans. apply round_UP_pt with (beta:=radix2).
                  apply fexp_correct. reflexivity. apply generic_format_0. auto.
                  Psatz.lra. }
          * destruct mx as [[]| | |[]], my as [|[]| |]; try discriminate.
            unfold RF_le. simpl in *. Psatz.lra. reflexivity.
        + pose proof Bsign_false_ge_0 _ _ _ SGNmx.
          destruct (is_finite _ _ My) eqn:FINMy.
          * assert (x*y <= B2R _ _ mx * B2R _ _ My)%R.
            { eapply Rle_trans, Ropp_le_cancel. eauto.
              rewrite <- !Ropp_mult_distr_r_reverse.
              apply Rmult_le_compat_l; Psatz.lra. }
            clear Hmx my HMy Hmy ySgn H.
            { pose proof Bmult_correct 53 1024 eq_refl eq_refl Float.binop_pl mode_UP mx My.
              rewrite FINmx, FINMy, SGNmx in H. Rlt_bool_case H.
              - destruct H as (? & ? & _).
                unfold RF_le. rewrite is_finite_not_is_nan, H3, H by auto.
                eapply Rle_trans. eauto. apply round_UP_pt. apply fexp_correct. reflexivity.
              - destruct (Bsign _ _ My) eqn:SGNMy, Bmult as [| |? []|]; inv H.
                2:reflexivity.
                eapply Rle_trans. eauto. clear x y H1. apply Bsign_true_le_0 in SGNMy.
                assert (B2R 53 1024 mx * B2R 53 1024 My <= 0)%R.
                { apply Ropp_le_cancel. rewrite Ropp_0, <- Ropp_mult_distr_r_reverse.
                  apply Rmult_le_pos; Psatz.lra. }
                apply Rabs_ge_inv in H2. destruct H2.
                + eapply Rle_trans. apply round_UP_pt with (beta:=radix2).
                  apply (fexp_correct 53). reflexivity.
                  eapply Rle_trans. apply H1. compute; Psatz.lra.
                + eapply Rle_trans with (r3:=0%R) in H1. compute in H1. Psatz.lra.
                  eapply Rle_trans. apply round_UP_pt with (beta:=radix2).
                  apply fexp_correct. reflexivity. apply generic_format_0. auto.
                  Psatz.lra. }
          * destruct mx as [[]| | |[]], My as [|[]| |]; try discriminate.
            unfold RF_le. simpl in *. Psatz.lra. reflexivity.
      - clear FINmx. destruct mx as [|[]| |]; inv Hmx.
        destruct my as [|[]| |]; try discriminate. simpl in *. Psatz.lra. reflexivity.
        simpl in Hmy. eapply Rle_trans in Hmy.
        apply Ropp_le_cancel, F2R_le_0_reg in Hmy. destruct b.
        reflexivity. destruct Hmy. auto. Psatz.lra. }
    clear - H. destruct (Fleb M1' M2') eqn:EQFleb. 2:apply H.
    unfold RF_le in H.
    destruct (is_finite _ _ M1') eqn:FINM1', (is_finite _ _ M2') eqn:FINM2'.
    + rewrite Fleb_Rle in EQFleb by auto. Rle_bool_case EQFleb. Psatz.lra. discriminate.
    + destruct M1' as [| | |[]], M2' as [|[]| |]; try discriminate; reflexivity.
    + rewrite H in EQFleb. destruct M2'; discriminate.
    + rewrite H in EQFleb. destruct M2' as [|[]| |]; try discriminate. auto.
  - assert (RF_le (-(x*y)) (Bopp _ _ Float.neg_pl m1')).
    { subst M1 M1' M2 M2' m1 m1' m2 m2'. clear Mx HMx.
      destruct (is_finite _ _ mx) eqn:FINmx.
      - assert (B2R _ _ mx*y <= x*y)%R by (apply Rmult_le_compat_r; Psatz.lra).
        destruct (Bsign _ _ mx) eqn:SGNmx.
        + pose proof Bsign_true_le_0 _ _ _ SGNmx.
          destruct (is_finite _ _ My) eqn:FINMy.
          * assert (B2R _ _ mx * B2R _ _ My <= x*y)%R.
            { eapply Rle_trans, H.
              apply Ropp_le_cancel. rewrite <- !Ropp_mult_distr_l_reverse.
              apply Rmult_le_compat_l; Psatz.lra. }
            clear Hmx my Hmy HMy ySgn H.
            { pose proof Bmult_correct 53 1024 eq_refl eq_refl Float.binop_pl mode_DN mx My.
              rewrite FINmx, FINMy, SGNmx in H. Rlt_bool_case H.
              - destruct H as (? & ? & _).
                unfold RF_le.
                rewrite is_finite_not_is_nan, is_finite_Bopp, H3, B2R_Bopp, H by auto.
                eapply Ropp_le_contravar, Rle_trans. 2:eauto.
                apply round_DN_pt. apply fexp_correct. reflexivity.
              - destruct (Bsign _ _ My) eqn:SGNMy, Bmult as [| |? []|]; inv H.
                2:reflexivity.
                eapply Rle_trans. apply Ropp_le_contravar, H1.
                clear x y H1. apply Bsign_true_le_0 in SGNMy.
                rewrite B2R_Bopp. apply Ropp_le_contravar.
                assert (0 <= B2R 53 1024 mx * B2R 53 1024 My)%R.
                { rewrite <- Rmult_opp_opp. apply Rmult_le_pos; Psatz.lra. }
                apply Rabs_ge_inv in H2. destruct H2.
                + eapply Rle_trans with (r1:=0%R) in H1. compute in H1. Psatz.lra.
                  eapply Rle_trans. 2:apply round_DN_pt with (beta:=radix2).
                  2:apply fexp_correct. 2:reflexivity. 2:apply generic_format_0.
                  Psatz.lra. auto.
                + eapply Rle_trans. 2:apply round_DN_pt with (beta:=radix2).
                  2:apply (fexp_correct 53). 2:reflexivity.
                  eapply Rle_trans. 2:apply H1. compute; Psatz.lra. }
          * destruct mx as [[]| | |[]], My as [|[]| |]; try discriminate.
            unfold RF_le. simpl in *. Psatz.lra. reflexivity.
        + pose proof Bsign_false_ge_0 _ _ _ SGNmx.
          destruct (is_finite _ _ my) eqn:FINmy.
          * assert (B2R _ _ mx * B2R _ _ my <= x*y)%R.
            { eapply Rle_trans, H. apply Rmult_le_compat_l; Psatz.lra. }
            clear Hmx My Hmy HMy ySgn H.
            { pose proof Bmult_correct 53 1024 eq_refl eq_refl Float.binop_pl mode_DN mx my.
              rewrite FINmx, FINmy, SGNmx in H. Rlt_bool_case H.
              - destruct H as (? & ? & _).
                unfold RF_le.
                rewrite is_finite_not_is_nan, is_finite_Bopp, B2R_Bopp, H3, H by auto.
                eapply Ropp_le_contravar, Rle_trans. 2:eauto.
                apply round_DN_pt. apply fexp_correct. reflexivity.
              - destruct (Bsign _ _ my) eqn:SGNmy, Bmult as [| |? []|]; inv H.
                reflexivity.
                eapply Rle_trans. eapply Ropp_le_contravar. eauto.
                clear x y H1. apply Bsign_false_ge_0 in SGNmy.
                assert (0 <= B2R 53 1024 mx * B2R 53 1024 my)%R.
                { apply Rmult_le_pos; Psatz.lra. }
                apply Rabs_ge_inv in H2. destruct H2.
                + eapply Rle_trans with (r1:=0%R) in H1. compute in H1. Psatz.lra.
                  eapply Rle_trans. 2:apply round_DN_pt with (beta:=radix2).
                  2:apply fexp_correct. 2:reflexivity. 2:apply generic_format_0.
                  Psatz.lra. auto.
                + rewrite B2R_Bopp. apply Ropp_le_contravar.
                  eapply Rle_trans. 2:apply round_DN_pt with (beta:=radix2).
                  2:apply (fexp_correct 53). 2:reflexivity.
                  eapply Rle_trans. 2:apply H1. compute; Psatz.lra. }
          * destruct mx as [[]| | |[]], my as [|[]| |]; try discriminate.
            unfold RF_le. simpl in *. Psatz.lra. reflexivity.
      - clear FINmx. destruct mx as [|[]| |], My as [|[]| |]; try discriminate.
        simpl in *. replace y with 0%R by Psatz.lra.
        rewrite Rmult_0_r, Ropp_0. apply Rle_refl. reflexivity.
        simpl in HMy. eapply Rle_trans in HMy.
        apply F2R_ge_0_reg in HMy. destruct b.
        destruct HMy. auto. reflexivity. auto. }
    clear - H. destruct (Fleb m1' m2') eqn:EQFleb. apply H.
    unfold RF_le in H. rewrite is_finite_Bopp, B2R_Bopp in *.
    destruct (is_finite _ _ m1') eqn:FINm1', (is_finite _ _ m2') eqn:FINm2'.
    + rewrite Fleb_Rle in EQFleb by auto. Rle_bool_case EQFleb. discriminate. Psatz.lra.
    + destruct m2 as [|[]| |]; try discriminate. now auto.
      destruct m1' as [| | |[]]; discriminate.
    + destruct m1' as [|[]| |]; try discriminate. destruct m2; discriminate.
    + destruct m2 as [|[]| |]; try discriminate. now auto.
      destruct m1' as [|[]| |]; discriminate.
  - assert (RF_le (-(x*y)) (Bopp _ _ Float.neg_pl m2')).
    { subst M1 M1' M2 M2' m1 m1' m2 m2'. clear mx Hmx.
      destruct (is_finite _ _ Mx) eqn:FINMx.
      - assert (B2R _ _ Mx*y <= x*y)%R.
        { apply Ropp_le_cancel. rewrite <- !Ropp_mult_distr_r_reverse.
          apply Rmult_le_compat_r; Psatz.lra. }
        destruct (Bsign _ _ Mx) eqn:SGNMx.
        + pose proof Bsign_true_le_0 _ _ _ SGNMx.
          destruct (is_finite _ _ My) eqn:FINMy.
          * assert (B2R _ _ Mx * B2R _ _ My <= x*y)%R.
            { eapply Rle_trans, H. eapply Ropp_le_cancel.
              rewrite <- !Ropp_mult_distr_l_reverse.
              apply Rmult_le_compat_l; Psatz.lra. }
            clear HMx my Hmy HMy ySgn H.
            { pose proof Bmult_correct 53 1024 eq_refl eq_refl Float.binop_pl mode_DN Mx My.
              rewrite FINMx, FINMy, SGNMx in H. Rlt_bool_case H.
              - destruct H as (? & ? & _).
                unfold RF_le.
                rewrite is_finite_not_is_nan, is_finite_Bopp, H3, B2R_Bopp, H by auto.
                eapply Ropp_le_contravar, Rle_trans. 2:eauto.
                apply round_DN_pt. apply fexp_correct. reflexivity.
              - destruct (Bsign _ _ My) eqn:SGNMy, Bmult as [| |? []|]; inv H.
                2:reflexivity.
                eapply Rle_trans. apply Ropp_le_contravar, H1.
                clear x y H1. apply Bsign_true_le_0 in SGNMy.
                rewrite B2R_Bopp. apply Ropp_le_contravar.
                assert (0 <= B2R 53 1024 Mx * B2R 53 1024 My)%R.
                { rewrite <- Rmult_opp_opp. apply Rmult_le_pos; Psatz.lra. }
                apply Rabs_ge_inv in H2. destruct H2.
                + eapply Rle_trans with (r1:=0%R) in H1. compute in H1. Psatz.lra.
                  eapply Rle_trans. 2:apply round_DN_pt with (beta:=radix2).
                  2:apply fexp_correct. 2:reflexivity. 2:apply generic_format_0.
                  Psatz.lra. auto.
                + eapply Rle_trans. 2:apply round_DN_pt with (beta:=radix2).
                  2:apply (fexp_correct 53). 2:reflexivity.
                  eapply Rle_trans. 2:apply H1. compute; Psatz.lra. }
          * destruct Mx as [[]| | |[]], My as [|[]| |]; try discriminate.
            unfold RF_le. simpl in *. Psatz.lra. reflexivity.
        + pose proof Bsign_false_ge_0 _ _ _ SGNMx.
          destruct (is_finite _ _ my) eqn:FINmy.
          * assert (B2R _ _ Mx * B2R _ _ my <= x*y)%R.
            { eapply Rle_trans, H. eapply Ropp_le_cancel.
              rewrite <- !Ropp_mult_distr_r_reverse.
              apply Rmult_le_compat_l; Psatz.lra. }
            clear HMx My Hmy HMy ySgn H.
            { pose proof Bmult_correct 53 1024 eq_refl eq_refl Float.binop_pl mode_DN Mx my.
              rewrite FINMx, FINmy, SGNMx in H. Rlt_bool_case H.
              - destruct H as (? & ? & _).
                unfold RF_le.
                rewrite is_finite_not_is_nan, is_finite_Bopp, B2R_Bopp, H3, H by auto.
                eapply Ropp_le_contravar, Rle_trans. 2:eauto.
                apply round_DN_pt. apply fexp_correct. reflexivity.
              - destruct (Bsign _ _ my) eqn:SGNmy, Bmult as [| |? []|]; inv H.
                reflexivity.
                eapply Rle_trans. eapply Ropp_le_contravar. eauto.
                clear x y H1. apply Bsign_false_ge_0 in SGNmy.
                assert (0 <= B2R 53 1024 Mx * B2R 53 1024 my)%R.
                { apply Rmult_le_pos; Psatz.lra. }
                apply Rabs_ge_inv in H2. destruct H2.
                + eapply Rle_trans with (r1:=0%R) in H1. compute in H1. Psatz.lra.
                  eapply Rle_trans. 2:apply round_DN_pt with (beta:=radix2).
                  2:apply fexp_correct. 2:reflexivity. 2:apply generic_format_0.
                  Psatz.lra. auto.
                + rewrite B2R_Bopp. apply Ropp_le_contravar.
                  eapply Rle_trans. 2:apply round_DN_pt with (beta:=radix2).
                  2:apply (fexp_correct 53). 2:reflexivity.
                  eapply Rle_trans. 2:apply H1. compute; Psatz.lra. }
          * destruct Mx as [[]| | |[]], my as [|[]| |]; try discriminate.
            unfold RF_le. simpl in *. Psatz.lra. reflexivity.
      - clear FINMx. subst Mx. destruct my as [|[]| |]; try discriminate.
        simpl in *. Psatz.lra. reflexivity.
        simpl in Hmy. apply Ropp_le_cancel in Hmy. eapply Rle_lt_trans in Hmy.
        apply F2R_lt_0_reg in Hmy. destruct b.
        reflexivity. discriminate. auto. }
    clear - H. destruct (Fleb m1' m2') eqn:EQFleb. 2:apply H.
    unfold RF_le in H. rewrite is_finite_Bopp, B2R_Bopp in *.
    destruct (is_finite _ _ m1') eqn:FINm1', (is_finite _ _ m2') eqn:FINm2'.
    + rewrite Fleb_Rle in EQFleb by auto. Rle_bool_case EQFleb. Psatz.lra. discriminate.
    + destruct m2 as [|[]| |]; try discriminate. destruct m1 as [| | |[]]; discriminate.
    + destruct m1' as [|[]| |]; try discriminate. reflexivity. destruct m2; discriminate.
    + destruct m1 as [|[]| |]; try discriminate. now auto.
      destruct m2' as [|[]| |]; discriminate.
Qed.

Definition itv_is_top (i:itv) : bool :=
  match i with
  | ITV (B754_infinity true) (B754_infinity false) => true
  | _ => false
  end.

Definition rρ (ρ : var -> ideal_num) (x:var) : option R := r_of_inum (ρ x).

Definition linear_expr := (itv * list (var * (itv * itv)))%type.

Definition eval_linear_expr (ρ:var -> ideal_num) (e:linear_expr): ℘(R) :=
  List.fold_right
    (fun (kitvs:var*(itv*itv)) X =>
       let '(k, (itvα, itvk)) := kitvs in
       fun y =>
         match rρ ρ k with
         | None => False
         | Some kx => (∃ α, α ∈ γ itvα /\ X (y - α * kx)) /\
                      rρ ρ k ∈ γ itvk
         end)%R
    (γ (fst e)) (snd e).

Definition T_linear_expr := (itv * T.t (itv * itv))%type.

Definition eval_T_linear_expr ρ (e:T_linear_expr) : ℘(R) :=
  eval_linear_expr ρ (fst e, T.elements (snd e)).

Definition T_linear_expr_is_top (le:T_linear_expr) : bool :=
  T.fold (fun acc x itvs => acc || itv_is_top (fst itvs))
         (snd le) false.
