(* *********************************************************************)
(*                                                                     *)
(*              The Verasco static analyzer                            *)
(*                                                                     *)
(*          Vincent Laporte, Université de Rennes 1                    *)
(*                                                                     *)
(*  Copyright Institut National de Recherche en Informatique et en     *)
(*  Automatique and Université de Rennes 1.  All rights reserved.      *)
(*  This file is distributed under the terms of the GNU General        *)
(*  Public License as published by the Free Software Foundatio,        *)
(*  either version 2 of the License, or (at your option) any later     *)
(*  version.  This file is also distributed under the terms of the     *)
(*  INRIA Non-Commercial License Agreement.                            *)
(*                                                                     *)
(* *********************************************************************)

Require Import Coqlib Utf8.
Require Import Maps ShareTree Util.
Require Import Integers.
Require Import Floats FloatLib.
Require Import Values.
Require Import AST.
Require Import Globalenvs.
Require Import Cminor.
Require Import Csharpminor.
Require Import Memory.
Require Import AdomLib.
Require Import Util.
Require Import PExpr NoError.
Require Import TypesDomain.
Require Import AbMachineEnv.
Require Import String ToString.

Local Unset Elimination Schemes.

Local Open Scope string_scope.

Module Nconvert(T:SHARETREE).

  Section nconvert.
    Variable ge: genv.
    Variable ε: ident → option block.
    Variable var : Type.
    Variable m: Mem.mem.
    Variable ρC: T.elt -> val.
    Variable ρ: var -> num_val.

    Variable elt2p: T.elt -> var.

    Definition convert_constant (cst: constant) : nexpr var :=
      match cst with
        | Ointconst n => NEconst _ (NVint n)
        | Olongconst n => NEconst _ (NVint64 n)
        | Ofloatconst f => NEconst _ (NVfloat f)
        | Osingleconst f => NEconst _ (NVsingle f)
        | Oaddrsymbol _ ofs => NEconst _ (NVint ofs)
      end.

    Inductive nconvert_msg_kind : Set :=
    | MSG_CONST_NONE
    | MSG_ILL_TYPED_UNOP
    | MSG_ILL_TYPED_BINOP
    .

    Definition nconvert_msg_type (k: nconvert_msg_kind) : Type :=
      match k with
      | MSG_CONST_NONE
        => string
      | MSG_ILL_TYPED_UNOP => Cminor.unary_operation → AbTy.t → string
      | MSG_ILL_TYPED_BINOP => Cminor.binary_operation → AbTy.t → AbTy.t → string
      end.

    Definition nconvert_msg (k: nconvert_msg_kind) : nconvert_msg_type k :=
      match k return nconvert_msg_type k with
      | MSG_CONST_NONE => "const None"
      | MSG_ILL_TYPED_UNOP => λ op ty, "ill-typed op: " ++ to_string op ++ to_string ty
      | MSG_ILL_TYPED_BINOP => λ op ty₁ ty₂, "ill-typed op: " ++ to_string ty₁ ++ to_string op ++ to_string ty₂
      end.
    Opaque nconvert_msg.

    Fixpoint nconvert (e:pexpr T.elt) : nexpr var + string :=
      match e with
        | PEvar c ty =>
          match ty with
            | VI|VP|VIP => inl (NEvar NVTint (elt2p c))
            | VL => inl (NEvar NVTint64 (elt2p c))
            | VF => inl (NEvar NVTfloat (elt2p c))
            | VS => inl (NEvar NVTsingle (elt2p c))
          end
        | PElocal _ => inl (NEconst _ (NVint Int.zero))
        | PEconst (Some cst) _ => inl (convert_constant cst)
        | PEconst None _ => inr (nconvert_msg MSG_CONST_NONE)
        | PEunop op e _ =>
          match pexpr_get_ty e, op with
            | VI,
              (Ocast8unsigned | Ocast8signed | Ocast16unsigned | Ocast16signed
               | Onegint | Onotint | Ofloatofint | Ofloatofintu
               | Osingleofint | Osingleofintu | Olongofint | Olongofintu )
            | VL,
              (Onegl | Onotl | Ointoflong | Ofloatoflong | Ofloatoflongu
               | Osingleoflong | Osingleoflongu)
            | VF,
              (Onegf | Oabsf | Osingleoffloat | Ointoffloat | Ointuoffloat
               | Olongoffloat | Olonguoffloat)
            | VS,
              (Onegfs | Oabsfs | Ofloatofsingle | Ointofsingle | Ointuofsingle
               | Olongofsingle | Olonguofsingle) =>
              match nconvert e with
                | inl e' => inl (NEunop op e')
                | q => q
              end
            | ty, op => inr (nconvert_msg MSG_ILL_TYPED_UNOP op ty)
          end
        | PEbinop op e1 e2 _ =>
          match pexpr_get_ty e1, pexpr_get_ty e2, op with
            | VI, VP, Oadd | VP, VI, Oadd
            | VP, VP, Osub
            | VP, VI, Osub
            | VI, VI, Ocmpu _
            | VP, VP, Ocmpu _
            | VI, VI,
              (Oadd | Osub | Omul | Odiv | Odivu | Omod | Omodu | Oand
               | Oor | Oxor | Oshl | Oshr | Oshru | Ocmp _)
            | VL, VL,
              (Oaddl | Osubl | Omull | Odivl | Odivlu | Omodl | Omodlu
               | Oandl | Oorl | Oxorl | Ocmpl _ | Ocmplu _)
            | VL, VI, (Oshll | Oshrl | Oshrlu)
            | VF, VF, (Oaddf | Osubf | Omulf | Odivf | Ocmpf _)
            | VS, VS, (Oaddfs | Osubfs | Omulfs | Odivfs | Ocmpfs _) =>
              match nconvert e1, nconvert e2 with
                | inl e1', inl e2' => inl (NEbinop op e1' e2')
                | inr q, _ => inr q
                | _, q => q
              end
            | ty₁, ty₂, op => inr (nconvert_msg MSG_ILL_TYPED_BINOP op ty₁ ty₂)
          end
      end.

    Lemma bind_inl {A B} (f: A → A + B) (ab: A + B) (a': A):
      match ab with inl a' => f a' | inr b => inr b end = inl a' →
      ∃ a : A, ab = inl a ∧ f a = inl a'.
    Proof.
      destruct ab as [ a | b ]. vauto. intro; eq_some_inv.
    Qed.

    Local Ltac ncinv :=
      try match goal with H : _ = Some _ |- _ => apply do_opt_some_inv in H; hsplit end;
      repeat match goal with H : _ = inl _ |- _ => apply bind_inl in H; hsplit end;
      eq_some_inv; subst; simpl;
      repeat match goal with H : _, K : _ → _ |- _ => specialize (K _ H) end.

    Lemma nconvert_fv pe :
      ∀ ne, nconvert pe = inl ne →
      ∀ x, x ∈ ne_free_var ne → ∃ y, y ∈ pe_free_var pe ∧ x = elt2p y.
    Proof.
      induction pe; simpl; intros ne Hne cc Hcc.
      - destruct t; eq_some_inv; subst ne; simpl in Hcc; subst cc; vauto.
      - eq_some_inv; subst ne. destruct Hcc.
      - destruct o as [[]|]; eq_some_inv; subst ne; simpl in *; destruct Hcc.
      - destruct (pexpr_get_ty pe); eq_some_inv;
        destruct u; eq_some_inv; ncinv; eq_some_inv; auto.
      - destruct (pexpr_get_ty pe1); eq_some_inv;
        destruct (pexpr_get_ty pe2); eq_some_inv;
        destruct b; eq_some_inv;
        ncinv; simpl in Hcc;
        destruct Hcc; ncinv; hsplit; vauto.
    Qed.

    Definition ncompat : val -> num_val -> Prop :=
      λ v, match v with
             | Vint i | Vptr _ i => eq (NVint i)
             | Vlong l => eq (NVint64 l)
             | Vsingle f => eq (NVsingle f)
             | Vfloat f => eq (NVfloat f)
             | Vundef => λ _, True
           end.
    Lemma ncompat_inv_int v i :
      ncompat v (NVint i) →
      v = Vint i ∨ (∃ b, v = Vptr b i) ∨ v = Vundef.
    Proof.
      destruct v; inversion 1; vauto.
    Qed.

    Lemma ncompat_inv_int64 v i :
      ncompat v (NVint64 i) →
      v = Vlong i ∨ v = Vundef.
    Proof.
      destruct v; inversion 1; vauto.
    Qed.

    Lemma ncompat_inv_float v f :
      ncompat v (NVfloat f) →
      v = Vfloat f ∨ v = Vundef.
    Proof.
      destruct v; inversion 1; vauto.
    Qed.

    Lemma ncompat_inv_single v f :
      ncompat v (NVsingle f) →
      v = Vsingle f ∨ v = Vundef.
    Proof.
      destruct v; inversion 1; vauto.
    Qed.

    Definition compat_fun (v:val) : num_val :=
      match v with
      | Vptr _ i | Vint i => NVint i
      | Vlong l => NVint64 l
      | Vundef => NVint Int.zero
      | Vfloat f => NVfloat f
      | Vsingle f => NVsingle f
      end.

    Lemma ncompat_compat_fun v :
      ncompat v (compat_fun v).
    Proof.
      destruct v; simpl; constructor.
    Qed.

    Ltac ncompat_inv H :=
      match type of H with
             | ncompat ?v (NVint ?i) =>
               destruct (ncompat_inv_int v i H) as [?|[(? & ?)|?]];
               clear H
             | ncompat ?v (NVint64 ?i) =>
               destruct (ncompat_inv_int64 v i H) as [?|?];
               clear H
             | ncompat ?v (NVfloat ?i) =>
               destruct (ncompat_inv_float v i H) as [?|?];
               clear H
             | ncompat ?v (NVsingle ?i) =>
               destruct (ncompat_inv_single v i H) as [?|?];
               clear H
             end.

    Definition wtype_num (v:num_val) (tp:AbTy.t) : Prop :=
      match v, tp with
        | NVint _, (VI|VP|VIP)
        | NVint64 _, VL
        | NVfloat _, VF
        | NVsingle _, VS => True
        | _, _ => False
      end.

    Definition compat (ρC:T.elt -> val) (ρ:var -> num_val) : Prop :=
      forall c, ncompat (ρC c) (ρ (elt2p c)).

    Variable Hcompat: compat ρC ρ.
    Hint Constructors is_bool.

    Lemma eval_pexpr_type_expr_ptr_VI: forall a b i,
      eval_pexpr ge m ρC ε a (Vptr b i) ->
      pexpr_get_ty a = VI -> False.
    Proof.
      intros a b i H H0. apply pexpr_get_ty_ok in H.
      rewrite H0 in H. auto. discriminate.
    Qed.

    Lemma eval_pexpr_type_expr_int_VP: forall a i,
      eval_pexpr ge m ρC ε a (Vint i) ->
      pexpr_get_ty a = VP -> False.
    Proof.
      intros a i H H0. apply pexpr_get_ty_ok in H.
      rewrite H0 in H. auto. discriminate.
    Qed.

    Lemma eval_pexpr_type_expr_float_VI: forall a f,
      eval_pexpr ge m ρC ε a (Vfloat f) ->
      pexpr_get_ty a = VI -> False.
    Proof.
      intros a f H H0. apply pexpr_get_ty_ok in H.
      rewrite H0 in H. auto. discriminate.
    Qed.

    Lemma val_cmp_exists_int: forall c i1 i2,
      exists i,
        Val.cmp c (Vint i1) (Vint i2) = Vint i /\
        of_bool (Int.cmp c i1 i2) = (NVint i).
    Proof.
      unfold Val.cmp, of_bool, Val.cmp_bool; intros.
      destruct Int.cmp; simpl; repeat econstructor.
    Qed.

    Lemma val_cmpu_exists_int: forall f c i1 i2,
      exists i,
        Val.cmpu f c (Vint i1) (Vint i2) = Vint i /\
        of_bool (Int.cmpu c i1 i2) =  NVint i.
    Proof.
      unfold Val.cmpu, of_bool, Val.cmpu_bool; intros.
      destruct Int.cmpu; simpl; repeat econstructor.
    Qed.

    Lemma val_cmpf_exists_int: forall c f1 f2,
      exists i,
        Val.cmpf c (Vfloat f1) (Vfloat f2) = Vint i /\
        of_bool (Float.cmp c f1 f2) =  NVint i.
    Proof.
      unfold Val.cmpf, of_bool, Val.cmpf_bool; intros.
      destruct Float.cmp; simpl; repeat econstructor.
    Qed.

    Lemma val_cmpfs_exists_int: forall c f1 f2,
      exists i,
        Val.cmpfs c (Vsingle f1) (Vsingle f2) = Vint i /\
        of_bool (Float32.cmp c f1 f2) =  NVint i.
    Proof.
      unfold Val.cmpfs, of_bool, Val.cmpfs_bool; intros.
      destruct Float32.cmp; simpl; repeat econstructor.
    Qed.

    Lemma val_of_bool_exists_int: forall b,
      exists i,
        Val.of_bool b = Vint i /\
        of_bool b = NVint i.
    Proof.
      unfold of_bool; intros.
      destruct b; repeat econstructor.
    Qed.

    Lemma ncompat_Vundef: forall n, ncompat Vundef n.
    Proof.
      destruct n; constructor.
    Qed.
    Hint Resolve ncompat_Vundef.

    Lemma ncompat_eval_nexpr: forall e v,
      match v with
        | Vint i => eval_nexpr ρ e (NVint i)
        | Vptr b i => eval_nexpr ρ e (NVint i)
        | Vlong i => eval_nexpr ρ e (NVint64 i)
        | Vundef => exists n, eval_nexpr ρ e n
        | Vfloat f => eval_nexpr ρ e (NVfloat f)
        | Vsingle f => eval_nexpr ρ e (NVsingle f)
      end ->
      exists n : num_val,
        ncompat v n /\  eval_nexpr ρ e n.
    Proof.
      intros e; destruct v;
        (intros (q & Hq) || intros Hq); vauto;
      (refine (ex_intro _ _ (conj _ _)); try eassumption; reflexivity).
    Qed.

    Lemma exist_ncompat_Vundef: forall P,
      (exists n, P n) ->
      exists n : num_val, ncompat Vundef n /\ P n.
    Proof.
      intros P (n & H).
      repeat econstructor; eauto.
    Qed.

    Lemma exist_ncompat_Vint: forall (P:num_val -> Prop) i,
      P (NVint i) ->
      exists n : num_val, ncompat (Vint i) n /\ P n.
    Proof.
      intros P H.
      repeat econstructor; eauto.
    Qed.

    Lemma exist_ncompat_Vptr: forall (P:num_val -> Prop) b i,
      P (NVint i) ->
      exists n : num_val, ncompat (Vptr b i) n /\ P n.
    Proof.
      intros P H.
      repeat econstructor; eauto.
    Qed.

    Lemma exist_ncompat_Vlong: forall (P:num_val -> Prop) l,
      P (NVint64 l) ->
      exists n : num_val, ncompat (Vlong l) n /\ P n.
    Proof.
      intros P H.
      repeat econstructor; eauto.
    Qed.

    Lemma exist_ncompat_Vfloat: forall (P:num_val -> Prop) f,
      P (NVfloat f) ->
      exists n : num_val, ncompat (Vfloat f) n /\ P n.
    Proof.
      intros P H.
      repeat econstructor; eauto.
    Qed.

    Lemma exist_ncompat_Vsingle: forall (P:num_val -> Prop) f,
      P (NVsingle f) ->
      exists n : num_val, ncompat (Vsingle f) n /\ P n.
    Proof.
      intros P H.
      repeat econstructor; eauto.
    Qed.

    Lemma of_bool_NVint : forall b,
      exists i, of_bool b = NVint i.
    Proof.
      unfold of_bool; destruct b; econstructor; reflexivity.
    Qed.

    Lemma nconvert_correct: forall e,
      forall v, eval_pexpr ge m ρC ε e v ->
      forall e', nconvert e = inl e' ->
      v ≠ Vundef →
      exists n, ncompat v n /\
                eval_nexpr ρ e' n /\
                wtype_num n (pexpr_get_ty e).
    Proof.
      intros e v Hv e' He' Hvundef.
      cut (∃ n : num_val, ncompat v n ∧ eval_nexpr ρ e' n).
      { intros (n & A & B). exists n. split. auto. split. auto.
        pose proof pexpr_get_ty_ok Hv Hvundef.
        destruct v, n, (pexpr_get_ty e); try contradiction; try congruence; constructor. }
      revert v Hv e' He' Hvundef. induction e; intros vv Hvv; inv Hvv; simpl.
      - intros. specialize (Hcompat v).
        exists (ρ (elt2p v)); repeat split; subst; auto.
        destruct t, (ρC v) eqn:EQρ; eq_some_inv; subst; try congruence;
        (constructor; [now auto|]); inv Hcompat; try constructor; now intuition.
      - intros e' E' NB; eq_some_inv. subst. exists (NVint Int.zero).
        repeat (constructor; auto).
      - destruct cst; simpl in *; eq_some_inv; intros e' ?; eq_some_inv; subst;
        (econstructor; split; [auto|repeat econstructor]);
        try apply Zle_refl.
        destruct Genv.find_symbol. reflexivity. elim (Hvundef eq_refl).
      - intros; eq_some_inv.
      - intros e' He' HB. revert He'.
        assert (v1 ≠ Vundef). { contradict HB. subst. destruct u; inv H4; auto. }
        assert (Hg:= pexpr_get_ty_ok H2 H). intro He'.
        destruct (pexpr_get_ty e) eqn:Eel;
        destruct u; try discriminate;
        (destruct (nconvert e) as [nconvert_e|]; eq_some_inv; subst;
         specialize (IHe _ H2 nconvert_e eq_refl));
        (destruct IHe as (n & E1 & E2); [
           destruct v1, vv; simpl in *; congruence
          | destruct n]);
        ncompat_inv E1; subst; simpl in *; try congruence; eq_some_inv; subst;
        try (apply exist_ncompat_Vundef; repeat econstructor);
        try match goal with
          | H : option_map _ ?o = Some ?x |- _ => destruct o eqn:?; inv H
        end;
        repeat econstructor; simpl; eauto.
      - intros e' He' HB.
        assert (v1 ≠ Vundef). { contradict HB. subst. destruct b; inv H6; auto. }
        assert (Hg1:= pexpr_get_ty_ok H3 H). red in Hg1.
        assert (v2 ≠ Vundef). { contradict HB. subst. destruct v1, b; inv H6; auto. }
        assert (Hg2:= pexpr_get_ty_ok H5 H0). red in Hg2.
        destruct (pexpr_get_ty e1) eqn:He_a1; simpl in Hg1;
        destruct (pexpr_get_ty e2) eqn:He_a2; simpl in Hg2;
        try discriminate;
        destruct b; try discriminate;
        (destruct (nconvert e1) as [nconvert_e1|],
                  (nconvert e2) as [nconvert_e2|]; eq_some_inv; subst;
         specialize (IHe1 _ H3 nconvert_e1 eq_refl);
         specialize (IHe2 _ H5 nconvert_e2 eq_refl));
        (destruct IHe1 as (n1 & E1 & E2); [
           intros ->; apply HB; destruct v2; vm_compute in H6; congruence
          | destruct n1]);
        (destruct IHe2 as (n2 & F1 & F2); [
           intros ->; apply HB; destruct v1; vm_compute in H6; congruence
          | destruct n2]);
        ncompat_inv E1; ncompat_inv F1; subst;
        try (elim Hg1; discriminate); try (elim Hg2; discriminate);
        simpl in H6;
        try (unfold Val.cmpl in H6; simpl in H6);
        try (unfold Val.cmplu in H6; simpl in H6);
        eq_some_inv; subst;
        repeat match goal with
           | id: (if ?b then _ else _) = Some _ |- _ => destruct b eqn:?; inv id
           | id: (_ || _) = false |- _ => apply orb_false_elim in id; destruct id
           | id: (_ && _) = false |- _ => apply andb_false_iff in id
           | |- context [if ?b then _ else _] => destruct b eqn:?
         end;
        try match goal with
          | |- context[Oshl] => destruct (Int.ltu i0 Int.iwordsize) eqn:?
          | |- context[Oshr] => destruct (Int.ltu i0 Int.iwordsize) eqn:?
          | |- context[Oshru] => destruct (Int.ltu i0 Int.iwordsize) eqn:?
          | |- context[Oshll] => destruct (Int.ltu i0 Int64.iwordsize') eqn:?
          | |- context[Oshrl] => destruct (Int.ltu i0 Int64.iwordsize') eqn:?
          | |- context[Oshrlu] => destruct (Int.ltu i0 Int64.iwordsize') eqn:?
          | |- context[Val.cmp ?c (Vint ?i1) (Vint ?i2)] =>
            destruct (val_cmp_exists_int c i1 i2) as (bi & B1 & B2); rewrite B1
          | |- context[Val.cmpu ?m ?c (Vint ?i1) (Vint ?i2)] =>
            destruct (val_cmpu_exists_int m c i1 i2) as (bi & B1 & B2); rewrite B1
          | |- context[Val.cmpf ?c (Vfloat ?f1) (Vfloat ?f2)] =>
            destruct (val_cmpf_exists_int c f1 f2) as (bi & B1 & B2); rewrite B1
          | |- context[Val.cmpfs ?c (Vsingle ?f1) (Vsingle ?f2)] =>
            destruct (val_cmpfs_exists_int c f1 f2) as (bi & B1 & B2); rewrite B1
          | |- context[Val.of_bool ?b] =>
            destruct (val_of_bool_exists_int b) as (bi & B1 & B2); rewrite B1
        end;
        try (rewrite He_a1, He_a2);
        try (elim (HB eq_refl));
        try (apply exist_ncompat_Vundef; econstructor; split;
             [ econstructor; try eassumption; econstructor; eauto
             | try match goal with
                 | |- context[of_bool ?b] =>
                   destruct (of_bool_NVint b) as (nb & Hb); rewrite Hb
               end; simpl; auto]; fail);
        try (apply exist_ncompat_Vint;
             econstructor; try eassumption; try rewrite <- B2;
             econstructor; eauto; fail);
        try (apply exist_ncompat_Vptr;
             econstructor; try eassumption;
             (econstructor || rewrite Int.add_commut; econstructor); eauto; fail);
        try (now repeat (econstructor (eauto))).
    Qed.

    Fixpoint nconvert_def_pre (pe: pexpr T.elt) : Prop :=
      match pe with
      | PEvar v _ => ρC(v) ≠ Vundef
      | PElocal i => ε(i) ≠ None
      | PEconst (Some (Oaddrsymbol i _)) _ =>
        Genv.find_symbol ge i ≠ None
      | PEconst _ _ => True
      | PEunop u pe' _ =>
        nconvert_def_pre pe'
      | PEbinop b pe1 pe2 _ =>
        nconvert_def_pre pe1 ∧
        nconvert_def_pre pe2 ∧
        match b with
        | Ocmpu c =>
          ∀ b1 i1 b2 i2,
            Vptr b1 i1 ∈ eval_pexpr ge m ρC ε pe1 →
            Vptr b2 i2 ∈ eval_pexpr ge m ρC ε pe2 →
            b1 = b2 ∧ Mem.weak_valid_pointer m b1 (Int.unsigned i1) && Mem.weak_valid_pointer m b2 (Int.unsigned i2) = true
        | Osub =>
          ∀ b1 i1 b2 i2,
            Vptr b1 i1 ∈ eval_pexpr ge m ρC ε pe1 →
            Vptr b2 i2 ∈ eval_pexpr ge m ρC ε pe2 →
            b1 = b2
        | _ => True
        end
      end.

    Ltac Split :=
      repeat
      match goal with
      | H : ∃ _, _ |- _ => destruct H as (H & ?)
      | H : _ ∨ _ |- _ => destruct H as [H|H]
      | H : _ ∧ _ |- _ => destruct H as (H & ?)
      end.

    Lemma error_pexpr_unop_inv u q ty :
      error_pexpr ge m ρC ε (PEunop u q ty) →
      (∃ v, eval_pexpr ge m ρC ε q v ∧ error_unop u v) ∨ (error_pexpr ge m ρC ε q).
    Proof. inversion 1; subst; vauto. Qed.

    Lemma error_pexpr_binop_inv b q r ty :
      error_pexpr ge m ρC ε (PEbinop b q r ty) →
      (∃ v w, eval_pexpr ge m ρC ε q v ∧ eval_pexpr ge m ρC ε r w ∧ error_pbinop m b v w) ∨
      (error_pexpr ge m ρC ε q) ∨
      (∃ v, eval_pexpr ge m ρC ε q v ∧ error_pexpr ge m ρC ε r).
    Proof. inversion 1; subst; vauto. Qed.

    Lemma nconvert_def pe :
      nconvert_def_pre pe →
      ∀ ne, nconvert pe = inl ne →
      ¬ error_nexpr ρ ne →
      ¬ error_pexpr ge m ρC ε pe.
    Proof.
      induction pe; intros PRE ne.
    - simpl.
      specialize (Hcompat v).
      intros NE _ B. inv B. apply PRE; assumption.
    - simpl. intros ?; eq_some_inv; subst ne.
      intros _. simpl in PRE. intros B; inv B. auto.
    - simpl. destruct o; intros K; eq_some_inv. subst ne.
      intros _. intros B; inv B; simpl in PRE; auto.
    - specialize (IHpe PRE).
      pose proof (@pexpr_get_ty_ok _ _ _ ge m ρC ε pe) as TY.
      simpl.
      intros K NB.
      destruct (pexpr_get_ty pe); eq_some_inv;
      destruct u;
      eq_some_inv;
      apply bind_inl in K;
      destruct K as (ne' & K & ?);
      eq_some_inv;
      subst ne;
      assert (¬ error_nexpr ρ ne') as NB' by (intro; elim NB; vauto);
      specialize (IHpe _ K NB');
      intros B; apply error_pexpr_unop_inv in B; destruct B as [ (? & EV & B) | B ];
      try (apply IHpe; assumption);
      match goal with
      | EV : eval_pexpr _ _ _ _ _ ?i,
        B : error_unop _ _ |- _ =>
        let DEF := fresh in
        assert (DEF := error_pexpr_def IHpe EV);
        specialize (TY _ EV DEF);
        destruct i; destruct TY;
        try (inversion B; intuition eauto; discriminate);
        let CPT := fresh in
        edestruct (nconvert_correct _ _ EV _ K) as (? & CPT & ? & _); eauto;
        simpl in CPT; subst; inversion B; intuition eauto;
        apply NB; vauto
      end.
    - assert (nconvert_def_pre pe1) as PRE1 by apply PRE.
      specialize (IHpe1 PRE1). clear PRE1.
      assert (nconvert_def_pre pe2) as PRE2 by apply PRE.
      specialize (IHpe2 PRE2). clear PRE2.
      pose proof (@pexpr_get_ty_ok _ _ _ ge m ρC ε pe1) as TY1.
      pose proof (@pexpr_get_ty_ok _ _ _ ge m ρC ε pe2) as TY2.
      simpl.
      intros K NB.
      destruct (pexpr_get_ty pe1); eq_some_inv;
      destruct (pexpr_get_ty pe2); eq_some_inv;
      destruct b; eq_some_inv;
      apply bind_inl in K; destruct K as (ne1 & NE1 & K);
      apply bind_inl in K; destruct K as (ne2 & NE2 & K);
      eq_some_inv; subst ne;
      assert (¬ error_nexpr ρ ne1) as NB1 by (intro; elim NB; vauto);
      specialize (IHpe1 _ NE1 NB1); clear NB1;
      specialize (IHpe2 _ NE2);
      intros B; apply error_pexpr_binop_inv in B;
      destruct B as [ ( ? & ? & EV1 & EV2 & B) | [ B | (? & EV1 & B) ] ];
      eauto;
      match goal with
      | EV1: eval_pexpr _ _ _ _ pe1 ?i1 |- _ =>
        assert (DEF1 := error_pexpr_def IHpe1 EV1);
        specialize (TY1 _ EV1 DEF1);
        destruct i1; destruct TY1;
        let T := fresh in
        destruct (nconvert_correct pe1 _ EV1 _ NE1) as (? & T & ? & _);[ easy | simpl in T ]
      end;
      assert (¬ error_nexpr ρ ne2) as NB2 by (intro; elim NB; vauto);
      specialize (IHpe2 NB2); clear NB2;
      eauto;
      match goal with
      | EV2: eval_pexpr _ _ _ _ pe2 ?i2 |- _ =>
        assert (DEF2 := error_pexpr_def IHpe2 EV2);
        specialize (TY2 _ EV2 DEF2);
        destruct i2; destruct TY2;
        let T := fresh in
        destruct (nconvert_correct pe2 _ EV2 _ NE2) as (? & T & ? & _);[ easy | simpl in T ]
      end;
      inversion B; Split; try discriminate;
      subst;
      eauto;
      try (apply NB; now vauto);
      simpl in PRE; destruct PRE as (PRE1 & PRE2 & PRE); eauto;
      edestruct PRE; eauto.
      InvBooleans. congruence.
    Qed.

  End nconvert.

  Import Utf8.

  Local Instance TeltEqDec : EqDec T.elt := T.elt_eq.
  Lemma compat_upd var `{EQ: EqDec var} f ρC ρ :
    (∀ x y, x ≠ y → f x ≠ f y) →
    compat var f ρC ρ →
    ∀ x v n ρC',
      ncompat v n →
      ρC' ∈ upd_spec ρC x v →
      compat var f ρC' (upd ρ (f x) n).
  Proof.
    intros INJ H x v n ρC' V UPD c. rewrite (UPD c). clear UPD.
    unfold upd. simpl.
    destruct (eq_dec c x). subst. rewrite dec_eq_true. assumption.
    rewrite dec_eq_false; auto.
  Qed.

End Nconvert.
