(* *********************************************************************)
(*                                                                     *)
(*              The Verasco static analyzer                            *)
(*                                                                     *)
(*          Jacques-Henri Jourdan, INRIA Paris-Rocquencourt            *)
(*                                                                     *)
(*  Copyright Institut National de Recherche en Informatique et en     *)
(*  Automatique and Université de Rennes 1.  All rights reserved.      *)
(*  This file is distributed under the terms of the GNU General        *)
(*  Public License as published by the Free Software Foundatio,        *)
(*  either version 2 of the License, or (at your option) any later     *)
(*  version.  This file is also distributed under the terms of the     *)
(*  INRIA Non-Commercial License Agreement.                            *)
(*                                                                     *)
(* *********************************************************************)

Require Import Utf8.
Require Import Reals.
Require Import IdealExpr.
Require Import AbIdealNonrel.
Require Import ZArith.
Require Import Coqlib.
Require Import ZIntervals.
Require Import Integers.
Require Import AdomLib.
Require Import AbIdealEnv ShareTree Util TreeAl.
Require Import Fcore_Raux.
Require Import Fcore_generic_fmt.
Require Import Fappli_IEEE.
Require Import Fappli_IEEE_bits.
Require Import Fappli_IEEE_extra.
Require Import Fourier.
Require Import Floats.
Require Import FloatLib.
Require Import Psatz.
Require Import LinearQuery AbLinearize.
Require Import FloatIntervals.
Require Import FactMsgHelpers.
Require Import ToString.
Require Import String.

Local Existing Instance itv_gamma_oR.
Local Existing Instance itv_gamma_R.

Inductive ovar :=
| ovPlus (v:var)
| ovMinus (v:var).

Instance ovar_tostring {_:ToString Cells.cell}: ToString ovar :=
  fun x =>
    match x with
    | ovPlus x => ("+"++to_string x)%string
    | ovMinus x => ("-"++to_string x)%string
    end.

Definition oρ (ρ : var -> ideal_num) (x:ovar) : option R :=
  match x with
  | ovPlus x => rρ ρ x
  | ovMinus x => option_map Ropp (rρ ρ x)
  end.

Definition ovar_opp (x:ovar) : ovar :=
  match x with
  | ovPlus v => ovMinus v
  | ovMinus v => ovPlus v
  end.

Lemma ovar_opp_opp:
  ∀ ρ v x, oρ ρ v = Some (-x)%R <-> oρ ρ (ovar_opp v) = Some x.
Proof.
  destruct v; simpl.
  - destruct rρ; simpl. 2:intuition congruence.
    split; intros; inv H; rewrite Ropp_involutive; auto.
  - destruct rρ; simpl. 2:intuition congruence.
    split; intros; inv H; f_equal; Psatz.lra.
Qed.

Module ovar_EQ <: TYPE_EQ
    with Definition s := ovar
    with Definition t := (var * (unit + unit))%type.

  Definition s := ovar.
  Definition t := (var * (unit + unit))%type.

  Definition t_of_s (x:s) : t :=
    match x with
    | ovPlus x => (x, inl tt)
    | ovMinus x => (x, inr tt)
    end.

  Definition s_of_t (x:t) : s :=
    match x with
    | (x, inl _) => ovPlus x
    | (x, inr _) => ovMinus x
    end.

  Lemma s_of_t_of_s : forall x : s, s_of_t (t_of_s x) = x.
  Proof. intros []; reflexivity. Qed.

  Lemma t_of_s_of_t: forall x : t, t_of_s (s_of_t x) = x.
  Proof. intros [? [[]|[]]]; reflexivity. Qed.

  Instance eq : EqDec s.
  Proof. unfold EqDec. decide equality; apply eq_dec. Defined.
End ovar_EQ.

Module SUU := SumShareTree(UnitShareTree)(UnitShareTree).
Module PTSUU := ProdShareTree(T)(SUU).
Module To := BijShareTree(ovar_EQ)(PTSUU).
Module Top := ShareTree_Properties(To).
Module Too := ProdShareTree(To)(To).
Module Toop := ShareTree_Properties(Too).

Instance ovar_eq : EqDec ovar := To.elt_eq.

Definition t0 := Too.t fin_float.
Definition iter_t := t0+⊥.

Instance t0_tostring {_:ToString Cells.cell}: ToString t0 :=
  fun x =>
    (Too.fold (fun acc xy b =>
       let '(x, y) := xy in
       acc ++ "[" ++ to_string x ++ to_string (ovar_opp y) ++ "<=" ++ to_string (proj1_sig b)++"]")
       x "{" ++ "}")%string.

Instance octitert_tostring {_:ToString Cells.cell} : ToString iter_t := _.

Instance ogamma0: gamma_op t0 (var -> ideal_num) :=
  fun ab ρ =>
    ∀ x y b, Too.get (x, y) ab = Some b ->
    ∃ xv yv, oρ ρ x = Some xv /\ oρ ρ y = Some yv /\ (xv-yv <= B2R _ _ (proj1_sig b))%R.

(******** Generic functions ********)

Definition best_bound (b1:option fin_float) (b2:option fin_float) : option fin_float * bool :=
  match b2, b1 with
  | None, _ => (b1, true)
  | _, None => (b2, false)
  | Some (exist b2' _), Some (exist b1' _) =>
    if Fleb b1' b2' then (b1, true) else (b2, false)
  end.

Lemma best_bound_either:
  ∀ b1 b2 (P:option fin_float * bool -> Prop),
    (P (b1, true)) ->
    (P (b2, false)) ->
    (P (best_bound b1 b2)).
Proof.
  intros.
  destruct b1 as [[]|], b2 as [[]|]; auto.
  simpl. destruct @Fleb; auto.
Qed.

Lemma reduct_rule_1:
  ∀ x y z (f1 f2:fin_float),
    (x - y <= B2R _ _ (proj1_sig f1))%R ->
    (y - z <= B2R _ _ (proj1_sig f2))%R ->
    let b := Bplus 53 1024 eq_refl eq_refl Float.binop_pl mode_UP
                   (proj1_sig f1) (proj1_sig f2) in
    is_finite _ _ b = true ->
    (x - z <= B2R _ _ b)%R.
Proof.
  intros.
  assert (x-z<=B2R 53 1024 (proj1_sig f1)+B2R 53 1024 (proj1_sig f2))%R by fourier.
  eapply Rle_trans. eauto.
  eapply Rle_trans.
  { edestruct (round_UP_pt radix2) as (A & B & C).
    apply (fexp_correct 53 1024 eq_refl). apply B. }
  pose proof Bplus_correct 53 1024 eq_refl eq_refl Float.binop_pl mode_UP
    (proj1_sig f1) (proj1_sig f2) (proj2_sig f1) (proj2_sig f2).
  Rlt_bool_case H3.
  - apply Req_le. now intuition.
  - destruct H3.
    destruct Bplus; try discriminate. destruct (Bsign 53 1024 (proj1_sig f1)); discriminate.
    simpl in H3. unfold binary_overflow in H3. simpl in H3.
    destruct (Bsign 53 1024 (proj1_sig f1)) eqn:SGN1. 2:discriminate.
    assert (B2R _ _ (proj1_sig f1) <= 0)%R.
    { destruct f1 as [[| | |[]] ?]; try discriminate; simpl. fourier.
      apply Fcore_float_prop.F2R_le_0_compat. discriminate. }
    assert (B2R _ _ (proj1_sig f2) <= 0)%R.
    { destruct f2 as [[| | |[]] ?]; try discriminate; simpl. fourier.
      apply Fcore_float_prop.F2R_le_0_compat. compute. discriminate. }
    rewrite Rabs_left1 in H4.
    2:apply round_le_generic. 2:apply fexp_correct, eq_refl. 2:apply valid_rnd_round_mode.
    2:apply generic_format_0. 2:fourier.
    apply Ropp_le_contravar in H4. rewrite Ropp_involutive in H4.
    eapply Rle_trans. eauto. apply (Rabs_le_inv (B2R 53 1024 b)).
    apply Rlt_le, abs_B2R_lt_emax.
Qed.

Definition half : float :=
  match Float.exact_inverse (BofZ 53 1024 eq_refl eq_refl 2) as x
        return match x with Some _ => float | None => True end
  with
  | Some x => x
  | None => I
  end.

Lemma half_half:
  B2R _ _ half = (1/2)%R.
Proof. compute. Psatz.lra. Qed.

Opaque Bmult.

Lemma reduct_rule_2:
  ∀ x y (f1 f2:fin_float),
    (2 * x <= B2R _ _ (proj1_sig f1))%R ->
    (- 2 * y <= B2R _ _ (proj1_sig f2))%R ->
    let b := Bplus 53 1024 eq_refl eq_refl Float.binop_pl mode_UP
                  (proj1_sig f1) (proj1_sig f2) in
    let b := Bmult 53 1024 eq_refl eq_refl Float.binop_pl mode_UP
                   half b in
    is_finite _ _ b = true ->
    (x - y <= B2R _ _ b)%R.
Proof.
  intros.
  assert (x-y<=(1/2)*(B2R 53 1024 (proj1_sig f1)+B2R 53 1024 (proj1_sig f2)))%R by Psatz.lra.
  eapply Rle_trans. eauto.
  eapply Rle_trans.
  { apply Rmult_le_compat_l. Psatz.lra.
    edestruct (round_UP_pt radix2) as (A & B & C).
    apply (fexp_correct 53 1024 eq_refl). apply B. }
  eapply Rle_trans.
  { edestruct (round_UP_pt radix2) as (A & B & C).
    apply (fexp_correct 53 1024 eq_refl). apply B. }
  pose proof Bplus_correct 53 1024 eq_refl eq_refl Float.binop_pl mode_UP
    (proj1_sig f1) (proj1_sig f2) (proj2_sig f1) (proj2_sig f2).
  Rlt_bool_case H3.
  - destruct H3 as (? & ? & ?). simpl round_mode in H3. rewrite <- H3. fold b. rewrite <- half_half.
    pose proof Bmult_correct 53 1024 eq_refl eq_refl Float.binop_pl mode_UP half b.
    Rlt_bool_case H7.
    + destruct H7 as (? & ? & ?). simpl round_mode in H7. rewrite <- H7. apply Rle_refl.
    + exfalso. pose proof abs_B2R_lt_emax _ _ b.
      assert (1/2*Rabs (B2R 53 1024 b) < bpow radix2 1024 * /2)%R by Psatz.lra.
      change (/2)%R with (bpow radix2 (-1)) in H10. rewrite <- bpow_plus in H10.
      rewrite <- (Rabs_right (1/2)), <- Rabs_mult, <- half_half in H10 by Psatz.lra.
      eapply Rlt_le, Rabs_le_inv in H10. destruct H10.
      eapply round_ge_generic in H10.
      2:apply (fexp_correct 53 1024 eq_refl). 2:apply (valid_rnd_round_mode mode_UP).
      2:apply generic_format_opp, generic_format_bpow; discriminate.
      eapply round_le_generic in H11.
      2:apply (fexp_correct 53 1024 eq_refl). 2:apply (valid_rnd_round_mode mode_UP).
      2:apply generic_format_bpow; discriminate.
      pose proof (Rabs_le _ _ (conj H10 H11)).
      assert (bpow radix2 1024 <= bpow radix2 (1024-1))%R.
      { eapply Rle_trans; eauto. }
      apply le_bpow in H13. omega.
  - destruct H3.
    destruct Bplus; try discriminate. destruct (Bsign 53 1024 (proj1_sig f1)); discriminate.
    simpl in H3. unfold binary_overflow in H3. simpl in H3.
    destruct (Bsign 53 1024 (proj1_sig f1)) eqn:SGN1. 2:discriminate.
    assert (B2R _ _ (proj1_sig f1) <= 0)%R.
    { destruct f1 as [[| | |[]] ?]; try discriminate; simpl. fourier.
      apply Fcore_float_prop.F2R_le_0_compat. discriminate. }
    assert (B2R _ _ (proj1_sig f2) <= 0)%R.
    { destruct f2 as [[| | |[]] ?]; try discriminate; simpl. fourier.
      apply Fcore_float_prop.F2R_le_0_compat. compute. discriminate. }
    rewrite Rabs_left1 in H4.
    2:apply round_le_generic. 2:apply fexp_correct, eq_refl. 2:apply valid_rnd_round_mode.
    2:apply generic_format_0. 2:fourier.
    apply Ropp_le_contravar in H4. rewrite Ropp_involutive in H4.
    eapply Rle_trans.
    { apply round_le_generic. apply fexp_correct. reflexivity. apply (valid_rnd_round_mode mode_UP).
      apply generic_format_opp, generic_format_bpow with (e:=-1+1024). discriminate.
      eapply Rle_trans. apply Rfourier_le. 2:Psatz.lra. apply H4.
      compute. Psatz.lra. }
    unfold b0, b. clear b0 b H1. inv H3. Transparent Bmult. compute. Opaque Bmult. Psatz.lra.
Qed.

Definition reduce_2 (ab:t0) : t0 :=
  let mapBnds :=
    To.combine (fun x _ => do_opt x <- x; x)
      (To.map (@To.get _ ∘ ovar_opp) ab) (To.empty unit)
  in
  To.fold (fun ab A (bndA:fin_float) =>
      match To.get_set A ab with
      | (None, _) => ab (* Should not happen, because A is in mapBnds *)
      | (Some abA, setter_abA) =>
        let '(abA, chgt) :=
          To.fold (fun abA_chgt B (bndB:fin_float) =>
            let '(abA, chgt) := abA_chgt in
            let '(old, setter) := To.get_set (ovar_opp B) abA in
            let new :=
              let b := Bplus 53 1024 eq_refl eq_refl Float.binop_pl mode_UP
                             (proj1_sig bndA) (proj1_sig bndB) in
              let b := Bmult 53 1024 eq_refl eq_refl Float.binop_pl mode_UP
                             half b in
              fin_float_of_float b
            in
            if snd (best_bound old new) then abA_chgt
            else match new with None => abA_chgt | Some new => (setter new, true) end)
          mapBnds (abA, false)
        in
        if chgt then setter_abA abA else ab
      end)
  mapBnds ab.

Lemma reduce_2_correct:
  ∀ ab ρ,
    ρ ∈ γ ab ->
    ρ ∈ γ (reduce_2 ab).
Proof.
  unfold reduce_2. intros.
  set (mapBnds:=To.combine (fun x _ => do_opt x <- x; x)
         (To.map (@To.get _ ∘ ovar_opp) ab) (To.empty unit)).
  assert (∀ X B, List.In (X, B) (To.elements mapBnds) ->
          ∃ Xv, oρ ρ X = Some Xv /\ 2*Xv <= B2R _ _ (proj1_sig B))%R.
  { unfold mapBnds. intros. apply To.elements_complete in H0.
    rewrite To.gcombine, To.gmap in H0 by auto.
    specialize (H X (ovar_opp X)). unfold Too.get in H.
    change PTSUU.t with To.t in *. destruct (To.get X ab). 2:discriminate.
    edestruct H as (xv & yv & Hxv & Hyv & Hxvb). apply H0. clear H H0.
    rewrite <- ovar_opp_opp, Hxv in Hyv. inv Hyv.
    eexists. split. eauto. Psatz.lra. }
  revert H0. generalize mapBnds. clear mapBnds.
  intros. rewrite To.fold_spec. generalize mapBnds at 2, H0.
  intros X HH. revert HH ab H. induction (To.elements X) as [|[A Ab]]. now auto.
  intros. apply IHl. intros; apply HH; right; now eauto.
  edestruct HH as (Av & HAv & H2Av). left. now eauto. clear HH IHl l.
  unfold fst at 1. destruct (To.get_set_spec A ab).
  change PTSUU.t with To.t. destruct (To.get_set A ab) as [[] setter_abA]. 2:now auto.
  match goal with
  | |- context [To.fold ?F ?A ?B] => destruct (To.fold F A B) as [? []] eqn:EQ
  end.
  2:now auto. rewrite H2. clear H2.
  repeat intro. unfold Too.get  in H2. rewrite To.gsspec in H2.
  destruct (To.elt_eq x A). 2:now auto. subst x.
  specialize (H A y). unfold Too.get in H. rewrite <- H1 in H. simpl in H. clear H1.
  rewrite To.fold_spec in EQ. revert H0 t H EQ. generalize false.
  induction (To.elements mapBnds) as [|[B Bb]]. { intros. inv EQ. auto. }
  intros. eapply IHl. intros; apply H0; right; now eauto.
  2:rewrite <- surjective_pairing; now apply EQ. clear t1 EQ IHl H2.
  unfold fst, snd. destruct (To.get_set_spec (ovar_opp B) t).
  destruct (To.get_set (ovar_opp B) t). unfold fst, snd in H1, H2. subst o.
  apply best_bound_either. now auto.
  pose proof fin_float_of_float_correct
    (Bmult 53 1024 eq_refl eq_refl Float.binop_pl mode_UP half
       (Bplus 53 1024 eq_refl eq_refl Float.binop_pl mode_UP
          (proj1_sig Ab) (proj1_sig Bb))).
  destruct fin_float_of_float. 2:now auto.
  rewrite H2, To.gsspec. clear t1 H2. destruct (To.elt_eq y (ovar_opp B)). 2:now auto.
  intros. inv H2. edestruct H1 as [? ->]. now auto. clear H H1.
  edestruct H0 as (Bv & HBv & H2Bv). left. now auto. clear H0.
  eexists. eexists. split. now eauto. split.
  rewrite <- ovar_opp_opp, HBv. f_equal. symmetry. apply Ropp_involutive.
  apply reduct_rule_2, H2. now auto. Psatz.lra.
Qed.

CoInductive Thunkt0 (x:t0) : Type := cThunk (x':t0) (Hx':γ x ⊆ γ x').
Definition t := sigT Thunkt0.

Definition get_unreduced (x:t) : t0 := projT1 x.
Definition get_reduced (x:t) : t0 := let 'existT _ (cThunk res _) := x in res.
Definition t_of_t0 (x:t0) : t := existT _ x (cThunk _ (reduce_2 x) (reduce_2_correct _)).
Definition t_of_t0_reduced (x:t0) : t := existT _ x (cThunk _ x (fun _ H => H)).

Instance octt_tostring {_:ToString Cells.cell} : ToString t :=
  fun x => to_string (get_unreduced x).

Instance ogamma: gamma_op t (var -> ideal_num) := fun ab => γ (get_unreduced ab).

Lemma t_of_t0_correct: ∀ x, γ x ⊆ γ (t_of_t0 x).
Proof. auto. Qed.

Lemma t_of_t0_reduced_correct: ∀ x, γ x ⊆ γ (t_of_t0_reduced x).
Proof. auto. Qed.

Lemma get_reduced_correct:
  ∀ x, γ x ⊆ γ (get_reduced x).
Proof. intros x ρ Hρ. destruct x as [? []]. auto. Qed.

Definition reduce_diag (x0 x:t0) : t0+⊥.
refine (
  let handle_bound f :=
      match Bcompare _ _ (proj1_sig f) Float.zero with
      | None => None (* Impossible *)
      | Some Lt => Some false (* Contradiction => will return Bot *)
      | Some Eq => None (* 0 : Nothing to do *)
      | Some Gt => Some true
      end
  in
  let diffDiag :=
      To.shcombine_diff
        (fun k x0 x =>
           match x0, x with
           | None, None => None
           | Some _, None => None
           | None, Some x =>
             if Top.is_empty x then None
             else
               match To.get k x with
               | None => Some true
               | Some f => handle_bound f
               end
           | Some x0, Some x =>
             if Top.is_empty x then None
             else
               match To.get k x0, To.get k x with
               | None, None => None
               | Some _, None => Some true
               | None, Some f => handle_bound f
               | Some f0, Some f =>
                 if fin_float_eqdec f f0 then None
                 else handle_bound f
               end
           end)
        _ x0 x
  in
  To.fold
    (fun oct k (b:bool) =>
       if b then
         do_bot oct <- oct;
         NotBot (Too.set (k, k) (exist _ Float.zero eq_refl) oct)
       else Bot)
    diffDiag (NotBot x)).
Proof.
  clear.
  abstract
    (destruct v; [|reflexivity]; destruct Top.is_empty; [reflexivity|];
     destruct To.get; [|reflexivity]; destruct fin_float_eqdec; congruence).
Defined.

Lemma reduce_diag_correct:
  ∀ x0 x ρ, ρ ∈ γ x -> ρ ∈ γ (reduce_diag x0 x).
Proof.
  intros. unfold reduce_diag.
  match goal with
  | |- ρ ∈ γ (To.fold _ ?diffDiag _) =>
    assert (HDIFF:∀ k,
               match To.get k diffDiag with
               | None => True
               | Some true => oρ ρ k <> None
               | Some false => False
               end);
    [|revert HDIFF; generalize diffDiag; intros]
  end.
  { intros. rewrite To.gshcombine_diff. pose proof (H k k). specialize (H k).
    unfold Too.get in H, H0. fold To.t.
    destruct (To.get k x). 2:destruct To.get; exact I.
    pose proof Top.is_empty_correct _ t1. destruct (Top.is_empty t1).
    destruct (To.get k x0); exact I.
    assert (oρ ρ k ≠ None). {
      intro. destruct H1. lapply H3. discriminate.
      intros. specialize (H i). destruct (To.get i t1). 2:auto.
      edestruct H as (xv & _ & Hxv & _ & _). eauto. congruence. }
    destruct (To.get k t1). 2:destruct (To.get k x0); [destruct (To.get k t2)|]; now auto.
    edestruct H0 as (xv & yv & Hxv & Hyv & Hxyb). reflexivity. clear H H0 H1.
    rewrite Hxv in Hyv. inv Hyv. replace (yv-yv)%R with 0%R in Hxyb by ring.
    rewrite Bcompare_correct. 2:apply proj2_sig. 2:reflexivity.
    apply Rcompare_not_Lt in Hxyb. change (B2R 53 1024 Float.zero) with 0%R.
    unfold float, binary64 in *.
    destruct To.get. 2:destruct Rcompare; try easy; congruence.
    destruct To.get. 2:destruct Rcompare; try easy; congruence.
    destruct fin_float_eqdec. now auto. destruct Rcompare; try easy; congruence. }
  rewrite To.fold_spec.
  assert (∀ k b, List.In (k, b) (To.elements t1) -> b = true /\ oρ ρ k <> None).
  { intros. apply To.elements_complete in H0. specialize (HDIFF k).
    rewrite H0 in HDIFF. destruct b; now intuition. }
  clear HDIFF.
  assert (γ (NotBot x) ρ) by auto. clear H.
  revert H1 H0. generalize (NotBot x). induction (To.elements t1). auto.
  intros. apply IHl.
  - destruct a as [k []]; destruct (H0 _ _ (or_introl eq_refl)); clear H0.
    2:discriminate.
    eapply botbind_spec; eauto.
    repeat intro. rewrite Too.gsspec in H3. destruct Too.elt_eq. inv e. inv H3.
    + destruct oρ. 2:congruence. eexists. eexists. split. eauto. split. eauto.
      simpl. change (B2R 53 1024 Float.zero) with 0%R. Psatz.lra.
    + apply H0. auto.
  - intros. apply H0. right. auto.
Qed.

(******** Top ********)

Definition otop (c:query_chan): (t * messages_chan)+⊥ :=
  NotBot (t_of_t0 (Too.empty _), List.nil).

Lemma otop_correct:
  ∀ c ρ, ρ ∈ γ c -> ρ ∈ γ (otop c).
Proof.
  split. apply t_of_t0_correct. repeat intro. rewrite Too.gempty in H0. discriminate. constructor.
Qed.

(******** Join ********)

Instance ojoin_inst: join_op t t.
intros x y.
refine (t_of_t0_reduced (
          Too.shcombine
            (fun _ (x y:option fin_float) =>
               do_opt x' <- x;
               do_opt y' <- y;
               if Fleb (proj1_sig x') (proj1_sig y') then y else x)
            _ (get_reduced x) (get_reduced y))).
Proof. abstract (destruct v; [destruct @Fleb|]; reflexivity). Defined.

Instance ojoin_inst_correct : join_op_correct t t (var -> ideal_num).
Proof.
  intros x y ρ Hρ. apply t_of_t0_reduced_correct.
  intros xv yv b Hb. rewrite Too.gshcombine in Hb.
  destruct (Too.get (xv, yv) (get_reduced x)) eqn:EQx, (Too.get (xv, yv) (get_reduced y)) eqn:EQy; inv Hb.
  rewrite Fleb_Rle in H0 by apply proj2_sig.
  destruct Hρ as [Hρ|Hρ]; apply get_reduced_correct in Hρ;
  edestruct Hρ as (xvv & yvv & Hxvv & Hyvv & Hb); eauto;
  eexists; eexists; split; eauto; split; eauto; Rle_bool_case H0; inv H0; fourier.
Qed.

Definition ojoin (x:t * query_chan) (y:t * query_chan) (c:query_chan) :
  (t * messages_chan)+⊥ :=
  NotBot (join (fst x) (fst y), List.nil).

Lemma ojoin_correct :
  forall x y c ρ,
    ρ ∈ γ x \/ ρ ∈ γ y -> ρ ∈ γ c ->
    ρ ∈ γ (ojoin x y c).
Proof. split. simpl. apply join_correct. destruct H as [[]|[]]; auto. constructor. Qed.

(******** Comparison ********)

(* TODO : do queries if we cannot prove x <= y *)
(*        is that a good idea ? *)
(*        pros : better precision, lower number of iterations *)
(*        cons : can lead to a lot of queries.
                 most of the time useless because octs
                 should be synchronized with itvs *)
Definition oleb (x:t * query_chan) (y:t) : bool.
refine (
  let x := get_unreduced (fst x) in
  let y := get_unreduced y in
  Too.shforall2 (fun k (xff yff:option fin_float) =>
    match yff with
    | None => true
    | Some fy =>
      match xff with
      | None => false
      | Some fx => Fleb (proj1_sig fx) (proj1_sig fy)
      end ||
      let x' :=
        let '(A, B) := k in
        do_opt bndA <- Too.get (A, ovar_opp A) x;
        do_opt bndB <- Too.get (ovar_opp B, B) x;
        let b := Bplus 53 1024 eq_refl eq_refl Float.binop_pl mode_UP
                       (proj1_sig bndA) (proj1_sig bndB) in
        let b := Bmult 53 1024 eq_refl eq_refl Float.binop_pl mode_UP
                       half b in
        fin_float_of_float b
      in
      match x' with
      | None => false
      | Some fx' => Fleb (proj1_sig fx') (proj1_sig fy)
      end
    end)
    _ x y).
Proof.
  clear.
  abstract (
      destruct v; [rewrite Fleb_refl; destruct f as [[] ?];
                   try discriminate; reflexivity
                  |reflexivity]).
Defined.

Lemma oleb_correct:
  forall x y ρ,
    oleb x y = true ->
    ρ ∈ γ x ->
    ρ ∈ γ y.
Proof.
  repeat intro.
  unfold oleb in H. eapply Too.shforall2_correct with (x:=(x0, y0)) in H.
  rewrite H1 in H. destruct H0. apply Bool.orb_true_iff in H. destruct H.
  - specialize (H0 x0 y0).
    destruct Too.get. 2:discriminate.
    edestruct H0 as (xb & yb & Hxb & Hyb & Hxyb). eauto.
    exists xb, yb. split. auto. split. auto.
    eapply Rle_trans. eauto. rewrite Fleb_Rle in H by apply proj2_sig.
    Rle_bool_case H. auto. discriminate.
  - destruct (Too.get (x0, ovar_opp x0) (get_unreduced (fst x))) eqn:EQx. 2:discriminate.
    destruct (Too.get (ovar_opp y0, y0) (get_unreduced (fst x))) eqn:EQy. 2:discriminate.
    apply H0 in EQx. destruct EQx as (xv & mxv & Hxv & Hmxv & Hxmxb).
    apply H0 in EQy. destruct EQy as (myv & yv & Hmyv & Hyv & Hmyyb).
    eexists. eexists. split. eauto. split. eauto.
    pose proof (fin_float_of_float_correct
      (Bmult 53 1024 eq_refl eq_refl Float.binop_pl mode_UP half
        (Bplus 53 1024 eq_refl eq_refl Float.binop_pl mode_UP
               (proj1_sig f) (proj1_sig f0)))).
    destruct fin_float_of_float. 2:discriminate.
    rewrite Fleb_Rle in H by apply proj2_sig.
    Rle_bool_case H. 2:discriminate. eapply Rle_trans, H4. clear H4 b H1 H y.
    edestruct H3 as [? ->]. now auto. apply reduct_rule_2, H.
    rewrite <- ovar_opp_opp, Hxv in Hmxv. inv Hmxv. Psatz.lra.
    rewrite <- ovar_opp_opp, Hyv in Hmyv. inv Hmyv. Psatz.lra.
Qed.

(******** Widenning & init_iter ********)

Definition oinit_iter (x:t+⊥) : iter_t := do_bot x <- x; NotBot (get_reduced x).

Lemma oinit_iter_sound:
  ∀ x, γ x ⊆ γ (oinit_iter x).
Proof. intros. eapply botbind_spec, H. intros. apply get_reduced_correct, H0. Qed.

(* TODO : better widenning strategy. *)
(* TODO : widenning with noise. *)
Definition do_widen (x:t0) (y:t0) : t0.
  refine (Too.shcombine
            (fun _ (x0 y0:option fin_float) =>
               match x0, y0 with
               | Some x, Some y => if Fleb (proj1_sig y) (proj1_sig x) then x0 else None
               | _, _ => None
               end
         ) _ x y).
Proof.
  abstract (destruct v; auto; rewrite Fleb_refl; auto; apply is_finite_not_is_nan, proj2_sig).
Defined.

Lemma do_widen_incr:
  ∀ x y, γ x ⊆ γ (do_widen x y).
Proof.
  intros x y ρ Hρ xv yv b Hb. unfold do_widen in Hb. rewrite Too.gshcombine in Hb.
  specialize (Hρ xv yv). destruct Too.get. 2:discriminate.
  destruct Too.get. 2:discriminate.
  destruct (Hρ _ eq_refl) as (xvv & yvv & Hxvv & Hyvv & Hb').
  rewrite Fleb_Rle in Hb by apply proj2_sig. Rle_bool_case Hb; inv Hb. eauto.
Qed.

Definition widen_reduce_1 (y0 y:t0) : t0 :=
  (To.fold
     (fun (y:t0) c _ =>
        match To.get (ovar_opp c) y, To.get c y with
        | None, _ | _, None => y
        | Some aa, Some bb =>
          Toop.map_changed fin_float_eqdec
             (fun ab yb =>
                let '(a, b) := ab in
                let bound :=
                  do_opt f1 <- To.get (ovar_opp a) aa;
                  do_opt f2 <- To.get b bb;
                  fin_float_of_float
                    (Bplus 53 1024 eq_refl eq_refl Float.binop_pl mode_UP
                           (proj1_sig f1) (proj1_sig f2))
                in
                fst (best_bound yb bound))
             y0 y
        end
     ) y y).

Lemma widen_reduce_1_correct:
  ∀ y0 y ρ, ρ ∈ γ y -> ρ ∈ γ (widen_reduce_1 y0 y).
Proof.
  intros y0 y ρ. unfold widen_reduce_1. rewrite To.fold_spec.
  generalize y at 1 3. induction (To.elements y) as [|[c ?]]. now auto.
  intros. apply IHl. unfold fst at 1 2.
  destruct (To.get (ovar_opp c) y1) as [aa|] eqn:EQaa. 2:now auto.
  destruct (To.get c y1) as [bb|] eqn:EQbb. 2:now auto.
  intros a b bd Hbd.
  edestruct Toop.map_changed_correct as [[]|[]];
    (eapply eq_trans in Hbd; [|symmetry; apply H1]).
  eapply H; now eauto.
  revert bd Hbd. eapply best_bound_either. now apply H.
  destruct (To.get (ovar_opp a) aa) as [aa'|] eqn:EQaa'. 2:discriminate.
  destruct (To.get b bb) as [bb'|] eqn:EQbb'. 2:discriminate.
  intros. apply fin_float_of_float_correct in Hbd. destruct Hbd.
  assert (Too.get (ovar_opp c, ovar_opp a) y1 = Some aa').
  { unfold Too.get. rewrite EQaa, EQaa'; auto. }
  eapply H in H4; eauto using ovar_opp_opp.
  destruct H4 as (xc & xa & Hxc & Hxa & Baa'). rewrite <- ovar_opp_opp in Hxc, Hxa.
  assert (Too.get (c, b) y1 = Some bb').
  { unfold Too.get. rewrite EQbb, EQbb'; auto. }
  eapply H in H4; eauto using ovar_opp_opp.
  destruct H4 as (xc' & xb & Hxc' & Hxb & Bbb').
  eexists. eexists. split. eauto. split. eauto.
  rewrite H3.
  apply reduct_rule_1 with (y:=xc'%R). rewrite Hxc in Hxc'. inv Hxc'. fourier. fourier. auto.
Qed.

Definition widen_reduce_2 (y0 y:t0) : t0 :=
  Toop.map_changed fin_float_eqdec
     (fun v1v2 yb =>
        let '(v1, v2) := v1v2 in
        let bound :=
          do_opt f1 <- Too.get (v1, ovar_opp v1) y;
          do_opt f2 <- Too.get (ovar_opp v2, v2) y;
          let b := Bplus 53 1024 eq_refl eq_refl Float.binop_pl mode_UP
                         (proj1_sig f1) (proj1_sig f2) in
          let b := Bmult 53 1024 eq_refl eq_refl Float.binop_pl mode_UP
                         half b in
          fin_float_of_float b
        in
        fst (best_bound yb bound))
     y0 y.

Lemma widen_reduce_2_correct:
  ∀ y0 y ρ, ρ ∈ γ y -> ρ ∈ γ (widen_reduce_2 y0 y).
Proof.
  intros y0 y ρ Hρ a b bd Hbd. unfold widen_reduce_2 in Hbd.
  edestruct Toop.map_changed_correct as [[]|[]];
    (eapply eq_trans in Hbd; [|symmetry; apply H0]).
  eapply Hρ; now eauto. clear H0.
  pose proof (Hρ a (ovar_opp a)). pose proof (Hρ (ovar_opp b) b).
  revert Hbd. apply best_bound_either. apply Hρ.
  unfold To.elt, ovar_EQ.s in *. clear H. destruct Too.get, Too.get; try discriminate.
  specialize (H0 _ eq_refl). destruct H0 as (xa & mxa & Hxa & Hmxa & H2xa).
  specialize (H1 _ eq_refl). destruct H1 as (xb & mxb & Hmxb & Hxb & H2xb).
  intros. apply fin_float_of_float_correct in Hbd. destruct Hbd.
  eexists. eexists. split. eauto. split. eauto. rewrite H0.
  apply reduct_rule_2. 3:now auto.
  - rewrite <- ovar_opp_opp, Hxa in Hmxa. inv Hmxa. Psatz.lra.
  - rewrite <- ovar_opp_opp, Hxb in Hmxb. inv Hmxb. Psatz.lra.
Qed.

(* TODO: stay in sync with the interval domain ? *)
Definition owiden (x:iter_t) (y:(t * query_chan)+⊥) (c:query_chan+⊥) :
  iter_t * (t * messages_chan) +⊥ :=
  match y with
  | Bot =>
    match x with
    | Bot => (Bot, Bot)
    | NotBot x =>
      (NotBot x,
       let w := widen_reduce_1 (Too.empty _) x in
       do_bot w <- reduce_diag (Too.empty _) w;
       NotBot (t_of_t0 w, List.nil))
    end
  | NotBot y =>
    let y := get_reduced (fst y) in
    match x with
    | Bot =>
      (NotBot y, NotBot (t_of_t0_reduced y, List.nil))
    | NotBot x =>
      let w0 := do_widen x y in
      (NotBot w0,
       let w := widen_reduce_1 y w0 in
       do_bot w <- reduce_diag y w;
       NotBot (existT _ w (cThunk _ (widen_reduce_2 y w)
                                  (widen_reduce_2_correct _ _)),
               List.nil))
    end
  end.

Lemma owiden_incr:
  forall x y c ρ,
    ρ ∈ γ x -> ρ ∈ γ c ->
    ρ ∈ γ (owiden x y c).
Proof.
  intros. unfold widen. destruct x, y; try contradiction.
  - split. auto. eapply botbind_spec, reduce_diag_correct, widen_reduce_1_correct, H.
    split. apply t_of_t0_correct, H1. constructor.
  - split. apply do_widen_incr. auto.
    eapply botbind_spec, reduce_diag_correct, widen_reduce_1_correct, do_widen_incr, H.
    split. apply H1. constructor.
Qed.

(******** Forget ********)

Definition oforget_t0 (x:var) (ab:t0) : t0 :=
  let ab := To.remove (ovPlus x) ab in
  let ab := To.remove (ovMinus x) ab in
  To.map1
    (fun ab =>
       let ab := To.remove (ovPlus x) ab in
       To.remove (ovMinus x) ab)
    ab.

Lemma oforget_t0_incl:
  ∀ x ab, γ ab ⊆ γ (oforget_t0 x ab).
Proof.
  repeat intro. unfold oforget_t0, Too.get in H0.
  rewrite To.gmap1, !To.grspec in H0.
  destruct To.elt_eq. discriminate.
  destruct To.elt_eq. discriminate.
  destruct (To.get x0 ab) eqn:EQ. 2:discriminate.
  simpl in H0. rewrite !To.grspec in H0.
  destruct To.elt_eq. discriminate.
  destruct To.elt_eq. discriminate.
  apply H. simpl. rewrite EQ. auto.
Qed.

Lemma oforget_t0_forget:
  ∀ x ab ρ n,
    ρ ∈ γ (oforget_t0 x ab) ->
    ρ+[x => n] ∈ γ (oforget_t0 x ab).
Proof.
  repeat intro. specialize (H _ _ _ H0).
  destruct H as (xb & yb & Hxb & Hyb & Hxyb).
  exists xb, yb. split. 2:split. 3:now auto.
  - unfold oρ, rρ.
    destruct x0; simpl; rewrite upd_o; auto; unfold oforget_t0, Too.get in H0; intro; subst;
    rewrite To.gmap1 in H0; [rewrite To.gro in H0 by discriminate|]; rewrite To.grs in H0;
    discriminate.
  - unfold oρ, rρ.
    destruct y; simpl; rewrite upd_o; auto; unfold oforget_t0, Too.get in H0; intro; subst;
    rewrite To.gmap1 in H0; destruct To.get; try discriminate; simpl in H0;
    [rewrite To.gro in H0 by discriminate|]; rewrite To.grs in H0; discriminate.
Qed.

(* reduce_2 cannot bring anything before a forget that we cannot compute after
   a forget. Thus, we use get_unreduced. *)
Definition oforget (x:var) (ab:t * query_chan) (c:query_chan) :
  (t * messages_chan)+⊥ :=
  NotBot (t_of_t0 (oforget_t0 x (get_unreduced (fst ab))), List.nil).

Lemma oforget_correct:
  forall x ρ n ab chan,
    ρ ∈ γ ab -> (ρ+[x => n]) ∈ γ chan ->
    (ρ+[x => n]) ∈ γ (oforget x ab chan).
Proof.
  intros. split. 2:constructor. apply oforget_t0_forget, oforget_t0_incl, H.
Qed.

(******** Assume - process messages ********)

Definition add_zerocycle (x:ovar) (ab:t0) : t0 :=
  match Too.get_set (x, x) ab with
  | (None, setter) => setter (exist _ Float.zero eq_refl)
  | (Some v, setter) =>
    if Fleb Float.zero (proj1_sig v) then setter (exist _ Float.zero eq_refl)
    else ab
  end.

Lemma add_zerocycle_correct:
  ∀ x ab ρ xv,
    ρ ∈ γ ab ->
    oρ ρ x = Some xv ->
    ρ ∈ γ (add_zerocycle x ab).
Proof.
  repeat intro. unfold add_zerocycle in H1.
  destruct (Too.get_set_spec (x, x) ab). destruct Too.get_set.
  unfold fst, snd in H2, H3. subst o. rewrite H3 in H1. clear H3.
  destruct (Too.get (x, x) ab); [destruct @Fleb|]; rewrite ?Too.gsspec in H1;
  try (destruct Too.elt_eq; [inv e|]); inv H1; auto; simpl; change (B2R 53 1024 Float.zero) with 0%R.
  - eexists. eexists. split. eauto. split. eauto. fourier.
  - eexists. eexists. split. eauto. split. eauto. fourier.
Qed.

Definition constraint_row := To.t fin_float.

Instance cr_tostring {_:ToString Cells.cell} : ToString constraint_row :=
  fun x =>
    (To.fold (fun acc y b =>
       acc ++ "[_" ++ to_string (ovar_opp y) ++ "<=" ++ to_string (proj1_sig b)++"]")
       x "{" ++ "}")%string.

Definition constraints_row_valid (row:constraint_row) (ρ:var -> ideal_num) : ℘ R :=
  fun v =>
  ∀ x b, To.get x row = Some b ->
  ∃ xv, oρ ρ x = Some xv /\ (v-xv <= B2R _ _ (proj1_sig b))%R.

(* Just for optimization: keep only the new constraints. *)
Definition trim_constraints_row (a:ovar) (row:constraint_row) (ab:t0) : constraint_row :=
  let mapa :=
      match To.get a ab with
      | None => To.empty _
      | Some mapa => mapa
      end
  in
  To.combine
    (fun old new => if snd (best_bound old new) then None else new)
    mapa row.

Lemma trim_constraints_row_correct:
  ∀ a row ab ρ av,
    ρ ∈ γ ab ->
    constraints_row_valid row ρ av ->
    constraints_row_valid (trim_constraints_row a row ab) ρ av.
Proof.
  unfold constraints_row_valid, trim_constraints_row. intros.
  rewrite To.gcombine in H1 by auto. revert H1.
  apply best_bound_either; simpl. discriminate. auto.
Qed.

Definition reduce_constraints_row_step (row:constraint_row) (x:ovar) (ax:fin_float) (ab:t0):
  constraint_row :=
  match To.get x ab with
  | None => row
  | Some mapx =>
    To.combine
      (fun (ay xy:option fin_float) =>
         let ay' :=
             match xy with
             | None => None
             | Some xy =>
               fin_float_of_float (
                   Bplus 53 1024 eq_refl eq_refl Float.binop_pl mode_UP
                         (proj1_sig ax) (proj1_sig xy))
             end
         in
         fst (best_bound ay ay'))
      row mapx
  end.

Lemma reduce_constraints_row_step_correct:
  ∀ row x ax ab ρ v,
    ρ ∈ γ ab ->
    (∀ xv, oρ ρ x = Some xv -> (v - xv <= B2R _ _ (proj1_sig ax))%R) ->
    constraints_row_valid row ρ v ->
    constraints_row_valid (reduce_constraints_row_step row x ax ab) ρ v.
Proof.
  unfold constraints_row_valid, reduce_constraints_row_step.
  intros. specialize (H x x0). unfold Too.get in H.
  destruct (To.get x ab). 2:now auto.
  rewrite To.gcombine in H2 by auto. revert H2. apply best_bound_either.
  - intros. edestruct H1 as (xv0 & Hxv0 & Hxv0'). eauto. eauto.
  - intros. destruct To.get. 2:discriminate. edestruct H as (? & ? & ? & ? & ?). now eauto.
    apply fin_float_of_float_correct in H2. destruct H2 as (? & ->).
    eauto using reduct_rule_1.
Qed.

Definition reduce_constraints_row_final (row:constraint_row) (a:ovar) (ab:t0):
  constraint_row:=
  let best_aa' :=
    T.fold (fun bnd z azaz' =>
              match azaz' with
              | (Some az, Some az') =>
                fst (best_bound bnd (
                  fin_float_of_float (
                      Bplus 53 1024 eq_refl eq_refl Float.binop_pl mode_UP
                            (proj1_sig az) (proj1_sig az'))))
              | _ => bnd
              end)
    row (To.get (ovar_opp a) row)
  in
  match best_aa' with
  | None => row
  | Some best_aa' =>
    let row := To.set (ovar_opp a) best_aa' row in
    reduce_constraints_row_step row (ovar_opp a) best_aa' ab
  end.

Lemma reduce_constraints_row_final_correct:
  ∀ row a ab ρ av,
    ρ ∈ γ ab ->
    oρ ρ a = Some av ->
    constraints_row_valid row ρ av ->
    constraints_row_valid (reduce_constraints_row_final row a ab) ρ av.
Proof.
  unfold constraints_row_valid, reduce_constraints_row_final. intros.
  match type of H2 with
  | To.get x match ?BESTAA' with _ => _ end = _ =>
    assert (∀ best_aa', BESTAA' = Some best_aa' -> av+av <= B2R _ _ (proj1_sig best_aa'))%R;
      [|destruct BESTAA']
  end.
  { clear H2. rewrite T.fold_spec with (A:=SUU.t fin_float).
    pose proof T.elements_complete row.
    revert H2. pose proof (H1 (ovar_opp a)). revert H2.
    generalize (To.get (ovar_opp a) row).
    induction (T.elements row) as [|[?[]] ?]; simpl; intros.
    - subst. destruct (H2 _ eq_refl) as (? & ? & ?).
      apply ovar_opp_opp in H4. rewrite H0 in H4. inv H4. Psatz.lra.
    - eapply IHl, H4. 2:now eauto.
      intros. destruct t1, t2; try now (subst; eauto).
      revert H5. apply best_bound_either; simpl; intro. now (subst; eauto).
      apply fin_float_of_float_correct in H5. destruct H5 as (? & ->).
      specialize (H3 _ _ (or_introl eq_refl)).
      pose proof (H1 (ovMinus e)). specialize (H1 (ovPlus e)).
      unfold To.get in H1, H6. simpl in H1, H6.
      rewrite H3 in H1, H6. edestruct H1 as (? & ? & ?), H6 as (? & ? & ?); try reflexivity.
      rewrite H7 in H9. inv H9.
      eexists. split. apply -> ovar_opp_opp. rewrite Ropp_involutive. now eauto.
      eapply reduct_rule_1; eauto. Psatz.lra. }
  2:now auto.
  specialize (H3 _ eq_refl).
  eapply reduce_constraints_row_step_correct. now eauto. 3:now eauto.
  - intros. rewrite <- ovar_opp_opp, H0 in H4. inv H4. now auto.
  - repeat intro. rewrite To.gsspec in H4. destruct To.elt_eq. 2:now auto.
    inv H4. eexists. rewrite <- ovar_opp_opp, H0, Ropp_involutive. split. eauto. Psatz.lra.
Qed.

Definition reduce_constraints_row (row:constraint_row) (a:ovar) (ab:t0) : constraint_row :=
  let row :=
    To.fold (fun (acc:constraint_row) (x:ovar) (ax:fin_float) =>
               reduce_constraints_row_step acc x ax ab)
      row row
  in
  reduce_constraints_row_final row a ab.

Lemma reduce_constraints_row_correct:
  ∀ row a ab ρ v,
    ρ ∈ γ ab ->
    oρ ρ a = Some v ->
    constraints_row_valid row ρ v ->
    constraints_row_valid (reduce_constraints_row row a ab) ρ v.
Proof.
  intros. apply reduce_constraints_row_final_correct; auto.
  rewrite To.fold_spec.
  generalize H1. generalize row at 1 3.
  specialize (fun i v H => H1 _ _ (To.elements_complete _ i v H)).
  induction To.elements as [|[x ax]]. now auto.
  intros. eapply IHl. intros. apply H1. right. now auto.
  edestruct H1 as (? & ? & ?). left. reflexivity.
  eapply reduce_constraints_row_step_correct; eauto.
  simpl. intros. assert (xv=x0) by congruence. subst. now auto.
Qed.

Definition add_constraints_row_noreduce (a:ovar) (row:constraint_row) (ab:t0) : t0 :=
  let ab := add_zerocycle a ab  in
  let ab := add_zerocycle (ovar_opp a) ab  in
  match To.get (ovar_opp a) ab with
  | None => ab (* Impossible *)
  | Some map_ya =>
    let ab :=
      To.fold (fun acc yopp (ya:fin_float) =>
        match To.get_set (ovar_opp yopp) acc with
        | (None, _) => acc
        | (Some map_yx, setter_yx) =>
          let '(map_yx, chgt) :=
            To.fold (fun (acc':To.t fin_float * bool) x (ax:fin_float) =>
              let '(map_yx, chgt) := acc' in
              let new :=
                fin_float_of_float (
                  Bplus 53 1024 eq_refl eq_refl Float.binop_pl mode_UP
                    (proj1_sig ya) (proj1_sig ax))
              in
              let (old, setter) := To.get_set x map_yx in
              if snd (best_bound old new) then acc'
              else match new with
                   | None => acc' (* Impossible *)
                   | Some new => (setter new, true)
                   end) row (map_yx, false)
          in
          if chgt then (setter_yx map_yx) else acc
        end) map_ya ab
    in
    match To.get (ovar_opp a) ab with
    | None => ab (* Impossible *)
    | Some map_ya =>
      To.fold (fun acc x (ax:fin_float) =>
        match To.get_set (ovar_opp x) acc with
        | (None, _) => acc
        | (Some map_xoppyopp, setter_xoppyopp) =>
          let '(map_xoppyopp, chgt) :=
            To.fold (fun (acc':To.t fin_float * bool) yopp (ya:fin_float) =>
              let '(map_xoppyopp, chgt) := acc' in
              let new :=
                fin_float_of_float (
                  Bplus 53 1024 eq_refl eq_refl Float.binop_pl mode_UP
                    (proj1_sig ya) (proj1_sig ax))
              in
              let (old, setter) := To.get_set yopp map_xoppyopp in
              if snd (best_bound old new) then acc'
              else match new with
                   | None => acc' (* Impossible *)
                   | Some new => (setter new, true)
                   end) map_ya (map_xoppyopp, false)
          in
          if chgt then (setter_xoppyopp map_xoppyopp) else acc
        end) row ab
    end
  end.

Lemma add_constraints_row_noreduce_correct:
  ∀ a row ab ρ av,
    oρ ρ a = Some av -> ρ ∈ γ ab ->
    constraints_row_valid row ρ av ->
    ρ ∈ γ (add_constraints_row_noreduce a row ab).
Proof.
  intros.
  unfold constraints_row_valid in H1. unfold add_constraints_row_noreduce.
  eapply add_zerocycle_correct in H0. 2:now eauto.
  eapply add_zerocycle_correct in H0.
  2:apply -> ovar_opp_opp; etransitivity; [apply H|]. 2:f_equal; symmetry; apply Ropp_involutive.
  revert H0 H1. generalize (add_zerocycle (ovar_opp a) (add_zerocycle a ab)) as ab'. clear ab. intros.
  destruct (To.get (ovar_opp a) ab') as [map_ya|] eqn:EQmap_ya. 2:now auto.
  assert (∀ yopp ya, List.In (yopp, ya) (To.elements map_ya) ->
          ∃ yv, oρ ρ yopp = Some yv ∧ (- yv - av <= B2R 53 1024 (proj1_sig ya))%R).
  { intros. specialize (H0 (ovar_opp a)). unfold Too.get in H0.
    rewrite EQmap_ya in H0. specialize (H0 _ _ (To.elements_complete _ _ _ H2)).
    destruct H0 as (? & ? & ? & ? & ?). eexists.
    split. eauto. rewrite <- ovar_opp_opp, H in H0. inv H0. Psatz.lra. }
  specialize (fun x ax H => H1 _ _ (To.elements_complete _ x ax H)). clear EQmap_ya.
  match goal with
  | |- context [To.fold ?F map_ya ab'] =>
    set (ab'':=To.fold F map_ya ab'); assert (ρ ∈ γ ab'')
  end.
  { clear H. subst ab''. rewrite To.fold_spec. revert ab' H0.
    induction (To.elements map_ya) as [|[]]; simpl; intros. now auto.
    apply IHl. intros. apply H2. right. auto. clear IHl.
    destruct (To.get_set_spec (ovar_opp e) ab'), To.get_set.
    destruct o. 2:now auto.
    pose proof (H0 (ovar_opp e)). unfold Too.get in H4.
    rewrite <- H in H4. clear H. simpl in H4. rewrite To.fold_spec.
    pattern t2 in H4.
    match goal with
    | H4 : ?F t2 |- γ match ?FOLD with _ => _ end ρ =>
      assert (F (fst FOLD)); [|destruct FOLD as [?[]]]
    end.
    { clear H3. change t2 with (fst (t2, false)) in H4. revert H4 H1.
      generalize (t2, false).
      induction (To.elements row) as [|[]]; simpl; intros. now auto.
      apply IHl0 in H. now auto. 2:now auto. clear IHl0. destruct p. intros.
      destruct (To.get_set_spec e0 t3), To.get_set.
      revert H3. apply best_bound_either; simpl; intros. now auto.
      match type of H3 with
      | context [match ?F with _ => _ end] => destruct F eqn:EQ
      end. 2:now auto.
      simpl in H3, H6. rewrite H6, To.gsspec in H3. destruct To.elt_eq. inv H3. 2:now auto.
      apply fin_float_of_float_correct in EQ. destruct EQ as [? ->].
      edestruct H1 as (? & ? & ?). now eauto.
      edestruct H2 as (? & ? & ?). left. now eauto.
      eexists. eexists. split. rewrite <- ovar_opp_opp, H9, Ropp_involutive. reflexivity.
      split. eauto. eapply reduct_rule_1; eauto. }
    2:now auto.
    rewrite H3. repeat intro. unfold Too.get in H5. rewrite To.gsspec in H5.
    destruct To.elt_eq. subst. now auto. apply H0, H5. }
  revert H3. generalize ab''. clear ab'' ab' H0. intros.
  destruct (To.get (ovar_opp a) ab'') eqn:EQ. 2:now auto.
  pose proof H3 (ovar_opp a). unfold Too.get in H0. rewrite EQ in H0. clear EQ.
  rewrite To.fold_spec. revert H1 ab'' H3.
  induction (To.elements row) as [|[]]; simpl; intros. now auto.
  apply IHl. now auto. destruct (To.get_set_spec (ovar_opp e) ab''), To.get_set as [[]?].
  clear IHl. 2:now auto.
  pose proof H3 (ovar_opp e). unfold Too.get in H6. rewrite <- H4 in H6. clear H4. simpl in H6.
  pattern t2 in H6.
  match goal with
  | H6 : ?F t2 |- γ match ?FOLD with _ => _ end ρ =>
    assert (F (fst FOLD)); [|destruct FOLD as [?[]]]
  end.
  { change t2 with (fst (t2, false)) in H6. rewrite To.fold_spec.
    specialize (fun y b H => H0 _ _ (To.elements_complete _ y b H)). revert H0.
    revert H6. generalize (t2, false).
    induction (To.elements t1) as [|[]]; simpl; intros. now auto. destruct p.
    eapply IHl0, H4. 2:now auto. intros y0 b1.
    destruct (To.get_set_spec e0 t4), To.get_set. apply best_bound_either. now auto.
    simpl. intro.
    match type of H9 with
    | context [match ?F with _ => _ end] => destruct F eqn:EQ
    end. 2:now auto. simpl in H9. rewrite H8, To.gsspec in H9.
    destruct To.elt_eq. 2:now auto. inv H9.
    apply fin_float_of_float_correct in EQ. destruct EQ as [? ->].
    edestruct H1 as (? & ? & ?). eauto.
    edestruct H0 as (? & ? & ? & ? & ?). eauto.
    eexists. eexists. split. rewrite <- ovar_opp_opp, H10, Ropp_involutive. reflexivity.
    split. now eauto. replace (-x-x1)%R with (-x1-x)%R by ring.
    rewrite <- ovar_opp_opp, H in H12. inv H12.
    eapply reduct_rule_1; eauto. Psatz.lra. }
  2:now auto. rewrite H5.
  repeat intro. unfold Too.get in H7. rewrite To.gsspec in H7.
  destruct To.elt_eq. 2:now auto. subst. now auto.
Qed.

Definition add_constraints_row (a:ovar) (row:constraint_row) (ab:t0) : t0 :=
  let row := trim_constraints_row a row ab in
  let row := reduce_constraints_row row a ab in
  let row := trim_constraints_row a row ab in
  add_constraints_row_noreduce a row ab.

Lemma add_constraints_row_correct :
  ∀ a row ab ρ av,
    oρ ρ a = Some av -> ρ ∈ γ ab ->
    constraints_row_valid row ρ av ->
    ρ ∈ γ (add_constraints_row a row ab).
Proof.
  intros. eapply add_constraints_row_noreduce_correct; eauto.
  apply trim_constraints_row_correct,
        reduce_constraints_row_correct, trim_constraints_row_correct; eauto.
Qed.

(* We assume y had no constrint before. *)
Definition add_constraint_to_row (r:constraint_row) (y:ovar) (b:float) : constraint_row :=
  match fin_float_of_float b with
  | None => r
  | Some b => To.set y b r
  end.

Lemma add_constraint_to_row_correct:
  ∀ row b ρ y yv v,
    oρ ρ y = Some yv ->
    (RF_le (v - yv) b)%R ->
    constraints_row_valid row ρ v ->
    constraints_row_valid (add_constraint_to_row row y b) ρ v.
Proof.
  intros. unfold add_constraint_to_row. repeat intro.
  pose proof fin_float_of_float_correct b.
  destruct fin_float_of_float. 2:now auto. edestruct H3. now eauto.
  rewrite To.gsspec in H2. destruct To.elt_eq. 2:now auto. inv H2.
  unfold RF_le in H0. rewrite H4 in H0. now eauto.
Qed.

(* We know x = i+terms. *)
(* What are the constraints such as x-_<=_ or -x-_ <= _? *)
(* What is an interval for terms ? *)
(* We assume x does not appear in terms, and that terms has not repeats. *)
(* TODO? : if needed: optimization when i = top *)
Fixpoint constraints_row_of_linear_expr (i:itv) (terms:list (var * (itv * itv))) :
  constraint_row * constraint_row * itv :=
  match terms with
  | nil => (To.empty _, To.empty _, ITV Float.zero Float.zero)
  | (y, (itvα, itvy)) :: q =>
    let iαy := mult_itv itvα itvy in
    let '(constr_p, constr_m, iq) :=
        constraints_row_of_linear_expr (add_itv iαy i) q
    in
    (* x-y =    i + q + (α-1)*y *)
    (* -x+y = -(i + q + (α-1)*y) *)
    (* x+y =    i + q + (α+1)*y *)
    (* -x-y = -(i + q + (α+1)*y) *)
    let i := add_itv iq i in
    let 'ITV xpy_a xpy_b :=
        add_itv (mult_itv (add_itv itvα (ITV one one)) itvy) i in
    let mone := Float.neg one in
    let 'ITV xmy_a xmy_b :=
        add_itv (mult_itv (add_itv itvα (ITV mone mone)) itvy) i in
    let constr_p := add_constraint_to_row constr_p (ovPlus y) xmy_b in
    let constr_m := add_constraint_to_row constr_m (ovMinus y) (Float.neg xmy_a) in
    let constr_p := add_constraint_to_row constr_p (ovMinus y) xpy_b in
    let constr_m := add_constraint_to_row constr_m (ovPlus y) (Float.neg xpy_a) in
    (constr_p, constr_m, add_itv iαy iq)
  end.

Transparent Float.zero Float.neg.

Lemma constraints_row_of_linear_expr_correct:
  ∀ terms x i constr_p constr_m iterms ρ,
    eval_linear_expr ρ (i, terms) x ->
    constraints_row_of_linear_expr i terms = (constr_p, constr_m, iterms) ->
    constraints_row_valid constr_p ρ x /\
    constraints_row_valid constr_m ρ (-x) /\
    eval_linear_expr ρ (ITV Float.zero Float.zero, terms) ⊆ γ iterms.
Proof.
  induction terms as [|[y [itvα itvy]]]; simpl; intros.
  - inv H0. split. 2:split.
    repeat intro. rewrite To.gempty in H0. discriminate.
    repeat intro. rewrite To.gempty in H0. discriminate.
    now eauto.
  - specialize (IHterms x (add_itv (mult_itv itvα itvy) i)).
    destruct constraints_row_of_linear_expr as [[constr_p0 constr_m0] iq].
    specialize (IHterms constr_p0 constr_m0 iq ρ).
    unfold eval_linear_expr in H. simpl in H.
    destruct (rρ ρ y) eqn:EQy; destruct H. destruct H as (α & Hα & FOLD).
    assert (eval_linear_expr ρ (add_itv (mult_itv itvα itvy) i, terms) x).
    { apply eval_linear_expr_change_cst in FOLD. destruct FOLD as (? & ? & ?).
      replace x with (x - α * r + (α * r + x0) - x0)%R by ring. apply H2.
      apply add_itv_correct. apply mult_itv_correct; now auto. now auto. }
    edestruct IHterms as (? & ? & ?). now auto. now auto. clear IHterms.
    assert (∃ x', x' ∈ γ i /\
                  eval_linear_expr ρ (ITV Float.zero Float.zero, terms) (x-x'-α*r)).
    { apply eval_linear_expr_change_cst in FOLD. destruct FOLD as (? & ? & ?).
      eexists. split. now eauto. replace (x - x0 - α * r)%R with (x - α * r + 0 - x0)%R by ring.
      apply H6. compute; Psatz.lra. }
    clear FOLD.
    destruct H5 as (? & ? & ?). pose proof (H4 _ H6).
    match type of H0 with
    | (let 'ITV _ _ := ?X in _) = _ =>
      refine ((fun H:(_ ∈ γ X) => _) _); [destruct X|]
    end.
    2:apply add_itv_correct; [apply mult_itv_correct;
      [apply add_itv_correct|]|apply add_itv_correct]; eauto.
    2:instantiate (1:=R1); compute; Psatz.lra.
    match type of H0 with
    | (let 'ITV _ _ := ?X in _) = _ =>
      refine ((fun H:(_ ∈ γ X) => _) _); [destruct X|]
    end.
    2:apply add_itv_correct; [apply mult_itv_correct;
      [apply add_itv_correct|]|apply add_itv_correct]; eauto.
    2:instantiate (1:=(-1)%R); compute; Psatz.lra.
    inv H0.
    replace ((α + 1) * r + (x - x0 - α * r + x0))%R with (x+r)%R in H8 by ring.
    replace ((α + - 1) * r + (x - x0 - α * r + x0))%R with (x-r)%R in H9 by ring.
    destruct H8, H9.
    split. 2:split.
    + eapply add_constraint_to_row_correct. simpl. rewrite EQy. reflexivity.
      replace (x- -r)%R with (x+r)%R by ring. auto.
      eapply add_constraint_to_row_correct; eauto.
    + eapply add_constraint_to_row_correct. eauto.
      replace (-x -r)%R with (-(x+r))%R by ring. auto.
      eapply add_constraint_to_row_correct. simpl. rewrite EQy. reflexivity.
      replace (-x - - r)%R with (-(x-r))%R by ring. auto. auto.
    + intros. red in H11. simpl in H11. rewrite EQy in H11.
      destruct H11 as [(? & ? & ?) ?].
      replace a with (x1*r+(a-x1*r))%R by ring.
      apply add_itv_correct; auto. apply mult_itv_correct; auto.
Qed.

Definition meet_constraints_row (x y:constraint_row) : constraint_row :=
  To.combine (fun x y => fst (best_bound x y)) x y.

Lemma meet_constraints_row_correct:
  ∀ x a b ρ,
    constraints_row_valid a ρ x ->
    constraints_row_valid b ρ x ->
    constraints_row_valid (meet_constraints_row a b) ρ x.
Proof.
  repeat intro. revert H1.
  unfold meet_constraints_row. rewrite To.gcombine by auto.
  apply best_bound_either; now auto.
Qed.

(* Add the constraint i + terms = 0 to the octagon. *)
Fixpoint assume_linear_constraint (i:itv) (terms:list (var * (itv * itv))) (ab:t0) : t0 :=
  match terms with
  | nil => ab
  | (x, (itvα, itvx)) :: q =>
    if itv_is_top i then ab
    else
      let ip := add_itv (mult_itv (add_itv itvα (ITV one one)) itvx) i in
      let '(constr_p0, constr_m0, iq) := constraints_row_of_linear_expr ip q in
      let ipq := add_itv ip iq in
      let 'ITV lx2 ux2 := add_itv ipq ipq in
      let constr_p0 := add_constraint_to_row constr_p0 (ovMinus x) ux2 in
      let constr_m0 := add_constraint_to_row constr_m0 (ovPlus x) (Float.neg lx2) in

      let im := add_itv (mult_itv (add_itv (opp_itv itvα) (ITV one one)) itvx)
                        (opp_itv i)
      in
      let qm := List.map (fun t =>
        let '(k, (itvα, itvk)) := t in (k, (opp_itv itvα, itvk))) q
      in
      let '(constr_p1, constr_m1, iq) := constraints_row_of_linear_expr im qm in
      let imq := add_itv im iq in
      let 'ITV lx2 ux2 := add_itv imq imq in
      let constr_p1 := add_constraint_to_row constr_p1 (ovMinus x) ux2 in
      let constr_m1 := add_constraint_to_row constr_m1 (ovPlus x) (Float.neg lx2) in

      let constr_p := meet_constraints_row constr_p0 constr_p1 in
      let ab := add_constraints_row (ovPlus x) constr_p ab in
      let constr_m := meet_constraints_row constr_m0 constr_m1 in
      let ab := add_constraints_row (ovMinus x) constr_m ab in

      let i := add_itv (mult_itv itvα itvx) i in
      assume_linear_constraint i q ab
  end.

Lemma assume_linear_constraint_correct:
  ∀ terms i ab ρ,
    ρ ∈ γ ab ->
    eval_linear_expr ρ (i, terms) 0%R ->
    ρ ∈ γ (assume_linear_constraint i terms ab).
Proof.
  induction terms as [|[x [itvα itvx]]]. now auto.
  intros. simpl. destruct itv_is_top. auto.
  unfold eval_linear_expr in H0. simpl in H0. destruct (rρ ρ x) eqn:EQx, H0.
  destruct H0 as (α & Hα & Hr).
  pose proof constraints_row_of_linear_expr_correct terms r
               (add_itv (mult_itv (add_itv itvα (ITV one one)) itvx) i).
  destruct constraints_row_of_linear_expr as [[constr_p0 constr_m0] ?].
  refine (let HH := H0 _ _ _ ρ _ eq_refl in _); clear H0.
  { apply eval_linear_expr_change_cst in Hr. destruct Hr as (? & ? & ?).
    replace r with (0-(α*r)+((α+1)*r+x0)-x0)%R by ring.
    apply H2, add_itv_correct, H0. apply mult_itv_correct, H1.
    apply add_itv_correct. apply Hα. compute; Psatz.lra. }
  destruct HH as (? & ? & ?).
  pose proof constraints_row_of_linear_expr_correct
       (map (fun t1 => let '(k, (itvα0, itvk)) := t1 in
                       (k, (opp_itv itvα0, itvk))) terms) r
       (add_itv
          (mult_itv (add_itv (opp_itv itvα) (ITV one one)) itvx)
          (opp_itv i)).
  destruct constraints_row_of_linear_expr as [[constr_p1 constr_m1] ?].
  refine (let HH := H4 _ _ _ ρ _ eq_refl in _); clear H4.
  { edestruct eval_linear_expr_change_cst as (? & ? & ?).
    2:replace r with (-(0-α*r)+((-α+1)*r+x0)-x0)%R by ring;
      apply H5, add_itv_correct, H4; apply mult_itv_correct, H1;
      apply add_itv_correct; [apply opp_itv_correct, Hα|compute; Psatz.lra].
    revert Hr. unfold eval_linear_expr. simpl. generalize (0-α*r)%R. clear.
    induction terms as [|[y [itvβ itvy]]]; simpl; intros. apply opp_itv_correct, Hr.
    destruct (rρ ρ y), Hr. split. 2:now auto.
    destruct H as (β & Hβ & Hr0). eexists. split. now apply opp_itv_correct, Hβ.
    apply IHterms in Hr0.
    replace (- r - - β * r0)%R with (- (r - β * r0))%R by ring. now auto. }
  destruct HH as (? & ? & ?).
  assert (∃ iv, iv ∈ γ i /\
                eval_linear_expr ρ (ITV Float.zero Float.zero, terms) (-α*r-iv)%R).
  { apply eval_linear_expr_change_cst in Hr. destruct Hr as (? & ? & ?). eexists. split. now eauto.
    replace (- α * r - x0)%R with (0 - α * r + 0 - x0)%R by ring. apply H8.
    compute; Psatz.lra. }
  destruct H7 as (iv & Hiv & EVAL).
  repeat match goal with
  | |- _ (let 'ITV lx2 ux2 := add_itv ?I ?I in _) _ =>
    assert (r ∈ γ I); [|destruct (add_itv I I) eqn:?]
  end.
  { replace r with ((α+1)*r + iv + (-α*r-iv))%R by ring.
    apply add_itv_correct. apply add_itv_correct. apply mult_itv_correct.
    apply add_itv_correct. now auto. compute; Psatz.lra. now auto. now auto. now auto. }
  { replace r with ((-α+1)*r - iv + (α*r+iv))%R by ring.
    apply add_itv_correct. apply add_itv_correct. apply mult_itv_correct.
    apply add_itv_correct. apply opp_itv_correct. now auto. compute; Psatz.lra.
    now auto. apply opp_itv_correct. now auto. apply H6. clear -EVAL.
    replace (-α*r-iv)%R with (-(α*r+iv))%R in EVAL by ring.
    revert EVAL. unfold eval_linear_expr. simpl. generalize (α * r + iv)%R.
    induction terms as [|[k[itvβ itvk]]]; simpl. compute; intros; Psatz.lra.
    destruct (rρ ρ k); intros ? []; intros. split. 2:now auto.
    destruct H as (α0 & Hα0 & FOLD). eexists. split. apply opp_itv_correct, Hα0.
    apply IHterms. replace (-(r1 - - α0 * r0))%R with (- r1 - α0 * r0)%R by ring. now auto. }
  assert ((r + r) ∈ γ (ITV f f0))%R.
  { rewrite <- Heqi2. apply add_itv_correct; auto. }
  assert ((r + r) ∈ γ (ITV f1 f2))%R.
  { rewrite <- Heqi0. apply add_itv_correct; auto. }
  assert (eval_linear_expr ρ (add_itv (mult_itv itvα itvx) i, terms) 0).
  { apply eval_linear_expr_change_cst in Hr. destruct Hr as (? & ? & ?).
    replace 0%R with (0 - α * r + (α*r+x0) - x0)%R by ring.
    auto using add_itv_correct, mult_itv_correct. }
  apply IHterms, H11.
  eapply add_constraints_row_correct, meet_constraints_row_correct; eauto.
  simpl. rewrite EQx. reflexivity.
  eapply add_constraints_row_correct, meet_constraints_row_correct; eauto.
  - eapply add_constraint_to_row_correct. simpl. rewrite EQx. reflexivity.
    replace (r - - r)%R with (r + r)%R by ring. apply H9. now auto.
  - eapply add_constraint_to_row_correct. simpl. rewrite EQx. reflexivity.
    replace (r - - r)%R with (r + r)%R by ring. apply H10. now auto.
  - eapply add_constraint_to_row_correct. apply EQx.
    replace (- r - r)%R with (-(r + r))%R by ring. apply H9. now auto.
  - eapply add_constraint_to_row_correct. apply EQx.
    replace (- r - r)%R with (-(r + r))%R by ring. apply H10. now auto.
Qed.

Definition msg_of_bnds
           (a b:option fin_float) (x:var)
           (ty:ideal_num_ty) (itv:(IdealIntervals.abs +⊤) +⊥) : message+⊤+⊥ :=
  let a :=
    match a with
    | None => B754_infinity _ _ true
    | Some a =>
      let a := proj1_sig a in
      let a := Bmult 53 1024 eq_refl eq_refl Float.binop_pl mode_UP half a in
      Float.neg a
    end
  in
  let b :=
    match b with
    | None => B754_infinity _ _ false
    | Some b =>
      let b := proj1_sig b in
      Bmult 53 1024 eq_refl eq_refl Float.binop_pl mode_UP half b
    end
  in
  do_bot itv' <- itv_of_itvabs itv ⊓ Just (ITV a b);
  let itv' := NotBot (toplift_bind (itvabs_of_itv ty) itv') in
  if itv ⊑ itv' then NotBot All
  else match itv' ⊓ itv with
       | NotBot All => NotBot All
       | Bot => Bot
       | NotBot (Just itv') =>
         NotBot (Just (Broadcasted_msg (
           match IdealIntervals.singleton itv' with
           | Some v => Known_value_msg x v
           | None => Itv_msg x itv'
           end)))
       end.

Lemma msg_of_bnds_correct:
  ∀ a b x ty itv ρ xv,
    ρ x ∈ γ ty ->
    ρ x ∈ γ itv ->
    rρ ρ x = Some xv ->
    (∀ a', a = Some a' -> -2*xv <= B2R _ _ (proj1_sig a'))%R ->
    (∀ b', b = Some b' -> 2*xv <= B2R _ _ (proj1_sig b'))%R ->
    ρ ∈ γ (msg_of_bnds a b x ty itv).
Proof.
  intros. unfold msg_of_bnds. eapply botbind_spec with (rρ ρ x), @meet_correct.
  2:exact _. 
  - intros. destruct (leb itv). constructor.
    match goal with
    | |- γ match ?X with _ => _ end _ => assert (ρ x ∈ γ X); [|destruct X as [|[]]; auto]
    end.
    { apply meet_correct. split. 2:now auto. apply toplift_bind_spec with (rρ ρ x), H4.
      intros. apply itvabs_of_itv_correct; auto. }
    pose proof (fun x => IdealIntervals.singleton_correct _ _ x H5).
    destruct IdealIntervals.singleton. 2:now auto. apply H6, eq_refl.
  - split. eapply itv_of_itvabs_correct, H0. rewrite H1. split.
    + destruct b. 2:now auto. specialize (H3 _ eq_refl). clear - H3. unfold RF_le.
      destruct s. simpl in *.
      match goal with
      | |- context [Bmult ?a ?b ?c ?d ?e ?f ?g ?h] =>
        pose proof Bmult_correct a b c d e f g h
      end.
      rewrite Rlt_bool_true in H.
      * destruct H as (-> & -> & _). rewrite e.
        apply Rle_trans with (B2R 53 1024 half * B2R 53 1024 x)%R, round_UP_pt, fexp_correct, eq_refl.
        clear -H3. compute in *. Psatz.lra.
      * clear H. eapply Rle_lt_trans, bpow_lt with (e1:=-1+1024), eq_refl.
        apply abs_round_le_generic. apply fexp_correct, eq_refl. apply valid_rnd_round_mode.
        apply generic_format_bpow. discriminate.
        rewrite bpow_plus, Rabs_mult. apply Rmult_le_compat.
        apply Rabs_pos. apply Rabs_pos.
        rewrite <- !B2R_Babs with (abs_nan:=Float.abs_pl). compute. Psatz.lra.
        apply Rlt_le, abs_B2R_lt_emax.
    + destruct a. 2:now auto. specialize (H2 _ eq_refl). clear - H2. unfold RF_le.
      destruct s. simpl in *.
      unfold Float.neg. rewrite !B2R_Bopp, !is_finite_Bopp, Ropp_involutive.
      match goal with
      | |- context [Bmult ?a ?b ?c ?d ?e ?f ?g ?h] =>
        pose proof Bmult_correct a b c d e f g h
      end.
      rewrite Rlt_bool_true in H.
      * destruct H as (-> & -> & _). rewrite e.
        apply Rle_trans with (B2R 53 1024 half * B2R 53 1024 x)%R, round_UP_pt, fexp_correct, eq_refl.
        clear -H2. compute in *. Psatz.lra.
      * clear H. eapply Rle_lt_trans, bpow_lt with (e1:=-1+1024), eq_refl.
        apply abs_round_le_generic. apply fexp_correct, eq_refl. apply valid_rnd_round_mode.
        apply generic_format_bpow. discriminate.
        rewrite bpow_plus, Rabs_mult. apply Rmult_le_compat.
        apply Rabs_pos. apply Rabs_pos.
        rewrite <- !B2R_Babs with (abs_nan:=Float.abs_pl). compute. Psatz.lra.
        apply Rlt_le, abs_B2R_lt_emax.
Qed.

Definition oprocess_msg (m:message) (ab:t * query_chan) : (t * messages_chan)+⊥.
refine
  (match m with
  | Linear_zero_msg e =>
    let c := snd ab in
    let ab0 := get_unreduced (fst ab) in
    let ab := assume_linear_constraint (fst e) (T.elements (snd e)) ab0 in
    do_bot ab <- reduce_diag ab0 ab;
    let diff :=
      To.shcombine_diff (fun x mapxy0 mapxy =>
        let a0 := do_opt mapxy0 <- mapxy0; To.get (ovar_opp x) mapxy0 in
        let a := do_opt mapxy <- mapxy; To.get (ovar_opp x) mapxy in
        match a0, a with
        | None, Some _ => a
        | Some a0', Some a' =>
          if Fleb (proj1_sig a0') (proj1_sig a') then None
          else a
        | _, _ => None
        end) _
        ab0 ab
    in
    do_bot msgs <-
      T.fold (fun acc x bnds =>
        match bnds with
        | (None, None) => acc
        | (a, b) =>
          match c.(get_var_ty) x with
          | Just ty =>
            let ichan := c.(get_itv) (IEvar ty x) in
            do_bot m <- msg_of_bnds b a x ty ichan;
            match m with
            | All => acc
            | Just m => do_bot acc <- acc; NotBot (m::acc)
            end
          | All => acc
          end
        end) diff (NotBot nil);
    NotBot (t_of_t0 ab, m::msgs)
  | _ => NotBot (fst ab, m::nil)
  end).
Proof.
  destruct v. 2:reflexivity. destruct To.get. 2:reflexivity.
  simpl. rewrite Fleb_refl. auto. apply is_finite_not_is_nan, proj2_sig.
Defined.

Lemma oprocess_msg_correct:
  ∀ m ρ ab,
    ρ ∈ γ ab -> ρ ∈ γ m ->
    ρ ∈ γ (oprocess_msg m ab).
Proof.
  intros [] ρ [ab chan] [Hab Hchan] Hmsg; simpl;
    try (split; [auto|constructor;[eauto|constructor]]).
  eapply botbind_spec, reduce_diag_correct, assume_linear_constraint_correct, Hmsg.
  2:now auto. intros.
  eapply botbind_spec.
  repeat constructor. apply t_of_t0_correct, H. auto. apply H0.
  rewrite T.fold_spec.
  assert (ρ ∈ γ (NotBot List.nil)) by constructor.
  revert H0. generalize (NotBot (@List.nil message)).
  match goal with
  | |- context [@T.elements ?A ?X] =>
    generalize (@T.elements_complete A X); induction (@T.elements A X)  as [|[?[]]]
  end. now auto.
  simpl. intros. apply IHl. now auto. clear IHl. specialize (H0 _ _ (or_introl eq_refl)).
  match type of H0 with
  | @T.get ?A e ?M = Some ?X =>
    pose proof (eq_refl:To.get (ovPlus e) M = do_opt m <- @T.get A e M; fst m);
    pose proof (eq_refl:To.get (ovMinus e) M = do_opt m <- @T.get A e M; snd m)
  end.
  rewrite H0 in H3, H2. rewrite To.gshcombine_diff in H2, H3. clear H0.
  unfold fst in H2, Hab. unfold snd in H3, Hchan. subst o o0.
  pose proof Hab (ovPlus e) (ovMinus e). pose proof Hab (ovMinus e) (ovPlus e).
  pose proof H (ovPlus e) (ovMinus e). pose proof H (ovMinus e) (ovPlus e).
  unfold Too.get in H0, H2, H3, H4. unfold ovar_opp.
  pose proof @get_var_ty_correct _ ρ Hchan e.
  destruct get_var_ty;
  repeat match goal with
  | H : context [@To.get ?A ?B ?C] |- context [@To.get ?A' ?B' ?C'] =>
    change A' with A; change B' with B; change C' with C;
    destruct (@To.get A B C)
  | H: ∀ b, Some _ = Some b -> _ |- _ =>
    edestruct H as (? & ? & ? & ? & ?); [reflexivity|clear H]
  | H: ?A = ?B, H': ?A = ?C |- _ => rewrite H in H'; inv H'
  | |- context [Fleb ?f ?f'] => destruct (Fleb f f')
  | |- (γ (do_bot x <- _; _) _) => eapply botbind_spec; [intros|]
  | |- (γ (msg_of_bnds _ _ _ _ _) _) => eapply msg_of_bnds_correct
  | |- (γ (get_itv _ _) _) => eapply get_itv_correct
  | H:oρ ρ (ovMinus ?e) = Some ?xm,
    H':oρ ρ (ovPlus ?e) = Some ?xp |- _ =>
    simpl in H, H'; rewrite H' in H; inv H
  | |- γ match ?A with _ => _ end _ => destruct A
  | |- γ (NotBot (_ :: _)) _ => constructor
  end; eauto; intros ff eq; inv eq;
  match goal with
  | H : (?A <= ?B)%R |- (?C <= ?D)%R => eapply Rle_trans, H; Psatz.lra
  end.
Qed.

(******** Assign ********)

Definition reduce_constraints_row_step_self (row:constraint_row) (x:ovar) (ab:t0):
  constraint_row :=
  match To.get x row with
  | None => row
  | Some ax => reduce_constraints_row_step row x ax ab
  end.

Lemma reduce_constraints_row_step_self_correct:
  ∀ row x ab ρ v,
    ρ ∈ γ ab ->
    constraints_row_valid row ρ v ->
    constraints_row_valid (reduce_constraints_row_step_self row x ab) ρ v.
Proof.
  unfold reduce_constraints_row_step_self. intros.
  pose proof (H0 x). destruct (To.get x row). 2:now auto.
  edestruct H1 as (? & ? & ?). now eauto.
  eapply reduce_constraints_row_step_correct; eauto.
  intros. assert (xv=x0) by congruence. subst. now auto.
Qed.

Definition oassign (x:var) (e0:iexpr) (ab:t * query_chan) (c:query_chan) :
  (t * messages_chan)+⊥:=
  let e := (snd ab).(linearize_expr) e0 in
  let ty := iexpr_ty e0 in
  match e with
  | None => oforget x ab c
  | Some e =>
    (* Assign should be equivalent to assume followed by forget with
       some artificial variables. Both these operations do not loose
       precision with get_unreduced. So we should not loose precision
       here too. *)
    let ab0 := get_unreduced (fst ab) in
    let ab := oforget_t0 x ab0 in
    let '(constr_p, constr_m, i) :=
        constraints_row_of_linear_expr (fst e) (T.elements (snd e))
    in
    let i := add_itv (fst e) i in
    let ichan := c.(get_itv) (IEvar (iexpr_ty e0) x) in
    do_bot i <-
      match itv_of_itvabs ichan with
      | All => NotBot i
      | Just i' => i' ⊓ i
      end;
    let 'ITV a b := add_itv i i in

    let constr_p := reduce_constraints_row_step_self constr_p (ovPlus x) ab0 in
    let constr_p := reduce_constraints_row_step_self constr_p (ovMinus x) ab0 in
    let x'mx := To.get (ovPlus x) constr_p in
    let x'px := To.get (ovMinus x) constr_p in
    let b' :=
      do_opt x'mx <- x'mx;
      do_opt x'px <- x'px;
      fin_float_of_float
        (Bplus 53 1024 eq_refl eq_refl Float.binop_pl mode_UP
          (proj1_sig x'mx) (proj1_sig x'px))
    in
    let constr_p := T.remove x constr_p in
    let constr_p :=
      match best_bound (fin_float_of_float b) b' with
      | (Some b, _) => To.set (ovMinus x) b constr_p
      | _ => constr_p
      end
    in

    let ab := add_constraints_row (ovPlus x) constr_p ab in

    let constr_m := reduce_constraints_row_step_self constr_m (ovPlus x) ab0 in
    let constr_m := reduce_constraints_row_step_self constr_m (ovMinus x) ab0 in
    let mx'mx := To.get (ovPlus x) constr_m in
    let mx'px := To.get (ovMinus x) constr_m in
    let a' :=
      do_opt mx'mx <- mx'mx;
      do_opt mx'px <- mx'px;
      fin_float_of_float
        (Bplus 53 1024 eq_refl eq_refl Float.binop_pl mode_UP
             (proj1_sig mx'mx) (proj1_sig mx'px))
    in
    let constr_m := T.remove x constr_m in
    let constr_m :=
      match best_bound (fin_float_of_float (Float.neg a)) a' with
      | (Some a, _) => To.set (ovPlus x) a constr_m
      | _ => constr_m
      end
    in

    let ab := add_constraints_row (ovMinus x) constr_m ab in

    do_bot ab <- reduce_diag ab0 ab;

    let a := Too.get (ovMinus x, ovPlus x) ab in
    let b := Too.get (ovPlus x, ovMinus x) ab in
    do_bot msg <- msg_of_bnds a b x ty ichan;
    let msgs :=
      match msg with
      | All => nil
      | Just m => m::nil
      end
    in
    NotBot (t_of_t0 ab, msgs)
  end.

Lemma oassign_correct:
  forall x e ρ n ab chan,
    ρ ∈ γ ab -> (ρ+[x => n]) ∈ γ chan ->
    n ∈ eval_iexpr ρ e ->
    (ρ+[x => n]) ∈ γ (oassign x e ab chan).
Proof.
  unfold oassign. intros.
  pose proof linearize_expr_correct (snd ab) e.
  destruct ((snd ab).(linearize_expr) e). 2:now auto using oforget_correct.
  edestruct H2 as (y & Hy & EVAL). now eauto. now apply H. now apply H1. clear H2.
  destruct (constraints_row_of_linear_expr (fst t1) (T.elements (snd t1)))
    as [[constr_p constr_m] i] eqn:CONSTR.
  eapply constraints_row_of_linear_expr_correct in CONSTR. 2:now apply EVAL.
  destruct CONSTR as (CONSTRp & CONSTRm & HITV).
  assert (∃ xx,
    eval_linear_expr ρ (ITV Float.zero Float.zero, T.elements (snd t1)) (y-xx) /\
    xx ∈ γ (fst t1)).
  { unfold eval_T_linear_expr in EVAL. apply eval_linear_expr_change_cst in EVAL.
    destruct EVAL as (? & ? & ?). eexists. split. 2:now eauto.
    replace (y - x0)%R with (y+0-x0)%R by ring. apply H3. compute; Psatz.lra. }
  destruct H2 as (xx & Hxx1 & Hxx2). apply HITV in Hxx1.
  match goal with
  | |- γ (do_bot i0 <- ?I0; _) _ => assert (y ∈ γ I0)
  end.
  { assert (eval_iexpr (ρ+[x=>n]) (IEvar (iexpr_ty e) x) n).
    { repeat constructor. unfold upd; rewrite eq_dec_true; auto.
      eapply iexpr_ty_correct, H1. }
    eapply get_itv_correct in H2. 2:now eauto.
    apply itv_of_itvabs_correct in H2.
    assert (y ∈ γ (add_itv (fst t1) i)).
    { replace y with (xx+(y-xx))%R by ring. apply add_itv_correct; auto. }
    destruct itv_of_itvabs. auto. apply meet_correct. rewrite Hy in H2. auto. }
  eapply botbind_spec. 2:now eauto. clear H2. intros.
  assert ((y + y)%R ∈ γ (add_itv a a)) by eauto using add_itv_correct.
  destruct add_itv.
  eapply botbind_spec, reduce_diag_correct.
  { intros. eapply botbind_spec, msg_of_bnds_correct. intros.
    split. eapply t_of_t0_correct, H4. destruct a1; constructor. eauto. constructor.
    eapply iexpr_ty_correct. rewrite upd_s. eauto.
    eapply get_itv_correct; eauto. constructor. auto.
    rewrite upd_s. eapply iexpr_ty_correct; eauto.
    unfold rρ. rewrite upd_s. eauto.
    intros. apply H4 in H5. unfold oρ, rρ in H5. rewrite upd_s, Hy in H5.
    destruct H5 as (? & ? & ? & ? & ?). inv H5. inv H6. eapply Rle_trans, H7. Psatz.lra.
    intros. apply H4 in H5. unfold oρ, rρ in H5. rewrite upd_s, Hy in H5.
    destruct H5 as (? & ? & ? & ? & ?). inv H5. inv H6. eapply Rle_trans, H7. Psatz.lra. }
  match goal with
  | |- context [add_constraints_row (ovMinus x) _ ?AB] =>
    assert ((ρ +[ x => n]) ∈ γ AB)
  end.
  { eapply add_constraints_row_correct.
    { unfold upd, oρ, rρ. rewrite eq_dec_true, Hy. reflexivity. }
    apply oforget_t0_forget, oforget_t0_incl, H.
    eapply reduce_constraints_row_step_self_correct with (x:=ovPlus x) in CONSTRp. 2:apply H.
    eapply reduce_constraints_row_step_self_correct with (x:=ovMinus x) in CONSTRp. 2:apply H.
    match type of CONSTRp with
    | constraints_row_valid ?X _ _ =>
      assert (constraints_row_valid (T.remove x X) (ρ+[x=>n]) y)
    end.
    { repeat intro. specialize (CONSTRp x0 b). unfold To.get in CONSTRp, H4.
      unfold upd, oρ, rρ.
      destruct x0; simpl in CONSTRp, H4; rewrite T.grspec in H4;
      destruct (T.elt_eq v x); try discriminate;
      (destruct CONSTRp as (xv & Hxv1 & Hxv2); [now auto|]); exists xv; split; auto;
      destruct (eq_dec v x); auto; congruence. }
    match goal with
    | |- constraints_row_valid (let (o, _) := ?BND in _) _ _ =>
      assert (∀ bnd, fst BND = Some bnd -> y+y <= B2R _ _ (proj1_sig bnd))%R;
        [|destruct BND as [[]?]]
    end.
    { intro. apply best_bound_either; intro.
      - apply fin_float_of_float_correct in H5. destruct H3, H5 as [? ->].
        unfold RF_le in H3. rewrite H5 in H3. apply H3.
      - pose proof (CONSTRp (ovPlus x)). pose proof (CONSTRp (ovMinus x)).
        destruct To.get, To.get; try discriminate.
        edestruct H6 as (? & ? & ?), H7 as (? & ? & ?). now eauto. now eauto. now eauto.
        simpl in H8, H10. rewrite H8 in H10. inv H10.
        apply fin_float_of_float_correct in H5. destruct H5 as [? ->].
        replace (y+y)%R with (y- -y)%R by ring. eapply reduct_rule_1; eauto. Psatz.lra. }
    2:now auto. specialize (H5 _ eq_refl). repeat intro.
    rewrite To.gsspec in H6. destruct To.elt_eq. inv H6.
    eexists. split. unfold upd, oρ, rρ. rewrite eq_dec_true, Hy. reflexivity. Psatz.lra.
    unfold To.get, PTSUU.get in H6. destruct (ovar_EQ.t_of_s x0) eqn:EQ. rewrite T.grspec in H6.
    destruct T.elt_eq. discriminate.
    specialize (CONSTRp x0). unfold To.get in CONSTRp. rewrite EQ in CONSTRp.
    apply CONSTRp in H6. unfold upd, oρ, rρ. destruct x0; inv EQ.
    destruct (eq_dec v x). congruence. now auto.
    destruct (eq_dec v x). congruence. now auto. }
  eapply add_constraints_row_correct.
  { unfold upd, oρ, rρ. rewrite eq_dec_true, Hy. reflexivity. }
  now apply H4. clear H4.
  eapply reduce_constraints_row_step_self_correct with (x:=ovPlus x) in CONSTRm. 2:apply H.
  eapply reduce_constraints_row_step_self_correct with (x:=ovMinus x) in CONSTRm. 2:apply H.
  match type of CONSTRm with
  | constraints_row_valid ?X _ _ =>
    assert (constraints_row_valid (T.remove x X) (ρ+[x=>n]) (-y))
  end.
  { repeat intro. specialize (CONSTRm x0 b). unfold To.get in CONSTRm, H4.
    unfold upd, oρ, rρ.
    destruct x0; simpl in CONSTRm, H4; rewrite T.grspec in H4;
    destruct (T.elt_eq v x); try discriminate;
    (destruct CONSTRm as (xv & Hxv1 & Hxv2); [now auto|]); exists xv; split; auto;
    destruct (eq_dec v x); auto; congruence. }
    match goal with
    | |- constraints_row_valid (let (o, _) := ?BND in _) _ _ =>
      assert (∀ bnd, fst BND = Some bnd -> -y-y <= B2R _ _ (proj1_sig bnd))%R;
        [|destruct BND as [[]?]]
    end.
    { intro. apply best_bound_either; intro.
      - apply fin_float_of_float_correct in H5. destruct H3, H5 as [? ->].
        unfold RF_le in H6. rewrite H5 in H6. Psatz.lra.
      - pose proof (CONSTRm (ovPlus x)). pose proof (CONSTRm (ovMinus x)).
        destruct To.get, To.get; try discriminate.
        edestruct H6 as (? & ? & ?), H7 as (? & ? & ?). now eauto. now eauto. now eauto.
        simpl in H8, H10. rewrite H8 in H10. inv H10.
        apply fin_float_of_float_correct in H5. destruct H5 as [? ->].
        eapply reduct_rule_1; eauto. Psatz.lra. }
    2:now auto. specialize (H5 _ eq_refl). repeat intro.
    rewrite To.gsspec in H6. destruct To.elt_eq. inv H6.
    eexists. split. unfold upd, oρ, rρ. rewrite eq_dec_true, Hy. reflexivity. Psatz.lra.
    unfold To.get, PTSUU.get in H6. destruct (ovar_EQ.t_of_s x0) eqn:EQ. rewrite T.grspec in H6.
    destruct T.elt_eq. discriminate.
    specialize (CONSTRm x0). unfold To.get in CONSTRm. rewrite EQ in CONSTRm.
    apply CONSTRm in H6. unfold upd, oρ, rρ. destruct x0; inv EQ.
    destruct (eq_dec v x). congruence. now auto.
    destruct (eq_dec v x). congruence. now auto.
Qed.

Program Instance oct_idenv : ab_ideal_env t iter_t :=
  {| id_leb := oleb;
     id_top := otop;
     id_join := ojoin;
     id_init_iter := oinit_iter;
     id_widen := owiden;
     assign := oassign;
     forget := oforget;
     enrich_query_chan ab c := c;
     process_msg := oprocess_msg |}.
Next Obligation. eapply otop_correct; eauto. Qed.
Next Obligation. eapply oleb_correct; eauto. Qed.
Next Obligation. eapply ojoin_correct; eauto. Qed.
Next Obligation. eapply oinit_iter_sound; auto. Qed.
Next Obligation. eapply owiden_incr; eauto. Qed.
Next Obligation. eapply oassign_correct; eauto. Qed.
Next Obligation. eapply oforget_correct; eauto. Qed.
Next Obligation. eapply oprocess_msg_correct; eauto. Qed.
