(* *********************************************************************)
(*                                                                     *)
(*              The Verasco static analyzer                            *)
(*                                                                     *)
(*          Vincent Laporte, Université de Rennes 1                    *)
(*          Jacques-Henri Jourdan, INRIA Paris-Rocquencourt            *)
(*                                                                     *)
(*  Copyright Institut National de Recherche en Informatique et en     *)
(*  Automatique and Université de Rennes 1.  All rights reserved.      *)
(*  This file is distributed under the terms of the GNU General        *)
(*  Public License as published by the Free Software Foundatio,        *)
(*  either version 2 of the License, or (at your option) any later     *)
(*  version.  This file is also distributed under the terms of the     *)
(*  INRIA Non-Commercial License Agreement.                            *)
(*                                                                     *)
(* *********************************************************************)

Require Import Coqlib Utf8 String ToString PrintPos.
Require Import Util AdomLib.
Require Import Integers.
Require Psatz.

Inductive itv : Type :=
| ITV: Z -> Z -> itv.
Instance ItvToString : ToString itv :=
  { to_string i :=
      match i with
        | ITV min max =>
          ("[" ++ to_string min ++ "," ++ to_string max ++ "]")%string
      end }.

Instance leb_itv : leb_op itv := fun (i1 i2: itv) =>
  match i1, i2 with
    | ITV min1 max1, ITV min2 max2 =>
      zle min2 min1 && zle max1 max2
  end.

Lemma leb_eq:
  ∀ i:itv, i ⊑ i = true.
Proof. destruct i; unfold leb; simpl; destruct zle, zle; auto; omega. Qed.

Instance join_itv : join_op itv itv := fun i1 i2 =>
  match i1, i2 with
    | ITV min1 max1, ITV min2 max2 =>
      match Z.compare min1 min2, Z.compare max1 max2 with
        | (Lt | Eq), (Eq | Gt) => i1
        | (Eq | Gt), Lt | Gt, Eq => i2
        | Lt, Lt => ITV min1 max2
        | Gt, Gt => ITV min2 max1
      end
  end.

Lemma join_eq:
  ∀ x:itv, x ⊔ x = x.
Proof. unfold join. destruct x. simpl. rewrite !Z.compare_refl. auto. Qed.

Instance meet_itv : meet_op itv (itv+⊥) :=  fun i1 i2 =>
  match i1, i2 with
    | ITV min1 max1, ITV min2 max2 =>
      let m := Zmax min1 min2 in
      let M := Zmin max1 max2 in
      if zlt M m then Bot
      else NotBot (ITV m M)
  end.

Definition widening_hints :=
  - Z.shiftl 1 63 ::
  - Z.shiftl 1 31 ::
  - Z.shiftl 1 15 ::
  - Z.shiftl 1 7 ::
  0 :: 1 ::
  Z.shiftl 1 7  - 1 ::
  Z.shiftl 1 8  - 1 ::
  Z.shiftl 1 15 - 1 ::
  Z.shiftl 1 16 - 1 ::
  Z.shiftl 1 31 - 1 ::
  Z.shiftl 1 32 - 1 ::
  Z.shiftl 1 63 - 1 :: nil.

Definition widening_hints_rev :=
  List.rev_append widening_hints nil.

Definition next_larger z :=
  List.find (Z.leb z) widening_hints.

Lemma next_larger_larger:
  ∀ n m, next_larger n = Some m -> n <= m.
Proof.
  unfold next_larger. induction widening_hints. discriminate.
  simpl. intros. destruct (Z.leb_spec n a). congruence. auto.
Qed.

Definition next_smaller z :=
  List.find (Z.geb z) widening_hints_rev.

Lemma next_smaller_smaller:
  ∀ n m, next_smaller n = Some m -> m <= n.
Proof.
  unfold next_smaller. induction widening_hints_rev. discriminate.
  simpl. intros. destruct (Z.geb_spec n a). congruence. auto.
Qed.

Definition widen_itv i1 i2 :=
  match i1, i2 with
    | ITV min1 max1, ITV min2 max2 =>
      let lemin := Z.leb min1 min2 in
      let lemax := Z.leb max2 max1 in
      if lemin && lemax then Just i1
      else
        let min :=
            if lemin then Some min1
            else next_smaller min2
        in
        let max :=
            if lemax then Some max1
            else next_larger max2
        in
        match min, max with
          | Some min, Some max => Just (ITV min max)
          | _, _ => All
        end
  end.

Lemma widen_eq:
  ∀ x:itv, widen_itv x x = Just x.
Proof. unfold widen_itv. destruct x. simpl. rewrite !Z.leb_refl. auto. Qed.

Instance gamma_itv : gamma_op itv Z := fun i v =>
  match i with
    | ITV m M => m <= v <= M
  end.

Instance leb_itv_correct : leb_op_correct itv Z.
Proof.
  intros [] []. unfold leb. simpl.
  intros. split; destruct H0, (zle z1 z), (zle z0 z2); discriminate || omega.
Qed.

Instance join_itv_correct : join_op_correct itv itv Z.
Proof.
  intros [? ?] [? ?] ? [[? ?]|[? ?]]; unfold join; simpl;
  destruct (Z.compare_spec z z1), (Z.compare_spec z0 z2); split;
  omega.
Qed.

Lemma widen_incr:
  ∀ ab1 ab2, γ ab1 ⊆ γ (widen_itv ab1 ab2).
Proof.
  unfold widen_itv. intros [m1 M1] [m2 M2] n H.
  destruct andb. auto.
  destruct (if m1 <=? m2 then Some m1 else next_smaller m2) as [m|] eqn:EQm. 2:constructor.
  destruct (if M2 <=? M1 then Some M1 else next_larger M2) as [M|] eqn:EQM. 2:constructor.
  destruct H. split.
  - destruct (Z.leb_spec m1 m2). congruence. etransitivity. apply next_smaller_smaller.
    eauto. omega.
  - destruct (Z.leb_spec M2 M1). congruence. etransitivity. 2:apply next_larger_larger, EQM.
    omega.
Qed.

Instance meet_itv_correct: meet_op_correct itv (itv+⊥) Z.
Proof.
  intros [mx Mx] [my My] z Hz; simpl in *; unfold meet; simpl.
  destruct Hz as [[][]].
  case (zlt (Zmin Mx My) (Zmax mx my)); apply Zmin_case; apply Zmax_case; intro; simpl; omega.
Qed.

Definition forward_neg (x:itv+⊤) : itv+⊤ :=
  match x with
    | Just (ITV mx Mx) => Just (ITV (-Mx) (-mx))
    | All => All
  end.

Lemma forward_neg_correct:
  ∀ x x_ab, x ∈ γ x_ab -> (-x) ∈ γ (forward_neg x_ab) .
Proof.
  intros x [|[mx Mx]] Hx; simpl in *; intuition.
Qed.

Definition forward_mult (x y:itv+⊤) : itv+⊤+⊥ :=
  match x, y with
    | Just (ITV mx Mx), Just (ITV my My) =>
      let b1 := mx * my in      let b2 := mx * My in
      let b3 := Mx * my in      let b4 := Mx * My in
      NotBot (Just (ITV (Zmin (Zmin b1 b4) (Zmin b3 b2))
                        (Zmax (Zmax b1 b4) (Zmax b3 b2))))
    | Just (ITV 0 0), _ | _, Just (ITV 0 0) => NotBot (Just (ITV 0 0))
    | All, _ | _, All => NotBot All
  end.

Lemma Mult_interval_correct_min_max :
  forall a b c d x y : Z,
    a <= x <= b ->
    c <= y <= d ->
    Zmin (Zmin (a*c) (b*d)) (Zmin (b*c) (a*d)) <= x * y
        <= Zmax (Zmax (a*c) (b*d)) (Zmax (b*c) (a*d)).
Proof.
  intros. split.
  repeat apply Z.min_case_strong;
    destruct (Z.nonpos_pos_cases x), (Z.nonpos_pos_cases y);
    destruct (Z.nonpos_pos_cases a), (Z.nonpos_pos_cases b);
    destruct (Z.nonpos_pos_cases c), (Z.nonpos_pos_cases c);
    Psatz.nia.
  repeat apply Z.max_case_strong;
    destruct (Z.nonpos_pos_cases x), (Z.nonpos_pos_cases y);
    destruct (Z.nonpos_pos_cases a), (Z.nonpos_pos_cases b);
    destruct (Z.nonpos_pos_cases c), (Z.nonpos_pos_cases c);
    Psatz.nia.
Qed.

Lemma forward_mult_correct:
  ∀ x x_ab, x ∈ γ x_ab ->
  ∀ y y_ab, y ∈ γ y_ab ->
    (x * y)%Z ∈ γ (forward_mult x_ab y_ab).
Proof.
  intros x [|[mx Mx]]  Hx y [|[my My]] Hy; simpl in Hx, Hy.
  - tauto.
  - destruct my, My; simpl; try tauto. Psatz.nia.
  - destruct mx, Mx; simpl; try tauto. Psatz.nia.
  - destruct mx, Mx; apply Mult_interval_correct_min_max; auto.
Qed.

Definition forward_div_pos (x:itv+⊤) (my My:positive) : itv+⊤ :=
  match x with
    | Just (ITV mx Mx) =>
      Just (ITV
              (Z.quot mx (if Zlt_bool mx 0 then Zpos my else Zpos My))
              (Z.quot Mx (if Zlt_bool Mx 0 then Zpos My else Zpos my)))
    | All => All
  end.

Lemma Zquot_le_compat_l_neg:
  ∀ p q r : Z, p < 0 -> 0 < q <= r -> p ÷ q <= p ÷ r.
Proof.
  intros.
  rewrite Z.opp_le_mono.
  rewrite <- !Z.quot_opp_l by omega.
  apply Z.quot_le_compat_l; omega.
Qed.

Lemma forward_div_pos_correct:
  ∀ x x_ab, x ∈ γ x_ab ->
  ∀ y my My, Zpos my <= y <= Zpos My ->
    (x ÷ y) ∈ γ (forward_div_pos x_ab my My).
Proof.
  intros x [|[mx Mx]] Hx [|y|] my My Hy; unfold forward_div_pos; try (exfalso; zify; omega).
  auto. simpl.
  pose proof (Zlt_cases mx 0). pose proof (Zlt_cases Mx 0).
  simpl in *.

  split.
  destruct (Zlt_bool mx 0).
  eapply Zle_trans; [apply Zquot_le_compat_l_neg|apply Z.quot_le_mono]; zify; omega.
  eapply Zle_trans; [apply Z.quot_le_compat_l|apply Z.quot_le_mono]; zify; omega.
  destruct (Zlt_bool Mx 0).
  eapply Zle_trans; [apply Z.quot_le_mono|apply Zquot_le_compat_l_neg]; zify; omega.
  eapply Zle_trans; [apply Z.quot_le_mono|apply Z.quot_le_compat_l]; zify; omega.
Qed.

Definition forward_div_neg (x:itv+⊤) (my My:positive) : itv+⊤ :=
  forward_neg (forward_div_pos x My my).

Lemma forward_div_neg_correct:
  ∀ x x_ab, x ∈ γ x_ab ->
  ∀ y my My, Zneg my <= y <= Zneg My ->
    (x ÷ y) ∈ γ (forward_div_neg x_ab my My).
Proof.
  intros x x_ab Hx [| |y] my My Hy; try (exfalso; zify; omega).
  rewrite <- Z.opp_involutive.
  apply forward_neg_correct.
  rewrite <- Z.quot_opp_r; auto.
  eapply forward_div_pos_correct. eauto. zify; omega. zify; omega.
Qed.

Definition forward_div (x y:itv+⊤) : itv+⊤+⊥ :=
  match x, y with
    | _, Just (ITV (Zpos _|0) (Zneg _|0)) => Bot
    | Just (ITV 0 0), _ => NotBot (Just (ITV 0 0))
    | _, All => NotBot All
    | x, Just (ITV (Zpos my) (Zpos My)) =>
      NotBot (forward_div_pos x my My)
    | x, Just (ITV 0 (Zpos My)) =>
      NotBot (forward_div_pos x 1 My)
    | x, Just (ITV (Zneg my) (Zpos My)) =>
      NotBot ((forward_div_neg x my 1) ⊔ (forward_div_pos x 1 My))
    | x, Just (ITV (Zneg my) 0) =>
      NotBot (forward_div_neg x my 1)
    | x, Just (ITV (Zneg my) (Zneg My)) =>
      NotBot (forward_div_neg x my My)
  end.

Lemma forward_div_correct:
  ∀ x x_ab, x ∈ γ x_ab ->
  ∀ y y_ab, y ∈ γ y_ab ->
    y ≠ 0 ->
    (x ÷ y) ∈ γ (forward_div x_ab y_ab).
Proof.
  simpl. unfold forward_div.
  intros x [|[[|mx|mx] [|Mx|Mx]]] Hx y [|[[|my|my] [|My|My]]] Hy Hy0;
    simpl in Hx, Hy; try (simpl; zify; omega);
  try (apply join_correct; auto; destruct (Zcompare_spec 0 y); [congruence|right|left]);
  (try apply forward_div_pos_correct with (x:=x) (y:=y); auto);
  (try apply forward_div_neg_correct with (x:=x) (y:=y); auto);
  try omega;
  try (assert (x = 0) by omega; subst; easy).
Qed.

Definition forward_mod_pos_pos (x:itv+⊤) (my My:positive) : itv+⊤+⊥ :=
  match x with
    | All => NotBot (Just (ITV 0 (Zpos My - 1)))
    | Just (ITV mx Mx) =>
      if Zle_bool 0 Mx then
        NotBot (Just (ITV (if Zlt_bool Mx (Zpos my) then Zmax 0 mx else 0)
                          (Zmin (Zpos My-1) Mx)))
      else Bot
  end.

Lemma forward_mod_pos_pos_correct:
  ∀ x x_ab, x ∈ γ x_ab -> 0 <= x ->
  ∀ y my My, Zpos my <= y <= Zpos My ->
    (Z.rem x y) ∈ γ (forward_mod_pos_pos x_ab my My).
Proof.
  intros x x_ab Hx Hxpos [|y|] my My Hy; unfold forward_mod_pos_pos; try (exfalso; zify; omega).

  assert (0 <= Z.rem x (Zpos y) <= Zpos My - 1).
  pose proof (Zquot.Zrem_lt_pos_pos x (Z.pos y)).
  zify; omega.

  destruct x_ab as [|[mx Mx]]. assumption. simpl in Hx.
  pose proof (Zle_cases 0 Mx). destruct (Zle_bool 0 Mx); [|simpl; omega].
  split.

  pose proof (Zlt_cases Mx (Zpos my)). destruct (Zlt_bool Mx (Zpos my)).
  apply Z.max_case_strong. omega. rewrite Z.rem_small; omega.
  omega.

  apply Z.min_case_strong. omega. intro.
  pose proof (Z.rem_le x (Zpos y)). zify; omega.
Qed.

Definition forward_mod_pos (x:itv+⊤) (my My:positive) : itv+⊤+⊥ :=
  forward_mod_pos_pos x my My ⊔
  botlift_fun1 forward_neg (forward_mod_pos_pos (forward_neg x) my My).

Lemma forward_mod_pos_correct:
  ∀ x x_ab, x ∈ γ x_ab ->
  ∀ y my My, Zpos my <= y <= Zpos My ->
    (Z.rem x y) ∈ γ (forward_mod_pos x_ab my My).
Proof.
  intros x x_ab Hx y my My Hy. unfold forward_mod_pos.
  apply join_correct.
  destruct (Coqlib.zlt 0 x); [left|right].
  eapply forward_mod_pos_pos_correct; auto. omega.
  eapply botlift_fun1_spec, forward_mod_pos_pos_correct, Hy.
  2:apply forward_neg_correct, Hx. 2:omega.
  intros. rewrite <- Z.opp_involutive, <- Z.rem_opp_l'.
  apply forward_neg_correct, H.
Qed.

Definition forward_mod_neg (x:itv+⊤) (my My:positive) : itv+⊤+⊥ :=
  forward_mod_pos x My my.

Lemma forward_mod_neg_correct:
  ∀ x x_ab, x ∈ γ x_ab ->
  ∀ y my My, Zneg my <= y <= Zneg My ->
    (Z.rem x y) ∈ γ (forward_mod_neg x_ab my My).
Proof.
  intros x x_ab Hx [| |y] my My Hy; try (exfalso; zify; omega).
  rewrite <- Z.rem_opp_r'.
  eapply forward_mod_pos_correct. auto. zify; omega.
Qed.

Definition forward_mod (x y:itv+⊤) : itv+⊤+⊥ :=
  match y with
    | All =>
      match x with
        | All => NotBot All
        | Just (ITV mx Mx) =>
          let M := Zmax (Zabs mx) (Zabs Mx) in
          NotBot (Just (ITV (-M) M))
      end
    | Just (ITV (Zpos _|0) (Zneg _|0)) => Bot
    | Just (ITV (Zpos my) (Zpos My)) =>
      forward_mod_pos x my My
    | Just (ITV 0 (Zpos My)) =>
      forward_mod_pos x 1 My
    | Just (ITV (Zneg my) (Zpos My)) =>
      forward_mod_neg x my 1 ⊔ forward_mod_pos x 1 My
    | Just (ITV (Zneg my) 0) =>
      forward_mod_neg x my 1
    | Just (ITV (Zneg my) (Zneg My)) =>
      forward_mod_neg x my My
  end.

Lemma forward_mod_correct:
  ∀ x x_ab, x ∈ γ x_ab ->
  ∀ y y_ab, y ∈ γ y_ab ->
    y ≠ 0 ->
    (Z.rem x y) ∈ γ (forward_mod x_ab y_ab).
Proof.
  unfold forward_mod.
  assert (∀ x y, Zabs (Z.rem x y) <= Zabs x).
  { intros.
    destruct (Z.eq_dec y 0).
    subst. rewrite Zquot.Zrem_0_r. apply Z.le_refl.
    rewrite <- Z.rem_abs; [apply Z.rem_le|]; try (zify; omega). }
  Opaque forward_mod_pos forward_mod_neg.
  intros x [|[[|mx|mx] [|Mx|Mx]]] Hx y [|[[|my|my] [|My|My]]] Hy Hy0;
    simpl in Hx, Hy;
    try tauto; try (simpl; zify; omega);
    repeat (simple apply join_correct; auto; destruct (Zcompare_spec 0 y); [congruence|right|left]);
    try apply forward_mod_pos_correct with (x:=x) (y:=y);
    try apply forward_mod_neg_correct with (x:=x) (y:=y);
    try (simpl; now auto); try (zify; omega);
    specialize (H x y).
  assert (x = 0) by omega. subst.
  compute; intuition.
  unfold Zmax. simpl. zify; omega.
  unfold Zmax. simpl. destruct (Pcompare_spec mx Mx); zify; omega.
  unfold Zmax. simpl. zify; omega.
  unfold Zmax. simpl. destruct (Pcompare_spec mx Mx); zify; omega.
  unfold Zmax. simpl. destruct (Pcompare_spec mx Mx); zify; omega.
Qed.

Definition forward_lnot (x:itv+⊤) : itv+⊤ :=
  match x with
    | Just (ITV mx Mx) => Just (ITV (Z.pred (-Mx)) (Z.pred (-mx)))
    | All => All
  end.

Lemma forward_lnot_correct:
  ∀ x x_ab, x ∈ γ x_ab -> (Z.lnot x) ∈ γ (forward_lnot x_ab) .
Proof.
  intros x [|[mx Mx]] Hx; simpl in *. auto. unfold Z.lnot. omega.
Qed.

Lemma le_common_bits :
  ∀ m p M, (m <= p <= M)%N ->
  ∀ i, (N.size (N.lxor m M) <= i)%N ->
  N.testbit m i = N.testbit p i /\ N.testbit p i = N.testbit M i.
Proof.
  intros.
  set (c:=N.size (N.lxor m M)) in *.
  replace i with (i-c + c)%N by (zify; omega).
  rewrite <- !N.shiftr_spec by (zify; omega).
  assert (N.shiftr m c <= N.shiftr p c <= N.shiftr M c)%N.
  { rewrite !N.shiftr_div_pow2.
    destruct H; split; apply N.div_le_mono; trivial; apply N.pow_nonzero; zify; omega. }
  assert (N.shiftr m c = N.shiftr M c).
  { apply N.lxor_eq.
    rewrite <- N.shiftr_lxor, N.shiftr_div_pow2.
    apply N.div_small, N.size_gt. }
  split; f_equal; zify; omega.
Qed.

Definition forward_and_pos (mx Mx my My:N) : N * N :=
  let cx := N.size (N.lxor mx Mx) in
  let cy := N.size (N.lxor my My) in
  let c := Nmax cx cy in
  let onesc := N.ones c in
  let high := N.ldiff (N.land mx my) onesc in
  (high, N.lor high (N.min (N.land Mx onesc) (N.land My onesc))).

Lemma forward_and_pos_correct:
  ∀ mx x Mx, (mx <= x <= Mx)%N ->
  ∀ my y My, (my <= y <= My)%N ->
  ∀ r R, forward_and_pos mx Mx my My = (r, R) ->
  (r <= N.land x y <= R)%N.
Proof.
  unfold forward_and_pos; intros.
  inversion H1; subst; clear H1.
  set (mask:=(N.ones (N.max (N.size (N.lxor mx Mx)) (N.size (N.lxor my My))))).
  assert (N.ldiff (N.land mx my) mask = N.ldiff (N.land x y) mask).
  { apply N.bits_inj; intro. rewrite !N.ldiff_spec, !N.land_spec.
    destruct (N.le_gt_cases (N.max (N.size (N.lxor mx Mx)) (N.size (N.lxor my My))) n).
    - f_equal. f_equal;
        (eapply proj1, le_common_bits; [eauto|];
         revert H1; apply N.max_case_strong; intros; zify; omega).
    - unfold mask; rewrite N.ones_spec_low by trivial.
      repeat (destruct N.testbit; try reflexivity). }
  assert (∀ a b c, a <= b -> a <= c -> b-a <= c-a -> a <= b <= c)%N by (intros; zify; omega).
  apply H2; clear H2.
  - apply N.ldiff_le, N.bits_inj_0; intro.
    rewrite H1, !N.ldiff_spec, !N.land_spec; simpl.
    repeat (destruct N.testbit; try reflexivity).
  - apply N.ldiff_le, N.bits_inj_0; intro.
    rewrite !N.ldiff_spec, N.lor_spec, N.ldiff_spec; simpl.
    repeat (destruct N.testbit; try reflexivity).
  - rewrite H1, !N.sub_nocarry_ldiff.
    + assert (N.land (N.land x y) mask <= N.min (N.land Mx mask) (N.land My mask))%N.
      { apply N.le_trans with (N.min (N.land x mask) (N.land y mask)).
        apply N.min_glb; zify; apply Int.Ztestbit_le; trivial; intros i Hi;
        rewrite !Z2N.inj_testbit, !N.land_spec in * by trivial;
        repeat (destruct N.testbit; try (simpl; congruence)).
        apply N.min_le_compat.
        - assert (N.ldiff x mask = N.ldiff Mx mask).
          { apply N.bits_inj; intro. rewrite !N.ldiff_spec.
            destruct (N.le_gt_cases (N.max (N.size (N.lxor mx Mx)) (N.size (N.lxor my My))) n).
            - f_equal.
              eapply proj2, le_common_bits. eauto.
              revert H2; apply N.max_case_strong; intros; zify; omega.
            - unfold mask; rewrite N.ones_spec_low by trivial.
              repeat (destruct N.testbit; try reflexivity). }
          apply proj2 in H.
          rewrite <- N.lor_ldiff_and with (a:=x) (b:=mask) in H.
          rewrite <- N.lor_ldiff_and with (a:=Mx) (b:=mask) in H.
          rewrite <- !N.lxor_lor, <- !N.add_nocarry_lxor in H;
            try now (apply N.bits_inj_0; intro;
                     rewrite !N.land_spec, N.ldiff_spec;
                     repeat (destruct N.testbit; try reflexivity)).
          zify; omega.
        - assert (N.ldiff y mask = N.ldiff My mask).
          { apply N.bits_inj; intro. rewrite !N.ldiff_spec.
            destruct (N.le_gt_cases (N.max (N.size (N.lxor mx Mx)) (N.size (N.lxor my My))) n).
            - f_equal.
              eapply proj2, le_common_bits. eauto.
              revert H2; apply N.max_case_strong; intros; zify; omega.
            - unfold mask; rewrite N.ones_spec_low by trivial.
              repeat (destruct N.testbit; try reflexivity). }
          apply proj2 in H0.
          rewrite <- N.lor_ldiff_and with (a:=y) (b:=mask) in H0.
          rewrite <- N.lor_ldiff_and with (a:=My) (b:=mask) in H0.
          rewrite <- !N.lxor_lor, <- !N.add_nocarry_lxor in H0;
            try now (apply N.bits_inj_0; intro;
                     rewrite !N.land_spec, N.ldiff_spec;
                     repeat (destruct N.testbit; try reflexivity)).
          zify; omega. }
      match goal with H:(?a <= ?b)%N |- (?a' <= ?b')%N => replace a' with a; [replace b' with b|] end. trivial.
      * apply N.min_case; apply N.bits_inj; intro;
        rewrite !N.ldiff_spec, N.lor_spec, N.ldiff_spec, !N.land_spec;
        repeat (destruct N.testbit; try reflexivity).
      * apply N.bits_inj; intro;
        rewrite !N.ldiff_spec, !N.land_spec;
        repeat (destruct N.testbit; try reflexivity).
    + apply N.bits_inj; intro;
      rewrite !N.ldiff_spec, !N.lor_spec, N.ldiff_spec;
      repeat (destruct N.testbit; try reflexivity).
    + apply N.bits_inj_0; intro;
      rewrite !N.ldiff_spec;
      repeat (destruct N.testbit; try reflexivity).
Qed.

Definition forward_or_pos (mx Mx my My:N) : N * N :=
  let N := N.ones (N.max (N.size Mx) (N.size My)) in
  let (m, M) := forward_and_pos (N-Mx) (N-mx) (N-My) (N-my) in
  (N-M, N-m)%N.

Lemma forward_or_pos_correct:
  ∀ mx x Mx, (mx <= x <= Mx)%N ->
  ∀ my y My, (my <= y <= My)%N ->
  ∀ r R, forward_or_pos mx Mx my My = (r, R) ->
  (r <= N.lor x y <= R)%N.
Proof.
  intros. unfold forward_or_pos in H1.
  set (N := N.ones (N.max (N.size Mx) (N.size My))) in *.
  destruct (forward_and_pos (N-Mx) (N-mx) (N-My) (N-my)) eqn:EQ.
  inversion H1; subst; clear H1.
  apply forward_and_pos_correct with (x:=(N-x)%N) (y:=(N-y)%N) in EQ; try (zify; omega).
  replace (N.lor x y) with (N - N.land (N - x) (N - y))%N. zify; omega.
  assert (∀ i, (Mx < 2 ^ i)%N -> N.testbit x i = false).
  { intros. rewrite N.testbit_false, N.div_small. auto. zify; omega. }
  assert (∀ i, (My < 2 ^ i)%N -> N.testbit y i = false).
  { intros. rewrite N.testbit_false, N.div_small. auto. zify; omega. }
  rewrite N.sub_nocarry_ldiff with (b:=x).
  rewrite N.sub_nocarry_ldiff with (b:=y).
  rewrite N.sub_nocarry_ldiff.
  - apply N.bits_inj; intro.
    rewrite N.ldiff_spec, N.land_spec, !N.ldiff_spec, N.lor_spec.
    destruct (N.le_gt_cases (N.max (N.size Mx) (N.size My)) n1).
     + unfold N; rewrite N.ones_spec_high by trivial. simpl.
       rewrite H1, H2; try reflexivity;
       (eapply N.lt_le_trans; [apply N.size_gt|apply N.pow_le_mono_r; [discriminate|zify; omega]]).
     + unfold N; rewrite N.ones_spec_low by trivial; simpl.
       repeat (destruct N.testbit; try reflexivity).
  - apply N.bits_inj_0; intro.
    rewrite N.ldiff_spec, N.land_spec, !N.ldiff_spec.
    repeat (destruct N.testbit; try reflexivity).
  - apply N.bits_inj_0; intro. rewrite N.ldiff_spec.
    destruct (N.le_gt_cases (N.max (N.size Mx) (N.size My)) n1).
    + rewrite H2. reflexivity.
      eapply N.lt_le_trans; [apply N.size_gt|apply N.pow_le_mono_r; [discriminate|zify; omega]].
    + unfold N; rewrite N.ones_spec_low by trivial.
      repeat (destruct N.testbit; try reflexivity).
  - apply N.bits_inj_0; intro. rewrite N.ldiff_spec.
    destruct (N.le_gt_cases (N.max (N.size Mx) (N.size My)) n1).
    + rewrite H1. reflexivity.
      eapply N.lt_le_trans; [apply N.size_gt|apply N.pow_le_mono_r; [discriminate|zify; omega]].
    + unfold N; rewrite N.ones_spec_low by trivial.
      repeat (destruct N.testbit; try reflexivity).
Qed.

Definition forward_diff_pos (mx Mx my My:N) : N * N :=
  let N := N.ones (N.max (N.size Mx) (N.size My)) in
  forward_and_pos mx Mx (N-My) (N-my).

Lemma forward_diff_pos_correct:
  ∀ mx x Mx, (mx <= x <= Mx)%N ->
  ∀ my y My, (my <= y <= My)%N ->
  ∀ r R, forward_diff_pos mx Mx my My = (r, R) ->
  (r <= N.ldiff x y <= R)%N.
Proof.
  intros. unfold forward_diff_pos in H1.
  set (N := N.ones (N.max (N.size Mx) (N.size My))).
  apply forward_and_pos_correct with (x:=x) (y:=(N.ones (N.max (N.size Mx) (N.size My)) - y)%N) in H1;
    try (zify; omega).
  fold N in H1.
  replace (N.land x (N-y)) with (N.ldiff x y) in H1. auto.
  assert (∀ i, (Mx < 2 ^ i)%N -> N.testbit x i = false).
  { intros. rewrite N.testbit_false, N.div_small. auto. zify; omega. }
  assert (∀ i, (My < 2 ^ i)%N -> N.testbit y i = false).
  { intros. rewrite N.testbit_false, N.div_small. auto. zify; omega. }
  rewrite N.sub_nocarry_ldiff.
  - apply N.bits_inj; intro.
    rewrite N.ldiff_spec, N.land_spec, !N.ldiff_spec.
    destruct (N.le_gt_cases (N.max (N.size Mx) (N.size My)) n).
    + unfold N; rewrite N.ones_spec_high by trivial. simpl.
      rewrite H2; try reflexivity.
      (eapply N.lt_le_trans; [apply N.size_gt|apply N.pow_le_mono_r; [discriminate|zify; omega]]).
    + unfold N; rewrite N.ones_spec_low by trivial; simpl.
      repeat (destruct N.testbit; try reflexivity).
  - apply N.bits_inj_0; intro.
    rewrite N.ldiff_spec.
    destruct (N.le_gt_cases (N.max (N.size Mx) (N.size My)) n).
    + unfold N; rewrite N.ones_spec_high by trivial. simpl.
      rewrite H3; try reflexivity.
      (eapply N.lt_le_trans; [apply N.size_gt|apply N.pow_le_mono_r; [discriminate|zify; omega]]).
    + unfold N; rewrite N.ones_spec_low by trivial; simpl.
      repeat (destruct N.testbit; try reflexivity).
Qed.

Definition positive_part (i:itv+⊤) : itv+⊤+⊥ :=
  match i with
    | Just (ITV _ (Zneg _)) => Bot
    | Just (ITV _ 0) => NotBot (Just (ITV 0 0))
    | Just (ITV x y) => NotBot (Just (ITV (Zmax 0 x) y))
    | All => NotBot All
  end.

Lemma positive_part_correct:
  ∀ i n, n ∈ γ i -> 0 <= n ->
  n ∈ γ (positive_part i).
Proof.
  destruct i as [|[]]; simpl; intros. auto.
  destruct z0; simpl; zify; omega.
Qed.

Definition pos_bind fJJ fJA fAJ x y : itv+⊤+⊥ :=
  botbind2 (fun x y =>
              match x, y with
                | Just (ITV mx Mx), Just (ITV my My) =>
                  let '(m, M) := fJJ (Z.to_N mx) (Z.to_N Mx) (Z.to_N my) (Z.to_N My) in
                  NotBot (Just (ITV (Z.of_N m) (Z.of_N M)))
                | Just (ITV mx Mx), All =>
                  match fJA (Z.to_N mx) (Z.to_N Mx) with
                  | None => NotBot All
                  | Some (m, M) =>
                    NotBot (Just (ITV (Z.of_N m) (Z.of_N M)))
                  end
                | All, Just (ITV my My) =>
                  match fAJ (Z.to_N my) (Z.to_N My) with
                  | None => NotBot All
                  | Some (m, M) =>
                    NotBot (Just (ITV (Z.of_N m) (Z.of_N M)))
                  end
                | All, All => NotBot All
              end)
           (positive_part x) (positive_part y).

Lemma pos_bind_spec:
  ∀ x_ab x, Z.of_N x ∈ γ x_ab ->
  ∀ y_ab y, Z.of_N y ∈ γ y_ab ->
  ∀ fJJ fAJ fJA f,
  (∀ mx x Mx, (mx <= x <= Mx)%N ->
   ∀ my y My, (my <= y <= My)%N ->
   ∀ r R, fJJ mx Mx my My = (r, R) ->
          (r <= f x y <= R)%N) ->
  (∀ mx x Mx, (mx <= x <= Mx)%N ->
   ∀ y,
   ∀ r R, fJA mx Mx = Some (r, R) ->
          (r <= f x y <= R)%N) ->
  (∀ x,
   ∀ my y My, (my <= y <= My)%N ->
   ∀ r R, fAJ my My = Some (r, R) ->
          (r <= f x y <= R)%N) ->
  Z.of_N (f x y) ∈ γ (pos_bind fJJ fJA fAJ x_ab y_ab).
Proof.
  unfold pos_bind. intros.
  eapply botbind2_spec.
  destruct x_ab0 as [|[]], y_ab0 as [|[]]; simpl; auto.

  4:apply positive_part_correct. 4:apply H. 4:zify; omega.
  4:apply positive_part_correct. 4:apply H0. 4:zify; omega.

  intros.
  destruct (fAJ (Z.to_N z) (Z.to_N z0)) as [[]|] eqn:EQ. 2:exact I.
  eapply H3 with (x:=x) (y:=y) in EQ. simpl; zify; auto.
  destruct z, z0; simpl; zify; omega.

  intros.
  destruct (fJA (Z.to_N z) (Z.to_N z0)) as [[]|] eqn:EQ. 2:exact I.
  eapply H2 with (x:=x) (y:=y) in EQ. simpl; zify; auto.
  destruct z, z0; simpl; zify; omega.

  intros.
  destruct (fJJ (Z.to_N z) (Z.to_N z0) (Z.to_N z1) (Z.to_N z2)) eqn:EQ.
  eapply H1 with (x:=x) (y:=y) in EQ. simpl; zify; auto.
  destruct z, z0; simpl; zify; omega.
  destruct z1, z2; simpl; zify; omega.
Qed.

Definition forward_land (x y:itv+⊤) : itv+⊤+⊥ :=
  pos_bind forward_and_pos (fun _ M => Some (0%N, M)) (fun _ M => Some (0%N, M)) x y ⊔
  pos_bind forward_diff_pos (fun _ M => Some (0%N, M)) (fun _ _ => None) x (forward_lnot y) ⊔
  pos_bind forward_diff_pos (fun _ M => Some (0%N, M)) (fun _ _ => None) y (forward_lnot x) ⊔
  botlift_fun1 forward_lnot
    (pos_bind forward_or_pos (fun _ _ => None) (fun _ _ => None)
              (forward_lnot x) (forward_lnot y)).

Lemma forward_land_correct:
  ∀ x x_ab, x ∈ γ x_ab ->
  ∀ y y_ab, y ∈ γ y_ab ->
  Z.land x y ∈ γ (forward_land x_ab y_ab).
Proof.
  intros. unfold forward_land.
  assert (∀ x, ∃ x', x = Z.of_N x' \/ x = Z.lnot (Z.of_N x')).
  { intro. unfold Z.lnot. destruct x0.
    exists 0%N. eauto.
    eexists (Npos _). simpl. eauto.
    eexists (Npos p - 1)%N. zify; omega. }
  destruct (H1 x) as [x' []], (H1 y) as [y' []]; subst.

  - replace (Z.land (Z.of_N x') (Z.of_N y')) with (Z.of_N (N.land x' y'))
      by (destruct x', y'; reflexivity).
    apply join_correct. left.
    apply join_correct. left.
    apply join_correct. left.
    apply pos_bind_spec; auto.
    + apply forward_and_pos_correct.
    + intros. inv H3. zify. split. omega.
      transitivity (Z.of_N x); try omega. apply Int.Ztestbit_le. auto.
      intros i Hi. rewrite !Z2N.inj_testbit, N.land_spec, Bool.andb_true_iff by auto.
      now intuition.
    + intros. inv H3. zify. split. omega.
      transitivity (Z.of_N y); try omega. apply Int.Ztestbit_le. auto.
      intros i Hi. rewrite !Z2N.inj_testbit, N.land_spec, Bool.andb_true_iff by auto.
      now intuition.
  - assert (γ (forward_lnot y_ab) (Z.of_N y')). {
      rewrite <- Z.lnot_involutive with (Z.of_N y').
      auto using forward_lnot_correct.
    }
    replace (Z.land (Z.of_N x') (Z.lnot (Z.of_N y'))) with (Z.of_N (N.ldiff x' y'))
      by (destruct x', y'; try reflexivity; simpl;
          replace (Pos.pred_N (p0 + 1)) with (Npos p0); [reflexivity|];
          rewrite Pos.add_1_r, Pos.pred_N_succ; reflexivity).
    apply join_correct. left.
    apply join_correct. left.
    apply join_correct. right.
    apply pos_bind_spec; auto.
    + apply forward_diff_pos_correct.
    + intros. inv H4. zify. split. omega.
      transitivity (Z.of_N x); try omega. apply Int.Ztestbit_le. auto.
      intros i Hi. rewrite !Z2N.inj_testbit, N.ldiff_spec, Bool.andb_true_iff by auto.
      now intuition.
    + discriminate.
  - assert (γ (forward_lnot x_ab) (Z.of_N x')). {
      rewrite <- Z.lnot_involutive with (Z.of_N x').
      auto using forward_lnot_correct.
    }
    replace (Z.land (Z.lnot (Z.of_N x')) (Z.of_N y')) with (Z.of_N (N.ldiff y' x'))
      by (destruct x', y'; try reflexivity; simpl;
          replace (Pos.pred_N (p + 1)) with (Npos p); [reflexivity|];
          rewrite Pos.add_1_r, Pos.pred_N_succ; reflexivity).
    apply join_correct. left.
    apply join_correct. right.
    apply pos_bind_spec; auto.
    + apply forward_diff_pos_correct.
    + intros. inv H4. zify. split. omega.
      transitivity (Z.of_N x); try omega. apply Int.Ztestbit_le. auto.
      intros i Hi. rewrite !Z2N.inj_testbit, N.ldiff_spec, Bool.andb_true_iff by auto.
      now intuition.
    + discriminate.
  - assert (γ (forward_lnot x_ab) (Z.of_N x')). {
      rewrite <- Z.lnot_involutive with (Z.of_N x').
      auto using forward_lnot_correct.
    }
    assert (γ (forward_lnot y_ab) (Z.of_N y')). {
      rewrite <- Z.lnot_involutive with (Z.of_N y').
      auto using forward_lnot_correct.
    }
    replace (Z.land (Z.lnot (Z.of_N x')) (Z.lnot (Z.of_N y'))) with (Z.lnot (Z.of_N (N.lor x' y')))
      by (rewrite <- Z.lnot_lor; destruct x', y'; reflexivity).
    apply join_correct. right. eapply botlift_fun1_spec.
    eauto using forward_lnot_correct.
    apply pos_bind_spec; auto.
    + apply forward_or_pos_correct.
    + discriminate.
    + discriminate.
Qed.

Definition forward_lor (x y:itv+⊤) : itv+⊤+⊥ :=
  botlift_fun1 forward_lnot (forward_land (forward_lnot x) (forward_lnot y)).

Lemma forward_lor_correct:
  ∀ x x_ab, x ∈ γ x_ab ->
  ∀ y y_ab, y ∈ γ y_ab ->
  Z.lor x y ∈ γ (forward_lor x_ab y_ab).
Proof.
  intros.
  rewrite <- Z.lnot_involutive, Z.lnot_lor.
  eapply botlift_fun1_spec; intros.
  apply forward_lnot_correct; eauto.
  apply forward_land_correct; apply forward_lnot_correct; auto.
Qed.

Definition forward_lxor (x y:itv+⊤) : itv+⊤+⊥ :=
  let A := forward_land x (forward_lnot y) in
  let B := forward_land (forward_lnot x) y in
  botbind2 forward_lor A B.

Lemma forward_lxor_correct:
  ∀ x x_ab, x ∈ γ x_ab ->
  ∀ y y_ab, y ∈ γ y_ab ->
  Z.lxor x y ∈ γ (forward_lxor x_ab y_ab).
Proof.
  intros.
  replace (Z.lxor x y) with (Z.lor (Z.land x (Z.lnot y)) (Z.land(Z.lnot x) y)).
  eapply botbind2_spec; intros.
  apply forward_lor_correct; eauto.
  apply forward_land_correct; eauto; apply forward_lnot_correct; auto.
  apply forward_land_correct; eauto; apply forward_lnot_correct; auto.
  apply Z.bits_inj'; intros.
  rewrite !Z.lor_spec, !Z.land_spec, !Z.lnot_spec, Z.lxor_spec; auto.
  repeat (destruct Z.testbit; try reflexivity).
Qed.

Definition forward_shl (x y:itv+⊤) : itv+⊤+⊥ :=
  match x with
    | All => NotBot All
    | Just (ITV mx Mx) =>
      if (Z.eqb mx 0 && Z.eqb Mx 0)%bool then NotBot (Just (ITV 0 0))
      else
        match y with
          | All => NotBot All
          | Just (ITV my My) =>
            if (Z.leb (-64) my && Z.leb My 64)%bool then
              NotBot (Just (ITV (Z.shiftl mx (if Z.leb 0 mx then my else My))
                                (Z.shiftl Mx (if Z.leb 0 Mx then My else my))))
            else
              NotBot All
        end
  end.

Lemma forward_shl_correct:
  ∀ x x_ab, x ∈ γ x_ab ->
  ∀ y y_ab, y ∈ γ y_ab ->
  Z.shiftl x y ∈ γ (forward_shl x_ab y_ab).
Proof.
  intros.
  destruct x_ab as [|[]]; unfold forward_shl. tauto.
  destruct ((z =? 0) && (z0 =? 0))%bool eqn:EQ.
  - rewrite Bool.andb_true_iff, !Z.eqb_eq in EQ. destruct EQ; subst.
    simpl in H. replace x with 0 by omega.
    rewrite Z.shiftl_0_l. split; reflexivity.
  - clear EQ.
    destruct y_ab as [|[]]. tauto.
    simpl in *.
    destruct ((-64 <=? z1) && (z2 <=? 64))%bool. 2:simpl; easy.
    simpl in *.
    assert (forall a b c, 0 <= a -> b <= c -> Z.shiftl a b <= Z.shiftl a c). {
      intros.
      destruct (Z.nonpos_pos_cases b), (Z.nonpos_pos_cases c); try omega.
      - rewrite !Z.shiftl_div_pow2 by omega.
        apply Z.div_le_compat_l. auto.
        split. apply Z.pow_pos_nonneg; omega. apply Z.pow_le_mono_r; omega.
      - rewrite Z.shiftl_div_pow2, Z.shiftl_mul_pow2 by omega.
        transitivity a.
        assert (a / 2 ^ (- b) <= a / 1).
        apply Z.div_le_compat_l. auto. pose proof (Z.pow_pos_nonneg 2 (-b)). omega.
        rewrite Zdiv_1_r in H5. auto.
        assert (a * 1  <= a * 2 ^ c).
        apply Zmult_le_compat_l. pose proof (Z.pow_pos_nonneg 2 c). omega. auto.
        rewrite Zmult_1_r in H5. auto.
      - rewrite !Z.shiftl_mul_pow2 by omega.
        apply Zmult_le_compat_l. apply Z.pow_le_mono_r; omega. auto.
    }
    assert (forall a b c, a < 0 -> b <= c -> Z.shiftl a c <= Z.shiftl a b). {
      intros.
      destruct (Z.nonpos_pos_cases b), (Z.nonpos_pos_cases c); try omega.
      - specialize (H1 (Z.lnot a) b c).
        replace (Z.shiftl (Z.lnot a) b) with (Z.lnot (Z.shiftl a b)) in H1.
        replace (Z.shiftl (Z.lnot a) c) with (Z.lnot (Z.shiftl a c)) in H1.
        unfold Z.lnot in H1. omega.
        apply Z.bits_inj'. intros. rewrite Z.lnot_spec, !Z.shiftl_spec, Z.lnot_spec by omega. reflexivity.
        apply Z.bits_inj'. intros. rewrite Z.lnot_spec, !Z.shiftl_spec, Z.lnot_spec by omega. reflexivity.
      - rewrite Z.shiftl_mul_pow2, Z.shiftl_div_pow2 by omega.
        transitivity a.
        assert ((-a) * 1  <= (-a) * 2 ^ c).
        apply Zmult_le_compat_l. pose proof (Z.pow_pos_nonneg 2 c). omega. omega.
        rewrite !Z.mul_opp_l in H6. omega.
        rewrite <- Z.div_mul with (a:=a) (b:=2^(-b)) at 1.
        apply Z.div_le_mono.
        apply Z.pow_pos_nonneg; omega.
        assert ((-a) * 1  <= (-a) * 2 ^ (-b)).
        apply Zmult_le_compat_l. pose proof (Z.pow_pos_nonneg 2 (-b)). omega. omega.
        rewrite !Z.mul_opp_l in H6. omega.
        pose proof (Z.pow_pos_nonneg 2 (-b)). omega.
      - rewrite !Z.shiftl_mul_pow2 by omega.
        assert ((-a)*2^b <= (-a)*2^c).
        apply Zmult_le_compat_l. apply Z.pow_le_mono_r; omega. omega.
        rewrite !Z.mul_opp_l in H6. omega.
    }
    assert (forall a b c, a <= b -> Z.shiftl a c <= Z.shiftl b c). {
      intros.
      destruct (Z.nonpos_pos_cases c).
      - rewrite !Z.shiftl_div_pow2 by omega.
        apply Z.div_le_mono. apply Z.pow_pos_nonneg; omega. auto.
      - rewrite !Z.shiftl_mul_pow2 by omega.
        apply Zmult_le_compat_r. auto. pose proof (Z.pow_pos_nonneg 2 c). omega.
    }
    destruct H0, H.
    split; [pose proof (Zle_cases 0 z)|pose proof (Zle_cases 0 z0)]; destruct Z.leb.
    + etransitivity. apply H1; eauto. apply H3; eauto.
    + etransitivity. apply H2. omega. eauto. apply H3; eauto.
    + etransitivity. apply H3; eauto. apply H1; eauto.
    + etransitivity. apply H3; eauto. apply H2. omega. eauto.
Qed.

Definition forward_shr (x y:itv+⊤) : itv+⊤+⊥ :=
  forward_shl x (forward_neg y).

Lemma forward_shr_correct:
  ∀ x x_ab, x ∈ γ x_ab ->
  ∀ y y_ab, y ∈ γ y_ab ->
  Z.shiftr x y ∈ γ (forward_shr x_ab y_ab).
Proof.
  intros.
  rewrite <- Z.shiftl_opp_r.
  apply forward_shl_correct. auto.
  apply forward_neg_correct, H0.
Qed.

Definition backward_cmp_l (c:comparison) (x y:itv+⊤) : itv+⊤+⊥ :=
  match c with
    | Ceq => x ⊓ y
    | Cne =>
      match y with
        | Just (ITV m M) =>
          if Z_eq_dec m M then
            match x with
              | Just (ITV mx Mx) =>
                if Z_eq_dec mx m then
                  if Z_eq_dec Mx m then Bot
                  else NotBot (Just (ITV (mx+1) Mx))
                else
                  if Z_eq_dec Mx m then NotBot (Just (ITV mx (Mx-1)))
                  else NotBot x
              | All => NotBot x
            end
          else NotBot x
        | All => NotBot x
      end
    | Cle =>
      match x, y with
        | Just (ITV mx Mx), Just (ITV _ My) =>
          if Zlt_bool My mx then Bot else NotBot (Just (ITV mx (Zmin Mx My)))
        | _, _ => NotBot x
      end
    | Cge =>
      match x, y with
        | Just (ITV mx Mx), Just (ITV my _) =>
          if Zlt_bool Mx my then Bot else NotBot (Just (ITV (Zmax mx my) Mx))
        | _, _ => NotBot x
      end
    | Clt =>
      match x, y with
        | Just (ITV mx Mx), Just (ITV _ My) =>
          let Mx' := My-1 in
          if Zlt_bool Mx' mx then Bot else NotBot (Just (ITV mx (Zmin Mx Mx')))
        | _, _ => NotBot x
      end
    | Cgt =>
      match x, y with
        | Just (ITV mx Mx), Just (ITV my _) =>
          let mx' := my+1 in
          if Zlt_bool Mx mx' then Bot else NotBot (Just (ITV (Zmax mx mx') Mx))
        | _, _ => NotBot x
      end
  end.

Definition Zcmp (c:comparison) (z1 z2:Z) : bool :=
  match c, z1 ?= z2 with
    | Ceq, Eq | Clt, Lt | Cgt, Gt
    | Cle, (Lt|Eq) | Cge, (Gt|Eq)
    | Cne, (Lt|Gt) => true
    | _, _ => false
  end.

Lemma backward_cmp_l_correct:
  ∀ x x_ab, x ∈ γ x_ab ->
  ∀ y y_ab, y ∈ γ y_ab ->
  ∀ c, Zcmp c x y = true ->
    x ∈ γ (backward_cmp_l c x_ab y_ab).
Proof.
  intros x x_ab Hx y y_ab Hy [] Hc; simpl in Hc; destruct (Zcompare_spec x y);
    try discriminate; subst; clear Hc; unfold backward_cmp_l.

  apply meet_correct. auto.

  destruct y_ab as [|[]]. auto.
  destruct (Z_eq_dec z z0); auto. subst.
  destruct x_ab as [|[]]. auto.
  destruct (Z_eq_dec z z0), (Z_eq_dec z1 z0); simpl in *; omega.

  destruct y_ab as [|[]]. auto.
  destruct (Z_eq_dec z z0); auto. subst.
  destruct x_ab as [|[]]. auto.
  destruct (Z_eq_dec z z0), (Z_eq_dec z1 z0); simpl in *; omega.

  destruct x_ab as [|[]], y_ab as [|[]]; auto.
  pose proof (Zlt_cases (z2-1) z). apply Zmin_case; destruct (Zlt_bool (z2-1) z); simpl in *; omega.

  destruct x_ab as [|[]], y_ab as [|[]]; auto.
  pose proof (Zlt_cases z2 z). apply Zmin_case; destruct (Zlt_bool z2 z); simpl in *; omega.

  destruct x_ab as [|[]], y_ab as [|[]]; auto.
  pose proof (Zlt_cases z2 z). apply Zmin_case; destruct (Zlt_bool z2 z); simpl in *; omega.

  destruct x_ab as [|[]], y_ab as [|[]]; auto.
  pose proof (Zlt_cases z0 (z1+1)). apply Zmax_case; destruct (Zlt_bool z0 (z1+1)); simpl in *; omega.

  destruct x_ab as [|[]], y_ab as [|[]]; auto.
  pose proof (Zlt_cases z0 z1). apply Zmax_case; destruct (Zlt_bool z0 z1); simpl in *; omega.

  destruct x_ab as [|[]], y_ab as [|[]]; auto.
  pose proof (Zlt_cases z0 z1). apply Zmax_case; destruct (Zlt_bool z0 z1); simpl in *; omega.
Qed.

Definition backward_mul_pos (res x:itv+⊤) (my My:positive) : itv+⊤+⊥ :=
  match res with
    | Just (ITV mres Mres) =>
      let mx := -((-mres) / Zpos (if Zlt_bool mres 0 then my else My)) in
      let Mx := Mres / Zpos (if Zlt_bool Mres 0 then My else my) in
      Just (ITV mx Mx) ⊓ x
    | All => NotBot x
  end.

Lemma backward_mul_pos_correct:
  ∀ x x_ab, x ∈ γ x_ab ->
  ∀ y my My, Zpos my <= y <= Zpos My ->
  ∀ res_ab, (x * y)%Z ∈ γ res_ab ->
    x ∈ γ (backward_mul_pos res_ab x_ab my My).
Proof.
  intros x x_ab Hx y my My Hy [|[mres Mres]] Hres. auto.
  apply meet_correct. split; auto.
  destruct Hres; split.

  apply Z.opp_le_mono. rewrite Zopp_involutive.
  eapply Zle_trans with ((-mres)/y).
  eapply Zdiv_le_lower_bound. zify; omega. replace (-x*y) with (-(x*y)) by ring. omega.
  pose proof (Zlt_cases mres 0).
  destruct (Zlt_bool mres 0).
  destruct (Z_eq_dec y (Zpos my)). subst; omega.
  apply Zdiv_le_compat_l. omega. zify; omega.
  apply Zdiv_le_lower_bound. reflexivity.
  rewrite (Z_div_mod_eq (-mres) y) at 2 by (zify; omega).
  rewrite Zmult_comm.
  eapply Zle_trans with (y*((-mres)/y)).
  replace ((-mres)/y) with (-1*(-((-mres)/y))) by ring.
  rewrite !Zmult_assoc. apply Zmult_le_compat_r. omega.
  pose proof (Z_div_pos mres y).
  destruct (Z_eq_dec (mres mod y) 0).
  rewrite Z_div_zero_opp_full; trivial. zify; omega.
  rewrite Z_div_nz_opp_full; trivial. zify; omega.
  pose proof (Z_mod_lt (-mres) y). zify; omega.

  eapply Zle_trans with (Mres/y).
  eapply Zdiv_le_lower_bound. zify; omega. auto.
  pose proof (Zlt_cases Mres 0).
  destruct (Zlt_bool Mres 0).
  destruct (Z_eq_dec y (Zpos My)). subst; omega.
  apply Zdiv_le_lower_bound. reflexivity.
  rewrite (Z_div_mod_eq Mres y) at 2 by (zify; omega).
  rewrite Zmult_comm.
  eapply Zle_trans with (y*(Mres/y)).
  replace (Mres/y) with (-1*(-(Mres/y))) by ring.
  rewrite !Zmult_assoc. apply Zmult_le_compat_r. omega.
  pose proof (Z_div_pos (-Mres) y).
  assert (Mres/y <= 0); [|omega].
  eapply Zdiv_le_upper_bound. zify; omega. omega.
  pose proof (Z_mod_lt Mres y). zify; omega.
  destruct (Z_eq_dec y (Zpos my)). subst; omega.
  apply Zdiv_le_compat_l. omega. zify; omega.
Qed.

Definition backward_mul_neg (res x:itv+⊤) (my My:positive) : itv+⊤+⊥ :=
  backward_mul_pos (forward_neg res) x My my.

Lemma backward_mul_neg_correct:
  ∀ x x_ab, x ∈ γ x_ab ->
  ∀ y my My, Zneg my <= y <= Zneg My ->
  ∀ res_ab, (x * y)%Z ∈ γ res_ab ->
    x ∈ γ (backward_mul_neg res_ab x_ab my My).
Proof.
  intros. apply backward_mul_pos_correct with (-y).
  auto. zify; omega.
  replace (x* -y) with (-(x*y)) by ring.
  apply forward_neg_correct, H1.
Qed.

Definition backward_mul_0 (res:itv+⊤) : itv+⊤+⊥ :=
  match res with
    | Just (ITV (Zpos _) _ | ITV _ (Zneg _)) => Bot
    | _ => NotBot All
  end.

Lemma backward_mul_0_correct:
  ∀ x,
  ∀ y, 0 <= y <= 0 ->
  ∀ res_ab, (x * y)%Z ∈ γ res_ab ->
    x ∈ γ (backward_mul_0 res_ab).
Proof.
  intros. replace y with 0 in * by omega.
  destruct res_ab as [|[[] []]]; simpl in *;
  try reflexivity;
  zify; omega.
Qed.

Definition backward_mul_l (res x y:itv+⊤) : itv+⊤+⊥ :=
  match y with
    | Just (ITV 0 0) => backward_mul_0 res
    | Just (ITV (Zpos _) (Zneg _|0) | ITV 0 (Zneg _)) => Bot
    | Just (ITV (Zpos my) (Zpos My)) => backward_mul_pos res x my My
    | Just (ITV 0 (Zpos My)) =>
      backward_mul_0 res ⊔ backward_mul_pos res x 1%positive My
    | Just (ITV (Zneg my) (Zpos My)) =>
      backward_mul_neg res x my 1 ⊔ backward_mul_0 res ⊔ backward_mul_pos res x 1 My
    | Just (ITV (Zneg my) 0) =>
      backward_mul_neg res x my 1 ⊔ backward_mul_0 res
    | Just (ITV (Zneg my) (Zneg My)) => backward_mul_neg res x my My
    | All => NotBot x
  end.

Lemma backward_mul_l_correct:
  ∀ x x_ab, x ∈ γ x_ab ->
  ∀ y y_ab, y ∈ γ y_ab ->
  ∀ res_ab, (x * y)%Z ∈ γ res_ab ->
    x ∈ γ (backward_mul_l res_ab x_ab y_ab).
Proof.
  intros x x_ab Hx y y_ab Hy res_ab Hres.
  destruct y_ab as [|[[|my|my] [|My|My]]], y as [|y|y];
    try reflexivity;
    simpl in Hy; try (exfalso; zify; omega);
    unfold backward_mul_l; auto.
  eapply backward_mul_0_correct; eauto.
  eapply join_correct, or_introl, backward_mul_0_correct; eauto. omega.
  eapply join_correct, or_intror, backward_mul_pos_correct; eauto; zify. omega.
  eapply backward_mul_pos_correct; eauto; zify; omega.
  eapply join_correct, or_intror, backward_mul_0_correct; eauto; omega.
  eapply join_correct, or_introl, backward_mul_neg_correct; eauto; zify; omega.
  eapply join_correct, or_introl. auto. eapply join_correct, or_intror, backward_mul_0_correct; eauto; omega.
  eapply join_correct, or_intror, backward_mul_pos_correct; eauto; zify; omega.
  eapply join_correct, or_introl. auto. eapply join_correct, or_introl, backward_mul_neg_correct; eauto; zify; omega.
  eapply backward_mul_neg_correct; eauto; zify; omega.
Qed.
