(* *********************************************************************)
(*                                                                     *)
(*              The Verasco static analyzer                            *)
(*                                                                     *)
(*          Jacques-Henri Jourdan, INRIA Paris-Rocquencourt            *)
(*                                                                     *)
(*  Copyright Institut National de Recherche en Informatique et en     *)
(*  Automatique and Université de Rennes 1.  All rights reserved.      *)
(*  This file is distributed under the terms of the GNU General        *)
(*  Public License as published by the Free Software Foundatio,        *)
(*  either version 2 of the License, or (at your option) any later     *)
(*  version.  This file is also distributed under the terms of the     *)
(*  INRIA Non-Commercial License Agreement.                            *)
(*                                                                     *)
(* *********************************************************************)

module ZArith = Z

let int_modulus = ZArith.(- ~$2 * of_int min_int)
let int32_modulus = ZArith.(shift_left ~$1 32)
let int64_modulus = ZArith.(shift_left ~$1 64)

module P = struct
  type t = ZArith.t

  let xI x = ZArith.(succ x + x)
  let xO x = ZArith.(x + x)
  let xH = ZArith.one

  let dispatch fxI fxO fxH x =
    let open ZArith in
    let div2 = shift_right x 1 in
    if equal (div2 + div2) x then fxO div2
    else if equal div2 zero then fxH () else fxI div2

  let one = ZArith.one
  let succ = ZArith.succ
  let add = ZArith.add
  let mul = ZArith.mul

  let pow a b = ZArith.(pow a (to_int b))

  let eq = ZArith.equal
  let lt = ZArith.lt
  let gt = ZArith.gt
  let le = ZArith.leq
  let ge = ZArith.geq
  let compare = ZArith.compare
  let compare_cont p q c =
    match ZArith.compare p q with
    | -1 -> Datatypes.Lt
    | 0 -> c
    | 1 -> Datatypes.Gt
    | _ -> assert false
  let to_int n =
    if ZArith.(gt n ~$max_int) then ZArith.(to_int (n - int_modulus))
    else ZArith.to_int n

  let of_int n =
    if n <= 0 then
      begin
        assert (n <> 0);
        ZArith.(of_int n + int_modulus)
      end
    else ZArith.of_int n

  let to_int32 n =
    if ZArith.(gt n (of_int32 Int32.max_int)) then ZArith.(to_int32 (n - int32_modulus))
    else ZArith.to_int32 n

  let of_int32 n =
    if n <= 0l then
      begin
        assert (n <> 0l);
        ZArith.(of_int32 n + int32_modulus)
      end
    else ZArith.of_int32 n

  let to_int64 n =
    if ZArith.(gt n (of_int64 Int64.max_int)) then ZArith.(to_int64 (n - int64_modulus))
    else ZArith.to_int64 n

  let of_int64 n =
    if n <= 0L then
      begin
        assert (n <> 0L);
        ZArith.(of_int64 n + int64_modulus)
      end
    else ZArith.of_int64 n

  let to_zarith n = n

  let of_zarith n =
    assert (ZArith.sign n = 1);
    n

  let (+) = add
  let (=) = eq
  let (<) = lt
  let (<=) = le
  let (>) = gt
  let (>=) = ge
end

module N = struct
  type t = ZArith.t

  let n0 = ZArith.zero
  let npos x = x

  let dispatch fN0 fNpos x =
    match ZArith.sign x with
    | 0 -> fN0 ()
    | 1 -> fNpos x
    | _ -> assert false

  let zero = n0
  let one = ZArith.one
  let succ = ZArith.succ
  let add = ZArith.add
  let mul = ZArith.mul

  let div_eucl a b =
    try ZArith.div_rem a b
    with Division_by_zero -> (zero, zero)

  let pow a b = ZArith.(pow a (to_int b))

  let logand = ZArith.logand
  let logxor = ZArith.logxor
  let logor = ZArith.logor
  let shiftl a b = ZArith.shift_left a (ZArith.to_int b)
  let shiftr a b = ZArith.shift_right a (ZArith.to_int b)
  let eq = ZArith.equal
  let lt = ZArith.lt
  let gt = ZArith.gt
  let le = ZArith.leq
  let ge = ZArith.geq
  let compare = ZArith.compare

  let to_int n =
    if ZArith.(gt n ~$max_int) then ZArith.(to_int (n - int_modulus))
    else ZArith.to_int n

  let of_int n =
    if n < 0 then ZArith.(of_int n + int_modulus)
    else ZArith.of_int n

  let to_int32 n =
    if ZArith.(gt n (of_int32 Int32.max_int)) then ZArith.(to_int32 (n - int32_modulus))
    else ZArith.to_int32 n

  let of_int32 n =
    if n < 0l then ZArith.(of_int32 n + int32_modulus)
    else ZArith.of_int32 n

  let to_int64 n =
    if ZArith.(gt n (of_int64 Int64.max_int)) then ZArith.(to_int64 (n - int64_modulus))
    else ZArith.to_int64 n

  let of_int64 n =
    if n < 0L then ZArith.(of_int64 n + int64_modulus)
    else ZArith.of_int64 n

  let to_zarith n = n

  let of_zarith n =
    assert (ZArith.sign n <> -1);
    n

  let (+) = add
  let ( * ) = mul
  let (=) = eq
  let (<) = lt
  let (<=) = le
  let (>) = gt
  let (>=) = ge
end

module Z = struct
  type t = ZArith.t

  let z0 = ZArith.zero
  let zpos x = x
  let zneg = ZArith.neg

  let dispatch fZ0 fZpos fZneg x =
    match ZArith.sign x with
    | -1 -> fZneg (ZArith.neg x)
    | 0 -> fZ0 ()
    | 1 -> fZpos x
    | _ -> assert false

  let zero = ZArith.zero
  let one = ZArith.one
  let mone = ZArith.minus_one
  let succ = ZArith.succ
  let pred = ZArith.pred
  let neg = ZArith.neg
  let add = ZArith.add
  let sub = ZArith.sub
  let mul = ZArith.mul

  let div_eucl a b =
    let open ZArith in
    match sign b with
    | 0 -> (zero, zero)
    | 1 -> ediv_rem a b
    | -1 ->
      let (q, r) = ediv_rem (neg a) (neg b) in
      (q, neg r)
    | _ -> assert false

  let quotrem a b =
    try ZArith.div_rem a b
    with Division_by_zero -> (zero, zero)

  let pow a b =
    let b = ZArith.to_int b in
    if b < 0 then zero
    else ZArith.(pow a b)

  let logand = ZArith.logand
  let logxor = ZArith.logxor
  let logor = ZArith.logor
  let shiftl a b =
    let open ZArith in
    match sign b with
    | 0 -> a
    | 1 -> shift_left a (to_int b)
    | -1 -> shift_right a (Pervasives.(~-) (to_int b))
    | _ -> assert false
  let shiftr a b =
    let open ZArith in
    match sign b with
    | 0 -> a
    | 1 -> shift_right a (to_int b)
    | -1 -> shift_left a (Pervasives.(~-) (to_int b))
    | _ -> assert false

  let eq = ZArith.equal
  let lt = ZArith.lt
  let gt = ZArith.gt
  let le = ZArith.leq
  let ge = ZArith.geq
  let compare = ZArith.compare

  let to_int n =
    if ZArith.(gt n ~$max_int) then ZArith.(to_int (n - int_modulus))
    else ZArith.to_int n

  let of_sint = ZArith.of_int

  let of_uint n =
    if n < 0 then ZArith.(of_int n + int_modulus)
    else ZArith.of_int n

  let to_int32 n =
    if ZArith.(gt n (of_int32 Int32.max_int)) then ZArith.(to_int32 (n - int32_modulus))
    else ZArith.to_int32 n

  let of_sint32 = ZArith.of_int32

  let of_uint32 n =
    if n < 0l then ZArith.(of_int32 n + int32_modulus)
    else ZArith.of_int32 n

  let to_int64 n =
    if ZArith.(gt n (of_int64 Int64.max_int)) then ZArith.(to_int64 (n - int64_modulus))
    else ZArith.to_int64 n

  let of_sint64 = ZArith.of_int64

  let of_uint64 n =
    if n < 0L then ZArith.(of_int64 n + int64_modulus)
    else ZArith.of_int64 n

  let to_zarith n = n

  let of_zarith n = n

  let of_N n = n

  let dec = ZArith.format "%d"

  let hex = ZArith.format "%X"

  let to_string = dec

  let (+) = add
  let (-) = sub
  let ( * ) = mul
  let (=) = eq
  let (<) = lt
  let (<=) = le
  let (>) = gt
  let (>=) = ge
end
