(* *********************************************************************)
(*                                                                     *)
(*              The Verasco static analyzer                            *)
(*                                                                     *)
(*          Vincent Laporte, Université de Rennes 1                    *)
(*                                                                     *)
(*  Copyright Institut National de Recherche en Informatique et en     *)
(*  Automatique and Université de Rennes 1.  All rights reserved.      *)
(*  This file is distributed under the terms of the GNU General        *)
(*  Public License as published by the Free Software Foundatio,        *)
(*  either version 2 of the License, or (at your option) any later     *)
(*  version.  This file is also distributed under the terms of the     *)
(*  INRIA Non-Commercial License Agreement.                            *)
(*                                                                     *)
(* *********************************************************************)

Require Import
  Utf8 Coqlib Util Maps.

Set Implicit Arguments.

Section ASSOC.

Context (A B: Type)
        `{EqDec A}
        (a: A).

Fixpoint assoc (l: list (A * B)) : option B :=
  match l with
  | cons (a', b) l' =>
    if eq_dec a' a then Some b else assoc l'
  | nil => None
  end.

Definition assoc_prop l pre (b: B) post : Prop :=
  l = pre ++ (a, b) :: post ∧
  ∀ b', ¬ In (a, b') pre.

Fixpoint assoc_dep (l : list (A * B)) :
  { res | assoc_prop l (fst (fst res)) (snd (fst res)) (snd res) } + { ∀ b, ¬ In (a, b) l}.
  refine(
  match l with
  | cons (a', b) l' =>
    match eq_dec a' a with
    | left EQ => inleft (exist _ (nil, b, l') _)
    | right NE => match assoc_dep l' with
                  | inleft (exist r R) =>
                    match r as r' return assoc_prop l' (fst (fst r')) (snd (fst r')) (snd r') → _ with
                    | (pre, b', post) => λ R', inleft (exist _ ((a', b) :: pre, b', post) _)
                    end R
                  | inright R => inright _
                  end
    end
  | nil => inright _
  end).
Proof.
  exact (λ _ K, K).
  clear -EQ. abstract (split; [rewrite EQ; reflexivity | exact (λ _ K, K)]).
  clear -NE R'. abstract (simpl in *; destruct R' as [R1 R2]; split;[simpl; congruence|
  (intros c [ K | K ];[eq_some_inv; intuition | exact (R2 _ K)])]).
  clear -NE R. abstract (intros c [K | K]; [eq_some_inv; intuition | exact (R _ K)]).
Defined.
Global Opaque assoc_dep.

Lemma assoc_in l :
  ∀ b, assoc l = Some b → In (a, b) l.
Proof.
  induction l as [|(a', b') l IH].
  exact (λ _ H, None_not_Some H _).
  intros b. simpl.
  destruct eq_dec. intro. eq_some_inv. left. subst. reflexivity.
  intuition.
Qed.

Lemma assoc_Some l :
  ∀ b,
    assoc l = Some b →
    ∃ pre post,
      l = pre ++ (a, b) :: post ∧
      ∀ b', ¬ In (a, b') pre.
Proof.
  induction l as [|(a', b') l IH].
  exact (λ _ H, None_not_Some H _).
  intros b. simpl.
  destruct eq_dec as [EQ|NE].
  intros EQ'.
  exists nil. exists l. simpl. eq_some_inv. intuition. subst. reflexivity.
  intros R. destruct (IH _ R) as (pre & post & X & Y).
  exists ((a', b') :: pre), post.
  split. simpl. subst. reflexivity.
  intros z [Q|Q]. apply pair_eq_inv in Q. intuition.
  intuition eauto.
Qed.

Lemma assoc_None l :
  assoc l = None ↔
  ∀ b, ¬ In (a, b) l.
Proof.
  induction l as [|(a', b') l IH]. intuition.
  simpl. destruct eq_dec. split. intro; eq_some_inv.
  intros K. destruct (K b'). left. congruence.
  split.
  intros K b [X|X]. eq_some_inv. auto.
  exact (proj1 IH K b X).
  intros K. firstorder.
Qed.

Lemma In_norepet_assoc (b: B) l :
  list_norepet (map fst l) →
  In (a, b) l → assoc l = Some b.
Proof.
  induction l as [|(a', b') l IH]. intros _ ().
  intros NR [IN|IN].
  eq_some_inv. simpl. subst. rewrite dec_eq_true. reflexivity.
  simpl. inv NR. rewrite dec_eq_false. apply IH; auto.
  intros ->. elim H2. rewrite in_map_iff. exists (a, b). intuition.
Qed.

Lemma assoc_app l m :
  assoc (l ++ m) = match assoc l with Some r => Some r | None => assoc m end.
Proof.
  induction l as [|(x,b) l IH]. reflexivity.
  simpl. destruct eq_dec; auto.
Qed.
  
Corollary assoc_iff l b :
  assoc l = Some b ↔ ∃ pre post, assoc_prop l pre b post.
Proof.
  split. apply assoc_Some.
  intros (pre & post & -> & K).
  rewrite assoc_app.
  generalize (assoc_in pre).
  destruct (assoc pre). intros X. elim (K _ (X _ eq_refl)).
  intros _. simpl. rewrite dec_eq_true. reflexivity.
Qed.

End ASSOC.

Lemma ptree_get_assoc {A: Type} (m: PTree.t A) :
  ∀ x, m ! x = assoc x (PTree.elements m).
Proof.
  intros x.
  generalize (PTree.elements_correct m x).
  generalize (PTree.elements_complete m x).
  generalize (PTree.elements_keys_norepet m).
  intros NR C I.
  destruct (m ! x). specialize (I _ eq_refl). symmetry. apply In_norepet_assoc; auto.
  generalize (assoc_in x (PTree.elements m)).
  destruct assoc; auto.
Qed.
