(* *********************************************************************)
(*                                                                     *)
(*              The Verasco static analyzer                            *)
(*                                                                     *)
(*          Vincent Laporte, Université de Rennes 1                    *)
(*                                                                     *)
(*  Copyright Institut National de Recherche en Informatique et en     *)
(*  Automatique and Université de Rennes 1.  All rights reserved.      *)
(*  This file is distributed under the terms of the GNU General        *)
(*  Public License as published by the Free Software Foundatio,        *)
(*  either version 2 of the License, or (at your option) any later     *)
(*  version.  This file is also distributed under the terms of the     *)
(*  INRIA Non-Commercial License Agreement.                            *)
(*                                                                     *)
(* *********************************************************************)

Require Import Utf8 Coqlib Util Maps.
Require Import AST Values Integers Memory Globalenvs Cminor Csharpminor.

Set Implicit Arguments.
Local Unset Elimination Schemes.

Lemma eval_expr_unop_ptr (ge: genv) e tmp m u q b i:
  ¬ eval_expr ge e tmp m (Eunop u q) (Vptr b i).
Proof.
  intros H.
  remember (Eunop u q) as U eqn: EQ. revert u q EQ.
  remember (Vptr b i) as v eqn: Hv. revert b i Hv.
  induction H; try congruence.
  intros b i -> u q K. inv K.
  destruct u; destruct v1; simpl in *; try congruence;
  match goal with H : context[option_map _ ?m] |- _ => destruct m; simpl in H end; congruence.
Qed.

Lemma eval_expr_binop_ptr (ge: genv) e tmp m o q1 q2 b i:
  eval_expr ge e tmp m (Ebinop o q1 q2) (Vptr b i) →
  (o = Oadd ∧ ∃ v1 v2, eval_expr ge e tmp m q1 v1 ∧ eval_expr ge e tmp m q2 v2 ∧ Val.add v1 v2 = Vptr b i)
  ∨
  (o = Osub ∧ ∃ v1 v2, eval_expr ge e tmp m q1 v1 ∧ eval_expr ge e tmp m q2 v2 ∧ Val.sub v1 v2 = Vptr b i).
Proof.
  intros H. inv H. revert H6. destruct o; simpl in *;
  try (destruct v1; destruct v2; simpl in *; try bif; discriminate);
  intro; eq_some_inv;
  try (
      unfold Val.cmp in *; unfold Val.cmpu in *;
      unfold Val.cmpf in *; unfold Val.cmpfs in *;
      unfold Val.cmpl in *; unfold Val.cmplu in *;
      match goal with
      | H : appcontext[Val.of_optbool ?b] |- _ => destruct b as [()|]; simpl in H; discriminate
      | H : appcontext[option_map ?f ?b] |- _ => destruct b as [()|]; simpl in H;
                                             eq_some_inv; discriminate
      end
    );
  intuition eauto.
Qed.

Lemma eval_expr_binop_ptr' (ge:genv) e tmp m o q1 q2 b i:
  eval_expr ge e tmp m (Ebinop o q1 q2) (Vptr b i) →
  ∃ j,
    (eval_expr ge e tmp m q1 (Vptr b j)
     ∨
     eval_expr ge e tmp m q2 (Vptr b j)).
Proof.
  intros H.
  destruct (eval_expr_binop_ptr H) as
      [(-> & v1 & v2 & H1 & H2 & K)
      |(-> & v1 & v2 & H1 & H2 & K)];
  destruct v1; simpl in K; try discriminate;
  destruct v2; simpl in K; inv K;
  intuition eauto.
  destruct eq_block; discriminate.
Qed.

Lemma create_undef_temps_get l x :
  match (create_undef_temps l) ! x with
  | Some v => v = Vundef ∧ In x l
  | None => ¬ In x l
  end.
Proof.
  induction l as [|y l IH]; simpl.
  rewrite PTree.gempty. exact id.
  rewrite PTree.gsspec. destruct peq. intuition.
  destruct ((create_undef_temps l) ! x). intuition.
  intuition.
Qed.

Lemma alloc_can_store m sz m' b :
  0 <= sz →
  Mem.alloc m 0 sz = (m', b) →
  ∃ m'', store_zeros m' b 0 sz = Some m''.
Proof.
  intros SZ H.
  pose proof (Mem.valid_access_alloc_same _ _ _ _ _ H Mint8unsigned) as K. clear H.
  assert (∀ o, 0 <= o → o + 1 <= sz → Mem.valid_access m' Mint8unsigned b o Writable) as H.
  intros o H H0. eapply Mem.valid_access_implies. eapply K; eauto. exists o; simpl; ring. vauto.
  clear K.
  cut (∀ p n, 0 <= p → 0 <= n → p + n <= sz → ∃ m'', store_zeros m' b p n = Some m'').
  intros K. apply (K 0 sz); intuition.
  revert m' H. clear.
  intros m' H p n.
  functional induction (store_zeros m' b p n). vauto.
  intros H0 H1 H2. eapply IHo; auto. intros o H3 H4.
  eapply Mem.store_valid_access_1; eauto. intuition. intuition. auto with zarith.
  intros H0 H1 H2. exfalso.
  edestruct (Mem.valid_access_store m Mint8unsigned b p Vzero) as (m' & M).
  apply H; auto. intuition.
  congruence.
Qed.

Lemma init_data_list_size_pos init :
  0 <= Genv.init_data_list_size init.
Proof.
  induction init as [|i init IH]. reflexivity.
  simpl. generalize (Genv.init_data_size_pos i). intuition.
Qed.

Lemma alloc_can_drop m lo hi m' b p :
  Mem.alloc m lo hi = (m', b) →
  ∃ m'', Mem.drop_perm m' b lo hi p = Some m''.
Proof.
  intros A.
  edestruct (Mem.range_perm_drop_2 m' b lo hi p). 2: vauto.
  intros o Ho. eapply Mem.perm_alloc_2; eauto.
Qed.

Lemma misaligned_load κ ofs :
  ¬ (align_chunk κ | ofs) →
  ∀ m b,
    Mem.load κ m b ofs = None.
Proof.
  intros H m b.
  generalize (Mem.load_valid_access m κ b ofs).
  destruct Mem.load. 2: exact (λ _, eq_refl).
  intros X. specialize (X _ eq_refl). destruct X as [_ X].
  exfalso. exact (H X).
Qed.
Arguments misaligned_load [κ][ofs] _ _ _.

Lemma not_valid_block_load_None m b :
  ¬ Mem.valid_block m b →
  ∀ κ ofs, Mem.load κ m b ofs = None.
Proof.
  intros H κ ofs.
  generalize (Mem.load_valid_access m κ b ofs).
  destruct (Mem.load κ m b ofs); auto.
  intros X. specialize (X _ eq_refl).
  elim H. eapply Mem.valid_access_valid_block.
  eapply Mem.valid_access_implies. eassumption. vauto.
Qed.
